/*
 *  build.c	DCM Wood  93/03/15
 *
 *  Build the tree
 *    Start at some gateway (allow user to specify)(pick cu-gw as default)
 *    Create top Tnode
 *    Place a pointer to the top Tnode on the FIFO work list
 *    Put level++ flag on FIFO work list
 *
 *    For each entry on the work list
 *      If level++,
 *        Start new level
 *        Put a level++ flag on FIFO work list
 *      else
 *        Process that Tnode
 *          For each connection in the SNGW structure
 *            Create a Tnode
 *            Put the Tnode on FIFO work list
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>

#define  TNODE
#include "jsdb.h"
#include "js2fig.h"

extern struct Node * FindNodeEnt();
extern struct Node * FindOrAddNode();

struct Qent {
	struct Tnode * Tnode;
	struct Qent * next;
	};

int	NumAtThisLevel = 0;

struct Qent * Qin = NULL;
struct Qent * Qout = NULL;

struct Tnode TopTnode = { NULL, NULL, NULL, NULL, 0, 0, 0, 0, NULL, NULL };


/*
 *  enFIFO
 *	Add entry to FIFO queue
 */

enFIFO( Tnode )
struct Tnode * Tnode;
{
    struct Qent * new;

    new = (struct Qent *) malloc( sizeof( struct Qent ) );
    new->Tnode = Tnode;
    new->next = NULL;

    if ( Qin == NULL )
	Qout = new;
    else
	Qin->next = new;
    Qin = new;

    return;
}


/*
 *  deFIFO
 *	Remove entry from FIFO queue
 */

struct Qent *
deFIFO()
{
    struct Qent * rp;
    struct Qent   rqe;

    if ( Qout == NULL )
	return( NULL );

    rqe.Tnode = Qout->Tnode;

    rp = Qout;
    Qout = Qout->next;
    if ( Qout == NULL )
	Qin = NULL;

    free( rp );

    return( &rqe );
}


/*
 *  AddToTree
 *	Add an SNGW node to the Tree
 */

struct Tnode *
AddToTree( sngwp )
struct Node * sngwp;
{
    u_long  sm;
    struct in_addr inaddr;
    struct list * lp;
    struct Node * netp;
    u_long  nm;
    int     snc;
    struct Tnode * tnp;
    struct Node * xp;

    xp = FindNodeEnt( sngwp->addr, sngwp->ntype );
    if ( xp != sngwp )
	{
	inaddr.s_addr = sngwp->addr;
	fprintf( stderr, "Node confusion for %s\n", inet_ntoa( inaddr ) );
	return( NULL );
	}

    if ( xp->ntype == NTYPE_INTF && xp->gwp == NULL && xp->nxtintfp == NULL )
	{
	if ( xp->sm != NULL )
	    sm = xp->sm;
	else
	    {
	    nm = GetNetMask( xp->addr );
	    netp = FindNodeEnt( xp->addr & nm, NTYPE_NET );
	    sm = netp->sm;
	    }
	snc = 0;
	for ( lp = sngwp->snlp; lp != NULL; lp = lp->next )
	    if ( ( ( lp->node->addr ^ xp->addr ) & sm ) != 0 )
		snc++;

	if ( snc == 0 )
	    return( NULL );
	}

    for ( xp = sngwp; xp != NULL; xp = xp->nxtintfp )
	{
	if ( xp->tnp != NULL )
	    return( NULL );

	if ( xp->gwp != NULL )
	    if ( xp->gwp->tnp != NULL )
		{
		xp->tnp = xp->gwp->tnp;
		return( NULL );
		}
	}

    tnp = (struct Tnode *) calloc( 1, sizeof(struct Tnode) );
    tnp->Datap = sngwp;
    sngwp->tnp = tnp;
    NumAtThisLevel++;
    TnCnt++;
    enFIFO( tnp );
    return( tnp );
}


/*
 *  ProcessTnode
 *	Process Tnode entry by adding all children
 */

ProcessTnode( Tnodep )
struct Tnode * Tnodep;
{
    u_long  csm;			/* Current Subnet Mask */
    struct list * lp;
    struct Tnode * lsp = NULL;
    u_long  nm;
    struct Node * netp;
    struct Node * sngwp;
    struct Node * snp;
    u_long  subnet;
    struct Tnode * tnp;
    struct Node * xp;

    sngwp = Tnodep->Datap;

    switch ( sngwp->ntype )
	{
	case NTYPE_INTF:  /*  was NTYPE_GATEWAY  */
	    lp = sngwp->snlp;
	    break;

	case NTYPE_SUBNET:
	    lp = sngwp->gwlp;
	    break;
	}

    for ( ; lp != NULL; lp = lp->next )
	{
	if ( Tnodep->Parent != NULL )
	    if ( Tnodep->Parent->Datap == lp->node )
		continue;

	tnp = AddToTree( lp->node );		/* Add children */
	if ( tnp != NULL )
	    {
	    if ( Tnodep->FirstChild == NULL )
		Tnodep->FirstChild = tnp;
	    else
		{
		lsp->RightSibling = tnp;
		tnp->LeftSibling = lsp;
		}
	    tnp->Parent = Tnodep;
	    lsp = tnp;
	    }
	}

    for ( xp = sngwp->nxtintfp; xp != NULL; xp = xp->nxtintfp )
	{
	if ( xp->tnp == NULL )
	    {
	    xp->tnp = Tnodep;
	    if ( xp->sm != NULL )
		csm = xp->sm;
	    else
		{
		nm = GetNetMask( xp->addr );
		netp = FindNodeEnt( xp->addr & nm, NTYPE_NET );
	        csm = netp->sm;
		}
	    subnet = xp->addr & csm;
	    if ( subnet != 0 )
		{
	        snp = FindNodeEnt( subnet, NTYPE_SUBNET );
	        if ( snp != NULL )
		    if ( snp->tnp == NULL )
		        AddToTree( snp );
		}
	    }
	else
	    {
	    fprintf( stderr, "Interface already on tree\n" );
	    }
	}

    return;
}


/*
 *  BuildTree
 *	Main program
 *	Called with a pointer to the top-level data node
 */

struct Tnode *
BuildTree( sngwp )
struct Node * sngwp;
{
    u_long  ipaddr;
    int     level = 0;
    struct Qent * rqep;
    struct Tnode * ttnp;

    TopTnode.Datap = sngwp;
    ttnp = &TopTnode;
    enFIFO( NULL );
    ProcessTnode( ttnp );
    enFIFO( NULL );

    while ( 1 )
	{
        rqep = deFIFO();
	if ( rqep == NULL )
	    break;

	ttnp = rqep->Tnode;
	if ( ttnp == NULL )
	    {
	    if ( NumAtThisLevel != 0 )
	        enFIFO( NULL );

	    NumAtThisLevel = 0;
	    level++;
	    }
	else
	    ProcessTnode( ttnp );
	}

    return( &TopTnode );
}

