/*
 *  js2fig.c	DCM Wood  93/03/05
 *
 *  Arguments:
 *	jshost, jsport, display, geometry
 *
 *  Read subnets (and networks) and gateways.
 *    Get NETWORKs
 *        For each entry,
 *            Create a struct subnet if one does not exist
 *                Add to hash table too
 *            Add list of attached gateways
 *    Get GATEWAYs
 *        For each entry,
 *            Create a struct gateway if one does not exist
 *                Add to hash table too
 *            Add list of connected subnets
 *
 *  Build the tree
 *    Start at some gateway (allow user to specify)(pick cu-gw as default)
 *      Place a pointer to each SN on the FIFO work list
 *      Create top struct node
 *      Put level++ flag on FIFO work list
 *    For each entry on the work list
 *      If level++,
 *        Start new level
 *        Create struct node
 *      else
 *        Add to struct node
 *        If entry has pointer list
 *          Put each pointer on FIFO work list
 *      
 *  Position the nodes
 *
 *  Output file for (x)fig
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <netdb.h>
#include <netinet/in.h>

#define MAIN
#include "js2fig.h"
#include "js.h"
#include "jsdb.h"
#include "PositionTree.h"

extern	struct Tnode *	BuildTree();
extern	struct Node *	FindNodeEnt();

int	PTsuccess;

static	char    ipstr[16] = "128.138.1.0";
static	int	ntype =     NTYPE_SUBNET;


main( argc, argv )
int	argc;
char **	argv;
{
    int     i;
    int     id;
    u_long  ipaddr;
    struct Node * sngwp;

    setlinebuf( stdout );
    jshostp = getenv( "JS_HOST" );

    /* get the command line arguments if they exist */

    while ( *(++argv) )
	{
	if ( **argv != '-' )
	    {
	    printf( "Use a - before arguments\n" );
	    exit( 1 );
	    }
	switch ( *(*argv + 1) )
	    {
	    case 'd':
		debug = 1;	/* set the debug flag */
		break;

	    case 'j':
		if ( *(++argv) )
		    jshostp = argv[0];
		else
		    {
		    Usage();
		    exit( -1 );
		    }
		break;

	    case 'n':
		if ( *(++argv) )
		    strcpy( ipstr, argv[0] );	/* set root node */
		break;

	    default:
		Usage();
		exit( -1 );
	    }
	}

    /* GetIntf();			/* Get interfaces on this machine  */

    ReadJS( INTF );
    ReadJS( NETWORK );
    ReadJS( GATEWAY );
        fprintf( stderr, "\n" );
        fprintf( stderr, " %5d Interface records read\n", IRCnt );
        fprintf( stderr, " %5d Network records read\n", NRCnt );
        fprintf( stderr, " %5d Gateway records read\n", GRCnt );
        fprintf( stderr, "\n" );
        fprintf( stderr, " %5d Interface entries created\n", IFCnt );
        fprintf( stderr, " %5d Name entries created\n", NameCnt );
        fprintf( stderr, " %5d Network entries created\n", NetCnt );
        fprintf( stderr, " %5d Subnet entries created\n", SNCnt );
        fprintf( stderr, "\n" );

    ipaddr = inet_addr( ipstr );
    sngwp = FindNodeEnt( ipaddr, ntype );
    if ( sngwp == NULL )
	{
	fprintf( stderr, "SNGW node not found for %s\n", ipstr );
	exit( 1 );
	}

    Tnp = BuildTree( sngwp );
        fprintf( stderr, " %5d Tree nodes created\n", TnCnt );
    Tnp->Xcoord = 100;
    Tnp->Ycoord = 700;

    LevelSeparation = 100;
    MaxDepth = 10;
    SiblingSeparation = 2;
    SubtreeSeparation = 4;

    xLimit = 8000;
    yLimit = 4000;

    Nodesize = 12;
    Orientation = O_WEST;

    PTsuccess = PositionTree( Tnp );
    if ( ! PTsuccess )
	fprintf( stderr, "PositionTree failed\n" );

    OutputFile( Tnp );
    exit( 0 );
}


Usage()
{
    fprintf( stderr, "js2fig: Usage: js2fig\n" );
}

