/*
 *  jsdb.c	DCM Wood  93/03/09
 */

/*
 *	Build data base structure from information retrieved from the JS
 *
 *  Procedures:
 *	AddGWIntf( gwjsnp, intfjsnp )
 *	AddGWIntfP( gwp, inp )
 *	AddIntf( jsNp )
 *	AddName( namep, Nodep )
 *	AddNet( jsNp )
 *	AddToLists( snp, gwp )
 *	CreateListElement( anp )
 *	CreateName( np )
 *	CreateNode( jsnp, ntype )
 *	FindNodeEnt( addr, ntype )
 *	FindOrAddNode( jsnp, ntype )
 *	FixIntfCounts()
 *	GetNetMask( addr )
 *	HashAddr( addr )
 *	HashName( np )
 *	MatchSNGW( snjsnp, gwjsnp )
 *	MFunc( p1, p2, p3 )

 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <strings.h>

#include <sys/param.h>
#include <sys/types.h>

#include <netdb.h>
#include <netinet/in.h>

#include "jsdb.h"
#include "js.h"

int	jsfd;
FILE *	jsfp;


#ifndef  TRUE
#define  TRUE  1
#endif

#ifndef  FALSE
#define  FALSE 0
#endif

int	htinit = FALSE;
int	nahtinit = FALSE;

	void			AddGWIntf();
static	void			AddGWIntfP();
	void			AddIntf();
static	struct Node *		AddName();
	void			AddNet();
static	void			AddToLists();
static	struct list *		CreateListElement();
static	struct NameEnt *	CreateName();
static	struct Node *		CreateNode();
	struct Node *		FindNodeEnt();
static	struct Node *		FindOrAddNode();
	void			FixIntfCounts();
	u_long			GetNetMask();
	int			HashAddr();
	int			HashName();
	void			MatchSNGW();
	void			MFunc();

/*
 *  AddGWIntf
 *	Connect a new interface into a gateway
 */

void
AddGWIntf( gwjsnp, intfjsnp )
struct jsNode *	gwjsnp;
struct jsNode *	intfjsnp;
{
    struct Node * gwp;
    struct in_addr inaddr;
    struct Node * inp;
    struct Node * np;
    struct Node * op;

    gwp = FindOrAddNode( gwjsnp, NTYPE_INTF );   /* was NTYPE_GATEWAY */
    inp = FindOrAddNode( intfjsnp, NTYPE_INTF );
    AddGWIntfP( gwp, inp );
    return;
}


/*
 *  AddGWIntfP
 *	Connect a new interface into a gateway
 */

static void
AddGWIntfP( gwp, inp )
struct Node *	gwp;
struct Node *	inp;
{
    struct in_addr inaddr;
    struct Node * np;
    struct Node * op;
    struct Node * xp;

    if ( gwp->ntype != NTYPE_INTF || inp->ntype != NTYPE_INTF )
	return;

    if ( gwp == inp )
	return;

    if ( inp->gwp == NULL && gwp->gwp != NULL )
	{
	xp = gwp;
	gwp = inp;
	inp = xp;
	}

    if ( inp->gwp == NULL )
	{
	if ( gwp->nxtintfp == NULL )
	    {
	    gwp->nxtintfp = inp;
	    inp->gwp = gwp;
	    }
	else
	    {
	    for ( np=gwp->nxtintfp; np!=NULL; op=np, np=np->nxtintfp )
		if ( np->addr == inp->addr )
		    {
	            inp->gwp = gwp;
		    return;
		    }

	    if ( np == NULL )
		{
		op->nxtintfp = inp;
		inp->nxtintfp = NULL;
	        inp->gwp = gwp;
		}
	    }
	}
    else
	{
	if ( inp->gwp != gwp )
	    {
	    inaddr.s_addr = inp->addr;
	    fprintf( stderr, "AddGWIntf: interface %s on multiple gateways\n",
							inet_ntoa( inaddr ) );
	    }
	}
    return;
}


/*
 *  AddIntf
 *	Add (or find) interface structure
 */

void
AddIntf( jsNp )
struct jsNode * jsNp;
{
    struct Node * gwp;
    int     hc;
    struct in_addr inaddr;
    struct in_addr insm;
    struct jsNode  jsn;
    struct Node * netp;
    u_long  nm;
    struct Node * oxp;
    char    smstr[16];
    u_long  sm = 0;
    struct Node * snp;
    struct Node * xp;

    inaddr.s_addr = jsNp->addr;
    if ( jsNp->addr == INADDR_LOOPBACK )
	return;

    xp = FindOrAddNode( jsNp, NTYPE_INTF );
    if ( ( xp->name == NULL ) && ( jsNp->name != NULL ) )
	 {
	 gwp = AddName( jsNp->name, xp );
	 if ( gwp != NULL )
	     AddGWIntfP( gwp, xp );
	 }

    if ( *(int *)xp->eaddr == 0 && *(int *)jsNp->eaddr != 0 )
        bcopy( jsNp->eaddr, xp->eaddr, 6 );

    nm = GetNetMask( jsNp->addr );
    jsn.addr  = jsNp->addr & nm;
    jsn.sm    = 0;
    netp = FindOrAddNode( &jsn, NTYPE_NET );

    if ( jsNp->sm != 0 )
	{
	insm.s_addr = jsNp->sm;
	strcpy( smstr, inet_ntoa( insm ) );
	if ( ( jsNp->sm & nm ) == nm )
	    {
	    sm = jsNp->sm;
	    if ( netp->sm == 0 )
		netp->sm = sm;
	    else
	    if ( netp->sm != jsNp->sm )
		{
		fprintf( stderr, " Subnet mask mismatch: %s (%s)\n",
						inet_ntoa( inaddr ), smstr );
		sm = netp->sm;
		}
	    }
	else
	    {
	    fprintf( stderr, " Bogus subnet mask: %s (%s)\n",
						inet_ntoa( inaddr ), smstr );
	    sm = netp->sm;
	    }
	}
    else
	if ( netp->sm != 0 )
	    sm = netp->sm;
/*
 *	else
 *	    fprintf( stderr, " Missing subnet mask: %s\n", inet_ntoa(inaddr ) );
 */

    if ( sm != 0 )
	{
	jsn.addr = jsNp->addr & sm;
	jsn.sm   = sm;
	snp = FindOrAddNode( &jsn, NTYPE_SUBNET );

	if ( snp->sm == 0 )
	    snp->sm = sm;
	else
	    if ( snp->sm != sm )
		{
		inaddr.s_addr = snp->addr;
		fprintf( stderr, "Masks conflict for subnet %s\n",
						inet_ntoa( inaddr ) );
		}
	AddToLists( snp, xp );
	if ( snp->icount == 0 )
	    netp->scount++;
	}

    if ( xp->icount == 0 )		/* if newly added interface */
	{
	netp->icount++;
	if ( sm != 0 )
	    snp->icount++;
	else
	    xp->scount = 1;
	xp->icount = 1;
	}
    else
	{
	if ( sm != 0 && xp->scount != 0 )
	    {
	    snp->icount++;
	    xp->scount = 0;
	    }
	}

    return;
}


/*
 *  AddName
 *	Return NULL if name not already created
 *	Return pointer Node pointer to interface if name already existed
 *	Fill in pointer to name
 */

static struct Node *
AddName( namep, Nodep )
char *	namep;
struct Node * Nodep;
{
    int     i;
    int     hc;
    struct NameEnt *  oxp;
    struct NameEnt *  xp;

    hc = HashName( namep );
    xp = NAHT[hc];

    if ( xp == NULL )
	{
	xp = CreateName( namep );
	NAHT[hc] = xp;
	Nodep->name = xp->name;
	xp->intfp = Nodep;
	return( NULL );
	}

    while ( xp != NULL )
	{
	if ( strcasecmp( xp->name, namep ) == 0 )
	    {
	    Nodep->name = xp->name;
	    return( xp->intfp );
	    }

	oxp = xp;
	xp = xp->SameHash;
	}

    xp = CreateName( namep );
    oxp->SameHash = xp;
    Nodep->name = xp->name;
    xp->intfp = Nodep;
    return( NULL );
}


/*
 *  AddNet
 *	Add (or find) network (and subnet) structure
 */

void
AddNet( jsNp )
struct jsNode * jsNp;
{
    int     hc;
    u_long  nm;
    struct Node * netp;
    int     ntype;
    struct Node * oxp;
    struct jsNode   jsn;
    struct Node * xp;

    nm = GetNetMask( jsNp->addr );
    jsn.addr  = jsNp->addr & nm;
    jsn.sm    = 0;
    netp = FindOrAddNode( &jsn, NTYPE_NET );

    if ( ( jsNp->addr & ~ nm ) == 0 )
	return;					/* ntype = NTYPE_NET; */
    else
	ntype = NTYPE_SUBNET;

    hc = HashAddr( jsNp->addr );
    xp = SNHT[hc];
    if ( xp == NULL )
	{
	xp = CreateNode( jsNp, ntype );
	SNHT[hc] = xp;
	netp->scount++;
	return;
	}

    while ( xp != NULL )
	{
	if ( ( xp->addr == jsNp->addr ) && ( xp->ntype == ntype ) )
	    return;

	oxp = xp;
	xp = xp->SameHash;
	}

    xp = CreateNode( jsNp, ntype );
    oxp->SameHash = xp;
    netp->scount++;
    return;
}


/*
 *  AddToLists
 *	Add subnet and gateways to lists for gateways and subnets
 */

static void
AddToLists( snp, gwp )
struct Node * snp;
struct Node * gwp;
{
    struct list * np;
    struct list * xp;
    struct list * oxp;

    xp = snp->gwlp;
    if ( xp == NULL )
	{
	np = CreateListElement( gwp );
	snp->gwlp = np;
	}
    else
	{
	while ( xp != NULL )
	    {
	    if ( xp->node->addr == gwp->addr )
		return;

	    oxp = xp;
	    xp = xp->next;
	    }
	np = CreateListElement( gwp );
	oxp->next = np;
	}

    xp = gwp->snlp;
    if ( xp == NULL )
	{
	np = CreateListElement( snp );
	gwp->snlp = np;
	}
    else
	{
	while ( xp != NULL )
	    {
	    if ( xp->node->addr == snp->addr )
		return;

	    oxp = xp;
	    xp = xp->next;
	    }
	np = CreateListElement( snp );
	oxp->next = np;
	}
    return;
}


/*
 *  CreateListElement
 *	Create an element for either subnet or gateway lists
 */

static struct list *
CreateListElement( anp )
struct Node * anp;
{
    struct list * np;

    np = (struct list *)malloc( sizeof( struct list ) );
    if ( np == NULL )
	{
	perror( "jsdb: CreateListElement: malloc" );
	exit( 1 );
	}

    np->node = anp;
    np->next = NULL;
    return( np );
}


/*
 *  CreateName
 */

static struct NameEnt *
CreateName( np )
char *	np;
{
    struct NameEnt * nep;

    nep = (struct NameEnt *) malloc( sizeof( struct NameEnt ) );
    if ( nep == NULL )
	{
	perror( "jsdb: CreateName: NameEnt malloc" );
	exit( -1 );
	}

    nep->Next = NULL;
    nep->SameHash = NULL;
    if ( NaFirst == NULL )
	{
	NaFirst = nep;
	NaLast = nep;
	}
    else
	{
	NaLast->Next = nep;
	NaLast = nep;
	}
    nep->name = (char *)malloc( strlen( np ) + 1 );
    if ( nep->name == NULL )
	{
	perror( "jsdb: CreateName: name malloc" );
	exit( -1 );
	}

    strcpy( nep->name, np );
    NameCnt++;
    return( nep );
}


/*
 *  CreateNode
 *	Create a node
 */

static struct Node *
CreateNode( jsnp, ntype )
struct jsNode * jsnp;
int	ntype;
{
    struct Node * np;

    np = (struct Node *)calloc( 1, sizeof(struct Node) );
    if ( np == NULL )
	{
	perror( "jsdb: CreateNode: calloc" );
	exit( -1 );
	}

    np->addr  = jsnp->addr;
    np->sm    = jsnp->sm;
    np->ntype = ntype;

    if ( fp[ntype] == NULL )
	fp[ntype] = np;
    else
	lp[ntype]->Next = np;
    lp[ntype] = np;

    return( np );
}


/*
 *  FindNodeEnt
 *	Find node with matching address and node type
 */

struct Node *
FindNodeEnt( addr, ntype )
u_long	addr;
int	ntype;
{
    int     hc;
    struct Node * xp = NULL;

    hc = HashAddr( addr );
    switch ( ntype )
	{
	case NTYPE_SUBNET:
	case NTYPE_NET:
	    xp = (struct Node *)SNHT[hc];
	    break;

	case NTYPE_GATEWAY:
	    xp = (struct Node *)GWHT[hc];
	    break;

	case NTYPE_INTF:
	    xp = (struct Node *)INHT[hc];
	    break;

	default:
	    return( NULL );
	}

    while ( xp != NULL )
	{
	if ( ( xp->addr == addr ) && ( xp->ntype == ntype ) )
	    return( xp );

	xp = xp->SameHash;
	}

    return( NULL );
}


/*
 *  FindOrAddNode
 *	Find (or create) a node with specified address and node type
 */

static struct Node *
FindOrAddNode( jsnp, ntype )
struct jsNode * jsnp;
int	ntype;
{
    int     hc;
    u_long  nm;
    struct Node * oxp;
    struct Node * xp = NULL;

    hc = HashAddr( jsnp->addr );
    switch ( ntype )
	{
	case NTYPE_NET:
	case NTYPE_SUBNET:
	    xp = SNHT[hc];
	    break;

	case NTYPE_INTF:
	    xp = INHT[hc];
	    break;

	case NTYPE_GATEWAY:
	    xp = GWHT[hc];
	    break;

	default:
	    return( NULL );
	}

    if ( xp == NULL )
	{
	xp = CreateNode( jsnp, ntype );
	switch ( ntype )
	    {
	    case NTYPE_NET:
		SNHT[hc] = xp;
		NetCnt++;
		break;

	    case NTYPE_SUBNET:
		SNHT[hc] = xp;
		SNCnt++;
		break;

	    case NTYPE_INTF:
		INHT[hc] = xp;
		IFCnt++;
		break;

	    case NTYPE_GATEWAY:
		GWHT[hc] = xp;
		GWCnt++;
		break;
	    }
	return( xp );
	}

    while ( xp != NULL )
	{
	if ( ( xp->addr == jsnp->addr ) && ( xp->ntype == ntype ) )
	    return( xp );

	oxp = xp;
	xp = xp->SameHash;
	}

    xp = CreateNode( jsnp, ntype );
    oxp->SameHash = xp;
    return( xp );
}


/*
 *  FixIntfCounts
 */

void
FixIntfCounts()
{
    u_long  nm;
    struct Node * netp;
    struct Node * np;
    struct Node * snp;

    for ( np = fp[NTYPE_INTF]; np != NULL; np = np->Next )
	{
	np->icount = 0;
	if ( np->scount == 1 )
	    {
	    nm = GetNetMask( np->addr );
	    netp = FindNodeEnt( np->addr & nm, NTYPE_NET );
	    if ( netp != 0 && netp->sm != 0 )
		{
		snp = FindNodeEnt( np->addr & netp->sm, NTYPE_SUBNET );
		snp->icount++;
		np->scount = 0;
		}
	    }
	}
    return;
}


/*
 *  GetNetMask
 *	Return the network mask (not subnet mask) for the specified network
 *	This is derived from the given address
 */

u_long
GetNetMask( addr )
u_long	addr;
{
    if ( IN_CLASSA( addr ) )
	return( IN_CLASSA_NET );

    if ( IN_CLASSB( addr ) )
	return( IN_CLASSB_NET );

    if ( IN_CLASSC( addr ) )
	return( IN_CLASSC_NET );

    return( 0xfffffff0 );
}


/*
 *  HashAddr
 *	Routine to generate hash code from IP address
 */

HashAddr( addr )
u_long	addr;
{
    u_char  byte[4];
    int     hc;
    int     i;

    if ( ! htinit )
	for ( i=0; i<HASH_TABLE_SIZE; i++ )
	    GWHT[i] = INHT[i] = SNHT[i] = 0;

    htinit = 1;
    bcopy( &addr, byte, 4 );
    hc = byte[0] + byte[1] + byte[2] + byte[3];
    hc = hc % HASH_TABLE_SIZE;
    return( hc );
}


/*
 *  HashName
 *	Routine to generate hash code from name
 */

HashName( np )
char *	np;
{
    int     hc;
    int     i;

    if ( ! nahtinit )
	for ( i=0; i<HASH_TABLE_SIZE; i++ )
	    NAHT[i] = 0;

    nahtinit = TRUE;
    hc = 0;
    for ( i=0; i<6; i++ )
	if ( np[i] != '\0' )
	    hc = ( hc << 4 ) + ( ( np[i] & 0x7f ) - '0' );
	else
	    break;

    hc = hc % HASH_TABLE_SIZE;
    return( hc );
}


/*
 *  MatchSNGW
 *	Connect Subnet with Gateway
 */

void
MatchSNGW( snjsnp, gwjsnp )
u_long	snjsnp;
u_long	gwjsnp;
{
    struct Node * gwp;
    struct Node * snp;

    snp = FindOrAddNode( snjsnp, NTYPE_SUBNET );
    gwp = FindOrAddNode( gwjsnp, NTYPE_INTF );   /* was NTYPE_GATEWAY */
    AddToLists( snp, gwp );
    return;
}


/*
 *  MFunc
 *	Routine to deal with SN/GW information from JS
 *      Called with three paramaters:
 *
 *  p1    p2				 p3
 * M_SNIP  <subnet_IP_addr>		<gateway_interface_IP_addr>
 * M_GWIP  <gateway_interface_IP_addr>	<another_gateway_interface_IP_addr>
 * M_GWSN  <gateway_interface_IP_addr>	<subnet_IP_addr>
 *    
 */

void
MFunc( p1, p2, p3 )
int	p1;
struct jsNode *	p2;
struct jsNode *	p3;
{
    switch ( p1 )
	{
	case M_SNIP:
	    MatchSNGW( p2, p3 );
	    break;

	case M_GWIP:
	    AddGWIntf( p2, p3 );
	    break;

	case M_GWSN:
	    MatchSNGW( p3, p2 );
	    break;
	}
    return;
}

