/*
 *  jslib.c	DCM Wood  93/03/05
 */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <strings.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <netdb.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#define  JSLIB
#include "js.h"

#include "pin.h"
char	pin[20]= PIN;

#define  MAXLINE  150

extern	int	debug;

static struct sockaddr_in * getaddr();

static struct jsNode *	ReadIntf();
static struct jsNode *	ReadNetwork();
static struct jsNode *	ReadGateway();

static int		Authenticate();

static u_long		ReadProtAddr();
static u_char *		ReadEtherAddr();
static char *		ReadName();


/*
 *  GetNextRecord
 *	Read next record from the specified file
 *    jsfp is the FILE pointer for the connection to the Journal Server
 *    mfunc is the address of the function to call to match subnets,
 *      gatetways, and interfaces.  It is called with three paramaters:
 *
 *  p1    p2				 p3
 * M_SNIP  <subnet_IP_addr>		<gateway_interface_IP_addr>
 * M_GWIP  <gateway_interface_IP_addr>	<another_gateway_interface_IP_addr>
 * M_GWSN  <gateway_interface_IP_addr>	<subnet_IP_addr>
 *    
 */

struct jsNode *
GetNextRecord( jsfp, mfunc )
FILE *	jsfp;
int	(*mfunc)();
{
    int     cc = 0;
    char *  cp;
    char    line[MAXLINE];
    struct jsNode * rp;
    char    scr[MAXLINE];

    while ( cc == 0 )
        cc = GetLine( line, MAXLINE, jsfp );

    if ( cc < 0 )
	return( NULL );

    strcpy( scr, line );
    Authenticate( scr );
    cp = strtok( scr, "\t\n " );
    if ( strcasecmp( cp, "INTF" ) == 0 )
	{
	rp = ReadIntf( jsfp );
	return( rp );
	}
    else
    if ( strcasecmp( cp, "NETWORK" ) == 0 )
	{
	rp = ReadNetwork( jsfp, mfunc );
	return( rp );
	}
    else
    if ( strcasecmp( cp, "GATEWAY" ) == 0 )
	{
	rp = ReadGateway( jsfp, mfunc );
	return( rp );
	}
    else
    	return( NULL );
}


/*
 *	Get a line from the specified file
 *	Return < 0 if error or EOF
 *	       = 0 if blank line
 *	       > 0 length of line.
 */

static int
GetLine( line, max, fp )
char *	line;
int	max;
FILE *	fp;
{
    char *  cp;
    int     i;
    int     nbc;

    cp = fgets( line, max, fp );
    if ( cp == NULL )
	return( -1 );					/* error or EOF */
    
    nbc = 0;
    for ( i=0; i<max; i++ )
	{
	if ( ( line[i] == '\n' ) || ( line[i] == '\0' ) )
	    break;

	if ( line[i] != ' ' )
	    {
	    nbc++;
	    break;
	    }
	}

    if ( nbc > 0 )
	return( strlen( line ) );			/* non-blank line */

    return( 0 );					/* blank line */
}


static struct jsNode *
ReadIntf( jsfp )
FILE *	jsfp;
{
    int     cc;
    char *  cp;
    int     i;
    static struct jsNode intf;
    char    line[MAXLINE];
    char    scr[MAXLINE];

    intf.addr = 0;
    intf.rtype = INTF;
    intf.sm = 0;
    intf.name = NULL;
    for ( i=0; i<6; i++ )
	intf.eaddr[i] = 0;

    while ( 1 )
	{
	cc = GetLine( line, MAXLINE, jsfp );
	if ( cc <= 0 )
	    return( &intf );

	strcpy( scr, line );
	cp = strtok( scr, ":\t\n " );
	if ( strlen( cp ) == 0 )
	    return( &intf );

	if ( strcasecmp( cp, "PROT-ADDR" ) == 0 )
	    {
	    intf.addr = ReadProtAddr();
	    }
	else
	if ( strcasecmp( cp, "ETHER-ADDR" ) == 0 )
	    {
	    cp = (char *)ReadEtherAddr();
	    bcopy( cp, intf.eaddr, 6 );
	    }
	else
	if ( strcasecmp( cp, "IN-NAME" ) == 0 )
	    {
	    intf.name = ReadName();
	    }
	else
	if ( strcasecmp( cp, "SUBNET-MASK" ) == 0 )
	    {
	    intf.sm = ReadProtAddr();
	    }
	else
	    continue;
	}
}


static struct jsNode *
ReadNetwork( jsfp, mfunc )
FILE *	jsfp;
int	(*mfunc)();
{
    int     cc;
    char *  cp;
    static struct jsNode gw;
    static struct jsNode intf;
    char    line[MAXLINE];
    char    scr[MAXLINE];

    gw.addr = 0;
    gw.rtype = INTF;
    gw.sm = 0;
    gw.name = NULL;

    intf.addr = 0;
    intf.rtype = NETWORK;
    intf.sm = 0;
    intf.name = NULL;

    while ( 1 )
	{
	cc = GetLine( line, MAXLINE, jsfp );
	if ( cc <= 0 )
	    return( &intf );

	strcpy( scr, line );
	cp = (char *)strtok( scr, ":\t\n " );
	if ( strlen( cp ) == 0 )
	    return( &intf );

	if ( strcasecmp( cp, "NET_NUMB" ) == 0 )
	    {
	    intf.addr = ReadProtAddr();
	    }
	else
	if ( strcasecmp( cp, "SUBNET_NUM" ) == 0 )
	    {
	    fprintf( stderr, "jslib: read SUBNET_NUM field\n" );
	    }
	else
	if ( strcasecmp( cp, "GATEWAY" ) == 0 )
	    {
	    gw.addr = ReadProtAddr();
	    if ( intf.addr != 0 )
		mfunc( M_SNIP, &intf, &gw );
	    }
	else
	    continue;
	}
}


static struct jsNode *
ReadGateway( jsfp, mfunc )
FILE *	jsfp;
int	(*mfunc)();
{
    int     cc;
    char *  cp;
    static struct jsNode gw;
    struct jsNode intf;
    char    line[MAXLINE];
    char    scr[MAXLINE];
    struct jsNode sn;

    gw.addr = 0;
    gw.rtype = GATEWAY;
    gw.sm = 0;
    gw.name = NULL;

    intf.addr = 0;
    intf.rtype = INTF;
    intf.sm = 0;
    intf.name = NULL;

    sn.addr = 0;
    sn.rtype = 0;
    sn.sm = 0;
    sn.name = NULL;

    while ( 1 )
	{
	cc = GetLine( line, MAXLINE, jsfp );
	if ( cc <= 0 )
	    return( &gw );

	strcpy( scr, line );
	cp = (char *)strtok( scr, ":\t\n " );
	if ( strlen( cp ) == 0 )
	    return( &gw );

	if ( strcasecmp( cp, "INTF" ) == 0 )
	    {
	    intf.addr = ReadProtAddr();
	    if ( gw.addr == 0 )
		gw.addr = intf.addr;
	    else
		mfunc( M_GWIP, &gw, &intf );
	    }
	else
	if ( strcasecmp( cp, "SUBNETS" ) == 0 )
	    {
	    sn.addr = ReadProtAddr();
	    if ( gw.addr != 0 )
		mfunc( M_GWSN, &gw, &sn );
	    }
	else
	    continue;
	}
}


static Authenticate( line )
char *	line;
{
    int     cc;
    char    type[10];
    char    scr[20];

    cc = sscanf( line, "%s %s", type, scr );
    switch ( cc )
	{
	case 2:
	    cc = strcmp( pin, scr );
	    return( cc );

	case 1:
	    return( 0 );

	default:
	    return( -1 );
	}
}


static u_long
ReadProtAddr()
{
    char *  cp;
    u_long  addr;

    cp = (char *)strtok( NULL, ":\t\n " );
    if ( strlen( cp ) == 0 )
	return( NULL );
    if ( strcasecmp( cp, "AF_INET" ) == 0 )
	{
	cp = (char *)strtok( NULL, ":\t\n " );
	if ( strlen( cp ) == 0 )
	    return( NULL );

	addr = inet_addr( cp );
	if ( addr == -1 )
	    return( NULL );

	return( addr );
	}
    else
	return( NULL );
}


static u_char *
ReadEtherAddr()
{
    int     cc;
    char *  cp;
    u_char  eaddr[6];
    int     es[6];
    int     i;

    cp = (char *)strtok( NULL, "\t\n " );
    if ( strlen( cp ) == 0 )
	return( NULL );

    cc = sscanf( cp, "%x:%x:%x:%x:%x:%x", &es[0], &es[1], &es[2],
					  &es[3], &es[4], &es[5] );
    if ( cc != 6 )
	return( NULL );

    for ( i=0; i<6; i++ )
	if ( es[i] < 256 )
	    eaddr[i] = es[i];
	else
	    return( NULL );

    return( eaddr );
}


static char *
ReadName()
{
    char *  cp;
    static  char    name[256];

    cp = (char *)strtok( NULL, ":\t\n " );
    if ( strlen( cp ) == 0 )
	return( NULL );

    strcpy( name, cp );
    return( name );
}


#include <errno.h>

/*
 *  JSconnect
 */

JSconnect( data_host, port )
char *	data_host;
int     port;
{
    int     cc;
    int     jsfd = -1;
    char *  jshenv;
    char    hostname[MAXHOSTNAMELEN];
    struct sockaddr_in * server_addr = NULL;
    extern int     errno, sys_nerr;    /* External Unix variables */
    extern char    *sys_errlist[];     /* External Unix variable */

    if ( port == 0 )
	port = DATA_PORT;

    /* get the hostname from the environment variable JS_HOST */

    jshenv = getenv( "JS_HOST" );

    if ( data_host )
	server_addr = getaddr( data_host, port );
    else
	{
	if ( jshenv != NULL )
	    {
	    server_addr = getaddr( jshenv, port );
	    if ( server_addr == NULL )
		{
		printf( "Unable to get correct server\n" );
		exit( 1 );
		}
	    }
	else
	    {
	    cc = gethostname( hostname, MAXHOSTNAMELEN );
	    if ( cc != 0 )
		{
		printf( "Could not get local hostname\n" );
		exit( 1 );
		}

	    /* found the Host name, so stick it away and
		find the IP address for the host */

	    server_addr = getaddr( hostname, port );
	    }

	if ( server_addr == NULL )
	    {
	    printf( "Unable to get journal server address\n" );
	    exit( 1 );
	    }
	}

    jsfd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( jsfd < 0 )
	{
	perror( "socket" );
	exit( 1 );
	}

    cc = connect( jsfd, (struct sockaddr *)server_addr,
						    sizeof(struct sockaddr) );
    if ( cc < 0 )
	{
        fprintf(stderr, "Unable to connect to Journal Server at %s: %s\n",
		data_host, errno < sys_nerr ? sys_errlist[errno] : "");
	exit( 1 );
	}

    return( jsfd );
}


/*
 * getaddr retrieves the ip address of the host where the data server
 * exists. The function is given the name host to look up the address.
 * It returns the sockaddr structure for the host.
 */

static struct sockaddr_in *
getaddr( host, port )
char *	host;			/* name of host to get address for */
int     port;
{
    static struct sockaddr_in serv_addr;	/* inet address field */
    struct hostent *hp;				/* name entry for host */

    /* check to see if the host name is an IP address  */

    serv_addr.sin_addr.s_addr = inet_addr( host );
    if ( serv_addr.sin_addr.s_addr == -1 )
	{ /* The address must be a name, so get IP address for the name */
	hp = gethostbyname( host );
	if ( hp == NULL )
	    {
	    printf( "Unable to get host information\n" );
	    exit( 1 );
	    }

	/* put address into the sockaddr_in structure */

	serv_addr.sin_addr.s_addr = *(int *)hp->h_addr_list[0];
	}

    /* Fill in the port and address family field */

    serv_addr.sin_port = htons( port );
    serv_addr.sin_family = AF_INET;

    return( &serv_addr );
}


