/*
 *  readjs.c	DCM Wood  93/03/09
 */

/*
 *  ReadJS
 *	Connect to Journal Server and retrieve specified information
 */

#include <stdio.h>
#include <stdlib.h>

#include <sys/param.h>
#include <sys/types.h>

#include <netdb.h>
#include <netinet/in.h>

#include "jsdb.h"
#include "js.h"

extern	int	debug;
extern	void	MFunc();

static	int	dbgflg = 0;
static	int	jsfd;
static	FILE *	jsfp;


/*
 *  ReadJS
 *	Read information from Journal Server
 *
 *  Called with req = one of the following:
 *	GATEWAY		Retrieve gateway info
 *	INTF		Retrieve interface info
 *	NETWORK		Retrieve network (and subnet) info
 */

ReadJS( req )
int	req;
{
    short   buf;
    int     cc;
    int     end;
    struct jsNode * jsNp;

    switch ( req )
	{
	case INTF:
	    buf = GET_IPADDR;	/* Retrieve interface info (sorted by IP addr)*/
	    break;

	case GATEWAY:
	    buf = GET_GW;	/* Retrieve gateway info */
	    break;

	case NETWORK:
	    buf = GET_NETWORK;	/* Retrieve network (and subnet) info */
	    break;

	default:
	    return;
	}

    if ( debug )
	printf( " Reading information from the Journal Server\n" );

    jsfd = JSconnect( jshostp, QUERY_PORT );

    if ( dbgflg == 0 )
	{
	dbgflg++;
	fprintf( stderr,
" The following represent potential network problems discovered by Fremont:\n");
	}

    cc = write( jsfd, &buf, sizeof(short) );
    if ( cc != sizeof(short) )
	{
	perror( "ReadJS" );
	close( jsfd );
	exit( 1 );
	}

    jsfp = fdopen( jsfd, "r" );

    end = 0;
    while ( ! end )
	{
	jsNp = GetNextRecord( jsfp, MFunc );
	if ( jsNp == NULL )
	    break;

	switch ( jsNp->rtype )
	    {
	    case NETWORK:
		AddNet( jsNp );
		NRCnt++;
		break;

	    case GATEWAY:
		AddIntf( jsNp );
		GRCnt++;
		break;

	    case INTF:
		AddIntf( jsNp );
		IRCnt++;
		break;
	    }
	}
    fclose( jsfp );
    return;
}

