/*
 *  Apportion.c
 */

/*
 *  This procedure cleans up the posistioning of small sibling subtrees,
 *    thus fixing the 'left-to-right gluing' problem evident in earlier
 *    algorithms.
 *  When moving a new subtree further and further to the right, gaps may
 *    open up among smaller subtrees that were previously sandwiched
 *    between larger subtrees.
 *  Thus, when moving the new, larger subtree to the right, the distance
 *    it is moved is also approtioned to the smaller, interior subtrees,
 *    creating a pleasing aesthetic placement.
 */

#include "PositionTree.h"

Apportion( Node, Level )
struct node * Node;
int	Level;
{
    struct node * AncestorLeftmost;
    struct node * AncestorNeighbor;
    struct node * GetLeftMost();
    struct node * Leftmost;
    struct node * Neighbor;
    struct node * TempPtr;
    int    CompareDepth;
    int    DepthToStop;
    int	   i;
    int    LeftModsum;
    int    LeftSiblings;
    int    MoveDistance;
    int    Portion;
    int    RightModsum;

    Leftmost = FIRSTCHILD( Node );
    Neighbor = LEFTNEIGHBOR( Leftmost );
    CompareDepth = 1;
    DepthToStop = MaxDepth - Level;

    while ( Leftmost != NULL &&
	    ( Neighbor = LEFTNEIGHBOR( Leftmost ) ) != NULL &&
	    CompareDepth <= DepthToStop )
	{
	/*  Compute the location of Leftmost and where it should	*/
	/*   be with respect to Neighbor.				*/
	LeftModsum = 0;
	RightModsum = 0;
	AncestorLeftmost = Leftmost;
	AncestorNeighbor = Neighbor;
	for ( i=0; i<CompareDepth; i=i+1 )
	    {
	    AncestorLeftmost = PARENT( AncestorLeftmost );
	    AncestorNeighbor = PARENT( AncestorNeighbor );
	    RightModsum = RightModsum + 
			  MODIFIER( AncestorLeftmost );
	    LeftModsum = LeftModsum +
			 MODIFIER( AncestorNeighbor );
	    }
	/*  Find the MoveDistance, and apply it to Node's subtree.	*/
	/*  Add appropriate portions to smaller interior subtrees.	*/
	MoveDistance = ( PRELIM( Neighbor ) +
			 LeftModsum +
			 SubtreeSeparation +
			 MeanNodeSize( Leftmost, Neighbor ) ) -
			 ( PRELIM( Leftmost ) + RightModsum );
	if ( MoveDistance > 0 )
	    {
	    /*  Count interior sibling subtrees in LeftSiblings		*/
	    TempPtr = Node;
	    LeftSiblings = 0;
	    while ( ( TempPtr != NULL ) &&
		    ( TempPtr != AncestorNeighbor ) )
		{
		LeftSiblings = LeftSiblings + 1;
		TempPtr = LEFTSIBLING( TempPtr );
		}
	    if ( ( TempPtr != NULL ) && ( LeftSiblings > 0 ) )
		/*  Apply portions of appropriate leftsibling		*/
		/*    subtrees.						*/
		{
		Portion = MoveDistance / LeftSiblings;
		TempPtr = Node;
		while ( TempPtr != AncestorNeighbor )
		    {
		    PRELIM( TempPtr ) =
			    PRELIM( TempPtr ) + MoveDistance;
		    MODIFIER( TempPtr ) =
			    MODIFIER( TempPtr ) + MoveDistance;
		    MoveDistance = MoveDistance -
			    Portion;
		    TempPtr = LEFTSIBLING( TempPtr );
		    }
		}
	    else
		/*  Don't need to move anything -- it needs to		*/
		/*    be done by an ancestor because			*/
		/*    AncestorNeighbor and AncestorLeftmost are		*/
		/*    not sibling of each other.			*/
		return;
	    }  /* if MoveDistance > 0 */
	/*  Determine the leftmost descendant of Node at the next	*/
	/*    lower level to compare its positioning against that of	*/
	/*    its Neighbor.						*/

	CompareDepth = CompareDepth + 1;
	if ( IsLeaf( Leftmost ) )
	    Leftmost = GetLeftMost( Node, 0, CompareDepth );
	else
	    Leftmost = FIRSTCHILD( Leftmost );
	}  /* while Leftmost etc */
}
