/*
 *  FirstWalk.c
 */

/*
 *  In this first post-order walk, every node of the tree is assigned
 *    a preliminary x-coordinate (held in the field Prelim(Node).
 *  In addition, internal nodes are given modifiers, which will be
 *    used to move their offspring to the right (held in the field
 *    Modifier(Node).
 */

#include "PositionTree.h"

FirstWalk( Node, Level )
struct node * Node;
int	      Level;
{
    struct node * GetPrevNodeAtLevel();
    struct node * Leftmost;
    struct node * Rightmost;
    int    Midpoint;

    /*  Set the pointer to the previous node at this level.		*/
    LEFTNEIGHBOR( Node ) = GetPrevNodeAtLevel( Level );
    SetPrevNodeAtLevel( Level, Node );  /* This is now the previous.	*/
    MODIFIER( Node ) = 0;	/* Set the default modifier value	*/
    if ( IsLeaf( Node ) || ( Level == MaxDepth ) )
	{
	if ( HasLeftSibling( Node ) )
	    {
	    /* Determine the preliminary x-coordinate based on:		*/
	    /*  the preliminary x-coordinate of the left sibling,	*/
	    /*  the separation between sibling nodes, and		*/
	    /*  the mean size of left sibline and current node.		*/
	    PRELIM( Node ) = PRELIM( LEFTSIBLING( Node ) ) +
			SiblingSeparation +
			MeanNodeSize( LEFTSIBLING( Node ), Node );
	    }
	else
	    /* No sibling on the left to worry about			*/
	    PRELIM( Node ) = 0;
	}
    else
	/* This Node is not a leaf, so call this procedure		*/
	/*  recursively for each of its offspring.			*/
	{
	Leftmost = Rightmost = FIRSTCHILD( Node );
	FirstWalk( Leftmost, Level + 1 );
	while ( HasRightSibling( Rightmost ) )
	    {
	    Rightmost = RIGHTSIBLING( Rightmost );
	    FirstWalk( Rightmost, Level + 1 );
	    }
	Midpoint = ( PRELIM( Leftmost ) + PRELIM( Rightmost ) ) / 2;
	if ( HasLeftSibling( Node ) )
	    {
	    PRELIM( Node ) = PRELIM( LEFTSIBLING( Node ) ) +
			SiblingSeparation +
			MeanNodeSize( LEFTSIBLING( Node ), Node );
	    MODIFIER( Node ) = PRELIM( Node ) - Midpoint;
	    Apportion( Node, Level );
	    }
	else
	    PRELIM( Node ) = Midpoint;
	}
}
