/*
 *  GetLeftMost.c
 */

/*
 *  This function return the leftmost descendant of a node at a given Depth.
 *  This is implemented using a post-order walk of the subtree under Node,
 *    down to the level of Depth.
 *  Level here is not the absolute tree level used in the two main tree
 *    walks;  it refers to the level below the node whose leftmost
 *    descendant is being found.
 */

#include "PositionTree.h"

struct node *
GetLeftMost( Node, Level, Depth )
struct node * Node;
int	Level;
int	Depth;
{
    struct node * Leftmost;
    struct node * Rightmost;

    if ( Level >= Depth )
	return( Node );

    else if ( IsLeaf( Node ) )
	return( NULL );

    else
	{
	Rightmost = FIRSTCHILD( Node );
	Leftmost = GetLeftMost( Rightmost, Level +1, Depth );
	/*  Do a postorder walk of the subtree below Node.		*/
	while ( ( Leftmost == NULL ) &&
		  HasRightSibling( Rightmost ) )
	    {
	    Rightmost = RIGHTSIBLING( Rightmost );
	    Leftmost = GetLeftMost( Rightmost, Level +1, Depth );
	    }
	return( Leftmost );
	}
}
