/*
 *  PositionTree.c
 */

/*  Determine the co-ordinates for each node in a tree.
 *  A pointer to the apex node is passed at input.
 *  This assumes that the x and YU co-ordinates of the apex node are set
 *    as desired, since the tree underneath will be positioned with respect
 *    to those co-ordinates.
 *  Returns TRUE if no errors, otherwise returns FALSE
 */

#include <stdio.h>

#include "PositionTree.h"

PositionTree( Node )
struct node * Node;
{
    if ( Node != NULL )
	{
	/* Initialize the list of previous nodes at each level */
	InitPrevNodeList();

	/* Do the preliminary positioning with a postorder walk */
	FirstWalk( Node, 0 );

	/* Determine how to adjust all the nodes with respect to */
	/*  the location of the root  */
	switch ( Orientation )
	    {
	    case O_NORTH:
		xTopAdjustment = XCOORD( Node ) - PRELIM( Node );
		yTopAdjustment = YCOORD( Node );
		break;

	    case O_SOUTH:
		xTopAdjustment = XCOORD( Node ) + PRELIM( Node );
		yTopAdjustment = YCOORD( Node );
		break;

	    case O_EAST:
		xTopAdjustment = XCOORD( Node );
		yTopAdjustment = YCOORD( Node ) + PRELIM( Node );
		break;

	    case O_WEST:
		xTopAdjustment = XCOORD( Node );
		yTopAdjustment = YCOORD( Node ) - PRELIM( Node );
		break;

	    default:
		fprintf( stderr, "Orientation = %d, not N, E, S, or W\n",
							Orientation );
		exit( 1 );
	    }

	/* Do the final positioning with a preorder walk */
	return( SecondWalk ( Node, 0, 0 ) );
	}
    else
	/* Trivial: return TRUE if a null pointer was passed  */
	return ( TRUE );
}
