/*
 *  SecondWalk.c
 */

/*
 *  During a second preorder walk, each node is given a final x-coordinate
 *    by summing its preliminary x-coordinate and the modifiers of all the
 *    node's ancestors.
 *  The y-coordinate depends on the height of the tree.
 *  If the actual position of an interior node is right of its preliminary
 *    place, the subtree rooted at the node must be moded right to centre
 *    the sons around the father.
 *  Rather than immediately readjust all the nodes in the subtree, each
 *    node remembers the distance to the provisional place in a modifier
 *    field (Modifier(Node)).
 *  In this second pass down the tree, modifiers are accumulated and
 *    applied to every node.
 *  Returns TRUE if no errors, otherwise returns FALSE.
 */

#include "PositionTree.h"

SecondWalk( Node, Level, Modsum )
struct node * Node;
int  Level;
int  Modsum;
{
    int    Result=TRUE;
    int    xTemp;
    int    yTemp;

    if ( Level <= MaxDepth )
	{
	switch ( Orientation )
	    {
	    case O_NORTH:
		xTemp = xTopAdjustment + PRELIM( Node ) + Modsum;
		yTemp = yTopAdjustment + ( Level * LevelSeparation );
		break;

	    case O_SOUTH:
		xTemp = xTopAdjustment - PRELIM( Node ) - Modsum;
		yTemp = yTopAdjustment - ( Level * LevelSeparation );
		break;

	    case O_EAST:
		xTemp = xTopAdjustment - ( Level * LevelSeparation );
		yTemp = yTopAdjustment - PRELIM( Node ) - Modsum;
		break;

	    case O_WEST:
		xTemp = xTopAdjustment + ( Level * LevelSeparation );
		yTemp = yTopAdjustment + PRELIM( Node ) + Modsum;
		break;
	    }
	/* Check to see that xTemp and xTemp are of the proper		*/
	/*  size for your application.					*/
	if ( CheckExtentsRange( xTemp, yTemp ) )
	    {
	    XCOORD( Node ) = xTemp;
	    YCOORD( Node ) = yTemp;
	    if ( HasChild( Node ) )
		/* Apply the Modifier value for this node to		*/
		/*  all its offspring.					*/
		Result = SecondWalk( FIRSTCHILD( Node ),
				Level + 1,
				Modsum + MODIFIER( Node ) );
	    if ( ( Result == TRUE ) &&
		 HasRightSibling( Node ) )
		Result = SecondWalk( RIGHTSIBLING( Node ),
				Level,
				Modsum );
	    }
	else
	    /* Continuing would put the tree outside of the		*/
	    /*  drawable extents reang.					*/
	    Result = FALSE;
	}
    else
	/* We are at a level deeper that what we want to draw		*/
	Result = TRUE;

    return( Result );
}
