/*
 * display.c
 */

#include <stdio.h>
#include <signal.h>	
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "xjsv.h"

#define STR_LEN		(256)
#define MAX_TITLE	100

Display	*Dp;
Window	MainWin;
Visual	*MVis;
GC	Text6x10GC;
Font	Fnt6x10;
int	screen;

int	ForGndPix;
int	BackGndPix;

static	char	HostName[STR_LEN];
static	char	mrt[MAX_TITLE];		/* Most recent title */

static	void	DisplayHost();

static	int	AfterXlib();
static	int	mrtl = 0;		/* Most recent title - length */
static	int     tx = 10;		/* Title X-coordinate */
static	int     ty = 35;		/* Title Y-coordinate */


static int
Hang( dep, xep )
Display * dep;
XErrorEvent * xep;
{
    int     hang = TRUE;
    char    msg[80];

    XGetErrorText( dep, xep->error_code, msg, 80 );
    printf( "X-Error code %s\n", msg );

    printf( "X Error of failed request:  %s\n",
					msg );
				/*  BadWindow (invalid Window parameter)    */
    printf( "  Major opcode of failed request:  %d\n",
					xep->request_code );
				/*  8 (X_MapWindow)			    */
    printf( "  Resource id in failed request:  %x (%s)\n",
					xep->resourceid,
					XrmNameToString( xep->resourceid ) );
				/*  0x0					    */
    printf( "  Serial number of failed request:  %d\n",
					xep->serial );
				/*  2921				    */
    printf( "  Current serial number in output stream:  %d\n",
						NextRequest( Dp ) );
				/*  3330				    */
    printf( "X error caught - hanging here\n" );

    while ( hang ) ;
    return;
}


void
InitWin( hostname )
char *	hostname;
{
    char    buf[256];
    XEvent  ev;

    if ( debug )
	printf( " Initializing window\n" );

    if( ( Dp = XOpenDisplay( NULL ) ) == NULL )
	{
	fprintf( stderr, "\n    Could not open display.\n" );
	exit( 1 );
	}

#ifdef XDEBUG
    /* XSynchronize( Dp, True ); */
    XSetAfterFunction( Dp, AfterXlib );
#endif XDEBUG

    screen = DefaultScreen( Dp );
    MVis = DefaultVisual( Dp, screen );

    ForGndPix = WhitePixel( Dp, screen );
    BackGndPix = BlackPixel( Dp, screen );

    MainWin = XCreateSimpleWindow( Dp, RootWindow( Dp, screen ), 0, 0,
				    MWIN_W, MWIN_H, 1, ForGndPix, BackGndPix );

    XSelectInput( Dp, MainWin,
			ExposureMask | StructureNotifyMask | ButtonPressMask );

    sprintf( buf, "xjsv: %s", hostname );

    XChangeProperty( Dp, MainWin, XA_WM_NAME, XA_STRING, 8,
			PropModeReplace, (unsigned char *)buf, strlen( buf ) );

    XMapWindow( Dp, MainWin );

    Text6x10GC = XCreateGC( Dp, MainWin, 0, NULL );
    Fnt6x10 = XLoadFont( Dp, "6x10" );
    XSetState( Dp, Text6x10GC, ForGndPix, BackGndPix, GXset, AllPlanes );
    XSetFont( Dp, Text6x10GC, Fnt6x10 );

    /* XSetErrorHandler( Hang ); */

    XFlush( Dp );

    while( 1 )
	{
	XNextEvent( Dp, &ev );

	if( ev.type == Expose )
	    break;
	}
}


void
FlushDpy()
{
    XFlush( Dp );
}


void
SyncDpy()
{
    XSync( Dp, False );
}


void
ResizeWindow( w, h )
int w, h;
{
    XResizeWindow( Dp, MainWin, w, h );
}


void
SetHost( host )
char *host;
{
    strncpy( HostName, host, STR_LEN );
    /* DisplayHost(); */
}


/*
 *  DrawTitle
 */

DrawTitle( cp )
char *	cp;
{
    int     len;

    if ( mrtl != 0 )
	XClearArea( Dp, MainWin, tx, ty, mrtl * 6, 10, False );

    mrtl = strlen( cp );
    if ( mrtl >= MAX_TITLE )
	{
	mrtl = MAX_TITLE - 1;
	cp[MAX_TITLE - 1] = '\0';
	}
    strcpy( mrt, cp );
    XDrawImageString( Dp, MainWin, Text6x10GC, tx, ty, mrt, mrtl );
}


/*
 *  RedrawTitle
 */

RedrawTitle()
{
    XDrawImageString( Dp, MainWin, Text6x10GC, tx, ty, mrt, mrtl );
}


/*
static void
DisplayHost()
{
    static int len = 0;
    char buf[256];
    int i;
    int l;

    sprintf( buf, "Journal Server: %s", HostName );
    l = strlen( buf );

    if( l < len )
	{
	for( i = l; i < len; i++ )
	    buf[i] = ' ';

	buf[i] = '\0';
	}
    else
	len = l;

    XDrawImageString( Dp, MainWin, Text6x10GC, 60, 17, buf, len );

    len = l;
}
 */


void
DoEvents( block )
int block;
{
    XEvent ev;
    void (*os)();
    int exp = FALSE;
    int     ql;

    ql = XPending( Dp );

/*  while( ( XPending( Dp ) > 0 ) || ( block == TRUE ) ) */

    while( ( ( ql = XPending( Dp ) ) > 0 ) || ( block == TRUE ) )
	{
	XNextEvent( Dp, &ev );

	switch( ev.type )
	    {
	case Expose:
	    exp = TRUE;
	    break;

	case ButtonPress:
	    os = signal( SIGALRM, SIG_IGN );
	    DoButtonPress( ev.xbutton.window );
	    signal( SIGALRM, os );
	    break;

/*
 *	case KeyPress:
 *	    os = signal( SIGALRM, SIG_IGN );
 *	    DoKeyInput( &(ev.xkey) );
 *	    signal( SIGALRM, os );
 *	    break;
 */
	    }

	block = FALSE;
	}

    if ( exp == TRUE )
	{
	RedrawAllBtns();
	/* DisplayHost(); */
	RedrawTitle();
	FlushDpy();
	}
}


static int
AfterXlib( dp )
Display *dp;
{
    long s;

    s = XNextRequest( Dp );
    fprintf( stderr, "%ld\n", s );
}

