/*
 * intf.c
 */

#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <strings.h>

#include "xjsv.h"
#include "jsdb.h"

extern		QuitStat();
extern	void	SetIndv();

static void	ExitIntf();
static void	HideIntfBtns();
static void	MakeIntfBtns();
static void	ShowIntfBtns();
static void	WipeIntfBtns();

static struct Btn * ButQuit;
static struct Btn * ButExit;

static int WindowWidth = 400;
static int WindowHeight = 100;


void MakeIntfDisplay()
{
    if ( debug )
	printf( " MakeIntfDisplay\n" );

    WindowHeight = 100;
    MakeIntfBtns();
    ResizeWindow( WindowWidth, WindowHeight );
}


void WipeIntfDisplay()
{
    if ( debug )
	printf( " WipeIntfDisplay\n" );

    WipeIntfBtns();
}


void ShowIntfWindow()
{
    char    cbuf[48];
    struct in_addr inaddr;

    if ( debug )
	printf( " ShowIntfWindow\n" );

    ResizeWindow( WindowWidth, WindowHeight );
    ShowIntfBtns();
    inaddr.s_addr = NewSubnet->addr;
    sprintf( cbuf, "Interfaces of %s", inet_ntoa( inaddr ) );
    DrawTitle( cbuf );
}


void HideIntfWindow()
{
    if ( debug )
	printf( " HideIntfWindow\n" );

    HideIntfBtns( ButExit );
}


static void
HideIntfBtns()
{
    struct Btn * bp;

    if ( debug )
	printf( " HideIntfBtns\n" );

    HideBtn( ButQuit );
    HideBtn( ButExit );

    for ( bp=BtnOne; bp!=NULL; bp=bp->next )
	HideBtn( bp );

    FlushDpy();
}


static void
ShowIntfBtns()
{
    struct Btn * bp;

    if ( debug )
	printf( " ShowIntfBtns\n" );

    ShowBtn( ButQuit );
    ShowBtn( ButExit );

    for ( bp=BtnOne; bp!=NULL; bp=bp->next )
	{
	ShowBtn( bp );
	/* RedrawBtn( bp ); */
	}

    FlushDpy();
}


static void
ExitIntf()
{
    if ( debug )
	printf( " ExitIntf\n" );

    ReturnToMain();
}


static void
MakeIntfBtns()
{
    int ncol = 1;
    int xpos = 0;
    int ypos = 50;
    int count = 0;

    struct Btn * bp;
    int     i;
    struct Node * np;

    if ( debug )
	printf( " MakeIntfBtns\n" );

    ButQuit = CreateBtn( "Quit", 0, 15, 5, 5, QuitStat, NULL );
    ShowBtn( ButQuit );
    /* RedrawBtn( ButQuit ); */

    ButExit = CreateBtn( "Return to Subnets", 0, 15, 54, 5, ExitIntf, NULL );
    ShowBtn( ButExit );
    /* RedrawBtn( ButExit ); */

    GetIntf();
    for ( i=0; i<MAX_BTNS; i++ )
	{
	np = GetNextIntf( NewSubnet->addr );
	if ( np == NULL )
	    break;

	bp = CreateBtn( NULL, 200, 17, xpos, ypos, SetIndv, np );
	ShowBtn( bp );
	/* RedrawBtn( bp ); */
	if( count < 50 )
	    {
	    ypos += 15;
	    count++;
	    }
	else 
	    {
	    ncol++;
	    xpos += 200;
	    ypos = 50;
	    count = 0;
	    }
	if ( ypos > WindowHeight )
	    WindowHeight = ypos + 25;
	if ( ncol * 200 > WindowWidth )
	    WindowWidth = ncol * 200 + 10;
	}
    FlushDpy();
}


static void
WipeIntfBtns()
{
    struct Btn * bp;
    int     i;
    struct Node * np;

    if ( debug )
	printf( " WipeIntfBtns\n" );

    RemvBtn( ButQuit );
    RemvBtn( ButExit );

    for ( bp=BtnOne; bp!=NULL; bp=bp->next )
	if ( bp->state == INTF_STATE )
	    RemvBtn( bp );

    FlushDpy();
}

