/*
 *  dm.c	92/11/17	DCM Wood
 */

/*
 *  Discovery Manager
 */

#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>

#include "arg.h"

#define MAIN
#include "dm.h"

int	NumChildren=0;
int	options = WNOHANG;
int	pid;
int	sockets[2];
/* int	status; */
union wait status;

u_long	dm_StartTime;
u_long	dm_alb=0;			/* addr lower bound */
u_long	dm_aub=0;			/* addr upper bound */
char	dm_inputf[64]="";
int	dm_ipi=2;			/* interpacket interval */
int	dm_max_retries=2;		/* retry count */
int	dm_max_ttl=30;
char	dm_outputf[64]="";		/* output file */
int	dm_ps=0;			/* packet size */
int	dm_runtime;
int	dm_snm=0;			/* mask request */
u_long	dm_subnet=0;			/* subnet to examine */
u_long	js_addr=0;			/* journal server address */
int	js_port;

struct arg argt[]={
    { ArgCS,	'd',	"default directory",	(int*)defdir,	 NULL	     },
    { ArgCS,	'f',	"input file",		(int*)dm_inputf, "dm_input"  },
    { ArgCS,	'o',	"output file",		(int*)dm_outputf,"dm_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&dm_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&dm_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&dm_aub,		NULL },

    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&dm_max_retries,"5"	},
    { ArgInt,	'v',	"interpacket interval",	&dm_ipi,	"25"	},
    { ArgInt,	'z',	"packet size",		&dm_ps,		NULL	},
    { ArgInt,	't',	"run time",		&dm_runtime,	"3"	},

    { ArgInt,	'm',	"mask request",		&dm_snm,	"1"	},

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};


/*
 *  ChildEnded
 *	Handler for SIGCHLD signal
 */

ChildEnded()
{
    int  i;
    struct em * emp;

    pid = wait3( &status, options, NULL );
    fprintf( stderr, " ChildEnded: pid= %d, status= %X",
					pid, status.w_status );
    if ( pid == -1 )
	{
	fprintf( stderr, "\n" );
	perror( " ChildEnded" );
	return;
	}

    if ( pid == 0 )
	{
	fprintf( stderr, "\n" );
	return;
	}

    for ( emp=emFirst; emp!=NULL; emp=emp->next )
	{
	if ( emp->pid == pid )
	    {
	    fprintf( stderr, ",  %s\n", emp->mname );
	    emp->r_status = status.w_status;
	    emp->flags.e_status = COMPLETED;
	    UpdateEMHF( emp );
            NumChildren--;
	    return;
	    }
	}
    return;
}


/*
 *  exprog 
 *	Fork, exec program using stdin to pass data.
 */

exprog( cmdline )
char * cmdline;          /* character string representing command line to run */
{
    int		cc;
    char	cmdscr[200];
    int		err_child;
    int		i;
    char *	newav[MaxAV];
    int		pid;
    char	prog[150];

    strcpy( cmdscr, cmdline );
    newav[0] = strtok( cmdscr, ", 	" );

    for ( i=1; i<MaxAV; i++ )
	{
	newav[i] = strtok( NULL, ",	 " );
	if ( newav[i] == NULL )
	    break;
	}

    if ( strchr( newav[0], '/' ) != 0 )
	strcpy( prog, newav[0] );
    else
        sprintf( prog, "%s/%s", defdir, newav[0] );

    fprintf( stderr, "\n dm: %s\n", prog );

    cc = pipe( sockets );
    if ( cc < 0 )
	{
	perror( "dm: opening pipe" );
	return( -1 );
	}

    pid = vfork();
    switch ( pid )
	{
	case -1 :	/* Fork failed */
	    perror( "dm: vfork failed" );
	    close( sockets[0] );
	    close( sockets[1] );
	    return( -2 );

	case 0 :	/* Child process */
	    fcntl( sockets[1], F_SETFD, 1 );	/* close "write" on execve */
	    dup2( sockets[0], 0 );		/* move pipe to stdin */
	    errno = 0;
	    execve( prog, newav, environ );	/* exec explorer module */

	    err_child = errno;
	    _exit( err_child );
	    /*NOTREACHED*/

	default :	/* Parent process */
            close( sockets[0] );		/* close "read" fd of pipe */
	}

    sleep( 1 );
    cc = wait4( pid, &status, options, NULL );
    if ( cc == pid )
	{
	if ( WIFEXITED( status ) )
	    {
	    if ( status.w_retcode == 0xff )
		fprintf( stderr, "dm: %s: child exited with -1 return code\n", 
									prog );
	    else
		fprintf( stderr, "dm: %s: child exited early: %s\n", 
					prog, sys_errlist[status.w_retcode] );
	    }
	else
	    fprintf( stderr, "dm: %s: child exited early: status=%x\n", 
				    prog, status.w_status );
	close( sockets[1] );	/* close the pipe entirely */
	return( -3 );
	}
    if ( cc == -1 )
	{
	perror( "dm: child failed:" );
	close( sockets[1] );	/* close the pipe entirely */
	return( -4 );
	}

    NumChildren++;
    return( pid );
}


/*
 *  em_kill
 *	Kill process executing specified explorer module
 */

em_kill( cmdline, emp )
char *	cmdline;
struct em * emp;
{
    struct em * empx;
    int     i;
    int     killed;
    int     sig;
    char *  tokp;

    tokp = strtok( cmdline, " \t" );
    tokp = strtok( NULL, " \t" );
    sig = atoi( tokp );
    tokp = strtok( NULL, " \t\n" );
    killed = 0;
    fprintf( stderr, "\n dm: killing %s\n", tokp );
    for ( empx=emFirst; empx!=NULL; empx=empx->next )
	{
	if ( strcmp( empx->mname, tokp ) == 0 )
	    {
	    if ( empx->flags.e_status == STARTED )
		{
		i = kill( empx->pid, sig );
		if ( i == 0 )
		    {
		    killed++;
		    empx->flags.e_status = COMPLETED;
		    UpdateEMHF( emp );
		    NumChildren--;
		    }
		else
		    perror( "dm: kill" );
		}
	    }
	}
    if ( killed > 0 )
	emp->flags.e_status = COMPLETED;
    else
	emp->flags.e_status = FAILED;
    emp->flags.u_status = UPDATED;
    return;
}


/*
 *  em_sleep
 *	Sleep for number of seconds specified on command line
 */

em_sleep( cmdline, emp )
char *	cmdline;
struct em * emp;
{
    struct em * empx;
    int     seconds;
    char *  tokp;

    tokp = strtok( cmdline, " \t" );
    tokp = strtok( NULL, " \t\n" );
    seconds = atoi( tokp );
    fprintf( stderr, "\n dm: sleeping for %d seconds\n", seconds );
    sleep( seconds );
    emp->flags.e_status = COMPLETED;
    emp->flags.u_status = UPDATED;
    return;
}


/*
 *  em_wait
 *	Wait for process executing specified explorer module to complete
 */

em_wait( cmdline, emp )
char *	cmdline;
struct em * emp;
{
    struct em * empx;
    int     cc;
    int     options = 0;
    char *  tokp;

    tokp = strtok( cmdline, " \t" );
    tokp = strtok( NULL, " \t\n" );
    fprintf( stderr, "\n dm: waiting for %s\n", tokp );
    for ( empx=emFirst; empx!=NULL; empx=empx->next )
	{
	if ( strcmp( empx->mname, tokp ) == 0 )
	    {
	    if ( empx->pid <= 0 )
		continue;

/*
 *	    if ( empx->flags.e_status == STARTED )
 *		{
 *		cc = wait4( empx->pid, &status, options, NULL );
 *		if ( cc == empx->pid )
 *		    {
 *		    empx->flags.e_status = COMPLETED;
 *		    UpdateEMHF( emp );
 *		    NumChildren--;
 *		    }
 *		}
 */
	    while ( empx->flags.e_status == STARTED )
		sleep ( 5 );

	    }
	}
    emp->flags.e_status = COMPLETED;
    emp->flags.u_status = UPDATED;
    return;
}


/*
 *  Discovery Manager main program
 */

main( argc, argv )
int	argc;
char **	argv;
{
    int     cc;
    char    cmdline[200];
    FILE *  dmcfp;
    FILE *  emhfp;
    struct hostent * hent;
    int     i;
    u_long  intmask;
    struct in_addr	jsa;
    char    jsac[200];
    int     jsp;
    struct tm * tm;
    struct timeval ts;
    char *  tokp;


    setlinebuf( stdout );
    gettimeofday( &ts, &tz );

    for (;;)
	{
        pid = wait3( &status, options, NULL );
        if ( pid == -1 )
	    break;
        fprintf( stderr, " CleanUp: pid= %d, status= %X\n", pid, status );
	}

    GetIntf();

    ProcessCL( argc, argv, argt );

    ReadDMCF();

    ReadEMHF();

    signal( SIGCHLD, ChildEnded );

    /*  MAIN LOOP  */

    for ( emp=emFirst; emp!=NULL; emp=emp->next )
	{
	if ( tv.tv_sec < ( emp->lrdt + emp->min +
		( emp->lre == 7 ) ? 0 :
		( emp->lre == 0 ) ? (emp->max - emp->min ) :
		( ( emp->max - emp->min ) / ( 1 << ( emp->lre - 1 ) ) ) ) )
	    continue;

	emp->nlre = emp->lre;
	emp->lre = 0;
	emp->nlrdt = emp->lrdt;
	strcpy( emp->nlrdts, emp->lrdts );
	gettimeofday( &tv, &tz );
	emp->lrdt = tv.tv_sec;
	tm = localtime( &tv.tv_sec );
	strcpy( emp->lrdts, asctime( tm ) );
	emp->lrdts[24] = '\0';
	emp->flags.e_status = STARTED;
	strcpy( cmdline, emp->cmdline );

	if ( strcmp( emp->mname, "kill" ) == 0 )
	    {
	    intmask = sigblock( sigmask(SIGCHLD) );
	    em_kill( cmdline, emp );			/* SIGCHLD blocked */
	    sigsetmask( intmask );
	    }
	else
	if ( strcmp( emp->mname, "wait" ) == 0 )
	    em_wait( cmdline, emp );
	else
	if ( strcmp( emp->mname, "sleep" ) == 0 )
	    em_sleep( cmdline, emp );
	else
	    {
	    intmask = sigblock( sigmask(SIGCHLD) );
	    emp->pid = exprog( cmdline );		/* SIGCHLD blocked */
	    if ( emp->pid < 0 )				/* SIGCHLD blocked */
		{					/* SIGCHLD blocked */
		emp->flags.e_status = FAILED;		/* SIGCHLD blocked */
		emp->lre = emp->r_status;		/* SIGCHLD blocked */
		UpdateEMHF( emp );			/* SIGCHLD blocked */
		}					/* SIGCHLD blocked */
	    else					/* SIGCHLD blocked */
		{					/* SIGCHLD blocked */
		if ( emp->ic != 'n' )			/* SIGCHLD blocked */
		    SendInputData( emp, sockets[1] );	/* SIGCHLD blocked */
		}					/* SIGCHLD blocked */
	    sigsetmask( intmask );
	    }

	if ( emp->wf == 'w' || emp->wf == 'W' )
	    {
	    if ( emp->flags.e_status == STARTED )
		{
		fprintf( stderr, " dm: waiting for %s\n", emp->mname );
		while ( emp->flags.e_status == STARTED )
		    sleep( 1 );
		}
	    }
	else
	    fprintf( stderr, " dm: not waiting for %s\n", emp->mname );

	}  /* end of for loop */

    if ( NumChildren > 0 )
	{
	fprintf( stderr, " dm: waiting for children\n" );
	for (;;)
	    {
	    if ( NumChildren < 1 )
		break;
	    sleep( 1 );
	    }
	}

    for ( emp=emFirst; emp!=NULL; emp=emp->next )
	{
	printf( " dm: %s status= %X\n", emp->mname, emp->flags.e_status );
	}
    gettimeofday( &tv, &tz );
    fprintf( stderr, " dm took %d seconds\n", tv.tv_sec - ts.tv_sec );
    fprintf( stdout, " dm took %d seconds\n", tv.tv_sec - ts.tv_sec );

    exit( 0 );
}

