/*
 *  em_cp.c  DCM Wood	93/02/03
 */

/*
 *	Copy Explorer Module
 *
 *  Get addresses of all interfaces on current system
 *  Determine connected network, subnet masks, and subnets
 *
 *  Set parameters from environment
 *  Set parameters from command line
 *  If Journal server location not set from env or cl,
 *    Get address from getpeername
 *  If input file specified,
 *    Open input file with  fd=0
 *  Read input file
 *    Set parameters from input file
 *    If EOF or blank line,
 *      Stop parameter processing
 *  If output file specified,
 *    Open output file with  fd=1
 *
 *  Open a connection to the Journal Server
 *  Read lines from specified file (default is stdin)
 *  Write lines to journal server
 *  Close the connection to the Journal Server
 */

#include <stdio.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/sockio.h>
#include <sys/errno.h>

#include <net/if.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <netdb.h>

#include "send_data.h"
#define SOCKADDR_IN
#include "data_server.h"
 
#define EM_MAIN

#include "em.h"


struct in_addr	jsa;		/* Journal Server Address  */
int     jsp;			/* Journal Server Pointer  */

struct in_addr thishost;	/* Address of this host  */
struct timeval	tv;		/* Time for various uses  */
struct timezone	tz;		/* Time zone  */

#define DELAY	1000

extern char ** environ;

u_long	em_StartTime;
u_long	em_alb=0;			/* addr lower bound */
u_long	em_aub=0;			/* addr upper bound */
char	em_inputf[64]="\n";
int	em_ipi=DELAY;			/* interpacket interval in ms */
int	em_max_retries=2;		/* retry count */
int	em_max_ttl=30;
char	em_outputf[64]="";		/* output file */
char	em_prog[64];			/* program name */
int	em_ps=0;			/* packet size */
int	em_runtime;
int	em_snm=0;			/* mask request */
u_long	em_subnet=0;			/* subnet to examine */

u_long	js_addr=0;			/* journal server address */
int	js_port=DATA_PORT;

FILE *	em_ifp;
FILE *	jsfp;

extern	char * strtok();

#define RBUFSIZE	1600
#define MAXPACKET	1500
#define MAX_INTF	20

struct net_intf {
	u_long	addr;
	u_long	mask;
	} net_intf[MAX_INTF];

int	debug=0;
int	datalen=64;
struct sockaddr_in from;
int	ident;
int	inflight;
char	line[MAXLINE];
int	num_net_intf=0;
int	retry;
u_char	rpacket[RBUFSIZE];
struct timeval td;
struct timeval ts;
int     len;

u_char	packet[MAXPACKET];
extern	int errno;

int	psfd;			/* packet socket file descriptor */

struct sockaddr_in whereto;	/* Who to ping */


/*
 *  GetIntf
 *	Get info for all interfaces on this system
 */

GetIntf()
{
    int     fd;
    int     host;
    int     i;
    struct ifconf  ifc;
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    u_long  max_host;
    u_long  ogw;

    /* Get some memory in which to store the ifreq entries      */

    ifc.ifc_len = MAX_INTF * sizeof( struct ifreq );

    ifc.ifc_req = (struct ifreq *)calloc( MAX_INTF, sizeof( struct ifreq ) );

    /* Open a socket to do ioctls on.                           */

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
        {        
        perror( "socket" );
        return (-1);
        }               
                        
    /* Do the ioctl, check the result.                          */
                 
    if ( ioctl( fd, SIOCGIFCONF, &ifc) == -1 )
        {
        perror( "ioctl SIOCGIFCONF");
        close( fd );
        return (-1);
        } 

    if ( ifc.ifc_len/sizeof(struct ifreq) == 1 )
        printf( " There is only one interface on this system\n" );
    else
        printf( " There are %d interfaces on this system\n",
                                        ifc.ifc_len/sizeof(struct ifreq) );
    ifreqp = ifc.ifc_req;
    ogw = 0;

    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
	{
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
        printf( " %s   %-16s", ifreqp->ifr_name, inet_ntoa( inaddr ) );
	net_intf[num_net_intf].addr = inaddr.s_addr;
        if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
            {
            perror( "ioctl SIOCGIFNETMASK" );
            continue;
            }
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
        printf( "   %-16s\n", inet_ntoa( inmask ) );
	net_intf[num_net_intf].mask = inmask.s_addr;
        if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
            continue;

	num_net_intf++;
        }

    close( fd );
    return( 0 );
}


/*
 *  ArgBit
 *	Process bit flag argument
 */

ArgBit( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep |= *altdflt;
    return( 0 );
}


/*
 *  ArgCS
 *	Process Character String argument
 */

ArgCS( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    strcpy( (char *)valuep, altdflt );
    return( 0 );
}


/*
 *  ArgInt
 *	Process Integer argument
 */

ArgInt( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = atoi( altdflt );
    return( 0 );
}


/*
 *  ArgIP
 *	Process Internet Protocol Address argument
 */

ArgIP( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = inet_addr( altdflt );
    if ( *valuep < 1 )
	return ( -1 );

    return( 0 );
}


/*
 *  ProcessCL
 *	Process Command Line arguments
 */

ProcessCL( ac, av )
int	ac;
char ** av;
{
    struct arg
	{
	int	(*func)();
	char	flag;
	char *	text;
	int *	value;
	char *	altdflt;
	};

    static struct arg arg[]={
    { ArgCS,	'd',	"default directory",	(int*)defdir,	 NULL	     },
    { ArgCS,	'f',	"input file",		(int*)em_inputf, "\0"  },
    { ArgCS,	'o',	"output file",		(int*)em_outputf,"em_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&em_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&em_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&em_aub,		NULL },

    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&em_max_retries,"5"	},
    { ArgInt,	'v',	"interpacket interval",	&em_ipi,	"2500"	},
    { ArgInt,	'z',	"packet size",		&em_ps,		NULL	},
    { ArgInt,	't',	"run time",		&em_runtime,	"3"	},

    { ArgInt,	'm',	"mask request",		&em_snm,	"1"	},

    { ArgBit,	'x',	"debug flag",		&debug,		"1"	},

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};

    char   fc;
    int    i;
    char * progp;
    struct arg * ap;

    progp = av[0];
    ac--, av++;				/* Skip over program name */
    while (ac && *av[0] == '-')
	{
	fc = *++av[0];
	for ( ap=&arg[0]; ap->func!=NULL; ap++ )
	    {
	    if ( fc == ap->flag )
		{
		if ( ( ac == 1 ) || ( ( ac > 1 ) && ( *av[1] == '-' ) ) )
		    {
		    if ( ap->altdflt == NULL )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    else
			{
		        i = (*ap->func)( ap->value, ap->altdflt );
			if ( i < 0 )
			    {
			    fprintf( stderr, "%s error\n", ap->text );
			    goto arg_error;
			    }
			break;
			}
		    }
		else if ( ( ac > 1 ) && ( *av[1] != '-' ) )
		    {
		    ac--;
		    i = (*ap->func)( ap->value, av[1] );
		    if ( i < 0 )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    av++;
		    break;
		    }
		else
		    {
		    fprintf( stderr, "%s error\n", ap->text );
		    goto arg_error;
		    }
		}  /* endif ( fc == ap->flag ) */
	    }  /* end for */
	ac--, av++;
	}  /* end while */

    arg_error:
	if (ac != 0)
	    {
	    fprintf( stderr, "Usage: %s\n", progp );
	    for ( i=0; arg[i].func!=NULL; i++ )
		fprintf( stderr, "	[-%c %s]\n", arg[i].flag, arg[i].text );
	    exit( -1 );
	    }

    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    char * cp;
    struct dest * dp;
    int    i;
    int    jsfd;		/* Journal Server socket file descriptor */
    struct dest * sdp;

    jsfd = getnetconnect( js_addr, js_port );
    jsfp = fdopen( jsfd, "w" );

    for ( i=0; ; i++ )
	{
	cp = fgets( line, MAXLINE, stdin );
	if ( cp == NULL )
	    break;

	fputs( line, jsfp );
	}

    fprintf( stderr, " %s copied %d lines to the Journal Server\n",
								em_prog, i );
    fclose( jsfp );
    return;
}


main( argc, argv )
int	argc;
char **	argv;
{
    int     i;

    strcpy( em_prog, argv[0] );
    printf( " %s: started\n", em_prog );

    ProcessCL( argc, argv );

    GetIntf();

    /*  Start of code to do Exploring	*	*	*	/* */
    {								/* */
    }								/* */
    /*  End of code to do Exploring	*	*	*	/* */

    /*  Print statistics  */

    UpdateJS();

    close( 0 );
    printf( " %s: completed\n", argv[0] );
    return( 0 );
}

