/*
 *  em_skel.c  DCM Wood	92/12/03
 */

/*
 *	Skeleton Explorer Module
 *
 *  Get addresses of all interfaces on current system
 *  Determine connected network, subnet masks, and subnets
 *
 *  Set parameters from environment
 *  Set parameters from command line
 *  If Journal server location not set from env or cl,
 *    Get address from getpeername
 *  If input file specified,
 *    Open input file with  fd=0
 *  Read input file
 *    Set parameters from input file
 *    If EOF or blank line,
 *      Stop parameter processing
 *  If output file specified,
 *    Open output file with  fd=1
 */

#include <stdio.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>

#define EM_MAIN
#include "em.h"


int	debug=0;		/* Debug flag  */
struct in_addr	jsa;		/* Journal Server Address  */
int     jsp;			/* Journal Server Pointer  */
u_long	netmask;		/* Network mask */
struct in_addr thishost;	/* Address of this host  */
struct timeval	tv;		/* Time for various uses  */
struct timezone	tz;		/* Time zone  */

/* various options */

#define	F_NUMERIC	0x004
#define	F_DEBUG		0x100
#define	F_ADDRONLY	0x400

int	options=0;		/* Various options */

extern char ** environ;

u_long	em_StartTime;
u_long	em_alb=0;			/* addr lower bound */
u_long	em_aub=0;			/* addr upper bound */
char	em_inputf[64]="";
int	em_ipi=2;			/* interpacket interval */
int	em_max_retries=2;		/* retry count */
int	em_max_ttl=30;
char	em_outputf[64]="";		/* output file */
char	em_prog[64];			/* Program name */
int	em_ps=0;			/* packet size */
int	em_runtime;
int	em_snm=0;			/* mask request */
u_long	em_subnet=0;			/* subnet to examine */
u_long	js_addr=0;			/* journal server address */
int	js_port;

FILE *	em_ifp;

struct dest {
	struct in_addr subnet;
	struct in_addr mask;
	struct dest * next;
	};

struct dest * destp;
struct dest * destFirst=NULL;

struct net_intf {
	u_long	addr;
	u_long	nmask;
	u_long	smask;
	} net_intf[MAXINTERFACES];

int	intfcnt;

extern	char * strtok();

/*
 *  AddDest
 *	Add sub-network and mask to list of destinations
 */

AddDest( addr, mask )
u_long	addr;
u_long	mask;
{
    struct dest * destp;
    struct dest * lastp;
    u_long  net;
    u_long  netmask;
    u_long  subnet;


    if ( IN_CLASSA( addr ) )
	netmask = IN_CLASSA_NET;
    else if ( IN_CLASSB( addr ) )
	netmask = IN_CLASSB_NET;
    else if ( IN_CLASSC( addr ) )
	netmask = IN_CLASSC_NET;
    else
	netmask = 0xfffffff0;

    net = addr & netmask;
    subnet = addr & mask;

    if ( destFirst == NULL )
	{
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	destFirst = destp;
	}
    else
	{
        for ( destp=destFirst; destp!=NULL; destp=destp->next )
	    {
	    if ( ( destp->subnet.s_addr & mask ) == subnet )
		return;

	    lastp = destp;
	    }
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	lastp->next = destp;
	}

    destp->subnet.s_addr = addr;
    destp->mask.s_addr = mask;
    destp->next = NULL;
    return;
}


/*
 *  AddIntfToGW
 *	Add Interface to Gateway
 */

AddIntfToGW( intf, gw )
u_long	intf;
u_long	gw;
{

}


/*
 *  AddSNGW
 *	Add to the gateway/subnet/interface structures
 */

AddSNGW( addr, subnet )
u_long	addr;
u_long	subnet;
{

}


/*
 *  GetIntf
 *	Get info for all interfaces on this system
 */

GetIntf()
{
    int  fd;
    int  i;
    struct ifconf  ifc;
    struct ifreq   ifreq[MAXINTERFACES];
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    u_long ogw;

    ifc.ifc_len = MAXINTERFACES * sizeof( struct ifreq );
    ifc.ifc_req = ifreq;

    /* Open a socket to do ioctls on.                           */

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
        {        
        perror( "socket" );
        return (-1);
        }               
                        
    /* Do the ioctl, check the result.                          */
                 
    if ( ioctl( fd, SIOCGIFCONF, &ifc) == -1 )
        {
        perror( "ioctl SIOCGIFCONF");
        close( fd );
        return (-1);
        } 

    if ( ifc.ifc_len/sizeof(struct ifreq) == 1 )
        if ( debug )
	    printf( " There is only one interface on this system\n" );
    else
        if ( debug )
	    printf( " There are %d interfaces on this system\n",
                                        ifc.ifc_len/sizeof(struct ifreq) );
    ifreqp = ifc.ifc_req;
    ogw = 0;
    intfcnt = 0;
    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
        {
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
        if ( debug )
	    printf( " %s   %-16s", ifreqp->ifr_name, inet_ntoa( inaddr ) );
        if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
            {
            perror( "ioctl SIOCGIFNETMASK" );
            continue;
            }
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
        if ( debug )
	    printf( "   %-16s\n", inet_ntoa( inmask ) );
        if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
            continue;

	AddDest( inaddr.s_addr, inmask.s_addr );
        if ( ogw != 0 )
            AddIntfToGW( inaddr.s_addr, ogw );
        AddSNGW( inaddr.s_addr, inaddr.s_addr & inmask.s_addr );
        ogw = inaddr.s_addr;

	if ( IN_CLASSA( inaddr.s_addr ) )
	    netmask = IN_CLASSA_NET;
	else
	if ( IN_CLASSB( inaddr.s_addr ) )
	    netmask = IN_CLASSB_NET;
	else
	if ( IN_CLASSC( inaddr.s_addr ) )
	    netmask = IN_CLASSC_NET;
	else
	    netmask = 0;

	net_intf[intfcnt].addr = inaddr.s_addr;
	net_intf[intfcnt].nmask = netmask;
	net_intf[intfcnt].smask = inmask.s_addr;
	intfcnt++;
        }
    close( fd );
    return( 0 );
}


/*
 *  ArgBit
 *	Process bit flag argument
 */

ArgBit( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep |= *altdflt;
    return( 0 );
}


/*
 *  ArgCS
 *	Process Character String argument
 */

ArgCS( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    strcpy( (char *)valuep, altdflt );
    return( 0 );
}


/*
 *  ArgInt
 *	Process Integer argument
 */

ArgInt( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = atoi( altdflt );
    return( 0 );
}


/*
 *  ArgIP
 *	Process Internet Protocol Address argument
 */

ArgIP( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = inet_addr( altdflt );
    if ( *valuep < 1 )
	return ( -1 );

    return( 0 );
}


/*
 *  ProcessCL
 *	Process Command Line arguments
 */

ProcessCL( ac, av )
int	ac;
char ** av;
{
    struct arg
	{
	int	(*func)();
	char	flag;
	char *	text;
	int *	value;
	char *	altdflt;
	};

    static struct arg arg[]={
    { ArgCS,	'f',	"input file",		(int*)em_inputf, "\n"	     },
    { ArgCS,	'o',	"output file",		(int*)em_outputf,"em_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&em_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&em_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&em_aub,		NULL },

    { ArgInt,	'd',	"debug flag",		&debug,		"1"     },
    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&em_max_retries,"5"	},
    { ArgInt,	'v',	"interpacket interval",	&em_ipi,	"25"	},
    { ArgInt,	'z',	"packet size",		&em_ps,		NULL	},
    { ArgInt,	't',	"run time",		&em_runtime,	"3"	},

    { ArgInt,	'm',	"mask request",		&em_snm,	"1"	},

    { ArgBit,	'a',	"address-only flag",	&options,	"F_ADDRONLY" },
    { ArgBit,	'd',	"socket-debug flag",	&options,	"F_DEBUG" },
    { ArgBit,	'n',	"numeric-only flag",	&options,	"F_NUMERIC" },

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};

/*
 *	"port must be >0"
 *	"em_max_retries must be >0"
 *	"run time must be >1 sec"
 */

    char   fc;
    int    i;
    char * progp;
    struct arg * ap;

    progp = av[0];
    ac--, av++;				/* Skip over program name */
    while (ac && *av[0] == '-')
	{
	fc = *++av[0];
	for ( ap=&arg[0]; ap->func!=NULL; ap++ )
	    {
	    if ( fc == ap->flag )
		{
		if ( ( ac == 1 ) || ( ( ac > 1 ) && ( *av[1] == '-' ) ) )
		    {
		    if ( ap->altdflt == NULL )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    else
			{
		        i = (*ap->func)( ap->value, ap->altdflt );
			if ( i < 0 )
			    {
			    fprintf( stderr, "%s error\n", ap->text );
			    goto arg_error;
			    }
			break;
			}
		    }
		else if ( ( ac > 1 ) && ( *av[1] != '-' ) )
		    {
		    ac--;
		    i = (*ap->func)( ap->value, av[1] );
		    if ( i < 0 )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    av++;
		    break;
		    }
		else
		    {
		    fprintf( stderr, "%s error\n", ap->text );
		    goto arg_error;
		    }
		}  /* endif ( fc == ap->flag ) */
	    }  /* end for */
	ac--, av++;
	}  /* end while */

    arg_error:
	if (ac != 0)
	    {
	    fprintf( stderr, "Usage: %s\n", progp );
	    for ( i=0; arg[i].func!=NULL; i++ )
		fprintf( stderr, "	[-%c %s]\n", arg[i].flag, arg[i].text );
	    exit( 1 );
	    }

    return;
}


/*
 *  ProcessInputFile
 *	Open and read input file
 */

ProcessInputFile()
{
    int    ac;
    u_long addr;
    char * av[MAXARGS];
    int    buf[MAXLINE];
    int    cc;
    char * cp;
    int    ic;
    int    ifd;
    int    ll;
    struct sockaddr_in name;
    int    namelen;
    char * sp;

    if ( em_inputf[0] == '\n' )
	return;

    if ( em_inputf[0] == '\0' )
	{
	em_ifp = fdopen( 0, "r" );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, " Could not fdopen input.\n" );
	    exit( 1 );
	    }

	if ( jsa.s_addr == 0 )
	    {
	    namelen = sizeof( name );
	    cc = getpeername( 0, &name, namelen );
	    if ( cc == 0 && namelen != 0 )
		if ( name.sin_addr.s_addr != 0 )
		    jsa.s_addr = name.sin_addr.s_addr;
	    }
	}
    else
	{
	em_ifp = freopen( em_inputf, "r", 0 );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, " Could not freopen input file %s.\n", em_inputf );
	    exit( 1 );
	    }
	}

/*
    ac = 0;
    av[0] = "em_bp";
    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	ll = strlen( buf );
	if ( ll <= 1 )
	    break;

	if ( strspn( buf, " \t\n" ) == ll )
	    break;

	cp = strtok( buf, " \t\n" );

	do {
	    if ( cp == NULL )
		break;

	    cc = strlen( cp );
	    sp = (char *)malloc( cc+1 );
	    strncpy( sp, cp, cc );
	    ac++;
	    av[ac] = sp;
	    cp = strtok( NULL, " \t\n" );
	    } while ( ac < MAXARGS );

	}
    
    if ( ac > 0 )
	ProcessCL( ac, av );
 */

    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	for(;;)
	    {
	    if ( *cp != ' ' )
		break;

	    cp++;
	    }

	addr = inet_addr( cp );
	if ( addr != -1 )
	    AddDest( addr, 0 );
	}

    fclose( em_ifp );
    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    return;
}


main( argc, argv )
int	argc;
char **	argv;
{
    gettimeofday( &tv, &tz);
    em_StartTime = tv.tv_sec;
    strcpy( em_prog, argv[0] );
    printf( " %s: started\n", em_prog );

    ProcessCL( argc, argv );

    ProcessInputFile();

    GetIntf();

    /*  Start of code to do Exploring	*	*	*	/* */
    {								/* */
    char buf[400];						/* */
    int  ic;							/* */
    int  i;							/* */
								/* */
    for ( i=0; ;i++ )						/* */
	{							/* */
	ic = read( 0, buf, 400 );				/* */
	if ( ic == 0 )						/* */
	    break;						/* */
								/* */
	write( 1, buf, ic );					/* */
	}							/* */
    								/* */
    fprintf( stderr, " %s did %d read/writes\n", em_prog, i );	/* */
    }								/* */
    /*  End of code to do Exploring	*	*	*	/* */

    /*  Print statistics  */

    gettimeofday( &tv, &tz);
    printf( " \nStatistics:\n" );
    printf( " %s took %d seconds\n", em_prog, tv.tv_sec - em_StartTime );

    UpdateJS();

    close( 0 );
    printf( " %s: completed\n", em_prog );
    return( 0 );
}

/*  
 *  Explorer Module command line arguments:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS
 *  
 *  -f   -f  -f  -f  -f  -f  -f  -f  -f     input file
 *  -o   -o  -o  -o  -o  -o  -o  -o  -o     output file
 *  -j   -j  -j  -j  -j  -j  -j  -j  -j     journal server address
 *  -s   -s  -s  -s  -s  -s  -s  -s         subnet to examine
 *  -l       -l  -l      -l                 addr lower bound
 *  -u       -u  -u      -u                 addr upper bound
 *  -r       -r  -r  -r  -r  -r      -r     retry count
 *  -v       -v  -v  -v  -v  -v      -v     interpacket interval
 *  -z           -z  -z  -z                 packet size
 *  -t   -t                      -t         time to run
 *  -m                   -m                 mask request
 *  
 */

/*  
 *  Explorer Module inputs:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS
 *  
 *           al  al      al                 List of IP addresses
 *               rg      rg                 Range of IP addresses
 *       sn  sn  sn  sn  sn  sn  sn         List of Subnets
 *                   nw      nw      nw     List of Networks
 */


/*  
 *  Explorer Module outputs:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS    
 *  
 *       ei  ei              ei  ei         Ethernet/IP address associations
 *                       sm                 Subnet Mask/IP address associations
 *       nm              nm                 Name/IP address associations
 *       ip  ip  ip  ip  ip  ip  ip  ip     Interfaces
 *                           gw  gw  gw     Gateways (collections of interfaces)
 *                           sn  sn  sn     Subnets
 *                           nw  nw         Networks
 *                           is             Interfaces connecting to Subnets
 *                           ig      ig     Interfaces connecting to Gateways
 */

/*  
 *  Explorer Modules History File Format:
 *  
 *  	ModuleName  ProtoNum  ProgNum  InputCode  OutputCode
 *  	Command Line
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	(blank line)
 *  	ModuleName  ProtoNum  ProgNum  InputCode  OutputCode
 *  	Command Line
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	(blank line)
 */

