/*
 *  em_wait.c  DCM Wood	92/12/03
 */

/*
 *	Timed Wait Explorer Module
 *
 *  Get addresses of all interfaces on current system
 *  Determine connected network, subnet masks, and subnets
 *
 *  Set parameters from environment
 *  Set parameters from command line
 *  If Journal server location not set from env or cl,
 *    Get address from getpeername
 *  If input file specified,
 *    Open input file with  fd=0
 *  Read input file
 *    Set parameters from input file
 *    If EOF or blank line,
 *      Stop parameter processing
 *  If output file specified,
 *    Open output file with  fd=1
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>

#define EM_MAIN

#include "em.h"

struct in_addr	jsa;
int     jsp;
struct in_addr thishost;
struct timeval	tv;
struct timezone	tz;

extern char ** environ;

u_long	em_StartTime;
u_long	em_alb=0;			/* addr lower bound */
u_long	em_aub=0;			/* addr upper bound */
char	em_inputf[64]="";
int	em_ipi=2;			/* interpacket interval */
int	em_max_retries=2;		/* retry count */
int	em_max_ttl=30;
char	em_outputf[64]="";		/* output file */
char	em_progname[64]="";		/* program name file */
int	em_ps=0;			/* packet size */
int	em_runtime=60;			/* time to run */
int	em_snm=0;			/* mask request */
u_long	em_subnet=0;			/* subnet to examine */
u_long	js_addr=0;			/* journal server address */
int	js_port;

FILE *	em_ifp;


/*
 *  AddNet
 *	Add network and mask to list of networks
 */

AddNet( addr, mask )
u_long	addr;
u_long	mask;
{
    u_long  netmask;
    u_long  net;

    if ( IN_CLASSA( addr ) )
	netmask = IN_CLASSA_NET;
    else if ( IN_CLASSB( addr ) )
	netmask = IN_CLASSB_NET;
    else if ( IN_CLASSC( addr ) )
	netmask = IN_CLASSC_NET;
    else
	netmask = 0xfffffff0;

    net = addr & netmask;

    /* Do something with net and netmask */

    return;
}


/*
 *  AddIntfToGW
 *	Add Interface to Gateway
 */

AddIntfToGW( intf, gw )
u_long	intf;
u_long	gw;
{

}


/*
 *  AddSNGW
 *	Add to the gateway/subnet/interface structures
 */

AddSNGW( addr, subnet )
u_long	addr;
u_long	subnet;
{

}


/*
 *  GetIntf
 *	Get info for all interfaces on this system
 */

GetIntf()
{
    int  fd;
    int  i;
    struct ifconf  ifc;
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    u_long ogw;

    /* Get some memory in which to store the ifreq entries      */

    ifc.ifc_len = 20 * sizeof( struct ifreq );

    ifc.ifc_req = (struct ifreq *)calloc( 20, sizeof( struct ifreq ) );

    /* Open a socket to do ioctls on.                           */

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
        {        
        perror( "socket" );
        return (-1);
        }               
                        
    /* Do the ioctl, check the result.                          */
                 
    if ( ioctl( fd, SIOCGIFCONF, &ifc) == -1 )
        {
        perror( "ioctl SIOCGIFCONF");
        close( fd );
        return (-1);
        } 

    if ( ifc.ifc_len/sizeof(struct ifreq) == 1 )
        printf( " There is only one interface on this system\n" );
    else
        printf( " There are %d interfaces on this system\n",
                                        ifc.ifc_len/sizeof(struct ifreq) );
    ifreqp = ifc.ifc_req;
    ogw = 0;
    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
        {
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
        printf( " %s   %-16s", ifreqp->ifr_name, inet_ntoa( inaddr ) );
        if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
            {
            perror( "ioctl SIOCGIFNETMASK" );
            continue;
            }
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
        printf( "   %-16s\n", inet_ntoa( inmask ) );
        if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
            continue;

	AddNet( inaddr.s_addr, inmask.s_addr );
        if ( ogw != 0 )
            AddIntfToGW( inaddr.s_addr, ogw );
        AddSNGW( inaddr.s_addr, inaddr.s_addr & inmask.s_addr );
        ogw = inaddr.s_addr;
        }
    close( fd );
    return( 0 );
}


/*
 *  ArgCS
 *	Process Character String argument
 */

ArgCS( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    strcpy( (char *)valuep, altdflt );
    return( 0 );
}


/*
 *  ArgInt
 *	Process Integer argument
 */

ArgInt( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = atoi( altdflt );
    return( 0 );
}


/*
 *  ArgIP
 *	Process Internet Protocol Address argument
 */

ArgIP( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = inet_addr( altdflt );
    if ( *valuep < 1 )
	return ( -1 );

    return( 0 );
}


/*
 *  ProcessCL
 *	Process Command Line arguments
 */

ProcessCL( ac, av )
int	ac;
char ** av;
{
    struct arg
	{
	int	(*func)();
	char	flag;
	char *	text;
	int *	value;
	char *	altdflt;
	};

    static struct arg arg[]={
    { ArgCS,	'f',	"input file",		(int*)em_inputf, "em_input"  },
    { ArgCS,	'o',	"output file",		(int*)em_outputf,"em_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&em_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&em_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&em_aub,		NULL },

    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&em_max_retries,"5"	},
    { ArgInt,	'v',	"interpacket interval",	&em_ipi,	"25"	},
    { ArgInt,	'z',	"packet size",		&em_ps,		NULL	},
    { ArgInt,	't',	"run time",		&em_runtime,	"3"	},

    { ArgInt,	'm',	"mask request",		&em_snm,	"1"	},

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};

    char   fc;
    int    i;
    char * progp;
    struct arg * ap;

    progp = av[0];
    ac--, av++;				/* Skip over program name */
    while (ac && *av[0] == '-')
	{
	fc = *++av[0];
	for ( ap=&arg[0]; ap->func!=NULL; ap++ )
	    {
	    if ( fc == ap->flag )
		{
		if ( ( ac == 1 ) || ( ( ac > 1 ) && ( *av[1] == '-' ) ) )
		    {
		    if ( ap->altdflt == NULL )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    else
			{
		        i = (*ap->func)( ap->value, ap->altdflt );
			if ( i < 0 )
			    {
			    fprintf( stderr, "%s error\n", ap->text );
			    goto arg_error;
			    }
			break;
			}
		    }
		else if ( ( ac > 1 ) && ( *av[1] != '-' ) )
		    {
		    ac--;
		    i = (*ap->func)( ap->value, av[1] );
		    if ( i < 0 )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    av++;
		    break;
		    }
		else
		    {
		    fprintf( stderr, "%s error\n", ap->text );
		    goto arg_error;
		    }
		}  /* endif ( fc == ap->flag ) */
	    }  /* end for */
	ac--, av++;
	}  /* end while */

    arg_error:
	if (ac != 0)
	    {
	    fprintf( stderr, "Usage: %s\n", progp );
	    for ( i=0; arg[i].func!=NULL; i++ )
		fprintf( stderr, "	[-%c %s]\n", arg[i].flag, arg[i].text );
	    exit( 1 );
	    }

    return;
}


/*
 *  ProcessInputFile
 *	Open and read input file
 */

ProcessInputFile()
{
    int    cc;
    int    ic;
    int    ifd;

    if ( em_inputf[0] != '\0' )
	{
	em_ifp = fopen( em_inputf, "r" );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, " Could not open input file %s.\n", em_inputf );
	    exit( 1 );
	    }
/*
 *	for ( ic=0; ; ic++ )
 *	    {
 *	    cc = fscanf( em_ifp, "%s", ipstr );
 *	    if ( cc != 1 )
 *		break;
 *
 *	    if ( IsNet( ipstr, &addr.s_addr ) || IsAddr( ipstr, &addr.s_addr ) )
 *		AddDestEnt( &addr );
 *	    }
 *
 *	if ( ic == 0 )
 *	    {
 *	    fprintf( stderr, " Empty input file.\n" );
 *	    exit( 1 );
 *	    }
 */
	fclose( em_ifp );
	}
    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    return;
}


finish()
{
    /*  Print statistics  */

    gettimeofday( &tv, &tz);
    printf( " %s took %d seconds\n", em_progname, tv.tv_sec - em_StartTime );

    UpdateJS();

    close( 0 );
    printf( " %s: interrupted\n", em_progname );
    exit( 1 );
}

main( argc, argv )
int	argc;
char **	argv;
{
    strcpy( em_progname, argv[0] );
    gettimeofday( &tv, &tz);
    em_StartTime = tv.tv_sec;
    printf( " %s: started\n", em_progname );

    ProcessCL( argc, argv );

    ProcessInputFile();

    GetIntf();

    signal( SIGINT, finish );

    fprintf( stdout, " em_wait:  sleeping for %d seconds\n", em_runtime );
    sleep( em_runtime );

    /*  Print statistics  */

    gettimeofday( &tv, &tz);
    printf( " \nStatistics:\n" );
    printf( " %s took %d seconds\n", em_progname, tv.tv_sec - em_StartTime );

    UpdateJS();

    close( 0 );
    printf( " %s: completed\n", em_progname );
    return( 0 );
}

/*  
 *  Explorer Module command line arguments:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS
 *  
 *  -f   -f  -f  -f  -f  -f  -f  -f  -f     input file
 *  -o   -o  -o  -o  -o  -o  -o  -o  -o     output file
 *  -j   -j  -j  -j  -j  -j  -j  -j  -j     journal server address
 *  -s   -s  -s  -s  -s  -s  -s  -s         subnet to examine
 *  -l       -l  -l      -l                 addr lower bound
 *  -u       -u  -u      -u                 addr upper bound
 *  -r       -r  -r  -r  -r  -r      -r     retry count
 *  -v       -v  -v  -v  -v  -v      -v     interpacket interval
 *  -z           -z  -z  -z                 packet size
 *  -t   -t                      -t         time to run
 *  -m                   -m                 mask request
 *  
 */

/*  
 *  Explorer Module inputs:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS
 *  
 *           al  al      al                 List of IP addresses
 *               rg      rg                 Range of IP addresses
 *       sn  sn  sn  sn  sn  sn  sn         List of Subnets
 *                   nw      nw      nw     List of Networks
 */


/*  
 *  Explorer Module outputs:
 *  
 *      ARP EHP SPG BPG SNM TRR RIP DNS    
 *  
 *       ei  ei              ei  ei         Ethernet/IP address associations
 *                       sm                 Subnet Mask/IP address associations
 *       nm              nm                 Name/IP address associations
 *       ip  ip  ip  ip  ip  ip  ip  ip     Interfaces
 *                           gw  gw  gw     Gateways (collections of interfaces)
 *                           sn  sn  sn     Subnets
 *                           nw  nw         Networks
 *                           is             Interfaces connecting to Subnets
 *                           ig      ig     Interfaces connecting to Gateways
 */

/*  
 *  Explorer Modules History File Format:
 *  
 *  	ModuleName  ProtoNum  ProgNum  InputCode  OutputCode
 *  	Command Line
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	(blank line)
 *  	ModuleName  ProtoNum  ProgNum  InputCode  OutputCode
 *  	Command Line
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	HostAddress  HostName
 *  	    LastRunEffectiveness  LastRunDateTime  LastRunDateString
 *  	    NextLastRunEffectiveness  NextLastRunDateTime  NextLastRunDateString
 *  	(blank line)
 */

