/*
 *  emhf.c	92/12/04	DCM Wood
 */

/*
 *  Explorer Module History File operations
 *	OpenEMHF	Open the Explorer Module History Files
 *	CompareEM	Check for match of EM entry
 *	FindEMmatch	Find matching EM entry
 *	DecodeEMline	Read first line of Explorer Module information
 *	ReadEMentry	Read one entry from the History File
 *	FillEM		Fill in an Explorer Module entry from the History File
 *	ReadEMHF	Read the History File and build array of EM structures
 *	WriteEMentry	Write the specified EM entry to the History File
 *	UpdateEMHF	Update the History File from the specified EM entry
 *
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <netinet/in.h>
#include <netdb.h>

#include "dm.h"

char	emhfn[] = "em.hf";
char	emhfn2[] = "em.hf.new";

int	cc;
char *	cp;
FILE *	emhfp1;
FILE *	emhfp2;
struct hostent * hent;
int	i;

char	cmdline[MAXLINE];
int	diff;
short	family;
struct in_addr hostaddr;
char	hostaddrs[20];
char	hostname[140];
char	ic;
char	line[MAXLINE];
u_long	lrdt;
char	lrdts[30];
int	lre;
int	max;
int	min;
char	mname[30];
u_long	nlrdt;
char	nlrdts[30];
int	nlre;
char	oc;
short	prog;
char	wf;

/*
 *  OpenEMHF
 *	Open the Explorer Module History Files
 */

OpenEMHF()
{
    emhfp1 = fopen( emhfn, "r" );
    if ( emhfp1 == NULL )
	{
	fprintf( stderr, " Can't read old Explorer Module History File\n" );
	fprintf( stderr, " Giving up\n" );
	exit( -1 );
	}

    emhfp2 = fopen( emhfn2, "w" );
    if ( emhfp2 == NULL )
	{
	fprintf( stderr, " Problem opening Explorer Module History File\n" );
	fprintf( stderr, " Giving up\n" );
	exit( -1 );
	}
}


/*
 *  CompareEM
 *	Check for match of EM entry with that in scratch location
 */

EMP
CompareEM( emp )
struct em * emp;
{
    if ( emp->hostaddr.s_addr == hostaddr.s_addr &&
    		  emp->family == family &&
		  emp->prog   == prog )
	return( emp );

    return( NULL );
}


/*
 *  FindEMmatch
 *	Check for match of EM entry in scratch location with either
 *  one particular entry or with all changed but unwritten entries.
 */

EMP
FindEMmatch( emp )
struct em * emp;
{
    struct em * emxp;
    struct em * matchp = NULL;

    if ( emp == NULL )
	{
	for ( emxp=emFirst; emxp!=NULL; emxp=emxp->next )
	    {
	    if ( emxp->flags.u_status != UPDATED )
		{
		matchp = CompareEM( emxp );
		if ( matchp != NULL )
		    break;
		}
	    }
	}
    else
	{
	matchp = CompareEM( emp );
	}
    return( matchp );
}


/*
 *  DecodeEMline
 *	Read first line of Explorer Module information
 */

DecodeEMline( line, mname, family, prog, waitflag, inputc, outputc, min, max )
char *	line;
char *	mname;
short *	family;
short *	prog;
char *	waitflag;
char *	inputc;
char *	outputc;
int *	min;
int *	max;
{
    int     mx=1;

    *mname = '\0';
    *family = 0;
    *prog = 0;
    *waitflag = '\0';
    *inputc = '\0';
    *outputc = '\0';
    *min = 0;
    *max = 604800 * 52;

    cp = strtok( line, " \t\n," );
    if ( cp == NULL )  return;

    strcpy( mname, cp );

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    *family = atoi( cp );

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    *prog = atoi( cp );

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    *waitflag = *cp;

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    *inputc = *cp;
    *outputc = *++cp;

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    if ( *cp == '-' )
	{
	switch( cp[1] )
	    {
	    case 's': mx =      1; break;	/*  Seconds	*/
	    case 'm': mx =     60; break;	/*  Minutes	*/
	    case 'h': mx =   3600; break;	/*  Hours	*/
	    case 'd': mx =  86400; break;	/*  Days	*/
	    case 'w': mx = 604800; break;	/*  Weeks	*/
	    default: mx = 1;
	    }
	cp = strtok( NULL, " \t\n," );
	if ( cp == NULL )  return;
	}

    *min = atoi( cp );
    *min *= mx;

    cp = strtok( NULL, " \t\n," );
    if ( cp == NULL )  return;

    *max = atoi( cp );
    *max *= mx;

    return;
}


/*
 *  ReadEMentry
 *	Read one entry from the History File
 */

ReadEMentry()
{
    int    n;

    cp = fgets( line, MAXLINE, emhfp1 );
    if ( cp == NULL )
	return( 0 );

    DecodeEMline( line, mname, &family, &prog, &wf, &ic, &oc, &min, &max );

    cp = fgets( cmdline, MAXLINE, emhfp1 );
    cmdline[strlen(cmdline)-1] = '\0';

    cp = fgets( line, MAXLINE, emhfp1 );
    cc = sscanf( line, "%s %s", hostaddrs, hostname );

    cp = fgets( line, MAXLINE, emhfp1 );
    if ( line[0] == ' ' || line[0] == '\t' )
	{
	cc = sscanf( line, "%d %d %n", &lre, &lrdt, &n );
	cp = strpbrk( &line[n], "\n" );
	if ( cp != NULL )
	    *cp = '\0';
	strncpy( lrdts, &line[n], 26 );
        lrdts[24] = '\0';
	}

    cp = fgets( line, MAXLINE, emhfp1 );
    if ( line[0] == ' ' || line[0] == '\t' )
	{
	cc = sscanf( line, "%d %d %n", &nlre, &nlrdt, &n );
	cp = strpbrk( &line[n], "\n" );
	if ( cp != NULL )
	    *cp = '\0';
	strncpy( nlrdts, &line[n], 26 );
        nlrdts[24] = '\0';
	}

    cp = fgets( line, MAXLINE, emhfp1 );

    hostaddr.s_addr = inet_addr ( hostaddrs );
    return( 1 );
}


/*
 *  FillEM
 *	Fill in an Explorer Module entry from the History File
 */

FillEM( emp )
struct em * emp;
{
/*
 *  emp->family = family;
 *  emp->prog = prog;
 *  emp->wf = wf;
 *  emp->ic = ic;
 *  emp->oc = oc;
 *  emp->min = min;
 *  emp->max = max;
 */
    emp->lre = lre;
    emp->nlre = nlre;
    emp->lrdt = lrdt;
    emp->nlrdt = nlrdt;
    strcpy ( emp->lrdts, lrdts, 26 );
    strncpy ( emp->nlrdts, nlrdts, 26 );
/*
 *  emp->hostaddr.s_addr = hostaddr.s_addr;
 *  strcpy ( emp->hostname, hostname );
 *  strcpy ( emp->mname, mname );
 *  strcpy ( emp->cmdline, cmdline );
 */
    return;
}


/*
 *  ReadEMHF
 *	Read the History File and build the array of EM structures
 */

ReadEMHF()
{
    struct em ** emlp;
    struct em *  matchp;

    OpenEMHF();
    fclose( emhfp2 );

    gettimeofday( &tv, &tz );
    emlp = &emFirst;
    for ( i=emc; i<EMMAX; i++ )
	{
	cc = ReadEMentry();
	if ( cc == 0 )
	    break;

	matchp = FindEMmatch( NULL );
	if ( matchp != NULL )
	    {
	    FillEM( matchp );
	    printf( " dm: Explorer Module %s used from em.hf\n",
							matchp->mname );
	    continue;
	    }
/*
 *	if ( hostaddr.s_addr == thishost.s_addr )
 *	    {
 *	    emp = (struct em *)calloc( 1, sizeof( struct em ) );
 *	    emp->mname = (char *)malloc( strlen( mname )+1 );
 *	    emp->cmdline = (char *)malloc( strlen( cmdline )+1 );
 *	    emp->hostname = (char *)malloc( strlen( hostname )+1 );
 *	    emp->lrdts = (char *)malloc( 26 );
 *	    emp->nlrdts = (char *)malloc( 26 );
 *	    FillEM( emp );
 *	    emp->next = NULL;
 *	    *emlp = emp;
 *	    emlp = &emp->next;
 *	    printf( " dm: Explorer Module %s read from HF\n", emp->mname );
 *	    }
 */
	}
    fclose( emhfp1 );
    printf( "\n" );
    return;
}


/*
 *  WriteEMentry
 *	Write the specified EM entry to the History File
 */

WriteEMentry( emp )
struct em * emp;
{
    if ( emp == NULL )
	{
	cc = fprintf( emhfp2, "%s %d %d %c %c%c %d %d\n",
				    mname, family, prog, wf, ic, oc, min, max );
	cc = fprintf( emhfp2, "%s\n",
				    cmdline );
	cc = fprintf( emhfp2, "%s  %s\n",
				    inet_ntoa( hostaddr),
				    hostname );
	cc = fprintf( emhfp2, "  %3d  %d  %s\n",
				    lre, lrdt, lrdts );
	cc = fprintf( emhfp2, "  %3d  %d  %s\n",
				    nlre, nlrdt, nlrdts );
	cc = fprintf( emhfp2, "\n" );
	}
    else
	{
	cc = fprintf( emhfp2, "%s %d %d %c %c%c %d %d\n",
				    emp->mname, emp->family, emp->prog, emp->wf,
				    emp->ic, emp->oc, emp->min, emp->max );
	cc = fprintf( emhfp2, "%s\n",
				    emp->cmdline );
	cc = fprintf( emhfp2, "%s  %s\n",
				    inet_ntoa( emp->hostaddr),
				    emp->hostname );
	cc = fprintf( emhfp2, "  %3d  %d  %s\n",
				    emp->lre, emp->lrdt, emp->lrdts );
	cc = fprintf( emhfp2, "  %3d  %d  %s\n",
				    lre, lrdt, lrdts );
	cc = fprintf( emhfp2, "\n" );
	}
    return;
}


/*
 *  UpdateEMHF
 *	Update the History File using data from the specified EM entry
 */

UpdateEMHF( emp )
struct em * emp;
{
    int    matched=0;
    struct em * matchp;

    OpenEMHF();

    for (;;)
	{
	cc = ReadEMentry();
	if ( cc == 0 )
	    break;

	matchp = FindEMmatch( emp );
	WriteEMentry( matchp );
	if ( matchp != NULL )
	    {
	    fprintf( stderr, " UpdateEMHF: %s\n", matchp->mname );
	    matchp->flags.u_status = UPDATED;
	    matched++;
	    }
	}

    if ( matched == 0 )
	{
	lre = lrdt = 0;
	strcpy( lrdts, "No next last run" );
	WriteEMentry( emp );
	}

    fclose( emhfp2 );

    cc = rename( emhfn2, emhfn );
    if ( cc != 0 )
	perror( "dm: rename:" );

    return;
}

