/*
 *  freedl.c	93/04/08	DCM Wood
 */

/*
 *  Free the indicated data_list
 *	FreeDL( dlp )		Data List
 *	FreeIntf( intfp )	Interface
 *	FreeSN( snp )		Subnet
 *	FreeGWL( gwlp )		Gateway List
 *	FreeGW( gwp )		Gateway
 *	FreeCSN( snlp )		Connected Subnets
 *	FreeIntfL( intflp )	Interface List
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <netinet/in.h>
#include <netdb.h>

#include "send_data.h"
#include "data_server.h"

	void	FreeDL();
static	void	FreeIntf();
static	void	FreeSN();
static	void	FreeGWL();
static	void	FreeGW();
static	void	FreeCSN();
static	void	FreeIntfL();

/*
struct data_list {
	struct data_list *next;
	struct data_record *data;
	};

struct data_record {
	int data_type;
	union {
		struct intf *intf;
		struct network *subnet;
		struct host *gateway;
		} dr_type;
	};

#define  Intfp  dr_type.intf
#define  SNp    dr_type.subnet
#define  GWp    dr_type.gateway
 */

/*
 *  FreeDataList
 */

void
FreeDL( dlp )
struct data_list * dlp;
{
    int     cc;
    struct data_list * dlsp;
    struct data_record * drp;

    while ( dlp != NULL )
	{
	dlsp = dlp->next;		/* Save pointer to next entry */
	drp  = dlp->data;

	switch ( drp->data_type )
	    {
	    case INTF:
		FreeIntf( drp->Intfp );	/* Free the interface entry */
		break;

	    case NETWORK:
		FreeSN( drp->SNp );	/* Free the subnet entry */
		break;

	    case GATEWAY:
		FreeGW( drp->GWp );	/* Free the gateway entry */
		break;

	    default:
		fprintf( stderr, " Bogus data_type encountered\n" );
	    }

	free( dlp );			/* Free data_list node */
	dlp = dlsp;			/* Use saved pointer to next entry */
	}
    
    return;
}


/*
struct intf {
	struct intf_mac_addr *in_mac_addr;
	struct intf_prot_addr *ip_addr;
	struct intf_prot_addr *submask;
	struct intf_name *hostname;
	struct host *host;
	};

struct intf_mac_addr {
	struct ether_addr *mac_addr;
	time_t disc_time;
	char *disc_m;
	};

struct intf_prot_addr {
	struct sockaddr_in *pa;
	int addr_family;
	time_t disc_time;
	char *disc_m;
	};

struct intf_name {
	char *local_name;
	char *domain;
	time_t disc_time;
	char *disc_m;
	};
 */

/*
 *  FreeIntf
 */

static void
FreeIntf( intfp )
struct intf * intfp;
{
    int     cc;

    if ( intfp->in_mac_addr != NULL )
	{
	if ( intfp->in_mac_addr->mac_addr != NULL )
	    free( intfp->in_mac_addr->mac_addr );
	if ( intfp->in_mac_addr->disc_m != NULL )
	    free( intfp->in_mac_addr->disc_m );
	free( intfp->in_mac_addr );
	}
    if ( intfp->ip_addr != NULL )
	{
	if ( intfp->ip_addr->pa != NULL )
	    free( intfp->ip_addr->pa );
	if ( intfp->ip_addr->disc_m != NULL )
	    free( intfp->ip_addr->disc_m );
	free( intfp->ip_addr );
	}
    if ( intfp->submask != NULL )
	{
	if ( intfp->submask->pa != NULL )
	    free( intfp->submask->pa );
	if ( intfp->submask->disc_m != NULL )
	    free( intfp->submask->disc_m );
	free( intfp->submask );
	}
    if ( intfp->hostname != NULL )
	{
	if ( intfp->hostname->local_name != NULL )
	    free( intfp->hostname->local_name );
	if ( intfp->hostname->domain != NULL )
	    free( intfp->hostname->domain );
	if ( intfp->hostname->disc_m != NULL )
	    free( intfp->hostname->disc_m );
	free( intfp->hostname );
	}
    if ( intfp->host != NULL )
	{
	if ( intfp->host->gw_name	    != NULL )
	    free( intfp->host->gw_name );
	if ( intfp->host->connected_subnets != NULL )
	    FreeCSN( intfp->host->connected_subnets );
	if ( intfp->host->intfs		    != NULL )
	    FreeIntfL( intfp->host->intfs );
	if ( intfp->host->device_type	    != NULL )
	    free( intfp->host->device_type );
	if ( intfp->host->os		    != NULL )
	    free( intfp->host->os );
	if ( intfp->host->mail_host	    != NULL )
	    free( intfp->host->mail_host );
	free( intfp->host );
	}
    return;
}


/*
struct network {
	struct intf_prot_addr  *net_numb;
	struct intf_prot_addr *subnet;
	int gw_count;
	struct gw_pointer *gwp_start ;
	};

struct gw_pointer {
	struct host *gw_host;
	struct gw_pointer *gw_next;
	};
 */

/*
 *  FreeSN
 */

static void
FreeSN( snp )
struct network * snp;
{
    if ( snp->net_numb	!= NULL )
        free( snp->net_numb );
    if ( snp->subnet	!= NULL )
        free( snp->subnet );
    if ( snp->gwp_start	!= NULL )
        FreeGWL( snp->gwp_start );
    free( snp );
    return;
}


/*
 *  FreeGWL
 */

static void
FreeGWL( gwlp )
struct gw_pointer * gwlp;
{
    struct gw_pointer * xp;

    while ( gwlp != NULL )
	{
	xp = gwlp->gw_next;
	if ( gwlp->gw_host != NULL )
	    FreeGW( gwlp->gw_host );
	gwlp = xp;
	}

    return;
}

/*
struct host {
	int no_intf;
	char *gw_name;
	struct subnet *connected_subnets;
	struct intf_list *intfs;
	char *device_type;
	int gateway;
	char *os;
	int nfs_fileserver;
	char *mail_host;
	};
 */

/*
 *  FreeGW
 */

static void
FreeGW( gwp )
struct host * gwp;
{
    if ( gwp->gw_name		!= NULL )
	free( gwp->gw_name );
    if ( gwp->connected_subnets	!= NULL )
	FreeCSN( gwp->connected_subnets );
    if ( gwp->intfs		!= NULL )
	FreeIntfL( gwp->intfs );
    if ( gwp->device_type	!= NULL )
	free( gwp->device_type );
    if ( gwp->os		!= NULL )
	free( gwp->os );
    if ( gwp->mail_host		!= NULL )
	free( gwp->mail_host );
    free( gwp );
    return;
}

/*
struct subnet {
	struct network *subnet_info;
	struct intf *net_intf;
	struct subnet *next_subnet;
	};
 */

/*
 *  FreeCSN
 */

static void
FreeCSN( snlp )
struct subnet * snlp;
{
    struct subnet * xp;

    while ( snlp != NULL )
	{
	xp = snlp->next_subnet;
	if ( snlp->subnet_info	!= NULL )
	    FreeSN( snlp->subnet_info );
	if ( snlp->net_intf	!= NULL )
	    FreeIntf( snlp->net_intf );
	snlp = xp;
	}

    return;
}

/*
struct intf_list {
	struct intf *subnet_intf;
	struct intf_list *next_intf;
	};
 */

/*
 *  FreeIntfL
 */

static void
FreeIntfL( intflp )
struct intf_list * intflp;
{
    struct intf_list * xp;

    while ( intflp != NULL )
	{
	xp = intflp->next_intf;
	if ( intflp->subnet_intf != NULL )
	    FreeIntf( intflp->subnet_intf );
	intflp = xp;
	}

    return;
}

