/*
 *  sid.c	93/01/16	DCM Wood
 */

/*
 * Explorer Module inputs:
 * 
 *     ARP EHP SPG BPG SNM TRR RIP DNS
 * 
 *      xx                                 No input
 *                              xx         No input
 *          ie                             List of IP addrs needing Ether Addrs
 *              si                         List of subnets to find interfaces
 *                  si                     List of subnets to find interfaces
 *                      im                 List of IP addrs needing masks
 *                          ne             List of networks to explore
 *                          se             List of subnets to explore
 *                                  ne     List of networks to explore
 * 
 *                                     in  List of ip addrs needing names
 */

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "dm.h"

#include "send_data.h"
#define SOCKADDR_IN
#include "data_server.h"

#define  XX  'x' << sizeof ( char ) | 'x'
#define  IE  'i' << sizeof ( char ) | 'e'
#define  IM  'i' << sizeof ( char ) | 'm'
#define  SE  's' << sizeof ( char ) | 'e'
#define  NE  'n' << sizeof ( char ) | 'e'
#define  IN  'i' << sizeof ( char ) | 'n'
#define  TD  't' << sizeof ( char ) | 'd'

#define  DATA  " Test data sent from Discovery Manager\n"

extern	struct data_list * ReadJS();

static	int	cc;			/* scratch - completion code */
static	char *	cp;			/* scratch - character pointer */
static	char	ic;			/* input code */
static	char	line[MAXLINE];
static	char	oc;			/* output code */
static	FILE *	sfp;			/* Socket file pointer */


/*
 *  SendInputData
 *	Send input data to explorer module
 *
 */

SendInputData( emp, sfd )
struct em * emp;		/* explorer module pointer */
int	sfd;			/* socket file descriptor */
{
    struct data_list * dlfp;
    struct data_list * dlp;
    struct data_record * drp;
    int     i;
    struct intf * intfp;
    int     io;
    struct network * subnetp;
    int     rc=0;		/* count of records written to em */

    io = emp->ic << sizeof( char) | emp->oc;

    switch ( io )
	{
	case TD:	/* Test Data */
	    rc = 1;
	    cc = write( sfd, DATA, sizeof( DATA ) );
	    if ( cc < 0 )
		{
		perror( "dm: writing messages to Explorer Module" );
	        close( sfd );		/* close the pipe */
		return;
		}

    case XX:	/* No input */
	close( sfd );		/* close the pipe */
	return;
    }

    sfp = fdopen( sfd, "w" );
    setlinebuf( sfp );

    switch ( io )
	{
	case IE:	/* IP addrs needing Ethernet addrs */
	case IM:	/* IP addrs needing Subnet Masks */
	case IN:	/* IP addrs needing names */
	    dlp = ReadJS( INTF );
	    break;

	case SE:	/* Subnets needing exploration */
	case NE:	/* Networks needing exploration */
	    dlp = ReadJS( NETWORK );
	    break;
	
	default:
	    close( sfd );		/* close the pipe */
	    return;
	}

    if ( dlp == NULL )
	{
	fclose( sfp );
	return;
	}

    dlfp = dlp;
    for ( ; dlp!=NULL; dlp=dlp->next )
        {
        drp = dlp->data;
        switch ( io )
	    {
	    case IE:		/* IP addrs needing Ethernet addrs */
	    case IM:		/* IP addrs needing Subnet Masks */
	    case IN:		/* IP addrs needing names */
		intfp = drp->dr_type.intf;
		if ( intfp == NULL )
		    continue;
		break;

	    case SE:		/* Subnets needing exploration */
	    case NE:		/* Networks needing exploration */
		subnetp = drp->dr_type.subnet;
		if ( subnetp == NULL )
		    continue;
		break;
	    }

	switch ( io )
	    {
	    case IE:		/* IP addrs needing Ethernet addrs */
		if ( intfp->in_mac_addr == NULL )
		    {
		    cc = fprintf( sfp, "%s\n",
				inet_ntoa( intfp->ip_addr->pa->sin_addr ) );
		    rc++;
		    }
		break;

	    case IM:		/* IP addrs needing Subnet Masks */
		if ( intfp->submask == NULL )
		    {
		    cc = fprintf( sfp, "%s\n",
				inet_ntoa( intfp->ip_addr->pa->sin_addr ) );
		    rc++;
		    }
		break;

	    case IN:		/* IP addrs needing names */
		if ( intfp->hostname == NULL )
		    {
		    cc = fprintf( sfp, "%s\n",
				inet_ntoa( intfp->ip_addr->pa->sin_addr ) );
		    rc++;
		    }
		break;

	    case SE:		/* Subnets needing exploration */
		for ( i=0; i<intfcnt; i++ )
		    {
		    if ( ( ( subnetp->net_numb->pa->sin_addr.s_addr ^
					net_intf[i].addr ) &
					net_intf[i].nmask ) == 0 )
			break;
		    }
		if ( i < intfcnt )
		    {
		    cc = fprintf( sfp, "%s\n",
				inet_ntoa( subnetp->net_numb->pa->sin_addr ) );
		    rc++;
		    }
		break;

	    case NE:		/* Networks needing exploration */
		if ( subnetp->net_numb != NULL )
		    {
		    cc = fprintf( sfp, "%s\n",
				inet_ntoa( subnetp->net_numb->pa->sin_addr ) );
		    rc++;
		    }
		break;
	    }
	if ( cc < 0 )
	    {
	    perror( "dm: trying to write to em" );
	    break;
	    }
	}

    fclose( sfp );		/* close the pipe */
    fprintf( stderr, " dm: %d records written to %s\n", rc, emp->mname );
    /* FreeDL( dlfp ); */
    return;
}

