/*
 *  arpwatch.c
 */

#include <stdio.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/stropts.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/packetfilt.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netdb.h>

#include "arpwatch.h"
#include "data_server.h"

#define DBUFSIZE 10000

/* %W% %D% */

struct hashtab * etherhash[HASHSIZE];	/* hash table for found ethernet
					   addresses */
struct cur_arps * recent_arp = NULL;	/* ethernet addresses not sent to
					   output */
struct cur_arps * arp_marker = NULL;	/* points to end of list of non
					   reported ea */
int	cnt=0;			/* # of messages sent to journal server */
int	dump_count = 0;		/* count of time that data has been dumped  */
int	need_dump_data;		/* flag to dump data  */
int	new_data;		/* new data received since last dump */
char	data_file[256]="";	/* name of file or server to put data */
char *	dfp = NULL;		/* pointer to data_file, if set */
int	send_names = 0;		/* if set, gethostbyaddr */

struct	net_intf {
	u_long	addr;
	u_long	nmask;
	u_long	smask;
	} net_intf[20];

int	intfcnt = 0;


/*
 *  finish
 *	Upon receipt of SIGTERM, send data to Journal Server and quit
 */

finish()
{
    if ( new_data > 0 )
	dump_data();

    fprintf( stderr,
		" em_arpwatch: %d INTF entries sent to Journal Server\n", cnt );
    exit( 0 );
}


/*
 *  ether_ntoa
 *    Format ethernet address as a string
 */

char *
ether_ntoa( eap )
struct ether_addr * eap;
{
     static char eastr[20];

     sprintf( eastr,  "%02x:%02x:%02x:%02x:%02x:%02x",
			eap->ether_addr_octet[0],
			eap->ether_addr_octet[1],
			eap->ether_addr_octet[2],
			eap->ether_addr_octet[3],
			eap->ether_addr_octet[4],
			eap->ether_addr_octet[5] );
    return( eastr );
}


/*
 *  arpwatch main program
 */

main( argc, argv )
int     argc;
char **	argv;
{
    struct ifreq * ifreq;	/* interface struct for getting info for
				   ether net interface */
    void    get_interface_name();	/* function to get the network
					   interface name  */
    int     nit_dev[20];	/* file descriptor for nit c-comment */
    int     time_to_dump = 0;	/* time value to next dump  */
    struct ifconf * get_netconf();
    struct ifconf * intf_list;
    char *  iflim;
    char *  if_var;
    int     ninterface;
    int     i = 0;

    setlinebuf( stdout );

    /* get the command line arguments if they exist */

    while ( *(++argv) )
	{
	if ( **argv != '-' )
	    {
	    fprintf( stderr, "em_arpwatch: Use a - before arguments\n" );
	    exit( 1 );
	    }

	switch ( *(*argv + 1) )
	    {
	    case 'f':
		if ( *(++argv) )
		    {
		    if ( ! index( *argv, '/ ' ) )
			strcpy( data_file, "./" );

		    strcat( data_file, *argv );
		    dfp = data_file;
		    break;
		    }

	    case 'h':
		if ( *(++argv) )
		    {
		    strcpy( data_file, *argv );
		    dfp = data_file;
		    break;
		    }

	    case 'n':
		    send_names++;
		    break;

	    case 't':
		if ( *(++argv) )
		    time_to_dump = atoi( *argv );
		else
		    fprintf( stderr,
				" em_arpwatch: no time given, default used\n" );
	    }
	}

    /* allocate space for the ifreq struct and zero it out */

    ifreq = ( struct ifreq *) malloc( sizeof(struct ifreq) );
    bzero( ifreq, sizeof(struct ifreq) );

    /* clear out hash table  */

    bzero( etherhash, sizeof(struct hash_tab *) );

    /* get interface names on current machine */

    intf_list = get_netconf();

    /* get the names of the ethernet ports on local host */

    /* go through list of interface and setup nit  */

    iflim = intf_list->ifc_buf + intf_list->ifc_len;
    if_var = intf_list->ifc_buf;

    while ( if_var < iflim )
	{

	struct ifreq * ifr;

	ifr = (struct ifreq *) if_var;

	/* open the nit device and return the file descriptor */

	nit_dev[i] = init_device( ifr );
	i++;
	if_var += sizeof(struct ifreq);
	}

    ninterface = i--;

    /* read the data from the nit device and process */
    /* this will never return */

    read_loop( nit_dev, time_to_dump, ninterface );
}


void 
get_interface_name( name )
char *	name;
{
    char    buf[256];
    struct ifconf ifc;
    int     s;

    if ( ( s = socket( AF_INET, SOCK_DGRAM, 0 ) ) < 0 )
	{
	perror( "Failure to establish socket to interface name" );
	exit( 1 );
	}

    ifc.ifc_len = sizeof(buf);
    ifc.ifc_buf = buf;

    if ( ioctl( s, SIOCGIFCONF, (char *) &ifc ) < 0 )
	{
	perror( "Error in getting the interface" );
	exit( 1 );
	}

    strcpy( name, ifc.ifc_req->ifr_name );
    close( s );
}


init_device( ifreq )
struct ifreq * ifreq;
{
    int     s;
    struct strioctl si;
    u_long  if_flags;

    if ( ( s = open( "/dev/nit", O_RDONLY ) ) < 0 )
	{
	perror( "can't open nit device" );
	exit( 1 );
	}

    if ( ioctl( s, I_SRDOPT, (char *) RMSGD ) < 0 )
	{
	perror( "ioctl of ISRDOPT" );
	exit( 1 );
	}

    set_packet_filt( s );

    si.ic_cmd = NIOCBIND;
    si.ic_timout = INFTIM;
    si.ic_len = sizeof(struct ifreq);
    si.ic_dp = (char *) ifreq;

    if ( ioctl( s, I_STR, (char *) &si ) < 0 )
	{
	perror( "ioctl of NIOCBIND" );
	exit( 1 );
	}

    if_flags = NI_PROMISC;

    si.ic_cmd = NIOCSFLAGS;
    si.ic_len = sizeof(if_flags);
    si.ic_timout = INFTIM;
    si.ic_dp = (char *) &if_flags;
    if ( ioctl( s, I_STR, (char *) &si ) < 0 )
	{
	perror( "Couldn't open promiscuous mode" );
	exit( 1 );
	}

    return ( s );
}


set_packet_filt( nit_dev )
{

    int     cc;
    struct packetfilt pf;
    struct ether_header eh;
    register u_short * fwp = pf.Pf_Filter;
    struct strioctl si;

    int     offset;

    offset = ( (u_int)&eh.ether_type - (u_int)&eh.ether_dhost ) /
							sizeof(u_short);
    *fwp++ = ENF_PUSHWORD + offset;
    *fwp++ = ENF_PUSHLIT;
    *fwp++ = htons( ETHERTYPE_ARP );
    *fwp++ = ENF_EQ;
    pf.Pf_FilterLen = fwp - &pf.Pf_Filter[0];

/*  pf.Pf_Priority = 5;  */

    cc = ioctl( nit_dev, I_PUSH, "pf" );
    if ( cc < 0 )
	{
	perror( "Can\'t push packetfilter" );
	exit( 1 );
	}

    si.ic_cmd = NIOCSETF;
    si.ic_timout = INFTIM;
    si.ic_len = sizeof(pf);
    si.ic_dp = (char *) &pf;

    cc = ioctl( nit_dev, I_STR, (char *) &si );
    if ( cc < 0 )
	{
	perror( "Can\'t activate packetfilter" );
	exit( 1 );
	}
}


bind_device( nit_dev, ifreq )
int     nit_dev;
struct ifreq * ifreq;
{
    struct strioctl si;
    u_long  if_flags;

    si.ic_cmd = NIOCBIND;
    si.ic_len = sizeof(struct ifreq);
    si.ic_timout = INFTIM;
    si.ic_dp = (char *) ifreq;
    if ( ioctl( nit_dev, I_STR, (char *) &si ) < 0 )
	{
	perror( "Couldn't bind to interface" );
	exit( 1 );
	}

    si.ic_cmd = NIOCSFLAGS;
    if_flags = NI_PROMISC;
    si.ic_len = sizeof(if_flags);
    si.ic_timout = INFTIM;
    si.ic_dp = (char *) &if_flags;
    if ( ioctl( nit_dev, I_STR, (char *) &si ) < 0 )
	{
	perror( "Couldn't open promiscuous mode" );
	exit( 1 );
	}

    ioctl( nit_dev, I_FLUSH, (char *) FLUSHR );
}


read_loop( fd, time_to_dump, ninterface )
int     fd[];			/* file descriptor where nit device is picked
				   up */
int     time_to_dump;
int     ninterface;
{
    char *  buf;		/* buffer for data collected from nit */
    int     cc;			/* completion code  */
    fd_set  des_mask;		/* file descriptor mask for select */
    fd_set  tmp_mask;		/* temporary mask to put file descriptor */
    int     n;			/* number of file descriptors read  */
    int     nif;		/* counter */
    int     nfsd;		/* total number of file descriptors -- use
				   gettablesize */
    void    set_dumpclock();
    int     dump_handler();

    signal( SIGALRM, dump_handler );
    signal( SIGINT, finish );
    signal( SIGTERM, finish );

    /* set the time for next alarm */

    set_dumpclock( time_to_dump );

    /* create space for buffer to read data  */

    buf = (char *) malloc( DBUFSIZE );

    nfsd = getdtablesize();

    /* clear the file descriptor mask and set the appropriate file descriptor */

    FD_ZERO( &des_mask );

    for ( nif = 0; nif < ninterface; nif++ )
	FD_SET( fd[nif], &des_mask );

    need_dump_data = 0;
    new_data = 0;

    for (;;)
	{
	extern int errno;

	tmp_mask = des_mask;

	if ( need_dump_data )
	    {
	    if ( new_data > 0 )
		{
	        dump_data();
		new_data = 0;
		}
	    need_dump_data = 0;
	    dump_count++;
	    if ( dump_count >= 60 )
		{
		/* clean_arp_table(); */
	        dump_count = 0;
		}
	    }
	n = select( nfsd, &tmp_mask, NULL, NULL, NULL );
	if ( n < 0 )
	    {
	    if ( errno == EINTR )
		continue;
	    perror( "select failed" );
	    exit( 0 );
	    }

	if ( n == 0 )
	    continue;

	for ( nif = 0; nif < ninterface; nif++ )
	    {
	    if ( FD_ISSET( fd[nif], &tmp_mask ) )
		{
		char   addrstr[20];
		struct ether_arp * arp_p;	/* arp packet info */
		char * bp;	/* pointer into collected buffer */
		struct ether_header * ehp;
		struct hashtab ** ether_tab_entry;	/* entry in etherhash
							   array */
		struct hashtab * ether_pointer;	/* walking pointer through
						   the linked list at
						   e_t_entry */
		struct in_addr inaddr;
		struct ip * ipp;

		u_short * i;	/* pointer to first two values of mac address */
		u_short * j;	/* pointer to second two values of mac
				   address */
		u_short * k;	/* pointer to third two values of mac address */
		int	n;


		cc = read( fd[nif], buf, (int) 8192 );
		if ( cc < 0 )
		    {
		    fprintf( stderr, "em_arpwatch: can't read data\n" );
		    exit( 1 );
		    }

		/* find the start of the arp section of the code  */

		bp = buf;
		bp += sizeof(struct ether_header);
		arp_p = (struct ether_arp *) bp;

		/* Require that ethernet addresses match */

		ehp = (struct ether_header *)buf;

		bcopy( arp_p->arp_spa, &inaddr.s_addr, 4 );
		cc = bcmp( (char *)&ehp->ether_shost,
			   (char *)&arp_p->arp_sha, 6 );
		if ( cc != 0 )
		    {
		    strcpy( addrstr, ether_ntoa( &ehp->ether_shost ) );
		    fprintf( stderr,
				"em_arpwatch: %s does not match %s for %s\n",
				addrstr, ether_ntoa( &arp_p->arp_sha ),
				inet_ntoa( inaddr ) );
		    continue;
		    }

		/* Require that ip address be on a directly attached subnet */

		for ( n=0; n<intfcnt; n++ )
		    if ( ( ( inaddr.s_addr ^ net_intf[n].addr ) &
				net_intf[n].smask ) == 0 )
			break;

		if ( n == intfcnt )
		    {
		    fprintf( stderr,
				"em_arpwatch: %s  %s  not on local subnet\n",
				ether_ntoa( &ehp->ether_shost ),
				inet_ntoa( inaddr ) );
		    continue;
		    }

		/* get the short parts of each ethernet address */

		i = (u_short *) & arp_p->arp_sha;
		j = (u_short *) & arp_p->arp_sha + 1;
		k = (u_short *) & arp_p->arp_sha + 2;

		/* find the proper location of ethernet address in hash table */

		ether_tab_entry = &etherhash[(*i ^ *j ^ *k) & HASHSIZE - 1];

		/* get contents of slot  */

		ether_pointer = *ether_tab_entry;

		/* do we have this address already */

		while ( ether_pointer )
		    {
		    /* compare the ethernet address of current packet with
		       that in the hash table */

		    cc = ether_cmp( &ether_pointer->data.mach_address,
							&arp_p->arp_sha );
		    if ( ! cc )
			{
			/* if found, increment the count on how many times
			   the address was found */
			ether_pointer->data.count++;
			break;
			}

		    ether_pointer = ether_pointer->next;
		    }

		/* we dont have an entry for this ethernet address allocate
		   space for one. */

		if ( ether_pointer == NULL )
		    {
		    struct sockaddr_in * ip_tmp;
		    struct hashtab * new_tab_entry;	/* memory for new
							   entry */
		    struct cur_arps * rec_arp_tmp;	/* memory for new
							   current arp */
		    struct hostent * hp;		/* host name entry */

		    /* create memory for new address */

		    new_tab_entry = (struct hashtab *)
					malloc(sizeof(struct hashtab) );
		    bzero( new_tab_entry, sizeof(struct hashtab) );

		    /* create space for ip address  */

		    ip_tmp = (struct sockaddr_in *)
					malloc( sizeof(struct sockaddr_in) );

		    /* load in address family and IP address */

		    ip_tmp->sin_family = AF_INET;
		    bcopy( (char *)arp_p->arp_spa, (char *)&ip_tmp->sin_addr,
								sizeof(u_long) );
		    /* put ip data into arp hash */

		    bcopy( (char *)ip_tmp, (char *)&new_tab_entry->data.address,
						sizeof(struct sockaddr_in) );

		    /* load in the hardware address */

		    bcopy( arp_p->arp_sha,
				(char *)&new_tab_entry->data.mach_address,
						sizeof(struct ether_addr) );

		    /* point new tab entry to existing member of
		       ether_tab_entry */

		    new_tab_entry->next = *ether_tab_entry;

		    if ( send_names )
			{
		        /* get the hostname for the ether entry */

		        hp = gethostbyaddr( (char *)&ip_tmp->sin_addr,
				sizeof(struct in_addr), ip_tmp->sin_family );
			}
		    else
			hp = NULL;

		    if ( hp )
			{
			new_tab_entry->data.hostname =
				(char *) malloc( strlen( hp->h_name) + 1 );
			strcpy( new_tab_entry->data.hostname, hp->h_name );
			}
		    else
			new_tab_entry->data.hostname = NULL;

		    /* increment the count for seeing this host  */

		    new_tab_entry->data.count++;

		    /* create an entry in the list of recent found arps  */

		    rec_arp_tmp = (struct cur_arps *)
					malloc( sizeof(struct cur_arps) );
		    bzero( rec_arp_tmp, sizeof(struct cur_arps) );

		    /* set the current arp to the new tab entry */

		    *ether_tab_entry = rec_arp_tmp->arp_hash = new_tab_entry;

		    printf( " em_arpwatch: %s  %s",
			 ether_ntoa( &((*ether_tab_entry)->data.mach_address) ),
			 inet_ntoa( ip_tmp->sin_addr ) );

		    if ( (*ether_tab_entry)->data.hostname )

			/* printf( " em_arpwatch: name of host is %s\n",
			   (*ether_tab_entry)->data.hostname ); */

			printf( " %s\n",
					(*ether_tab_entry)->data.hostname );
		    else
			printf( "\n" );

		    new_data++;

		    /* put marker at end of recently acquired arps */

		    if ( arp_marker )
			{
			arp_marker->next = rec_arp_tmp;
			arp_marker = rec_arp_tmp;
			continue;
			}

		    /* new list so set the start and end of chain */

		    recent_arp = arp_marker = rec_arp_tmp;
		    }
		}
	    }
	}
}


void 
set_dumpclock( time )
int     time;
{
    struct itimerval timer;

    if ( ! time )
	time = 120;

    timer.it_value.tv_sec = time;
    timer.it_value.tv_usec = 0;

    timer.it_interval.tv_sec = time;
    timer.it_interval.tv_usec = 0;

    setitimer( ITIMER_REAL, &timer, NULL );
}


dump_handler()
{
    need_dump_data++;
}


dump_data()
{
    int     i;
    u_long  oldmask;
    struct cur_arps * recent_arp_tmp;

    i = 0;
    while ( recent_arp )
	{
	struct data_record * arp_data;

	arp_data = (struct data_record *)make_data( INTF );
	arp_data->dr_type.intf->in_mac_addr = (struct intf_mac_addr *)
					malloc( sizeof(struct intf_mac_addr) );
	if ( arp_data->dr_type.intf->in_mac_addr == NULL )
	    {
	    fprintf( stderr,
		   "em_arpwatch: unable to allocate memory for in_mac addr\n" );
	    exit( 1 );
	    }

	arp_data->data_type = INTF;
	arp_data->dr_type.intf->in_mac_addr->mac_addr =
				&recent_arp->arp_hash->data.mach_address;

	/* if( recent_arp->arp_hash->data.hostname ) */
	/* arp_data->dr_type.intf =(u_char *)
	   recent_arp->arp_hash->data.hostname; */

	arp_data->dr_type.intf->ip_addr = (struct intf_prot_addr *)
					malloc( sizeof(struct intf_prot_addr) );
	if ( arp_data->dr_type.intf->ip_addr == NULL )
	    {
	    fprintf( stderr,
			"em_arpwatch: Can't allocate space for ip_addr\n" );
	    exit( 1 );
	    }

	arp_data->dr_type.intf->ip_addr->addr_family = AF_INET;
	arp_data->dr_type.intf->ip_addr->pa =
					&(recent_arp->arp_hash->data).address;

	pack_data( arp_data );

	recent_arp_tmp = recent_arp->next;

	/* free( recent_arp ); */

	recent_arp = recent_arp_tmp;
	i++;
	}
    oldmask = sigblock( sigmask( SIGALRM ) );
    send_data( dfp );
    sigsetmask( oldmask );
    arp_marker = NULL;

    printf( " em_arpwatch:%4d INTF entries sent to Journal Server\n", i );
    cnt += i;
    return;
}


clean_arp_table()
{
    int     x;			/* marker into arp_table */
    struct hashtab * etherhash_pos;	/* entry in hashtable */

    printf( " em_arpwatch: clearing ARP data\n" );

    for ( x = 0; x < HASHSIZE; x++ )
	{
	etherhash_pos = etherhash[x];
	while ( etherhash_pos )
	    {
	    struct hashtab * etherhash_tmp;	/* pointer to next etherhash */

	    etherhash_tmp = etherhash_pos->next;
	    free( etherhash_pos );
	    etherhash_pos = etherhash_tmp;
	    }
	etherhash[x] = NULL;
	}

    return;
}


struct ifconf *
get_netconf()
{
    char  * buf;
    u_long  addr;
    int     cc;
    int     fake_socket;
    int     i;
    struct ifconf * ifc;
    struct ifreq * ifreqp;
    struct ifreq   ifrscr;
    int     maxintf;

    fake_socket = socket( AF_INET, SOCK_DGRAM, 0 );
    if ( fake_socket < 0 )
	{
	perror( "Failure to establish socket" );
	exit( 1 );
	}

    ifc = (struct ifconf *) malloc( sizeof(struct ifconf) );
    if ( ifc == NULL )
	{
	perror( "malloc ifc" );
	exit( 1 );
	}

    ifc->ifc_len = 20 * sizeof(struct ifreq);

    buf = (char *) malloc( ifc->ifc_len );
    if ( buf == NULL )
	{
	perror( "Failure to malloc space for buf" );
	exit( 1 );
	}

    ifc->ifc_buf = buf;
    cc = ioctl( fake_socket, SIOCGIFCONF, (char *) ifc );
    if ( cc < 0 )
	{
	perror( "ioctl SIOCGIFCONF" );
	exit( 1 );
	}

    maxintf = ifc->ifc_len / sizeof(struct ifreq);
    if ( maxintf > 20 )
	maxintf = 20;

    ifreqp = (struct ifreq *)buf;
    intfcnt = 0;

    for ( i=0; i<maxintf; i++, ifreqp++ )
	{
	bcopy( &ifreqp->ifr_addr.sa_data[2], &addr, 4 );
	if ( ( addr & 0xff000000 ) == 0x7f000000 )
	    continue;

	net_intf[intfcnt].addr = addr;
	if ( IN_CLASSA( addr ) )
	    net_intf[intfcnt].nmask = IN_CLASSA_NET;
	else
	if ( IN_CLASSB( addr ) )
	    net_intf[intfcnt].nmask = IN_CLASSB_NET;
	else
	if ( IN_CLASSC( addr ) )
	    net_intf[intfcnt].nmask = IN_CLASSC_NET;
	else
	    net_intf[intfcnt].nmask = 0;

	bcopy( ifreqp, &ifrscr, sizeof(ifrscr) );
	cc = ioctl( fake_socket, SIOCGIFNETMASK, &ifrscr );
	if ( cc == -1 )
	    {
	    perror( "ioctl SIOCGIFNETMASK" );
	    continue;
	    }
	bcopy( &ifrscr.ifr_addr.sa_data[2], &addr, 4 );
	net_intf[intfcnt].smask = addr;
	intfcnt++;
	}

    fprintf( stderr, " em_arpwatch: started, %d interface(s)\n", intfcnt );
    close( fake_socket );
    return ( ifc );
}

