/*
 *  em_bp.c  DCM Wood	92/12/03
 */

/*
 *	Broadcast Ping Explorer Module
 *
 *  Get addresses of all interfaces on current system
 *  Determine connected network, subnet masks, and subnets
 *
 *  Set parameters from environment
 *  Set parameters from command line
 *  If Journal server location not set from env or cl,
 *    Get address from getpeername
 *  If input file specified,
 *    Open input file with  fd=0
 *  Read input file
 *    Set parameters from input file
 *    If EOF or blank line,
 *      Stop parameter processing
 *  If output file specified,
 *    Open output file with  fd=1
 */

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/file.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netdb.h>

#include "send_data.h"
#define SOCKADDR_IN
#include "data_server.h"

#define EM_MAIN
#include "em.h"

/*
#ifndef PIN
#include "pin.h"
#endif

char	pin[20] = PIN;
*/

extern	char	pin[];

#define	DEFDATALEN	(64 - 8)	/* default data length */
#define ETHERMIN	(60-14)
#define	MAXIPLEN	60
#define	MAXICMPLEN	76
#define	MAXPACKET	(65536 - 60 - 8)/* max packet size */
#define	MAXWAIT		30		/* max seconds to wait for response */
#define MAXADDR		1000		/* max number of received addresses */

#define IP_HDR_LEN	20
#define ICMP_HDR_LEN	8

struct sockaddr_in	from;
struct sockaddr whereto;	/* who to ping */
struct sockaddr_in * to = (struct sockaddr_in *)&whereto;
int	rsock;			/* receive socket file descriptor */
int	ssock;			/* send socket file descriptor */
char *	hostname;
int	ident;			/* process id to identify our packets */

u_long	rcvaddr[MAXADDR];	/* Source addresses of received packets */
int	rcvindex;		/* Index into rcvaddr array */

	/* counters */
int	bp=0;			/* # of updates sent to Journal Server */
long	nreceived = 0;		/* # of packets we got back */
long	ntransmitted;		/* sequence # for outbound packets = #sent */
int	interval = 10;		/* interval between packets */

	/* timing */
int	timing= 1;		/* flag to do timing */
long	tmin = -1;		/* minimum round trip time */
long	tmax;			/* maximum round trip time */
u_long	tsum;			/* sum of all times, for doing average */

int	finish();

char	addrstr[20];
int	datalen;
u_char * datap;
u_long	defmask;
u_long	destsnet;
struct in_addr	jsa;
int	jsp;
char	maskstr[20];
u_char * outpack;
u_char * packet;
int	packlen;
struct in_addr thishost;
struct timeval	ts;
struct timeval	tv;
struct timezone	tz;

extern char ** environ;

u_long	em_StartTime;
u_long	em_alb=0;			/* addr lower bound */
u_long	em_aub=0;			/* addr upper bound */
char	em_inputf[64]="\n";
int	em_ipi=5;			/* interpacket interval */
int	em_max_retries=2;		/* retry count */
int	em_ttl=0;
int	em_max_ttl=10;
char	em_outputf[64]="";		/* output file */
int	em_ps=0;			/* packet size */
int	em_runtime;
int	em_snm=0;			/* mask request */
u_long	em_subnet=0;			/* subnet to examine */
u_long	js_addr=0;			/* journal server address */
int	js_port=DATA_PORT;

FILE *	em_ifp;

struct dest {
	struct in_addr subnet;
	struct in_addr mask;
	struct dest * next;
	};

struct dest * destp;
struct dest * destFirst=NULL;

/* various options */

#define	F_NUMERIC	0x004
#define	F_DEBUG 	0x100
#define F_ADDRONLY	0x400

int	options= F_NUMERIC | F_ADDRONLY;


/*
 *  AddDest
 *	Add sub-network and mask to list of destinations
 */

AddDest( addr, mask )
u_long	addr;
u_long	mask;
{
    u_long  subnet;
    struct dest * destp;
    struct dest * lastp;

    subnet = addr & mask;

    if ( destFirst == NULL )
	{
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	destFirst = destp;
	}
    else
	{
	for ( destp=destFirst; destp!=NULL; destp=destp->next )
	    {
	    if ( ( destp->subnet.s_addr & mask ) == subnet )
		return;

	    lastp = destp;
	    }
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	lastp->next = destp;
	}

    destp->subnet.s_addr = addr;
    destp->mask.s_addr = mask;
    destp->next = NULL;
    return;
}


/*
 *  GetIntf
 *	Get info for all interfaces on this system
 */

GetIntf( addflag )
int	addflag;
{
    int  fd;
    int  i;
    struct ifconf  ifc;
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    u_long ogw;

    /* Get some memory in which to store the ifreq entries	*/

    ifc.ifc_len = 20 * sizeof( struct ifreq );

    ifc.ifc_req = (struct ifreq *)calloc( 20, sizeof( struct ifreq ) );

    /* Open a socket to do ioctls on.				*/

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
	{
	perror( "socket" );
	return ( -1 );
	}

    /* Do the ioctl, check the result.				*/

    if ( ioctl( fd, SIOCGIFCONF, &ifc ) == -1 )
	{
	perror( "ioctl SIOCGIFCONF" );
	close( fd );
	return ( -1 );
	}

    if ( ifc.ifc_len/sizeof( struct ifreq ) == 1 )
	if ( debug )
	    printf( " em_bp: There is only one interface on this system\n" );
    else
	if ( debug )
	    printf( " em_bp: There are %d interfaces on this system\n",
					ifc.ifc_len/sizeof( struct ifreq ) );
    ifreqp = ifc.ifc_req;
    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
	{
	bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
	if ( debug )
	    printf( " em_bp: %s   %-16s", ifreqp->ifr_name,
							inet_ntoa( inaddr ) );
	if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
	    {
	    perror( "ioctl SIOCGIFNETMASK" );
	    continue;
	    }
	bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
	strcpy( maskstr, inet_ntoa( inmask ) );
	if ( debug )
	    printf( "   %-16s\n", maskstr );
	if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
	    continue;

	if ( defmask == 0 )
	    defmask = inmask.s_addr;
	else
	    if ( defmask != inmask.s_addr )
		printf( " em_bp: Interface mask (%s) != default mask (%s)\n",
						maskstr, inet_ntoa( defmask ) );
	if ( addflag == 0 )
	    AddDest( inaddr.s_addr, inmask.s_addr );
	}
    close( fd );
    return( 0 );
}


/*
 *  ArgBit
 *	Process bit flag argument
 */

ArgBit( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep |= *altdflt;
    return( 0 );
}


/*
 *  ArgCS
 *	Process Character String argument
 */

ArgCS( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    strcpy( (char *)valuep, altdflt );
    return( 0 );
}


/*
 *  ArgInt
 *	Process Integer argument
 */

ArgInt( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = atoi( altdflt );
    return( 0 );
}


/*
 *  ArgIP
 *	Process Internet Protocol Address argument
 */

ArgIP( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = inet_addr( altdflt );
    if ( *valuep == -1 )
	return ( -1 );

    return( 0 );
}


/*
 *  ProcessCL
 *	Process Command Line arguments
 */

ProcessCL( ac, av )
int	ac;
char ** av;
{
    struct arg
	{
	int	(*func)();
	char	flag;
	char *	text;
	int *	value;
	char *	altdflt;
	};

    static struct arg arg[]={
    { ArgCS,	'd',	"default directory",	(int*)defdir,	 NULL	     },
    { ArgCS,	'f',	"input file",		(int*)em_inputf, ""	     },
    { ArgCS,	'o',	"output file",		(int*)em_outputf,"em_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&em_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&em_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&em_aub,		NULL },

    { ArgInt,	'h',	"ttl",			&em_ttl,	NULL	},
    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&em_max_retries,"1"	},
    { ArgInt,	'v',	"interpacket interval",	&em_ipi,	"20"	},
    { ArgInt,	't',	"max ttl",		&em_max_ttl,	"30"	},
    { ArgInt,	'z',	"packet size",		&em_ps,		NULL	},

    { ArgInt,	'm',	"mask request",		&em_snm,	"1"	},

    { ArgBit,	'a',	"address-only flag",	&options,	"F_ADDRONLY" },
    { ArgBit,	'd',	"socket-debug flag",	&options,	"F_DEBUG" },
    { ArgBit,	'n',	"numeric-only flag",	&options,	"F_NUMERIC" },

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};

    char   fc;
    int    i;
    char * progp;
    struct arg * ap;

    progp = av[0];
    ac--, av++;				/* Skip over program name */
    while ( ac && *av[0] == '-' )
	{
	fc = *++av[0];
	for ( ap=&arg[0]; ap->func!=NULL; ap++ )
	    {
	    if ( fc == ap->flag )
		{
		if ( ( ac == 1 ) || ( ( ac > 1 ) && ( *av[1] == '-' ) ) )
		    {
		    if ( ap->altdflt == NULL )
			{
			fprintf( stderr, "em_bp: %s error\n", ap->text );
			goto arg_error;
			}
		    else
			{
			i = (*ap->func)( ap->value, ap->altdflt );
			if ( i < 0 )
			    {
			    fprintf( stderr, "em_bp: %s error\n", ap->text );
			    goto arg_error;
			    }
			break;
			}
		    }
		else if ( ( ac > 1 ) && ( *av[1] != '-' ) )
		    {
		    ac--;
		    i = (*ap->func)( ap->value, av[1] );
		    if ( i < 0 )
			{
			fprintf( stderr, "em_bp: %s error\n", ap->text );
			goto arg_error;
			}
		    av++;
		    break;
		    }
		else
		    {
		    fprintf( stderr, "em_bp: %s error\n", ap->text );
		    goto arg_error;
		    }
		}  /* endif ( fc == ap->flag ) */
	    }  /* end for */
	ac--, av++;
	}  /* end while */

    arg_error:
	if ( ac != 0 )
	    {
	    fprintf( stderr, "Usage: %s\n", progp );
	    for ( i=0; arg[i].func!=NULL; i++ )
		fprintf( stderr, "	[-%c %s]\n", arg[i].flag, arg[i].text );
	    exit( 1 );
	    }

    if ( em_subnet != 0 )
	AddDest( em_subnet, -1 );

    return;
}


/*
 *  ProcessInputFile
 *	Open and read input file
 */

ProcessInputFile()
{
    int    ac;
    u_long addr;
    char * av[MAXARGS];
    int    buf[MAXLINE];
    int    cc;
    char * cp;
    int    ic;
    int    ifd;
    int    ll;
    u_long mask;
    struct sockaddr_in name;
    int    namelen;
    char * sp;

    if ( em_inputf[0] == '\n' )
	return;

    if ( em_inputf[0] == '\0' )
	{
	em_ifp = fdopen( 0, "r" );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, "em_bp: Could not fdopen input.\n" );
	    exit( 1 );
	    }

	if ( jsa.s_addr == 0 )
	    {
	    namelen = sizeof( name );
	    cc = getpeername( 0, &name, namelen );
	    if ( cc == 0 && namelen != 0 )
		if ( name.sin_addr.s_addr != 0 )
		    jsa.s_addr = name.sin_addr.s_addr;
	    }
	}
    else
	{
	em_ifp = freopen( em_inputf, "r", 0 );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr,
		    "em_bp: Could not freopen input file %s.\n", em_inputf );
	    exit( 1 );
	    }
	}

    ac = 0;
    av[0] = "em_bp";
/*
    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	ll = strlen( buf );
	if ( ll <= 1 )
	    break;

	if ( strspn( buf, " \t\n" ) == ll )
	    break;

	cp = strtok( buf, " \t\n" );

	do {
	    if ( cp == NULL )
		break;

	    cc = strlen( cp );
	    sp = (char *)malloc( cc+1 );
	    strncpy( sp, cp, cc );
	    ac++;
	    av[ac] = sp;
	    cp = strtok( NULL, " \t\n" );
	    } while ( ac < MAXARGS );

	}
    
    if ( ac > 0 )
	ProcessCL( ac, av );
 */

    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	cc = sscanf( buf, "%s %s", addrstr, maskstr );
	if ( cc == 2 )
	    {
	    mask = inet_addr( maskstr );
	    addr = inet_addr( addrstr );
	    }
	else if ( cc == 1 )
	    {
	    mask = -1;
	    addr = inet_addr( addrstr );
	    }
	else
	    {
	    mask = -1;
	    addr = -1;
	    }

	if ( addr != -1 )
	    AddDest( addr, mask );
	}

    fclose( em_ifp );
    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    int     i, j;
    struct in_addr inaddr;
    int     n;
    int     jsfd;		/* Journal Server socket file descriptor */
    FILE *  jsfp;

    if ( rcvindex == 0 )
	return;

    jsfd = getnetconnect( js_addr, js_port );
    jsfp = fdopen( jsfd, "w" );

    n = 0;
    for ( i=0; i<rcvindex; i++ )
	{
	for ( j=0; j<i; j++ )
	    if ( rcvaddr[j] == rcvaddr[i] )
		break;

	if ( i != j )
	    continue;

	inaddr.s_addr = rcvaddr[i];

	fprintf( jsfp, "INTF  %s\n", pin );
	fprintf( jsfp, "PROT-ADDR:\tAF_INET %s\n", inet_ntoa( inaddr ) );
	fprintf( jsfp, "\n\n" );
	n++;
	}

    fclose( jsfp );
    inaddr.s_addr = destsnet;
    printf( " em_bp:%4d INTF entries sent to Journal Server for subnet %s\n",
						n, inet_ntoa( inaddr ) );
    bp += n;
    return;
}


/*
 *  SockInit -
 *	Initialize send/receive sockets
 */

SockInit()
{
    int     hold;
    int     i;
    int     on = 1;
    struct protoent *proto;

    ident = getpid() & 0xFFFF;

    datalen = em_ps;
    packlen = datalen + IP_HDR_LEN + ICMP_HDR_LEN;
    if ( packlen < ETHERMIN )
	{
	datalen += ETHERMIN - packlen;
	packlen = ETHERMIN;
	}
    outpack = (u_char *)malloc( packlen + 20 );
    bzero( outpack, packlen );

    datap = outpack + IP_HDR_LEN + ICMP_HDR_LEN;
    for ( i = IP_HDR_LEN + ICMP_HDR_LEN; i < packlen; ++i )
	*datap++ = i;

    bzero( (char *)&whereto, sizeof( struct sockaddr ) );
    to->sin_family = AF_INET;
    from.sin_family = AF_INET;

    if ( ( rsock = socket( AF_INET, SOCK_RAW, IPPROTO_ICMP ) ) < 0 )
	{
	perror( "em_bp: receive socket" );
	exit( 1 );
	}

    hold = 1;
    if ( options & F_DEBUG )
	setsockopt( rsock, SOL_SOCKET, SO_DEBUG,
						(char *)&hold, sizeof( hold ) );
#ifdef SO_RCVBUF
    hold = 48 * 1024;
    setsockopt( rsock, SOL_SOCKET, SO_RCVBUF,
						(char *)&hold, sizeof( hold ) );
#endif SO_RCVBUF

    if ( ( ssock = socket( AF_INET, SOCK_RAW, IPPROTO_RAW ) ) < 0 )
	{
	perror( "em_bp: send socket" );
	exit( 1 );
	}

#ifdef SO_SNDBUF
    if ( setsockopt( ssock, SOL_SOCKET, SO_SNDBUF, (char *)&packlen,
						sizeof( packlen ) ) < 0 )
	{
	perror( "em_bp: SO_SNDBUF" );
	exit( 6 );
	}
#endif SO_SNDBUF

#ifdef IP_HDRINCL
    if ( setsockopt( ssock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
						sizeof( on ) ) < 0 )
	{
	perror( "em_bp: IP_HDRINCL" );
	exit( 6 );
	}
#endif IP_HDRINCL

    if ( options & SO_DEBUG )
	 setsockopt( ssock, SOL_SOCKET, SO_DEBUG,
						(char *)&on, sizeof( on ) );
#ifndef IP_HDRINCL
	if ( bind( ssock, (struct sockaddr *)&from, sizeof( from ) ) < 0 )
	    {
	    perror ( "em_bp: bind:" );
	    exit ( 1 );
	    }
#endif IP_HDRINCL

    if ( ! ( packet = (u_char *)malloc( (u_int)packlen + 20 ) ) )
	{
	fprintf( stderr, "em_bp: out of memory.\n" );
	exit( 1 );
	}

    return;
}


main( argc, argv )
int	argc;
char **	argv;
{
    int     ch;
    u_long  destaddr;
    int     host;
    int     multiple_replies;
    int     nreplies;
    int     one_reply;
    int     retry;
    int     ettl=0;
    int     rttl;
    int     sttl=1;
    int     ttl;

    setlinebuf( stdout );

    gettimeofday( &tv, &tz );
    em_StartTime = tv.tv_sec;
    printf( " %s: started\n", argv[0] );

    ProcessCL( argc, argv );

    ProcessInputFile();

    GetIntf( destFirst );

    /*  Start of code to do Exploring	*	*	*	/* */

    SockInit();

    signal( SIGINT, finish );

    for ( destp=destFirst; destp!=NULL; destp=destp->next )
	{
	if ( em_ttl != 0 )
	    sttl = ettl = em_ttl;
	else
	    {
	    sttl = 1;
	    ettl = em_max_ttl;
	    }
	multiple_replies = 0;
	one_reply = 0;
	rttl = 0;
	for ( ttl=sttl; ttl<=ettl; ttl++ )
	    {
	    for ( host=-1; host<=0; host++ )
		{
		if ( destp->mask.s_addr == -1 )
		    destp->mask.s_addr = defmask;
		destsnet = destp->subnet.s_addr & destp->mask.s_addr;
		destaddr = ( destp->subnet.s_addr & destp->mask.s_addr ) |
						( host & ~ destp->mask.s_addr );
	    	for ( retry=0; retry<em_max_retries; retry++ )
		    {
		    SendPing( destaddr, ttl );
		    nreplies = CollectReplies( em_ipi );
		    if ( nreplies > 0 )
			one_reply++;
		    if ( nreplies > 1 )
			multiple_replies++;

		    }  /* retry  */

		}  /*  host  */

	    if ( rttl > 0 )
		break;
	    if ( one_reply )
		rttl++;
	    if ( multiple_replies )
		break;
	    }  /*  ttl  */

	UpdateJS();
	rcvindex = 0;
	}  /* destp  */

    finish();
    /* NOTREACHED */

    /*  End of code to do Exploring	*	*	*	/* */

}


CollectReplies( waittime )
int	waittime;
{
    int     cc;
    fd_set fds;
    struct sockaddr_in from;
    int     fromlen = sizeof( from );
    int     ic;
    int     nreplies=0;
    struct timeval wait;

    FD_ZERO( &fds );
    FD_SET( rsock, &fds );

    while ( TimeToWait( &ts, waittime, &wait ) > 0 )
	{
	ic = select( rsock+1, &fds, (fd_set *)0, (fd_set *)0, &wait );
	if ( ic < 0 )
	    {
	    perror( "em_bp: select error" );
	    exit( 1 );
	    }

	if ( ic == 0 )
	    {
	    if ( debug )
		printf( " em_bp:   %d replies\n\n", nreplies );
	    return( nreplies );
	    }

	if ( ic > 0 )
	    {
	    cc = recvfrom( rsock, (char *)packet, packlen, 0,
				(struct sockaddr *)&from, &fromlen );
	    if ( cc < 0 )
		{
		perror( "em_bp: recvfrom" );
		finish();
		}

	    ic = pr_pack( (char *)packet, cc, &from );
	    if ( ic == 0 )
		nreplies++;
	    }
	}
    if ( debug )
	printf( " em_bp:   %d replies\n\n", nreplies );
    return( nreplies );
}


/*
 * SendPing --
 *	Compose and transmit an ICMP ECHO REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */

SendPing( addr, ttl )
u_long	addr;
int	ttl;
{
    struct ip *ip = (struct ip *)outpack;
    register struct icmp * icp = (struct icmp *)(outpack + IP_HDR_LEN);
    register int cc;
    int i;

    ip->ip_v = 4;
    ip->ip_hl = 5;
    ip->ip_tos = 0;
    ip->ip_len = packlen;
    ip->ip_id = 0;
    ip->ip_off = 0;
    ip->ip_ttl = ttl;
    ip->ip_p = IPPROTO_ICMP;
    ip->ip_sum = 0;
    ip->ip_src.s_addr = 0;
    ip->ip_dst.s_addr = addr;

    icp->icmp_type = ICMP_ECHO;
    icp->icmp_code = 0;
    icp->icmp_cksum = 0;
    icp->icmp_id = ident;			/* ID */
    icp->icmp_seq = ntransmitted++;

    to->sin_addr.s_addr = addr;

    gettimeofday( &ts, &tz );
    bcopy( &ts, outpack+IP_HDR_LEN+ICMP_HDR_LEN, sizeof( struct timeval ) );

    cc = packlen;

    /* compute ICMP checksum here */
    icp->icmp_cksum = in_cksum( (u_short *)icp, cc );

    i = sendto( ssock, (char *)outpack, cc, 0, &whereto,
						sizeof( struct sockaddr ) );
    if ( i < 0 || i != cc )
	{
	if ( i < 0 )
	    perror( "em_bp: sendto" );
	fprintf( stderr, "em_bp: wrote %s %d chars, ret=%d\n",
							hostname, cc, i );
	}
    if ( debug )
	printf( " em_bp:   %s sent with ttl %d\n",
					inet_ntoa( to->sin_addr ), ttl );
    return;
}


/*
 * pr_pack --
 *	Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */

pr_pack( buf, cc, from )
char *buf;
int cc;
struct sockaddr_in *from;
{
    register struct icmp *icp;
    register u_long l;
    register int i;
    register int j;
    register u_char * cp;
    register u_char * dp;
    static int old_rrlen;
    static char old_rr[MAX_IPOPTLEN];
    struct ip *ip;
    struct timeval   tv;
    struct timeval * tp;
    long triptime;
    int hlen;

    gettimeofday( &tv, (struct timezone *)NULL );

    /* Check the IP header */

    ip = (struct ip *)buf;
    hlen = ip->ip_hl << 2;
    if ( cc < hlen + ICMP_MINLEN )
	{
	if ( options & F_DEBUG )
	    fprintf( stderr, "em_bp: packet too short (%d bytes) from %s\n", cc,
			inet_ntoa(*(struct in_addr *)&from->sin_addr.s_addr ) );
	return( -1 );
	}

    /* Now the ICMP part */

    cc -= hlen;
    icp = (struct icmp *)( buf + hlen );
    if ( icp->icmp_type == ICMP_ECHOREPLY )
	{
	if ( icp->icmp_id != ident )
	    return( -1 );			/* 'Twas not our ECHO */

	++nreceived;
	rcvaddr[rcvindex] = from->sin_addr.s_addr;
	rcvindex++;
	if ( rcvindex >= MAXADDR )
	    {
	    fprintf( stderr, "em_bp: Too many received addresses\n" );
	    finish();
	    }

	if ( timing )
	    {
#ifndef icmp_data
	    tp = (struct timeval *)&icp->icmp_ip;
#else
	    tp = (struct timeval *)icp->icmp_data;
#endif
	    tvsub( &tv, tp );
	    triptime = tv.tv_sec * 1000 + ( tv.tv_usec / 1000 );
	    tsum += triptime;
	    if ( triptime < tmin )
		tmin = triptime;
	    if ( triptime > tmax )
		tmax = triptime;
	    }

	if ( options & F_ADDRONLY )
	    {
	    if ( debug )
/*
 *		printf( " em_bp: %s",
 *				    inet_ntoa(
 *				    *(struct in_addr *)
 *				    &from->sin_addr.s_addr ) );
 */
		printf( " em_bp: %s", inet_ntoa( from->sin_addr ) );
	    }
	else
	    {
	    if ( debug )
		printf( " em_bp: %d bytes from %s: icmp_seq=%u",
				    cc, inet_ntoa(
				    *(struct in_addr *)
				    &from->sin_addr.s_addr ), icp->icmp_seq );
	    if ( debug )
		printf( " ttl=%d", ip->ip_ttl );

	    if ( timing )
		if ( debug )
		    printf( " time=%ld ms", triptime );
	    }

	/* check the data */

	cp = packet + IP_HDR_LEN + ICMP_HDR_LEN;
	dp = outpack + IP_HDR_LEN + ICMP_HDR_LEN;
	for ( i = IP_HDR_LEN + ICMP_HDR_LEN; i < packlen; ++i, ++cp, ++dp )
	    {
	    if ( *cp != *dp )
		{
		printf( "\n wrong data byte #%d should be 0x%x but was 0x%x",
			i, *dp, *cp );
		cp = (u_char*)&icp->icmp_data[0];
		for ( i = IP_HDR_LEN + ICMP_HDR_LEN; i < packlen; ++i, ++cp )
		    {
		    if ( (i % 32 ) == 8 )
			printf( "\n\t" );
		    printf( "%x ", *cp );
		    }
		break;
		}
	    }
	}
    else
	{
	/* We've got something other than an ECHOREPLY */
	return( -1 );
	}
    if ( debug )
	printf( "\n" );
    return( 0 );
}


/*
 * in_cksum --
 *	Checksum routine for Internet Protocol family headers (C Version)
 */

in_cksum( addr, len )
u_short *addr;
int len;
{
    register int nleft = len;
    register u_short *w = addr;
    register int sum = 0;
    u_short answer = 0;

    /*
     * Our algorithm is simple, using a 32 bit accumulator (sum), we add
     * sequential 16 bit words to it, and at the end, fold back all the
     * carry bits from the top 16 bits into the lower 16 bits.
     */

    while ( nleft > 1 )
	{
	sum += *w++;
	nleft -= 2;
	}

    /* mop up an odd byte, if necessary */
    if ( nleft == 1 )
	{
	*(u_char *)(&answer) = *(u_char *)w ;
	sum += answer;
	}

    /* add back carry outs from top 16 bits to low 16 bits */

    sum = ( sum >> 16 ) + ( sum & 0xffff );	/* add hi 16 to low 16 */
    sum += ( sum >> 16 );			/* add carry */
    answer = ~sum;				/* truncate to 16 bits */
    return( answer );
}


/*
 * tvsub --
 *	Subtract 2 timeval structs:  out = out - in.  Out is assumed to
 * be >= in.
 */

tvsub( out, in )
register struct timeval * out;
register struct timeval * in;
{
    if ( ( out->tv_usec -= in->tv_usec ) < 0 )
	{
	--out->tv_sec;
	out->tv_usec += 1000000;
	}
    out->tv_sec -= in->tv_sec;
}


/*
 * TimeToWait --
 *	Return a pointer to a new timeval for waiting and
 *	the amount of time yet to wait.
 */

TimeToWait( past, duration, wait )
struct timeval * past;
int	duration;
struct timeval * wait;
{
    struct timeval scr;

    scr.tv_sec = duration;
    scr.tv_usec = 0;
    gettimeofday( wait, &tz ) ;
    tvsub( wait, past );
    if ( ( wait->tv_usec = 0 - wait->tv_usec ) < 0 )
	{
	wait->tv_sec++;
	wait->tv_usec += 1000000;
	}
    return( wait->tv_sec = duration - wait->tv_sec );
}


/*
 * finish --
 *	Print out statistics, and give up.
 */

finish()
{
    signal( SIGINT, SIG_IGN );

    UpdateJS();

    fprintf( stderr, " em_bp:%4d INTF entries sent to Journal Server\n", bp );
    if ( ! ( options & F_ADDRONLY ) )
	{
	printf( "--- %s em_bp statistics ---\n", hostname );
	printf( "%ld packets transmitted, ", ntransmitted );
	printf( "%ld packets received, ", nreceived );
	putchar( '\n' );
	if ( nreceived && timing )
	    printf( "round-trip min/avg/max = %ld/%lu/%ld ms\n",
			    tmin, tsum / nreceived, tmax );
	}

    gettimeofday( &tv, &tz ) ;
    printf( " em_bp took %d seconds\n\n", tv.tv_sec - em_StartTime );
    close( 0 );
    return;
}


usage()
{
    fprintf( stderr, "usage: em_bp [-Rdfnqrvwa] [-c count] [-i wait]\n" );
    fprintf( stderr, "\t[-p pattern] [-s packetsize] host\n" );
    exit( 1 );
}

