/************************************************************************
 *									*
 *	File:  etherhostprobe.c						*
 *									*
 *	ETHERnet HOST PROBE routine.  Probes the ethernet (by forcing	*
 *	ARP's) for the specified range of IP addresses.			*
 *	This program is unable to probe the IP address of the host on	*
 *	which it is running.						*
 *									*
 *	Usage:								*
 *		etherhostprobe begin end				*
 *	where `begin' is the first IP address (or host name) in the	*
 *	range to be probed, and `end' is the last IP address (or host	*
 *	name in the range to be probed.					*
 *									*
 *	Written 25-Nov-87 by Paul E. McKenney, SRI International.	*
 *	Copyright (c) 1987 SRI International.				*
 *									*
 ************************************************************************/
 /* %W% %D% */

/*
 *	$Log:	etherhostprobe.c,v $
 * Revision 1.1  87/12/02  08:12:34  mckenney
 * Initial revision
 * 
 */

/* Include files.							*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netdb.h>

#define SOCKADDR_IN
#include "data_server.h"
#include "send_data.h"

/* Type definitions local to this file.					*/

#define INCR_IP(ip_addr) \
	{ \
	if ((++ip_addr.s_b4 == 0) && \
	    (++ip_addr.s_b3 == 0) && \
	    (++ip_addr.s_b2 == 0)) \
		ip_addr.s_b1++; \
	}

/* Functions exported from this file.					*/

/* Functions local to this file.					*/

/* Variables exported from this file.					*/

int	debug=0;

/* Variables local to this file.					*/



main(argc, argv)

int     argc;
char   *argv[];

{
    struct sockaddr_in bsa;	/* Beginning Socket Address.	 */
    struct ether_addr ea;	/* Ethernet Address.		 */
    char   *em;			/* Error Message pointer.	 */
    struct sockaddr_in esa;	/* Ending Socket Address.	 */
    int     i;
    int     intfcnt=0;		/* # of entries sent to JS	 */
    int     retval;		/* RETurn VALue.		 */
    struct sockaddr_in sa;	/* current Socket Address.	 */
    struct data_record *data;
    struct data_record *make_data();
    extern char *first_node;

    /* Check command line arguments.				 */

    if (argc > 3)
	{
	fprintf(stderr, "Usage: %s [begin-IP [end_IP]]\n", argv[0]);
	exit(-1);
	}

    /* Translate beginning IP address into socket address struct.	 */

    bzero((caddr_t) & bsa, sizeof(bsa));
    if (argc > 1)
	{
	switch (tx_hostname(argv[1], &bsa, &em))
	    {
	    case -2:
		perror(em);
		exit(-1);
	    case -1:
		fprintf(stderr, "%s", em);
		exit(-1);
	    }
	}

    /* Translate ending IP address into socket address struct.	 */

    bzero((caddr_t) & esa, sizeof(esa));
    if (argc > 2)
	{
	switch (tx_hostname(argv[2], &esa, &em))
	    {
	    case -2:
		perror(em);
		exit(-1);
	    case -1:
		fprintf(stderr, "%s", em);
		exit(-1);
	    }
	}

    /* Set limits from interface if not already set */

    if (bsa.sin_addr.s_addr == 0)
	GetIntf(&bsa, &esa);

    /* Initialize for scan loop.					 */

    bcopy((caddr_t) & bsa, (caddr_t) & sa, sizeof(sa));

    /* Each pass through the following loop looks for one IP	 */
    /* address, with up to one second delay.  This means 18 hours	 */
    /* to completely scan a class B network.			 */

#define SA_IP	(sa.sin_addr.S_un.S_un_b)
#define ESA_IP	(esa.sin_addr.S_un.S_un_b)
    while ( ( SA_IP.s_b1 < ESA_IP.s_b1 ) ||
	    ( ( SA_IP.s_b1 == ESA_IP.s_b1 ) &&
	      ( ( SA_IP.s_b2 < ESA_IP.s_b2 ) ||
		( ( SA_IP.s_b2 == ESA_IP.s_b2 ) &&
		  ( ( SA_IP.s_b3 < ESA_IP.s_b3 ) ||
		    ( ( SA_IP.s_b3 == ESA_IP.s_b3 ) &&
		      ( SA_IP.s_b4 <= ESA_IP.s_b4 ) ) ) ) ) ) )
	{

	/* Initiate the arp.  If this fails immediately, ditch	 */
	/* and go to the next address.				 */

	if ((retval = tx_arp_init(&sa, &ea, &em)) < 0)
	    {
	    INCR_IP(SA_IP);
	    continue;
	    }

	/* If we did not get an immediate answer, wait a bit	 */
	/* and check again.					 */

	if (retval != 1)
	    for (i = 0; i < 4; i++)
		{
		fastsleep(0L, 250000L);
		if ((retval = tx_arp_chk(&sa, &ea, &em)) == 1)
		    break;
		}

	/* If we got something, print it out.			 */

	if ( retval == 1 )
	    {
	    struct hostent *hostent;
	    u_char *hostname;
	    struct ether_addr *data_ea;
	    struct sockaddr_in *data_in;

	    hostent = gethostbyaddr( &sa.sin_addr,
					sizeof(sa.sin_addr), AF_INET);
	    if ( hostent != NULL )
		{
		hostname = (u_char *) malloc(strlen(hostent->h_name) + 1);
		if ( hostname == NULL )
		    {
		    fprintf( stderr, "Unable to allocate space for hostname\n");
		    exit(1);
		    }
		}
	    strcpy(hostname, hostent->h_name);

	    if ( debug )
		(void)fprintf( stderr, "%-17s %-15s %s\n",
			ethertoa(&ea), inet_ntoa(sa.sin_addr), hostname);
	    data = make_data(INTF);
	    data->dr_type.intf->ip_addr = (struct intf_prot_addr *)
					malloc(sizeof(struct intf_prot_addr));
	    if ( data->dr_type.intf->ip_addr == NULL )
		{
		fprintf( stderr, "Unable to allocate space for ip_addr\n");
		exit(1);
		}
	    data->dr_type.intf->in_mac_addr = (struct intf_mac_addr *)
					malloc(sizeof(struct intf_mac_addr));
	    if ( data->dr_type.intf->in_mac_addr == NULL )
		{
		fprintf( stderr, "unable to allocate memory for in_mac_addr\n");
		exit(1);
		}
	    data_ea = (struct ether_addr *) malloc(sizeof(struct ether_addr));
	    if ( data_ea == NULL )
		{
		fprintf( stderr, "unable to allocate space for data_ea\n");
		exit(1);
		}
	    data_in = (struct sockaddr_in *) malloc(sizeof(struct sockaddr_in));
	    if ( data_in == NULL )
		{
		fprintf( stderr, "Unable to allocate space for data_in\n");
		exit(1);
		}

	    bcopy( &ea, data_ea, sizeof(struct ether_addr) );
	    bcopy( &sa, data_in, sizeof(struct sockaddr_in) );

	    data->dr_type.intf->in_mac_addr->mac_addr = data_ea;

	    /* data->in_name = hostname; */

	    data->dr_type.intf->ip_addr->addr_family = AF_INET;
	    data->dr_type.intf->ip_addr->pa = (struct sockaddr_in *) data_in;
	    pack_data(data);
	    fflush(stdout);
	    intfcnt++;
	    }

	/* Increment to the next IP address.			 */

	INCR_IP(SA_IP);
	}

    /* When done, leave!  */

    send_data(first_node, NULL);

    fprintf( stderr,
	" etherhostprobe:  %d INTF entries sent to Journal Server\n", intfcnt );

    exit(0);
}

