
#ifndef lint
static char sccsid[] = "@(#)data.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  data.c --
 *
 *	Miscellaneous subroutines for accessing the data structures
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>
#include "res.h"

#define DATA
#include "data.h"


/*
 *  fncmp --
 *	Full Name Compare
 */

fncmp( nap, name )
struct NameAddr * nap;
char *		  name;
{
    int cc, i, j;

    for ( i=0; name[i] != '.' && nap->name[i] != NULL; i++ )
	{
	cc = nap->name[i] - name[i];
	if ( cc != 0 )
	    return( cc );
	}
    if ( name[i] != '.' )
	return ( 1 );

    if ( nap->name[i] != NULL )
	return ( -1 );

    for ( j=0, i++; name[i] != NULL && nap->domain[j] != NULL; i++, j++ )
	{
	cc = nap->domain[j] - name[i];
	if ( cc != 0 )
	    return( cc );
	}
    if ( name[i] == NULL && nap->domain[j]  == NULL )
	return( 0 );
    
    if ( name[i] == '.' &&
	    name[i+1] == NULL &&
	    nap->domain[j]  == NULL )
	return( 0 );

    if ( name[i] == NULL &&
	    nap->domain[j]  == '.' &&
	    nap->domain[j+1]  == NULL )
	return( 0 );

    return( nap->domain[j] - name[i] );
}


/*
 *  fncpy --
 *	Full Name Copy
 */

fncpy( nap, name )
struct NameAddr * nap;
char *		  name;
{
    int cc, i, j;

    for ( i=0; name[i] != '.'; i++ )
	 nap->name[i] = name[i];

    for ( j=0, i++; name[i] != NULL; i++, j++ )
	nap->domain[j] = name[i];

    if ( nap->domain[j--] == '.' )
	nap->domain[j] = '\0';
    return;
}


/*
 *  CreateDomainEnt --
 *
 *	Create a Domain Entity to the DomainTree structure
 */

DomainEntP
CreateDomainEnt( dnp )
char * dnp;
{
    char * cp;
    int    domainlength;
    struct DomainEnt * dp;
    int i;

    DomainEntCnt++;
    dp = (struct DomainEnt *)Malloc( sizeof(struct DomainEnt) );
    dp->CreationTime = now;
    dp->Next = NULL;
    dp->Down = NULL; 
    for ( i=0; i<5; i++ ) 
	dp->NS[i] = NULL;

    domainlength = strlen( dnp );
    domainlength++;
    cp = Malloc( domainlength );
    dp->name = cp;
    strcpy( cp, dnp );

    return( dp );
}


/*
 *  AddDomainEnt --
 *
 *	Add a Domain Entity to the DomainTree structure
 */

DomainEntP
AddDomainEnt( dnp )
char * dnp;
{
    char * cp;
    struct DomainEnt * dep;
    struct DomainEnt * dp;
    int i, match;

    if ( DTree == NULL )
	{
	dp = CreateDomainEnt( dnp );
	DTree = dp;
	return( dp );
	}

    for ( dep = DTree; dep->Next != NULL; dp = dep, dep = dep->Next )
    	if ( strcmp( dnp, dep->name ) == 0 )
	    return( dep );

    if ( strcmp( dnp, dep->name ) == 0 )
	return( dep );

    dp = CreateDomainEnt( dnp );
    dep->Next = dp;
    return( dp );
}


/*
 *  CreateNameAddrEntry --
 */

NameAddrP
CreateNameAddrEntry( name, addr )
char * name;
u_long addr;
{
    struct in_addr inaddr;
    int    dl;
    char * dn;
    struct DomainEnt * dp;
    char * ep;
    int    i, j;
    int    namelength;
    struct NameAddr * np;
    char * nnp;

    ep = strchr( name, '.' );
    if ( ep == NULL )
	return( NULL );

    namelength = ep - name;
    dn = ep + 1;
    dl = strlen( dn );
    if ( dn[dl-1] == '.' )
	dn[dl-1] = '\0';

    np = (struct NameAddr *)Malloc( sizeof(struct NameAddr) );

    np->CreationTime = now;
    np->flags = 0;
    np->addr.s_addr = addr;
    np->Next = NULL;
    np->List = NULL;
    np->LO = NULL;
    np->NextIntf = NULL;

    nnp = (char *)Malloc( namelength );
    np->name = nnp;
    strncpy( nnp, name, namelength );
    nnp[namelength] = '\0';

    dp = AddDomainEnt( dn );
    np->domain = dp->name;

    if ( addr != 0 )
	{
        NameAddrEntCnt++;
	if ( NAListp == NULL )
	    NAListp = np;
	else
	    NAListEndp->List = np;
	NAListEndp = np;
	inaddr.s_addr = addr;
	fprintf( dbgfp, "%-16s %s\n", inet_ntoa( inaddr ), name );
	}
    return( np );
}


/*
 *  AddNATbl --  Add name and address to table
 *
 *	If addr is not in "this" network,
 *	    Return
 *
 *	If no pointers with this hash,
 *	  Create new entry
 *	  Set name and addr hash pointers
 *	  Return
 *
 *	If name but no addr pointer,
 *	  Move to end of list
 *	    If matching addr before the end,
 *	      Complain
 *	      Set addr hash pointer from name hash pointer
 *	      Return
 *	  Create new entry
 *	  Set addr hash pointer from name hash pointer
 *	  Set "Next" pointer
 *	  Return
 *	    
 *	If addr but no name pointer,
 *	  Move to end of list
 *	    If matching name before the end,
 *	      Complain
 *	      Set name hash pointer from addr hash pointer
 *	      Return
 *	  Create new entry
 *	  Set name hash pointer from addr hash pointer
 *	  Set "Next" pointer
 *	  Return
 *	    
 *	If addr and name pointers,
 *	  Move to addr entry
 *	  Move to end list
 *	    If matching name and addr, remember that
 *	    If matching name pointer, remember that
 *	  If matching name pointer not found,
 *	    Move to name entry
 *	    Move to end of list
 *	      If matching name and addr, remember that and stop
 *	      If matchine addr pointer, stop
 *	  If no matching name and addr,
 *	    Create new entry
 *	    Set last "Next" pointer
 *	    Return
 */

NameAddrP
AddNATbl( name, addr )
char * name;
u_long addr;
{
    u_long  namehash;
    u_long  addrhash;
    int i;
    int match;
    char * cp;
    struct NameAddr * bp;		/* Beginning pointer */
    struct NameAddr * ep;		/* Ending pointer    */
    struct NameAddr * matchp;		/* Match pointer     */
    struct NameAddr * np;		/* returned from CreateNameAddrEntry */
    struct NameAddr * xp;		/* Moving pointer    */

    if ( ( ( addr ^ CurNet.s_addr ) & CurMask.s_addr ) != 0 )
	return( NULL );			/* Not in same net */

    cp = name + strlen( name ) - 1;
    if ( *cp == '.' )
	*cp = '\0';
    for ( namehash=0, cp=name; *cp != '.' && *cp != '\0'; cp++ )
	namehash = ( namehash << 2 ) | ( *cp & 0x3 );
    
    namehash %= HASH_TABLE_SIZE;

/*
    for ( addrhash=0, cp=(char *)&addr, i=0; i<4 ; cp++, i++ )
	addrhash += *cp & 0xff;
 */
    addrhash = addr & ~( CurMask.s_addr );
    
    addrhash %= HASH_TABLE_SIZE;

    if ( NameHashTbl[namehash] == NULL && AddrHashTbl[addrhash] == NULL )
	{
	np = CreateNameAddrEntry( name, addr );
	NameHashTbl[namehash] = AddrHashTbl[addrhash] = np;
	return( np );
	}

    if ( NameHashTbl[namehash] == NULL )
	{ /* addr pointer, but no name pointer */
	bp = xp = AddrHashTbl[addrhash];
	while ( ! ( match =
			( xp->addr.s_addr == addr &&
			  fncmp( xp, name ) == 0 ) ) &&
		  ( xp->Next != NULL ) )
	    xp = xp->Next;
	if ( ! match )
	    {
	    np = CreateNameAddrEntry( name, addr );
	    xp->Next = np;
	    }
	NameHashTbl[namehash] = np;
	return( np );
	}

    if ( AddrHashTbl[addrhash] == NULL )
	{ /* name pointer, but no addr pointer */
	bp = xp = NameHashTbl[namehash];
	while ( ! ( match =
			( xp->addr.s_addr == addr &&
			  fncmp( xp, name ) == 0 ) ) &&
		  ( xp->Next != NULL ) )
	    xp = xp->Next;
	if ( ! match )
	    {
	    np = CreateNameAddrEntry( name, addr );
	    xp->Next = np;
	    }
	AddrHashTbl[addrhash] = np;
	return( np );
	}

    /* Then neither is NULL */

    bp = xp = NameHashTbl[namehash];
    matchp = NULL;
    while ( xp != NULL )
	{
/*
	if ( xp != bp )
	    if ( AddrHashTbl[addrhash] == xp )
		{
		AddrHashTbl[addrhash] = bp;
		fprintf( stderr, "moved AddrHashTbl entry %s %s\n",
			AddrHashTbl[addrhash]->name,
			inet_ntoa( AddrHashTbl[addrhash]->addr ) );
		}
 */
	if ( match =
			( xp->addr.s_addr == addr &&
			  fncmp( xp, name ) == 0 ) )
	    matchp = xp;
	ep = xp;
	xp = xp->Next;
	}
    if ( matchp == NULL )
	{
	np = CreateNameAddrEntry( name, addr );
	ep->Next = np;
	}
    return( np );
}


/*
 *  gfn --
 *	Get full name of an entry
 */

char *
gfn( nap )
struct NameAddr * nap;
{
    char * cp;
    char FullName[MAXDNAME];

    cp = strcpy( FullName, nap->name );
    cp += strlen( nap->name );
    *cp++ = '.';
    strcpy( cp, nap->domain );
/*
 *  cp = FullName + strlen( FullName ) -1;
 *  if ( *cp != '.' )
 *	{
 *	cp++;
 *	*cp = '.';
 *	cp++;
 *	*cp = '\0';
 *	}
 */

    return( FullName );
}


/*
 *  AddNetEnt --
 *
 *	Add a Network Entity to the structure
 */

NetEntP
AddNetEnt( addr, mask, name )
u_long	addr;
u_long	mask;
char *  name;
{
    int  i;
    struct NetEnt * np;
    struct NameAddr * nap;

    np = NetEnts;
    while ( np != NULL && np->na->addr.s_addr != addr )
	np = np->Next;

    if ( np != NULL )
	return( np );

    NetEntCnt++;
    if ( debug )
        fprintf( stdout, " Adding net %s\n", name );
    np = (struct NetEnt *)Malloc( sizeof(struct NetEnt) );
    np->CreationTime = now;
    np->flags = 0;
    np->mask.s_addr = mask;
    np->SubNetMask.s_addr = 0;
    np->ifcount = 0;
    np->Next = NULL;
    np->SubnetList = NULL;
    np->origin = NULL;
    for ( i=0; i<5; i++ )
	 np->NS[i] = NULL;

    np->na = AddNATbl( name, addr );

    if ( LastNetEnt != NULL )
        LastNetEnt->Next = np;
    LastNetEnt = np;

    return( np );
}

