
/*
 *  data.h
 */

struct netaddr {
	struct in_addr	addr;
	struct in_addr	mask;
	};

struct	DomainEnt {
	u_long	CreationTime;
	char *	name;
	u_long	flags;
	struct DomainEnt *	Next;	/* Same level */
	struct DomainEnt *	Down;	/* First sub-domain */
	struct NameAddr *	NS[5];	/* Server for this level */
	};

struct Gateway {
	int	intfcnt;
	struct NameAddr * intfFirst;
	struct Gateway * Next;
	};

struct NameAddr {
	u_long	CreationTime;
	u_long	flags;
	struct in_addr	addr;
	char *	name;
	char *  domain;
	struct NameAddr * Next;		/* Pointer to another entry with
					   same hash codes but different
					   name or addr */
	struct NameAddr * List;		/* Pointer to next entry on list */
	struct NameAddr * NextIntf;	/* Next NA on same subnet */
	struct NameAddr * NextSameGW;	/* Next NA on same gateway */
	struct NameAddr * LO;		/* Next NA in length order */
	struct Gateway * Gateway;	/* Gateway to which this belongs */
	};

struct NetEnt {
	u_long	CreationTime;
	u_long	flags;
	struct NameAddr *	na;	/* Network Name/Addres */
	struct in_addr	mask;
	struct in_addr	SubNetMask;
	int	ifcount;
	struct NetEnt *	Next;		/* Next net in list */
	struct NetEnt * SubnetList;	/* Start of list of subnets */
	struct NameAddr * origin;	/* Primary server for this net */
	struct NameAddr * NS[5];	/* Other servers for this net */
	};

struct SubNetEnt {
	u_long	CreationTime;
	u_long	flags;
	struct in_addr	addr;
	struct in_addr	SubNetMask;
	int	ifcount;
	char *	name;
	char *  domain;
	struct NameAddr * FirstIntf;	/* Pointer to first interface  */
	struct NameAddr * LastIntf;	/* Pointer to last interface  */
	struct NameAddr * HighIntf;	/* Pointer to highest # interface  */
	struct NameAddr * LowIntf;	/* Pointer to lowest # interface  */
	struct GWEnt	* FirstGW;	/* Pointer to first gateway */
	};

struct net_intf {
	u_long	addr;
	u_long	nmask;
	u_long	smask;
	};

typedef struct DomainEnt * DomainEntP;
typedef struct NetEnt * NetEntP;
typedef struct NameAddr * NameAddrP;
typedef struct SubNetEnt * SubNetEntP;

#define  HASH_TABLE_SIZE  1024
#define  DUMPFILE  "nsl.dump"		/* dump file name  */
#define  DEBUGFILE  "nsl.debug"		/* debug file name  */

#ifdef MAIN

/*
 *  Accept any of the following to start a search:
 */

char	DomainName[256];	/* eg, colorado.edu  */
char	SubDomainName[256];	/* eg, cs.colorado.edu (is also a host alias) */
struct in_addr	NetworkNumber;	/* eg, 128.138  */
struct netaddr	SubnetNumber;	/* eg, 128.138.129  */
char	HostName[64];		/* eg, spot or spot.colorado.edu  */
struct in_addr	HostAddress;	/* eg, 128.138.129.2  */
struct in_addr	SubNetMask;	/* eg, 255.255.255.0  */
struct in_addr	IntfNetMask;

struct in_addr	CurNet;		/* Current Network */
struct in_addr	CurMask;	/* Current Mask */

struct NameAddr * AddrHashTbl[HASH_TABLE_SIZE];
struct NameAddr * NameHashTbl[HASH_TABLE_SIZE];

struct NameAddr * NAListp;	/* Pointer to list of all hashed entries */
struct NameAddr * NAListEndp;	/* Pointer to end of list */

struct DomainEnt * CurrentDomainP;	/* Pointer to current domain */
char * CurrentDomainName;		/* Pointer to current domain name */
struct DomainEnt *	DTree;		/* Top of domain tree to fill in */

struct NetEnt *	NetEnts;	/* Start of list of nets to analyze */
struct NetEnt *	LastNetEnt;	/* End of list of nets to analyze */
struct NetEnt *	nep;		/* Pointer to current net being analyzed */

struct Gateway * gwpStart;	/* Pointer to first gateway entry */

struct net_intf net_intf[20];	/* Interface descriptors */

u_long now;			/* Current time in seconds */

int	DomainEntCnt=0;			/* Count of Domain Entries */
int	NameAddrEntCnt=0;		/* Count of Name/Address Entries */
int	NetEntCnt=0;			/* Count of Network Entries */
int	SubNetCnt=0;			/* Count of SubNet Entries */
int	MaxSubNets=0;			/* Max number of subnets */
int	DGWCnt=0;			/* Count of distinguished gateways */
int	intfcnt=0;			/* Number of interfaces */

int	ssc=0;			/* Subnet shift count */
u_long	SNM=0;			/* SubNet Mask bits */
struct SubNetEnt ** SNP;	/* Array of SubNet pointers */

FILE *	dbgfp;			/* Debug file pointer */
int	debug=0;		/* Debug flag */

#else  /* MAIN */

extern char	DomainName;	/* eg, colorado.edu  */
extern char	SubDomainName;	/* eg, cs.colorado.edu (is also a host alias) */
extern struct in_addr	NetworkNumber;	/* eg, 128.138  */
extern struct netaddr	SubnetNumber;	/* eg, 128.138.129  */
extern char	HostName[];		/* eg, spot or spot.colorado.edu  */
extern struct in_addr	HostAddress;	/* eg, 128.138.129.2  */
extern struct in_addr	SubNetMask;	/* eg, 255.255.255.0  */
extern struct in_addr	IntfNetMask;

extern struct in_addr	CurNet;		/* Current Network */
extern struct in_addr	CurMask;	/* Current Mask */

extern struct NameAddr * AddrHashTbl[];
extern struct NameAddr * NameHashTbl[];

extern struct NameAddr * NAListp;	/* Pointer to list of all NA entries */
extern struct NameAddr * NAListEndp;	/* Pointer to end of list */

extern struct DomainEnt * CurrentDomainP;	/* Pointer to current domain */
extern char * CurrentDomainName;	/* Pointer to current domain name */
extern struct DomainEnt *	DTree;	/* Top of domain tree to fill in */

extern struct NetEnt *	NetEnts;	/* Start of list of nets to analyze */
extern struct NetEnt *	LastNetEnt;	/* End of list of nets to analyze */
extern struct NetEnt *	nep;	/* Pointer to current net being analyzed */

extern struct Gateway * gwpStart;	/* Pointer to first gateway entry */

extern struct net_intf net_intf[];	/* Interface descriptors */

extern u_long now;			/* Current time in seconds */

extern int	DomainEntCnt;		/* Count of Domain Entries */
extern int	NameAddrEntCnt;		/* Count of Name/Address Entries */
extern int	NetEntCnt;		/* Count of Network Entries */
extern int	SubNetCnt;		/* Count of SubNet Entries */
extern int	MaxSubNets;		/* Max number of subnets */
extern int	DGWCnt;			/* Count of distinguished gateways */
extern int	intfcnt;		/* Number of interfaces */

extern int	ssc;			/* Subnet shift count */
extern u_long	SNM;			/* SubNet Mask bits */
extern struct SubNetEnt ** SNP;		/* Array of SubNet pointers */

extern FILE *	dbgfp;			/* Debug file pointer */
extern int	debug;			/* Debug flag */

#endif /* MAIN */

#ifndef DATA

extern DomainEntP AddDomainEnt();
extern NameAddrP  CreateNameAddrEntry();
extern NameAddrP  AddNATbl();
extern NetEntP    AddNetEnt();

#endif  /* DATA */
