/*
 *  find.c --
 *
 *	Miscellaneous subroutines for finding things
 */

#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include <sys/wait.h>

#include "data.h"

/* FILE	filePtr; */

static struct {
    struct NameAddr * First;
    struct NameAddr * Last;
    int		  count;
    } LA[64];

/*
 *  findsm --
 *	Get subnet mask from first three hosts on list
 *	If not all the same, complain
 *
 *	Count and list number of addresses in each subnet
 *
 *	Look thru AddrHashTbl for all entries with different addresses.
 *
 */

findsm()
{
    char	address[256];
    char	oldname[256];
    u_long	oldaddr;
    char	*newav[5];
    extern char **environ;
    int		i;
    int		pid;
    struct NameAddr * nap;
    int		ncb;		/* NonContiguous Bits flag */
    union wait	status;
    u_long	tmpmask;

    nap = NAListp;
    oldname[0] = '\0';
    oldaddr = 0;
    i = 0;
    while ( ( SubNetMask.s_addr == 0 ) && ( nap != NULL ) )
	{
	if ( ( nap->addr.s_addr & 127 ) == 0 )
	    {
	    nap = nap->List;
	    continue;
	    }

	strcpy( address, gfn( nap ) );
	if ( ( strcmp( oldname, address ) == 0 ) ||
	    ( nap->addr.s_addr == oldaddr ) )
	    {
	    nap = nap->List;
	    continue;
	    }

	switch ( pid = vfork() )
	    {
	case -1 :	/* Fork failed */
	    perror( "vfork: ");
	    exit( 1 );

	    /*NOTREACHED*/

	case 0 :	/* Child process */
	    newav[ 0 ] = "sm";
	    newav[ 1 ] = address;
	    newav[ 2 ] = NULL;
/*
 *		dup2( filePtr, 0 );
 *		dup2( filePtr, 1 );
 *		dup2( filePtr, 2 );
 *		(void) close( filePtr );
 */
	    execve( "./sm", newav, environ );
	    perror( "execve" );
	    _exit( 1 );

	    /*NOTREACHED*/

	default :	/* Parent process */
	    if ( pid != wait( &status ) )
		fprintf( stderr, "em_nsl: Unknown child returned to me!" );
	    if ( WIFEXITED( status ) )
		{
		if ( status.w_retcode >= 8 && status.w_retcode
		    < 32 )
		    SubNetMask.s_addr = 0xffffffff <<
			(32 - status.w_retcode);
		if ( status.w_retcode == 2 )
		    break;
		}
	    }
	strcpy( oldname, address );
	oldaddr = nap->addr.s_addr;
	nap = nap->List;
	i++;
	if ( i >= 40 )
	    break;
	}

    printf( " em_nsl: Returned SubNetMask is %X\n", SubNetMask.s_addr );
    if ( SubNetMask.s_addr == 0 )
	{
	for ( i=0; i<intfcnt; i++ )
	    if ( ( ( net_intf[i].addr ^ CurNet.s_addr ) &
						net_intf[i].nmask ) == 0 )
		if ( net_intf[i].smask != 0 )
		    break;
					    
	if ( i < intfcnt )
	    {
	    SubNetMask.s_addr = net_intf[i].smask;
	    printf( " em_nsl: Assuming SubNetMask (%s) from interface\n",
		inet_ntoa( SubNetMask ) );
	    }
	else
	    {
	    printf( " em_nsl: Assuming 24 bits in SubNetMask ---\n" );
	    SubNetMask.s_addr = 0xffffff00;
	    }
	}
    SNM = CurMask.s_addr ^ SubNetMask.s_addr;
    MaxSubNets = 0;
    tmpmask = SNM;
    ncb = 0;
    for ( i=0; i<32; i++ )
	{
	if ( tmpmask == 0 )
	    break;

	if ( tmpmask & 1 == 1 )
	    {
	    if ( MaxSubNets == 0 )
		{
		MaxSubNets = tmpmask;
		ssc = i;
		}
	    }
	else
	    if ( MaxSubNets != 0 )
		ncb++;
	tmpmask = tmpmask >> 1;
	}
    MaxSubNets++;
    if ( ncb != 0 )
	fprintf( stderr,
		    "em_nsl: WARNING: Noncontiguous bits in subnet mask\n" );
}


/*
 *  CreateSubNetEnt
 *	Create an entry in the subnet table
 */

SubNetEntP
CreateSubNetEnt( nap )
struct NameAddr * nap;
{
    struct SubNetEnt * snp;

    snp = (struct SubNetEnt *) Malloc( sizeof( struct SubNetEnt ) );
    snp->SubNetMask.s_addr = SubNetMask.s_addr;
    snp->addr.s_addr = nap->addr.s_addr & SubNetMask.s_addr;
    snp->CreationTime = now;
    snp->flags = 0;
    snp->name = NULL;
    snp->domain = NULL;
    snp->FirstGW = NULL;
    snp->FirstIntf = nap;
    snp->LastIntf = nap;
    snp->HighIntf = nap;
    snp->LowIntf = nap;
    snp->ifcount=1;
    SubNetCnt++;
    return( snp );
}


/*
 *  AddIntfToList
 *	Add an interface (NA Entry) to the list for this subnet.
 */

AddIntfToList( nap, snp )
struct NameAddr * nap;
struct SubNetEnt * snp;
{

/*
 *  if ( snp->FirstIntf == NULL )
 *	{
 *	snp->FirstIntf = nap;
 *	snp->LastIntf = nap;
 *	snp->HighIntf = nap;
 *	snp->LowIntf = nap;
 *	return;
 *	}
 */

    if ( nap->addr.s_addr < snp->LowIntf->addr.s_addr )
	snp->LowIntf = nap;
    if ( nap->addr.s_addr > snp->HighIntf->addr.s_addr )
	snp->HighIntf = nap;
    snp->LastIntf->NextIntf = nap;
    snp->LastIntf = nap;
    snp->ifcount++;
    return;
}


/*
 *  findsn --
 *	Find subnets
 *	Determine subnet mask bits from net mask and subnet mask
 *	Create array of pointers for all possible subnets.
 *	Look thru all NameAddr entries for all subnets.
 *	  Create a subnet entry for each subnet found
 */

findsn()
{
    int  i;
    struct NameAddr * nap;
    char    ipstr[20];
    int  sni;
    struct SubNetEnt * snp;


    printf( "\n" );
    printf( "The following subnets have the indicated count of hosts:\n" );

    SNP = (struct SubNetEnt **)
	Calloc( MaxSubNets, sizeof( struct SubNetEnt * ) );
    for ( nap=NAListp; nap != NULL; nap=nap->List )
	{
	if ( ( nap->addr.s_addr & ~SubNetMask.s_addr ) == 0 )
	    continue;

	sni = (nap->addr.s_addr & SNM ) >> ssc;
	snp = *(SNP+sni);
	if ( snp == NULL )
	    *(SNP+sni) = snp = CreateSubNetEnt( nap );
	else
	    AddIntfToList( nap, snp );
	}
    for ( sni=0; sni<MaxSubNets; sni++ )
	if ( ( snp = *(SNP+sni) ) != NULL )
	    {
	    printf( " %-16s %4d  ",
		inet_ntoa( snp->addr ),
		snp->ifcount );
	    strcpy( ipstr, inet_ntoa( snp->LowIntf->addr ) );
	    printf( " %-16s  %-16s\n", ipstr,
		inet_ntoa( snp->HighIntf->addr ) );
	    }
}


/*
 *  AddToGW
 *	Add an interface to a gateway
 */

AddToGW( nap, gwp )
struct NameAddr * nap;
struct Gateway * gwp;
{
    struct Gateway * gwpx;
    struct NameAddr * ip;

    if ( nap->Gateway == gwp )
	return;				/* if already on this gw */

    if ( nap->Gateway == NULL )
	{				/* if not on any gw */
	nap->Gateway = gwp;
	gwp->intfcnt++;
	for ( ip=gwp->intfFirst; ip->NextSameGW!=NULL; ip=ip->NextSameGW
	    )
	    if ( ip->NextSameGW == nap )
		return;

	ip->NextSameGW = nap;
	return;
	}

					/* on some other gw, combine them */
    if ( nap->Gateway == gwpStart )
	{
	gwpStart = gwpStart->Next;
	}
    else
	{
	for ( gwpx=gwpStart; gwpx->Next!=NULL; gwpx=gwpx->Next )
	    {
	    if ( gwpx->Next == nap->Gateway )
		{
		gwpx->Next = gwpx->Next->Next;
		break;
		}
	    }
	}

    gwp->intfcnt += nap->Gateway->intfcnt + 1;
    ip = nap->Gateway->intfFirst;
    free( nap->Gateway );
    for ( ; ip!= NULL; ip=ip->NextSameGW )
	ip->Gateway = gwp;

    return;
}


/*
 *  FindGWp
 *	Find the pointer to the gateway to which this interface is a part.
 *	Create the Gateway if necessary
 */

struct Gateway *
FindGWp( nap )
struct NameAddr * nap;
{
    struct Gateway * gwp;
    struct Gateway * gwpx;

    if ( nap->Gateway != NULL )
	return( nap->Gateway );

    gwp = (struct Gateway *) malloc( sizeof( struct Gateway ) );
    gwp->intfFirst = nap;
    gwp->intfcnt = 1;
    gwp->Next = NULL;
    nap->Gateway = gwp;

    if ( gwpStart == NULL )
	gwpStart = gwp;
    else
	{
	for ( gwpx=gwpStart; gwpx->Next!=NULL; gwpx=gwpx->Next )
	    ;
	gwpx->Next = gwp;
	}
    return( gwp );
}


/*
 *  findgw --
 *	Find Gateways
 *	Look thru NameHashTbl for all entries with multiple NameAddr's
 *	  with same name but different addr.  These are the probable gateways.
 *
 *	Look thru AddrHashTbl for all entries with multiple NameAddr's
 *	  with same addr but different name.  Use the multiple names to
 *	  look for multiple addresses.  These are also probable gateways.
 *
 *	For each possible gateway,
 *	    List address and multiple names
 *	For each likely gateway,
 *	    List name and multiple addresses
 *	For each subnet,
 *	    List likely gateways
 *	    List possible gateways
 */

findgw()
{
    struct Gateway * gwp;
    struct in_addr  inaddr;
    int  hash;
    int  i;
    char ipstr[20];
    struct NameAddr * bap;
    struct NameAddr * nap;
    char *  name;
    char    namex[256];
    u_long  paddr;
    struct NameAddr * yettomatch;

    /*  List all entries marked as gateways  */

    printf( "\n" );
    if ( DGWCnt != 0 )
	{
	printf( "The following %d entries were identified as \"designated\" \
	    gateways:\n",
	    DGWCnt );
	for ( nap=NAListp; nap != NULL; nap=nap->List )
	    {
	    if ( ( nap->flags & 2 ) == 2 )
		printf( " %-16s  %s\n",
		    inet_ntoa( nap->addr ),
		    gfn ( nap ) );
	    }
	}
    else
	printf( "No entries were identified as \"designated\" gateways.\n" );


    /*  List all entries with multiple addresses for the same name  */

    printf( "\n" );
    printf(
	"The following entries have multiple addresses for the same name:\n" );

    for ( hash=0; hash<HASH_TABLE_SIZE; hash++ )
	{
	if ( ( nap = NameHashTbl[hash] ) == NULL )
	    continue;

	do  {
	    gwp = NULL;
	    yettomatch = NULL;
	    inaddr.s_addr = nap->addr.s_addr;
	    paddr = 0;
	    strcpy( namex, gfn( nap ) );
	    nap->flags |= 4;
	    nap = nap->Next;

	    for ( ; nap != NULL; nap = nap->Next )
		{
		if ( fncmp( nap, namex ) == 0 )
		    {
		    nap->flags |= 4;
		    if ( inaddr.s_addr == nap->addr.s_addr
			)
			continue;

		    if ( paddr == 0 )
			{
			printf( "%s\n", namex );
			strcpy( ipstr, inet_ntoa(
			    inaddr ) );
			printf( " %s\n", ipstr );
			paddr++;
			gwp = FindGWp( nap );
			}
		    printf( " %s\n", inet_ntoa( nap->addr ) );
		    if ( gwp != NULL )
			AddToGW( nap, gwp );
		    }
		else
		    {
		    if ( ( nap->flags & 4 == 0 ) && (
			yettomatch == NULL ) )
			yettomatch = nap;
		    }
		}
	    } while ( ( nap = yettomatch ) != NULL );
	}



    /*  List all entries with multiple names for the same address  */

    printf( "\n" );
    printf( "The following entries have multiple names for the same address:\n");

    for ( hash=0; hash<HASH_TABLE_SIZE; hash++ )
	{
	if ( ( nap = AddrHashTbl[hash] ) == NULL )
	    continue;

	do  {
	    yettomatch = NULL;
	    inaddr.s_addr = nap->addr.s_addr;
	    paddr = 0;
	    strcpy( namex, gfn( nap ) );
	    nap->flags |= 8;
	    nap = nap->Next;

	    for ( ; nap != NULL; nap = nap->Next )
		{
		if ( inaddr.s_addr == nap->addr.s_addr )
		    {
		    nap->flags |= 8;
		    if ( fncmp( nap, namex ) == 0 )
			continue;

		    if ( paddr == 0 )
			{
			strcpy( ipstr, inet_ntoa(
			    inaddr ) );
			printf( "%s\n", ipstr );
			printf( " %s\n", namex );
			paddr++;
			}
		    printf( " %s\n", gfn( nap ) );
		    }
		else
		    {
		    if ( ( nap->flags & 8 == 0 ) && (
			yettomatch == NULL ) )
			yettomatch = nap;
		    }
		}
	    } while ( ( nap = yettomatch ) != NULL );
	}


    /*  List all entries with the same name but different domains  */

    printf( "\n" );
    printf("The following entries have the same name, but different domains:\n");

    for ( hash=0; hash<HASH_TABLE_SIZE; hash++ )
	{
	if ( ( nap = NameHashTbl[hash] ) == NULL )
	    continue;

	do  {
	    yettomatch = NULL;
	    inaddr.s_addr = nap->addr.s_addr;
	    name = nap->name;
	    bap = nap;
	    paddr = 0;
	    nap->flags |= 16;
	    nap = nap->Next;

	    for ( ; nap != NULL; nap = nap->Next )
		{
		if ( strcmp( nap->name, bap->name ) == 0
		    )
		    {
		    nap->flags |= 16;
		    if ( bap->domain == nap->domain )
			continue;

		    if ( paddr == 0 )
			{
			printf( "%s\n", name );
			printf( " %-16s  %s\n", inet_ntoa(
			    inaddr ), gfn( bap ) );
			paddr++;
			}
		    printf( " %-16s  %s\n", inet_ntoa(
			nap->addr ),gfn( nap ) );
		    }
		else
		    {
		    if ( ( nap->flags & 16 == 0 ) &&
			( yettomatch == NULL ) )
			yettomatch = nap;
		    }
		}
	    } while ( ( nap = yettomatch ) != NULL );
	}
}


/*
 *  findss --
 *	Find names which are super strings of other names
 */

findss()
{
    struct NameAddr * cap;
    struct NameAddr * nap;
    int  i;
    int  pf;
    int  strl;

    /*  List all entries with the same initial substrings  */

    printf( "\n" );
    printf(
	"The following entries are related by having same initial substrings:\n"
	);

    for ( strl=3; strl<64; strl++ )
	{
	for ( cap = LA[strl].First; cap != NULL; cap = cap->LO )
	    {
	    pf = 0;
	    for ( i = strl+1; i<64; i++ )
		{
		nap = LA[i].First;
		for ( ; nap!=NULL; nap=nap->LO )
		    {
		    if ( ( ( cap->addr.s_addr ^ nap->addr.s_addr
			) &
			SubNetMask.s_addr ) == 0 )
			continue;

		    if ( strncmp( cap->name, nap->name,
			strl+1 ) == 0 )
			{
			if ( pf == 0 )
			    {
			    pf++;
			    printf( "\n" );
			    printf( " %-16s  %s\n",
				inet_ntoa( cap->addr
				),
				gfn( cap ) );
			    }
			printf( " %-16s  %s\n",
			    inet_ntoa( nap->addr ),
			    gfn( nap ) );
			}
		    }
		}
	    if ( cap == LA[strl].Last )
		break;
	    }
	}
    return;
}


/*
 *  findstr --
 *	Find names which with -gw or -gateway as variations
 */

findstr()
{
    struct NameAddr * cap;
    char * cc;
    struct Gateway * gwp;
    int    i;
    int    match;
    int    n;
    int    namehash;
    char   namescr[64];
    struct NameAddr * nap;
    int    pf;
    int    strl;

    /*  List all entries related by -gw or -gateway	*/
    /*  <name>						*/
    /*  <name>-gw					*/
    /*	<name>-gw[0-9]*   (not completely reliable	*/
    /*  <name>-gw-<stuff>				*/
    /*  <name>-gw-{0,1}*<stuff>				*/

    printf(
	"\nThe following entries are related by -gw or -gateway suffixes:\n" );

    for ( strl=5; strl<64; strl++ )
	{
	for ( cap = LA[strl].First; cap != NULL; cap = cap->LO )
	    {
	    pf = 0;
	    cc = cap->name;
	    while ( cc != NULL )
		{
		cc = strchr( cc, '-' );
		if ( cc == NULL )
		    break;

		if ( strncmp( cc, "-gw", 3 ) == 0 )
		    {
		    cc += 3;
		    break;
		    }

		if ( strncmp( cc, "-gateway", 8 ) == 0 )
		    {
		    cc += 8;
		    break;
		    }

		cc++;
		}
	    if ( cc == NULL )
		continue;

	    if ( *cc == '\0' )
		continue;

	    n = cc - cap->name + 1;
	    strncpy( namescr, cap->name, n );
	    namescr[n-1] = '.';
	    namescr[n] = '\0';
	    strcat( namescr, cap->domain );

	    namehash = 0;
	    for ( i=0; i<n-1; i++ )
		namehash = ( namehash << 2 ) | ( namescr[i]
		    & 0x3 );

	    namehash %= HASH_TABLE_SIZE;

	    nap = NameHashTbl[namehash];
	    while ( ! ( match =
		( nap->addr.s_addr != cap->addr.s_addr &&
		fncmp( nap, namescr ) == 0 )
		) &&
		( nap->Next != NULL ) )
		nap = nap->Next;
	    if ( match )
		{
		printf( "\n %-16s  %s\n", inet_ntoa( cap->addr ), gfn( cap ) );
		gwp = FindGWp( nap );
		printf( " %-16s  %s\n", inet_ntoa( nap->addr ), gfn( nap ) );
		AddToGW( cap, gwp );
		}

	    if ( cap == LA[strl].Last )
		break;
	    }
	}
    return;
}


/*
 *  sortnl --
 *	link names by name length
 */

sortnl()
{
    struct NameAddr * cap;
    struct NameAddr * nap;
    int  i;
    int  pf;
    int  strl;

    printf(
	"\nSorting entries by name length.\n"
	);

    for ( strl=0; strl<64; strl++ )
	{
	LA[strl].First = NULL;
	LA[strl].Last  = NULL;
	LA[strl].count = 0;
	}

    for ( nap = NAListp; nap != NULL; nap = nap->List )
	{
	strl = strlen( nap->name );
	if ( strl == 0 )
	    continue;

	if ( strl >= 64 )
	    strl = 64;
	strl--;
	if ( LA[strl].First == NULL )
	    {
	    LA[strl].First = nap;
	    LA[strl].Last  = nap;
	    LA[strl].count = 1;
	    }
	else
	    {
	    LA[strl].Last->LO = nap;
	    LA[strl].Last     = nap;
	    LA[strl].count++;
	    }
	}
    return;
}

