#ifndef lint
static char sccsid[] = "@(#)sm.c	4.3 (Berkeley) 8/6/85";
#endif

/*
 *			I C M P I N G . C
 *
 * Using the InterNet Control Message Protocol (ICMP) facility,
 * discover the subnet mask for a network.
 *
 * Author -
 *	Mike Muuss
 *	U. S. Army Ballistic Research Laboratory
 *	December, 1983
 * Modified at Uc Berkeley
 * Modified by DCMW for subnet mask retreval
 *
 * Status -
 *	Public Domain.  Distribution Unlimited.
 *
 * Bugs -
 *	More statistics could always be gathered.
 *	This program has to run SUID to ROOT to access the ICMP socket.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/time.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netdb.h>

#define	MAXWAIT		10	/* max time to wait for response, sec. */
#define	MAXPACKET	4096	/* max packet size */
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64
#endif

#define  TRUE	1
#define  FALSE	0

struct	ip_icmp {
	struct ip ip;
	struct icmp icmp; };

u_long	mask;
int	verbose;
union {
long	l;
u_char	p[MAXPACKET];
      }up;
#define packet up.p
int	options;
extern	int errno;

int s;			/* Socket file descriptor */
struct hostent *hp;	/* Pointer to host info */
struct timezone tz;	/* leftover */

struct sockaddr whereto;/* Who to ping */
int datalen = 100;		/* How much data */

char usage[] = "Usage:  sm [-drv] host\n";

char *hostname;
char hnamebuf[MAXHOSTNAMELEN];

int npackets;
int Npackets=1;
int ident;

int nreceived = 0;		/* # of packets we got back */
int timing = 0;
int tmin = 999999999;
int tmax = 0;
int tsum = 0;			/* sum of all times, for doing average */
int waittime = 0;
int finish();

/*
 * 			M A I N
 */
main(argc, argv)
char *argv[];
{
	struct sockaddr_in from;
	char **av = argv;
	struct sockaddr_in *to = (struct sockaddr_in *) &whereto;
	int on = 1;
	struct protoent *proto;
	int len = sizeof (packet);
	int fromlen = sizeof (from);
	int cc;
	int i;

	setlinebuf( stdout );

	argc--, av++;
	while (argc > 0 && *av[0] == '-') {
		while (*++av[0]) switch (*av[0]) {
			case 'd':
				options |= SO_DEBUG;
				break;
			case 'r':
				options |= SO_DONTROUTE;
				break;
			case 'v':
				verbose++;
				break;
		}
		argc--, av++;
	}
	if( argc < 1)  {
		printf( usage );
		exit( 2 );
	}

	bzero( (char *)&whereto, sizeof(struct sockaddr) );
	to->sin_family = AF_INET;
	to->sin_addr.s_addr = inet_addr(av[0]);
	if (to->sin_addr.s_addr != -1) {
		strcpy(hnamebuf, av[0]);
		hostname = hnamebuf;
	} else {
		hp = gethostbyname(av[0]);
		if (hp) {
			to->sin_family = hp->h_addrtype;
			bcopy(hp->h_addr, (caddr_t)&to->sin_addr, hp->h_length);
			hostname = hp->h_name;
		} else {
			printf( "em_nsl:%s: unknown host %s\n", argv[0], av[0]);
			exit(1);
		}
	}

	if( argc > 2 )
		Npackets = atoi( av[1] );

	ident = getpid() & 0xFFFF;

	if ((proto = getprotobyname("icmp")) == NULL) {
		fprintf( stderr, "em_nsl: icmp: unknown protocol\n" );
		exit( 2 );
	}
	if ((s = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
		perror("em_nsl: sm: socket");
		exit(2);
	}
	if (options & SO_DEBUG)
		setsockopt(s, SOL_SOCKET, SO_DEBUG, &on, sizeof(on));
	if (options & SO_DONTROUTE)
		setsockopt(s, SOL_SOCKET, SO_DONTROUTE, &on, sizeof(on));

	printf(" em_nsl: sm: %s: %d packets\n", hostname, Npackets );

	signal( SIGINT, finish );
	signal(SIGALRM, finish);

	for (i=0; i<3; i++)
	    {
	    pinger();
	    alarm( 10 );
	    if ( (cc=recvfrom(s, packet, len, 0, &from, &fromlen)) < 0)
		{
		if( errno != EINTR )
		    perror("sm: recvfrom");
		}
	    else
		{
		if ( pr_pack( packet, cc, &from ) != 0 )
		    continue;
		finish();
		}
	    }
	/*NOTREACHED*/
}


/*
 * 			P I N G E R
 * 
 * Compose and transmit an ICMP SUBNET MASK REQUEST packet.  The IP packet
 * will be added on by the kernel.  The ID field is our UNIX process ID,
 * and the sequence number is an ascending integer.  The first 8 bytes
 * of the data portion are used to hold a UNIX "timeval" struct in VAX
 * byte-order, to compute the round-trip time.
 */
pinger()
{
	union {
	long	l;
	u_char o[MAXPACKET];
	      } uo;
#define outpack  uo.o
	register struct icmp *icp = (struct icmp *) outpack;
	int i, cc;
	register struct timeval *tp = (struct timeval *) &outpack[8];
	register u_char *datap = &outpack[8+sizeof(struct timeval)];

	icp->icmp_type = ICMP_MASKREQ;
	icp->icmp_code = 0;
	icp->icmp_cksum = 0;
	icp->icmp_seq = 1;
	icp->icmp_id = ident;		/* ID */
	icp->icmp_mask = 0;		/* mask */

	cc = datalen+8;			/* skips ICMP portion */

	if (timing)
		gettimeofday( tp, &tz );

	for( i=8; i<datalen; i++)	/* skip 8 for time */
		*datap++ = i;

	/* Compute ICMP checksum here */
	icp->icmp_cksum = in_cksum( icp, cc );

	/* cc = sendto(s, msg, len, flags, to, tolen) */
	i = sendto( s, outpack, cc, 0, &whereto, sizeof(struct sockaddr) );

	if( i < 0 || i != cc )  {
		if( i<0 )  perror("sendto");
		printf( "em_nsl: sm: wrote %s %d chars, ret=%d\n",
			hostname, cc, i );
	}
}

/*
 * 			P R _ T Y P E
 *
 * Convert an ICMP "type" field to a printable string.
 */
char *
pr_type( t )
register int t;
{
	static char *ttab[] = {
		"Echo Reply",
		"ICMP 1",
		"ICMP 2",
		"Dest Unreachable",
		"Source Quence",
		"Redirect",
		"ICMP 6",
		"ICMP 7",
		"Echo",
		"ICMP 9",
		"ICMP 10",
		"Time Exceeded",
		"Parameter Problem",
		"Timestamp",
		"Timestamp Reply",
		"Info Request",
		"Info Reply",
		"Mask Request",
		"Mask Reply"
	};

	if( t < 0 || t > 18 )
		return("OUT-OF-RANGE");

	return(ttab[t]);
}

/*
 *			P R _ P A C K
 *
 * Print out the packet, if it came from us.  This logic is necessary
 * because ALL readers of the ICMP socket get a copy of ALL ICMP packets
 * which arrive ('tis only fair).  This permits multiple copies of this
 * program to be run without having intermingled output (or statistics!).
 */
pr_pack( icp, cc, from )
register struct ip_icmp *icp;
int cc;
struct sockaddr_in *from;
{
	register long *lp = (long *) packet;
	register int i;
	struct timeval tv;
	struct timeval *tp = (struct timeval *) &icp->icmp.icmp_otime;
	int triptime;
	char *inet_ntoa();

	from->sin_addr.s_addr = ntohl( from->sin_addr.s_addr );
	gettimeofday( &tv, &tz );

	mask = icp->icmp.icmp_mask;
	if( icp->icmp.icmp_type != ICMP_MASKREPLY )
		{
		printf( "sm: %d bytes from %s: ",
			cc,
			inet_ntoa( from->sin_addr ) );
		printf("icmp_type=%d (%s)\n",
			icp->icmp.icmp_type,
			pr_type(icp->icmp.icmp_type) );
/*
 *		for( i=0; i<12; i++)
 *		    printf("x%2.2x: x%8.8x\n", i*sizeof(long), *lp++ );
 */
		printf("icmp_code=%d\n", icp->icmp.icmp_code );
		return(-1);
		}
	if( icp->icmp.icmp_id != ident )
		return(-1);			/* 'Twas not our ECHO */

	printf( " em_nsl: sm: %d bytes from %s: ", cc,
		inet_ntoa( from->sin_addr ) );
/*	printf("icmp_seq=%d. ", icp->icmp.icmp_seq );  */
	printf( "mask=%X ", icp->icmp.icmp_mask );
	if (timing) {
		tvsub( &tv, tp );
		triptime = tv.tv_sec*1000+(tv.tv_usec/1000);
		printf("time=%d. ms\n", triptime );
		tsum += triptime;
		if( triptime < tmin )
			tmin = triptime;
		if( triptime > tmax )
			tmax = triptime;
	} else
		putchar('\n');
	nreceived++;
	return(0);
}


/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
in_cksum(addr, len)
u_short *addr;
int len;
{
	register int nleft = len;
	register u_short *w = addr;
	register u_short answer;
	register int sum = 0;

	/*
	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
	 *  we add sequential 16 bit words to it, and at the end, fold
	 *  back all the carry bits from the top 16 bits into the lower
	 *  16 bits.
	 */
	while( nleft > 1 )  {
		sum += *w++;
		nleft -= 2;
	}

	/* mop up an odd byte, if necessary */
	if( nleft == 1 )
		sum += *(u_char *)w;

	/*
	 * add back carry outs from top 16 bits to low 16 bits
	 */
	sum += (sum >> 16);	/* add hi 16 to low 16 */
	answer = ~sum;		/* truncate to 16 bits */
	return (answer);
}

/*
 * 			T V S U B
 * 
 * Subtract 2 timeval structs:  out = out - in.
 * 
 * Out is assumed to be >= in.
 */
tvsub( out, in )
register struct timeval *out, *in;
{
	if( (out->tv_usec -= in->tv_usec) < 0 )   {
		out->tv_sec--;
		out->tv_usec += 1000000;
	}
	out->tv_sec -= in->tv_sec;
}

/*
 *			F I N I S H
 *
 * Print out statistics, and give up.
 * Heavily buffered STDIO is used here, so that all the statistics
 * will be written with 1 sys-write call.  This is nice when more
 * than one copy of the program is running on a terminal;  it prevents
 * the statistics output from becomming intermingled.
 */
finish()
{
int	bits;
int	m;

    if ( nreceived == 0 )
	exit( 1 );

    if ( mask == 0 || mask == 0xffffffff )
	bits = 0;
    else
	{
	for ( m=1, bits=32; bits>0; m=m<<1, bits-- )
	    {
	    if ( (mask & m) != 0 )  break;
	    }
	}

    printf( " em_nsl: sm: There are %d bits in the subnet mask.\n", bits );
    exit( bits );
}
