
#ifndef lint
static char sccsid[] = "@(#)subr2.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  subr.c --
 *	Miscellaneous subroutines for the analysis section
 */

#include <sys/types.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "res.h"

#include <ctype.h>
#include <sys/types.h>


/*
 *  Malloc --
 *  Calloc --
 *
 *      Calls the malloc library routine with SIGINT blocked to prevent
 *	corruption of malloc's data structures. We need to do this because
 *	a control-C doesn't kill the program -- it causes a return to the
 *	main command loop.
 *
 *	NOTE: This method doesn't prevent the pointer returned by malloc
 *	from getting lost, so it is possible to get "core leaks".
 *
 *	If malloc fails, the program exits.
 *
 *  Results:
 *	(address)	- address of new buffer.
 */

char *
Malloc(size)
    int size;
{
    char	*ptr;

#ifdef SYSV
#ifdef SVR3
    sighold(SIGINT);
    ptr = malloc((unsigned) size);
    sigrelse(SIGINT);
#else
    { int (*old)();
      old = signal(SIGINT, SIG_IGN);
      ptr = malloc((unsigned) size);
      signal(SIGINT, old);
    }
#endif
#else
    { int saveMask;
      saveMask = sigblock(sigmask(SIGINT));
      ptr = malloc((unsigned) size);
      (void) sigsetmask(saveMask);
    }
#endif
    if (ptr == NULL) {
	fflush(stdout);
	fprintf(stderr, "*** Can't allocate memory\n");
	fflush(stderr);
	abort();
	/*NOTREACHED*/
    } else {
	return(ptr);
    }
}

char *
Calloc(num, size)
    register int num, size;
{
    char *ptr = Malloc(num*size);
    bzero(ptr, num*size);
    return(ptr);
}


/*
 * IsNet --
 *
 *	Returns TRUE if the string looks like an Internet network number.
 *	A string with a trailing dot is not a network number,
 *	even if it looks like one.
 *
 *	XXX doesn't treat 255.255.255.255 as a network number.
 */

Boolean
IsNet( host, addrPtr )
char *host;
unsigned long *addrPtr;		/* If return TRUE, contains IP address */
{
    register char *cp;
    unsigned long addr;

    if (isdigit(host[0])) {
	    /* Make sure it has only digits and dots. */
	    for (cp = host; *cp; ++cp) {
		if (!isdigit(*cp) && *cp != '.') 
		    return FALSE;
	    }
	    /* If it has a trailing dot, don't treat it as an address. */
	    if (*--cp != '.') { 
		if ((addr = inet_network(host)) != (unsigned long) -1) {
		    while ( ( addr & 0xff000000 ) == 0  && addr != 0 )
			addr = addr << 8;
		    *addrPtr = addr;
		    return TRUE;
		}
	    }
    }
    return FALSE;
}


/*
 *  MakeInAddr --
 *
 *	Create an in-addr.arpa address from a network number
 */

MakeInAddr( str1, str2 )
char * str1;
char * str2;
{
    int b[4];
    int i;
    u_long  net;

    IsNet ( str1, &net );

    b[3] = net & 0xff;
    net = net >> 8;
    b[2] = net & 0xff;
    net = net >> 8;
    b[1] = net & 0xff;
    net = net >> 8;
    b[0] = net & 0xff;

    for ( i=0; i<4; i++ )
	if ( b[i] == 0 )  break;
    switch (i)
	{
	case 4:
	sprintf( str2, "%d.%d.%d.%d.in-addr.arpa.\0", b[3],b[2],b[1],b[0] );
	break;

	case 3:
	sprintf( str2, "%d.%d.%d.in-addr.arpa.\0", b[2],b[1],b[0] );
	break;

	case 2:
	sprintf( str2, "%d.%d.in-addr.arpa.\0", b[1],b[0] );
	break;

	case 1:
	case 0:
	sprintf( str2, "%d.in-addr.arpa.\0", b[0] );
	}
    return;
}


/*
 *  MakeMask --
 *
 *	Return a mask derived from the given address
 */

u_long
MakeMask( addr )
u_long	addr;
{
    if ( IN_CLASSA( addr ) )
	return( IN_CLASSA_NET );

    if ( IN_CLASSB( addr ) )
	return( IN_CLASSB_NET );

    if ( IN_CLASSC( addr ) )
	return( IN_CLASSC_NET );

    return( 0xfffffff0 );
}

