
#ifndef lint
static char sccsid[] = "@(#)util.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  util.c --
 *
 *	Miscellaneous subroutines for the name server lookup program.
 */

#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>
#include "res.h"


/*
 *  MakeInAddr --
 *
 *	Create an in-addr.arpa address from a network number
 */

MakeInAddr( str1, str2 )
char * str1;
char * str2;
{
    int b[4];
    int i;
    u_long  net;

    IsNet ( str1, &net );

    b[3] = net & 0xff;
    net = net >> 8;
    b[2] = net & 0xff;
    net = net >> 8;
    b[1] = net & 0xff;
    net = net >> 8;
    b[0] = net & 0xff;

    for ( i=0; i<4; i++ )
	if ( b[i] == 0 )  break;
    switch (i)
	{
	case 4:
	sprintf( str2, "%d.%d.%d.%d.in-addr.arpa.\0", b[3],b[2],b[1],b[0] );
	break;

	case 3:
	sprintf( str2, "%d.%d.%d.in-addr.arpa.\0", b[2],b[1],b[0] );
	break;

	case 2:
	sprintf( str2, "%d.%d.in-addr.arpa.\0", b[1],b[0] );
	break;

	case 1:
	case 0:
	sprintf( str2, "%d.in-addr.arpa.\0", b[0] );
	}
    return;
}


/*
 * IsNet --
 *
 *	Returns TRUE if the string looks like an Internet network number.
 *	A string with a trailing dot is not a network number,
 *	even if it looks like one.
 *
 *	XXX doesn't treat 255.255.255.255 as a network number.
 */

Boolean
IsNet( host, addrPtr )
char *host;
unsigned long *addrPtr;		/* If return TRUE, contains IP address */
{
    register char *cp;
    unsigned long addr;

    if (isdigit(host[0])) {
	    /* Make sure it has only digits and dots. */
	    for (cp = host; *cp; ++cp) {
		if (!isdigit(*cp) && *cp != '.') 
		    return FALSE;
	    }
	    /* If it has a trailing dot, don't treat it as an address. */
	    if (*--cp != '.') { 
		if ((addr = inet_network(host)) != (unsigned long) -1) {
		    while ( ( addr & 0xff000000 ) == 0  && addr != 0 )
			addr = addr << 8;
		    *addrPtr = addr;
		    return TRUE;
		}
	    }
    }
    return FALSE;
}


/*
 * CvtNetToPtr --
 *
 *	Convert a dotted-decimal Internet network number into the standard
 *	PTR format (reversed address with .in-arpa. suffix).
 *
 *	Assumes the argument buffer is large enougth to hold the result.
 */

CvtNetToPtr( name )
char *name;
{
    char *p;
    int ip[4];
    struct in_addr addr;

    if (IsAddr(name, &addr.s_addr)) {
	p = inet_ntoa( addr );
	if (sscanf(p, "%d.%d.%d.%d", &ip[0], &ip[1], &ip[2], &ip[3]) == 4) {
	    sprintf(name, "%d.%d.%d.%d.in-addr.arpa.", 
		ip[3], ip[2], ip[1], ip[0]);
	}
    }
}



/*
 * CvtPtrToNet --
 *
 *	Convert a PTR format reverse address (witn .in-addr.arpa) suffix
 *	to an Internet network number (u_long).
 */

CvtPtrToNet( name, net, bytes )
char *	 name;
u_long * net;
int *	 bytes;
{
    char * cp;
    int    ov;

    for ( cp=name, *bytes=0, *net=0; *bytes<4 ; (*bytes)++ )
	{
	ov = 0;
	while ( isdigit( *cp ) )
	    ov = 10* ov+ (*cp++ - '0' );
	if ( *cp != '.' )
	    break;

	*net = *net >>8 | ov <<24;
	cp++;
	}
    if ( ! strncmp( cp, "in-addr.arpa", 12 ) == 0 )
	*net = 0;
    return;
}


/*
 *  MakeMask --
 *
 *	Return a mask derived from the given address
 */

u_long
MakeMask( addr )
u_long	addr;
{
    if ( IN_CLASSA( addr ) )
	return( IN_CLASSA_NET );

    if ( IN_CLASSB( addr ) )
	return( IN_CLASSB_NET );

    if ( IN_CLASSC( addr ) )
	return( IN_CLASSC_NET );

    return( 0xfffffff0 );
}


/*
 *  Make Lower Case --
 *	Convert a string to lower-case
 */

mlc( cp )
char *	cp;
{
    while ( *cp != NULL )
	{
	if ( isupper( *cp ) )
	    *cp = tolower( *cp );
	cp++;
	}
    return;
}

