
/*
 *  dumpload.c	D.C.M.Wood 92/04/13
 *
 *  Dump and load data to/from history file
 */

#include <sys/file.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <protocols/routed.h>
#include <signal.h>
#include <stdio.h>
#include <time.h>

#define SOCKADDR_IN
#include "data_server.h"
#include "send_data.h"

#include "ripwatch.h"
#include "externs.h"

/*
#ifndef PIN
#include "pin.h"
#endif

char	pin[20]= PIN;
*/
extern	char	pin[];

u_long	js_addr=0;			/* Journal server address */
u_long	js_port=DATA_PORT;		/* Journal Server port */


/*
 *  ether_ntoa
 *    Format ethernet address as a string
 */

char *
ether_ntoa( eap )
struct ether_addr * eap;
{
     static char eastr[20];

     sprintf( eastr,  "%02x:%02x:%02x:%02x:%02x:%02x",
			eap->ether_addr_octet[0],
			eap->ether_addr_octet[1],
			eap->ether_addr_octet[2],
			eap->ether_addr_octet[3],
			eap->ether_addr_octet[4],
			eap->ether_addr_octet[5] );
    return( eastr );
}


/*
 *  dumptable - Dump the data file to disk
 *
 *	gw= <ip address>
 *	  <dst ip address> <metric> <date/time>
 *	  ...
 */

int dumptable()
{
int	cc;
char	dstipstr[20];
struct route_Dst *dstp;
struct route_Gw * gwp;
char	gwipstr[20];
int	hash;
FILE *	hfp;
struct tm * lt;
struct route_ent * rep;
char	when[20];


    UpdateJS();

    hfp = fopen( DUMPFILE, "w" );

    if (hfp == NULL )
	{
	perror( " Error opening dump file for writing" );
	return;
	}

    for( hash=0; hash<HASHSIZE; hash++ )
	{
	gwp = gwtab[hash];
	while ( gwp != NULL )
	    {
	    strcpy( gwipstr, inet_ntoa(  gwp->Gw_addr ) );
	    rep = gwp->first_Dst;
	    while ( rep != NULL )
		{
		lt = localtime( &(rep->tv_seconds) );
		lt->tm_mon++;
		sprintf( when, "%02d/%02d/%02d %02d%02d",
				    lt->tm_year,
				    lt->tm_mon,
				    lt->tm_mday,
				    lt->tm_hour,
				    lt->tm_min );
		strcpy( dstipstr, inet_ntoa(  rep->ipRouteDest ) );
		cc = fprintf( hfp, "%-16s  %-16s  %2d  %s\n",
				    gwipstr,
				    dstipstr,
				    rep->ipRouteMetric1,
				    when);
		if ( cc < 0 )
		    {
		    perror( " Error writing dump file" );
		    return;
		    }
		rep = rep->next_Dst;
		}
	    gwp = gwp->next_Gw;
	    }
	}
    fclose( hfp );
    return;
}

/*
 *  loadtable - Load the data file to disk
 */

int loadtable()
{
int	cc;
char	dstipstr[20];
u_long	dstn;
struct route_Dst *dstp;
struct route_Gw * gwp;
char	gwipstr[20];
FILE *	hfp;
int	i;
char	line[200];
int	metric;
struct netinfo netinfo;
struct route_ent * rep;
u_long	newsrc = 0;
u_long	src;
time_t	seconds;
struct tm lt;


    lt.tm_sec = 0;

/*  Read the previous version of the data file  */

    hfp = fopen( DUMPFILE, "r" );

    if (hfp == NULL )
	{
	perror( " Error opening load file for reading" );
	return;
	}

/*
*<gw ip address>  <dst ip address> <metric> <date time>
*/

    for ( ;; )
	{
	cc = (int)fgets( line, 199, hfp );
	if( cc == NULL )
	    {
	    if ( gwcnt == 0 )
		perror( " Error at start of load" );
	    break;
	    }

	cc = sscanf( line, "%s  %s %d %2d/%2d/%2d %2d%2d",
		    gwipstr,
		    dstipstr,
		    &metric,
		    &lt.tm_year,
		    &lt.tm_mon,
		    &lt.tm_mday,
		    &lt.tm_hour,
		    &lt.tm_min );

	if ( cc < 7 )
	    {
	    perror( " Error reading load file destination" );
	    continue;
	    }

	newsrc = inet_addr( gwipstr );
	if ( newsrc != src )
	    {
	    src = newsrc;
	    gwp = locate_Gw( src );
	    }
	lt.tm_mon--;
	now = tm_to_time( &lt );
	dstn = inet_addr( dstipstr );
	bcopy ( &dstn, &(netinfo.rip_dst.sa_data[2]), 4 );
	netinfo.rip_metric = metric;
/*
	{
	int ip1,ip2,ip3,ip4;

	ip1 = (u_char)netinfo.rip_dst.sa_data[2];
	ip2 = (u_char)netinfo.rip_dst.sa_data[3];
	ip3 = (u_char)netinfo.rip_dst.sa_data[4];
	ip4 = (u_char)netinfo.rip_dst.sa_data[5];

	printf( "%s	%d.%d.%d.%d	%d\n",
				gwipstr, ip1, ip2, ip3, ip4, metric );
	}
 */
	rep = NULL;
	dstp = locate_Dst( gwp, rep, &netinfo );
	rep = locate_route_ent ( gwp, dstp, rep, &netinfo );
	update_route_ent ( gwp, dstp, rep, &netinfo );
	}

    fclose( hfp );
    printf(
	" em_ripwatch: %d gateways, %d destinations, and %d routing entries\n",
				gwcnt, dstcnt, recnt );
    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    struct route_Dst * dstp;
    struct route_Gw * gwp;
    int     hash;
    int     i;
    int     j;
    char    scrstr[20];
    int     jsfd;		/* Journal Server socket file descriptor */
    FILE *  jsfp;

    jsfd = getnetconnect( js_addr, js_port );
    jsfp = fdopen( jsfd, "w" );

    i = 0;
    for( hash=0; hash<HASHSIZE; hash++ )
	{
	for ( gwp = gwtab[hash]; gwp != NULL; gwp = gwp->next_Gw )
	    {
	    i++;
	    fprintf( jsfp, "INTF  %s\n", pin );

	    strcpy( scrstr, inet_ntoa(  gwp->Gw_addr ) );
	    fprintf( jsfp, "PROT-ADDR:\tAF_INET %s\n", scrstr );

	    strcpy( scrstr, ether_ntoa( gwp->eaddr ) );
	    fprintf( jsfp, "ETHER-ADDR:\t%s\n", scrstr );

	    fprintf( jsfp, "\n\n" );
	    }
	}
    fprintf( stderr,
	    " em_ripwatch:%4d INTF entries sent to Journal Server\n", i );

    i = 0;
    for( hash=0; hash<HASHSIZE; hash++ )
	{
	for ( dstp = dsttab[hash]; dstp != NULL; dstp = dstp->next_Dst )
	    {
	    if ( dstp->ipRouteDest.s_addr == 0 )
		continue;

	    for ( j=0; j<num_net_intf; j++ )
		if ( ( ( net_intf[j].addr ^ dstp->ipRouteDest.s_addr ) &
						net_intf[j].nmask ) == 0 )
		    if ( ( dstp->ipRouteDest.s_addr & ~ net_intf[j].smask )
							!= 0 )
			break;

	    if ( j < num_net_intf )
		continue;

	    i++;
	    fprintf( jsfp, "NETWORK  %s\n", pin );

	    strcpy( scrstr, inet_ntoa(  dstp->ipRouteDest ) );
	    fprintf( jsfp, "NET_NUMB:\tAF_INET %s\n", scrstr );

	    if ( dstp->NetMask.s_addr != 0 )
		{
		strcpy( scrstr, inet_ntoa(  dstp->NetMask ) );
		fprintf( jsfp, "SUBNET-MASK:\tAF_INET %s\n", scrstr );
		}

	    fprintf( jsfp, "\n\n" );
	    }
	}
    fprintf( stderr,
	    " em_ripwatch:%4d NETWORK entries sent to Journal Server\n", i );

    fclose( jsfp );
    return;
}

