/*
 * logfile.c - routines for updating log and snapshot files
 *
 */

#include <sys/param.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <stdio.h>

#include "ripwatch.h"
#include "externs.h"


/*
 * update_logfile - put out the information to the log file.
 */

void
update_logfile()
{
    char *tstr;
    char *ctime();
    float percent;
    char buf[BUFSIZ];
    struct timeval tv;
    register int i;

    (void) gettimeofday(&tv, (struct timezone *) 0);

    /*
     * Start a log entry.
     */
    (void) fprintf(logfp, "#\n# begin\n#\n");
    (void) fprintf(logfp, "Date: %.24s\n", ctime(&tv.tv_sec));

    tv.tv_sec -= starttime.tv_sec;
    tstr = prtime(tv.tv_sec);

    (void) fprintf(logfp, "Run Time: %d\n", runtime);
    (void) fprintf(logfp, "Elapsed Time: %.8s\n", tstr+11);

    /*
     * Print total packet counters.
     */
    (void) fprintf(logfp, "#\n# total packets     %8s %8s\n#\n",
		"network", "dropped");
    (void) fprintf(logfp, "Interval Packets:   %8d %8d\n",
		int_pkt_total, int_pkt_drops);
    (void) fprintf(logfp, "Total Packets:      %8d %8d\n",
		pkt_total, pkt_drops);

    /*
     * Put out a header for the packet counters.
     */
    (void) fprintf(logfp, "#\n# packet counters         %8s %8s %8s\n#\n",
		"int", "pct", "total");

    /*
     * Print the packet counters.  Percentage is calculated as
     * this interval counter over total packets this interval.
     */
    for (i = 0; i < PKT_NCOUNTERS; i++)
	{
	if (int_pkt_total)
	    percent = ((float) pkt_counters[i].pc_interval /
				  (float) int_pkt_total) * 100.0;
	else
	    percent = 0.0;

	(void) sprintf(buf, "%s:", pkt_counters[i].pc_name);
	(void) fprintf(logfp, "%-25s %8d %7.0f%% %8d\n",
			buf, pkt_counters[i].pc_interval, percent,
			pkt_counters[i].pc_total);
	}


    /*
     * End the log entry.
     */
    (void) fprintf(logfp, "#\n# end\n#\n");
    (void) fflush(logfp);
}

