/*
 * nit.c - routines for messing with the network interface tap.
 *
 */

#ifdef sun
#include <sys/param.h>
#include <sys/stropts.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <net/if.h>
#include <signal.h>
#include <stdio.h>

#ifndef ultrix
#include <net/if_arp.h>
#endif /* ultrix */

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>

#ifdef ultrix
#include "ultrix.map.h"
#include "ipports.h"
#endif /* ultrix */

#include <net/nit_if.h>
#include <net/nit_pf.h>
#include <net/packetfilt.h>
#include <net/nit_buf.h>

#include "ripwatch.h"
#include "externs.h"

#if NOFILE > 64
#define SUNOS40 1
#endif /* NOFILE */

/*
 * setup_nit_dev - set up the network interface tap.
 */
int
setup_nit_dev(device)
char **device;
{
int s, fd;
u_int chunksz;
char *strdup();
u_long if_flags;
char buf[BUFSIZ];
struct ifreq ifr;
struct ifconf ifc;
struct strioctl si;
struct timeval timeout;

struct packet_template {
	struct ether_header ether_header;
	struct ip iphdr;
	struct udphdr udphdr;
	} pt;
#define eh pt.ether_header

/* int	i; */
struct packetfilt pf;

u_short *fwp = pf.Pf_Filter;

/*
u_short offset_Type = {((u_int)&eh.ether_type - (u_int)&eh.ether_dhost)/
							sizeof(u_short)};
u_short offset_Proto = {((u_int)&pt.iphdr.ip_p - (u_int)&eh.ether_dhost)/
							sizeof(u_short)};
u_short offset_Port = {((u_int)&pt.udphdr - (u_int)&eh.ether_dhost)/
							sizeof(u_short)};
 */

u_short offset_Type = { 2 * sizeof(struct ether_addr) /	sizeof(u_short) };
u_short offset_Proto = { ( sizeof(struct ether_header) +
			   ( (char *)&pt.iphdr.ip_p - (char *)&pt.iphdr ) )/
			  sizeof(u_short)};
u_short offset_Port = {( sizeof(struct ether_header) +
			 sizeof(struct ip) ) /	sizeof(u_short)};
/*
 *  printf( "Offsets %d  %d  %d\n", offset_Type, offset_Proto, offset_Port);
 */
	/*
	 * Set up the packet filter to select only IP/UDP/RIP packets.
	 * Offsets are from the beginning of the Ethernet header
	 *  and in units of u_shorts.
	 */
	
	*fwp++ = ENF_PUSHWORD + offset_Port;
	*fwp++ = ENF_PUSHLIT | ENF_CAND;
	*fwp++ = IPPORT_ROUTESERVER;

	*fwp++ = ENF_PUSHWORD + offset_Proto;
	*fwp++ = ENF_PUSHLIT | ENF_AND;
	*fwp++ = htons(0x00FF);
	*fwp++ = ENF_PUSHLIT | ENF_CAND;
	*fwp++ = htons(IPPROTO_UDP);

	*fwp++ = ENF_PUSHWORD + offset_Type;
	*fwp++ = ENF_PUSHLIT | ENF_EQ;
	*fwp++ = htons(ETHERTYPE_IP);

	pf.Pf_FilterLen = fwp - &pf.Pf_Filter[0];

	/*
	 * If the interface device was not specified,
	 * get the default one.
	 */

	if (*device == NULL) {
		/*
		 * Grab a socket.
		 */
		if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			error("socket");
			finish(-1);
		}

		ifc.ifc_buf = buf;
		ifc.ifc_len = sizeof(buf);

		/*
		 * See what device it's attached to.
		 */
/*		if (ioctl(s, SIOCGIFCONF, (char *) &ifc) < 0) { */
		if (ioctl(s, SIOCGIFCONF, &ifc) < 0) {
			error("ioctl: SIOCGIFCONF");
			finish(-1);
		}

		*device = strdup(ifc.ifc_req->ifr_name);
		(void) close(s);
	}

	/*
	 * We want the ethernet in promiscuous mode, and
	 * we want to know about dropped packets.
	 */
	if ( non_prom == 0 )
	    if_flags = NI_DROPS | NI_PROMISC;
	else
	    if_flags = NI_DROPS;

	/*
	 * Open the network interface tap.
	 */
	if ((fd = open(NIT_DEV, O_RDONLY)) < 0) {
		error("nit: open");
		finish(-1);
	}

	/*
	 * Arrange to get discrete messages.
	 */
/*	if (ioctl(fd, I_SRDOPT, (char *) RMSGD) < 0) { */
	if (ioctl(fd, I_SRDOPT, RMSGN) < 0) {
		error("ioctl: I_SRDOPT");
		finish(-1);
	}


	/*
	 * Push and configure the nit packet filtering module.
	 */
	if (ioctl(fd, I_PUSH, NIT_PF) < 0) {
		error("ioctl: I_PUSH NIT_PF");
		finish(-1);
	}

	/*
	 * Activate the packet filter
	 */
	si.ic_cmd = NIOCSETF;
	si.ic_len = sizeof(struct packetfilt);
	si.ic_dp = (char *) &pf;

/*	if (ioctl(fd, I_STR, (char *) &si) < 0) { */
	if (ioctl(fd, I_STR, &si) < 0) {
		error("ioctl: I_STR NIOCSETF");
		finish(-1);
	}

	/*
	 * Push and configure the nit buffering module.
	 */
	if (ioctl(fd, I_PUSH, NIT_BUF) < 0) {
		error("ioctl: I_PUSH NIT_BUF");
		finish(-1);
	}

	/*
	 * Set the read timeout.
	 */
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;

	si.ic_cmd = NIOCSTIME;
	si.ic_timout = INFTIM;
	si.ic_len = sizeof(timeout);
	si.ic_dp = (char *) &timeout;

	if (ioctl(fd, I_STR, (char *) &si) < 0) {
		error("ioctl: I_STR NIOCSTIME");
		finish(-1);
	}

	/*
	 * Set the chunk size.
	 */
	chunksz = NIT_CHUNKSIZE;

	si.ic_cmd = NIOCSCHUNK;
	si.ic_len = sizeof(chunksz);
	si.ic_dp = (char *) &chunksz;

/*	if (ioctl(fd, I_STR, (char *) &si) < 0) { */
	if (ioctl(fd, I_STR, &si) < 0) {
		error("ioctl: I_STR NIOCSCHUNK");
		finish(-1);
	}

	/*
	 * Configure the network interface tap by binding it
	 * to the underlying interface, setting the snapshot
	 * length, and setting the flags.
	 */
	(void) strncpy(ifr.ifr_name, *device, sizeof(ifr.ifr_name));
	ifr.ifr_name[sizeof(ifr.ifr_name)-1] = '\0';

	si.ic_cmd = NIOCBIND;
	si.ic_len = sizeof(ifr);
	si.ic_dp = (char *) &ifr;

	/*
	 * If the bind fails, there's no such device.
	 */
/*	if (ioctl(fd, I_STR, (char *) &si) < 0) { */
	if (ioctl(fd, I_STR, &si) < 0) {
		/*
		error("ioctl: I_STR NIOCBIND");
		finish(-1);
		 */
		close(fd);
		return(-1);
	}

	/*
	 * SNAP is buggy on SunOS 4.0.x
	 */
/*
#ifndef SUNOS40
	si.ic_cmd = NIOCSSNAP;
	si.ic_len = sizeof(truncation);
	si.ic_dp = (char *) &truncation;

	if (ioctl(fd, I_STR, (char *) &si) < 0) {
		error("ioctl: I_STR NIOCSSNAP");
		finish(-1);
	}
#endif /* SUNOS40 */
/*  */
	si.ic_cmd = NIOCSFLAGS;
	si.ic_len = sizeof(if_flags);
	si.ic_dp = (char *) &if_flags;

/*	if (ioctl(fd, I_STR, (char *) &si) < 0) { */
	if (ioctl(fd, I_STR, &si) < 0) {
		error("ioctl: I_STR NIOCSFLAGS");
		finish(-1);
	}

	return(fd);
}

/*
 * flush_nit - flush data from the nit.
 */
void
flush_nit(fd)
int fd;
{
/*	if (ioctl(fd, I_FLUSH, (char *) FLUSHR) < 0) { */
	if (ioctl(fd, I_FLUSH, FLUSHR) < 0) {
		error("ioctl: I_FLUSH");
		finish(-1);
	return;
	}
}
#endif /* sun */
