
/*
 * pktfilter.c - filters to count the packets.
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <net/if.h>

#ifndef ultrix
#include <net/if_arp.h>
#endif /* ultrix */

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include <protocols/routed.h>

#ifdef sun
#  define NND	1
#  include <sun/ndio.h>
#endif /* sun */

#include <stdio.h>

#include "ripwatch.h"
#include "externs.h"

#ifdef ultrix
#include "ultrix.map.h"
#include "ipports.h"
#endif /* ultrix */


struct ether_header eheader;

/*
 * Ethernet broadcast address.
 */

static	struct ether_addr ether_broadcast = {
	0xff, 0xff, 0xff, 0xff, 0xff, 0xff
};


/*
 * pkt_filter - count a packet, and pass it off to the appropriate filter.
 */

void
pkt_filter(cp, length)
u_int length;
char *cp;
{
    int len;
    int bdcst;
    int packet[PACKETSIZE];
    struct ip *ip = (struct ip *)packet;

    /*
     * Count this packet in the network totals.
     */
    int_pkt_total++;
    pkt_total++;

    /*
     * Extract the ethernet header.
     */
    (void) bcopy(cp, (char *) &eheader, sizeof(struct ether_header));

    /*
     * Align the entire packet!  (a bit of overkill?)
     */
    if ( ( len = length - sizeof(struct ether_header) ) > PACKETSIZE )
    	len = PACKETSIZE;
    (void) bcopy( cp + sizeof(struct ether_header), (char *) packet, len);

    /*
     * See if it's a broadcast packet, and count it if it is.
     */
#ifdef	ultrix
    bdcst = !bcmp((char *) eheader.ether_dhost, (char *) &ether_broadcast,
			sizeof(struct ether_addr));
#else
    bdcst = !bcmp((char *) &eheader.ether_dhost, (char *) &ether_broadcast,
			sizeof(struct ether_addr));
#endif	ultrix

    if (bdcst)
	{
	pkt_counters[PKT_BROADCAST].pc_interval++;
	pkt_counters[PKT_BROADCAST].pc_total++;
	}
/*
    else
	{
        u_short s1,s2,s3, d1,d2,d3;
	char srcip[20];
	char dstip[20];

	s1 = 256 * eheader.ether_shost.ether_addr_octet[0]
		 + eheader.ether_shost.ether_addr_octet[1];
	s2 = 256 * eheader.ether_shost.ether_addr_octet[2]
		 + eheader.ether_shost.ether_addr_octet[3];
	s3 = 256 * eheader.ether_shost.ether_addr_octet[4]
		 + eheader.ether_shost.ether_addr_octet[5];
	strcpy ( srcip, inet_ntoa(  ip->ip_src ) );
	d1 = 256 * eheader.ether_dhost.ether_addr_octet[0]
		 + eheader.ether_dhost.ether_addr_octet[1];
	d2 = 256 * eheader.ether_dhost.ether_addr_octet[2]
		 + eheader.ether_dhost.ether_addr_octet[3];
	d3 = 256 * eheader.ether_dhost.ether_addr_octet[4]
		 + eheader.ether_dhost.ether_addr_octet[5];
	strcpy ( dstip, inet_ntoa(  ip->ip_dst ) );

	printf( "src= %04x.%04x.%04x(%s), dst= %04x.%04x.%04x(%s)\n",
		s1,s2,s3, srcip,  d1,d2,d3, dstip );
	}
 */

    /*
     * Figure out what kind of packet it is, and pass
     * it off to the appropriate filter.
     */
    if ( ntohs(eheader.ether_type) != ETHERTYPE_IP )
	{
	pkt_counters[PKT_OTHER].pc_interval++;
	pkt_counters[PKT_OTHER].pc_total++;
	return;
	}

    /*
     * Count it in the totals and pass it off.
     */
    int_dst_pkt_total++;
    dst_pkt_total++;

    ip_filter( ip, length, &ip->ip_src, &ip->ip_dst );
}

/*
 * ip_filter - strip off the IP header and pass off to the appropriate
 *	       filter.
 */
void
ip_filter(ip, length, src, dst)
struct ip *ip;
u_int length;
struct in_addr * src;
struct in_addr * dst;
{
u_int *data;
int datalength;
int type;

    data = (u_int *) ip;
    data += ip->ip_hl;
    datalength = ntohs(ip->ip_len) - (4 * ip->ip_hl);

    if ( ip->ip_p != IPPROTO_UDP )
	return;

    udp_filter((struct udphdr *) data, datalength, src, dst);
    return;
}

/*
 * udp_filter - count UDP packets, pass RPC packets to the RPC filter.
 */
void
udp_filter(udp, length, src, dst)
struct udphdr *udp;
struct in_addr * src;
struct in_addr * dst;
int length;
{
    int     default_metric;
    int     i;
    struct route_ent * rep = NULL;
    struct rip * rip;
    struct route_Gw * gwp;
    struct route_Dst * dstp;
    struct netinfo * netinfo;

    /*
     * Count as a UDP packet.
     */
    pkt_counters[PKT_UDP].pc_interval++;
    pkt_counters[PKT_UDP].pc_total++;

    if ( ntohs(udp->uh_sport) != IPPORT_ROUTESERVER )
	return;

    rip = (struct rip *)((char *)udp + sizeof( struct udphdr ) );
    if ( rip->rip_cmd != RIPCMD_RESPONSE || rip->rip_vers < 1 )
	return;

    pkt_counters[PKT_ROUTING].pc_interval++;
    pkt_counters[PKT_ROUTING].pc_total++;

    gwp = locate_Gw( src->s_addr );

    bcopy( eheader.ether_shost, &gwp->eaddr, sizeof(struct ether_addr) );

    /* Process advertisements */

    default_metric = -1;

    for ( netinfo = (struct netinfo *)rip->rip_nets;
			(int)(netinfo + 1 ) <= (int)udp + length;
			netinfo++ )
			/*
			(int)netinfo + sizeof(struct netinfo) <=
							(int)udp + length;
			netinfo += sizeof(struct netinfo) )
			 */
	{
	if ( netinfo->rip_dst.sa_family != (u_short)2 )
	    {
	    int data[14], family, i;
	    char srcip[20];

	    family = netinfo->rip_dst.sa_family;
	    if ( family == 0 )
		continue;

	    for ( i=0; i<14; i++ )
		data[i] = netinfo->rip_dst.sa_data[i];
	    strcpy ( srcip, inet_ntoa(  src ) );

	    printf( "%s  %d  %2x.%2x.%2x.%2x.%2x.%2x Non-IP routing info\n",
				srcip, family, data[0], data[1], data[2],
				data[3], data[4], data[5] );
	    continue;
	    }

	dstp = locate_Dst( gwp, rep, netinfo );
	rep = locate_route_ent ( gwp, dstp, rep, netinfo );
	update_route_ent ( gwp, dstp, rep, netinfo );
	}

    return;
}
