/*
 * ripwatch.h - definitions for ripwatch.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <net/if.h>
#include <netinet/if_ether.h>

#define HASHSIZE 256

struct route_Gw {
	struct in_addr Gw_addr;
	struct route_Gw * next_Gw;	/* Next GW with same hash */
	struct route_ent * first_Dst;	/* First DST for this GW */
	struct route_ent * last_Dst;	/* Last  DST for this GW */
	struct ether_addr eaddr		/* Ethernet address */
	};

struct route_Dst {
	struct in_addr ipRouteDest;
	struct in_addr NetMask;
	struct route_Dst * next_Dst;	/* Next DST with same hash */
	struct route_ent * first_Gw;	/* Best route to DST */
	struct route_ent * last_Gw;	/* Worst route to DST */
	};

struct route_ent {
	struct in_addr ipRouteDest;
	short ipRouteMetric1;
	short ipRouteMetric2;
	short ipRouteMetric3;
	short ipRouteMetric4;
	struct in_addr ipRouteNextHop;
	struct in_addr Gw_addr;
	short ipRouteProto;
	struct in_addr NetMask;
	u_int	tv_seconds;		/* Date/Time */
	struct route_ent * next_Dst;	/* Next DST for this GW */
	struct route_ent * better_Gw;	/* Better GW for this DST */
	struct route_ent * next_Gw;	/* Next GW for this DST */
	};

struct net_intf {
	u_long	addr;
	u_long	nmask;
	u_long	smask;
	};

int   gwcnt;				/* count of entries in hash table */
int   dstcnt;				/* count of entries in hash table */
int   recnt;				/* count of entries in routing array */

struct route_Dst * dsttab[HASHSIZE];
struct route_Gw  * gwtab[HASHSIZE];

#define RP_RIP	1

/*
 * General definitions.
 */

#ifndef TRUE
#define TRUE		1
#define FALSE		0
#endif /* TRUE */

#define LOGFILE		"ripwatch.log"	/* log file name		*/
#define RUNTIME		125		/* default run time		*/
#define PACKETSIZE	4096		/* max size of a packet		*/
#define MAXINTERFACES	16		/* Max. number of interfaces	*/
#define DUMPFILE	"ripwatch.dump"	/* dump file name		*/


/*
 * Network Interface Tap (NIT) definitions.
 */
#define NIT_DEV		"/dev/nit"	/* network interface tap device	*/
#define NIT_BUF		"nbuf"		/* nit stream buffering module	*/
#define NIT_PF		"pf"		/* nit packet filtering module	*/
#define NIT_CHUNKSIZE	8192		/* chunk size for grabbing pkts	*/

/*
 * Packet counter definitions.
 */

#define PKT_NCOUNTERS	4		/* number of packet counters	*/

#define PKT_UDP		0		/* UDP packets			*/
#define PKT_ROUTING	1		/* routing control packets	*/
#define PKT_BROADCAST	2		/* ethernet broadcast packets	*/
#define PKT_OTHER	3		/* none of the above packets	*/

typedef unsigned long	Counter;

/*
 * Packet counting structure.
 */

typedef struct {
	char	*pc_name;		/* name of counter		*/

	Counter	pc_interval;		/* packets this interval	*/
	Counter	pc_total;		/* packets since start		*/

	short	pc_intx, pc_inty;	/* screen coords of pc_interval	*/
	short	pc_totx, pc_toty;	/* screen coords of pc_total	*/
	short	pc_pctx, pc_pcty;	/* screen coords of percentage	*/
	short	pc_namex, pc_namey;	/* screen coords of pc_name	*/
	} PacketCounter;

/*
 * Definitions for earlier systems which don't have these from 4.3BSD.
 */

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN	64
#endif /* MAXHOSTNAMELEN */

#ifndef NFDBITS
  typedef long		fd_mask;

# define NFDBITS	(sizeof(fd_mask) * NBBY)

# define FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
# define FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
# define FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
# define FD_ZERO(p)	(void) bzero((char *)(p), sizeof(*(p)))
#endif /* NFDBITS */

