
/*
 * storageman.c - storage management
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <net/if.h>

#ifndef ultrix
#include <net/if_arp.h>
#endif /* ultrix */

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include <protocols/routed.h>

#ifdef sun
#  define NND	1
#  include <sun/ndio.h>
#endif /* sun */

#include <stdio.h>

#include "ripwatch.h"
#include "externs.h"

#ifdef ultrix
#include "ultrix.map.h"
#include "ipports.h"
#endif /* ultrix */


/*
 *  order_re  - Place route_ent into proper order for this destination
 */

/*
order_re ( dstp, rep )
struct route_Dst * dstp;
struct route_ent * rep;
{
struct route_ent ** tmpp;
struct route_ent * xx;

    if ( dstp->first_Gw == NULL )
	{
	dstp->first_Gw = rep;
	dstp->last_Gw  = rep;
	return;
	}

    tmpp = dst->first_Gw;
    if ( tmpp != rep )
	{
	if ( rep->ipRouteMetric1 < tmpp->ipRouteMetric1 )
	    {
	    xx = rep->next_Gw;
	    rep->next_Gw = tmpp->next_Gw;
	    tmpp->next_Gw = xx;
	    rep->next_Gw->better_Gw = rep;
	    rep->better_Gw = NULL;
	    return;
	    }
	}
    for ( tmpp = tmpp->next_Gw; tmpp != NULL; tmpp = tmpp->next_Gw )
	{
	if ( tmpp != rep )
	    if ( rep->ipRouteMetric1 < tmpp->ipRouteMetric1 )
		{
		xx = rep->next_Gw;
		rep->next_Gw = tmpp->next_Gw;
		tmpp->next_Gw = xx;
		rep->next_Gw->better_Gw = rep;
		rep->better_Gw = tmpp;
		}
	
	}
}
 */

/*
 *  add_new_re  - Add new route_ent entry
 */

struct route_ent *
add_new_re ( gwp, dstp, netinfo )
struct route_Gw * gwp;
struct route_Dst * dstp;
struct netinfo * netinfo;
{
struct route_ent * rep;

    rep = (struct route_ent *)malloc( sizeof( struct route_ent ) );
    bcopy( &netinfo->rip_dst.sa_data[2], &rep->ipRouteDest, 4 );
    rep->ipRouteMetric1 = netinfo->rip_metric;
    rep->ipRouteMetric2 = 0;
    rep->ipRouteMetric3 = 0;
    rep->ipRouteMetric4 = 0;
    rep->ipRouteNextHop.s_addr = 0;
    rep->Gw_addr.s_addr = gwp->Gw_addr.s_addr;
    rep->ipRouteProto = RP_RIP;
    rep->NetMask.s_addr = 0;
    rep->tv_seconds = now;
    rep->next_Dst = 0;
    rep->better_Gw = 0;
    rep->next_Gw = 0;

    if ( rep->ipRouteDest.s_addr == 0 )
	printf( "  default metric %2d", netinfo->rip_metric );

    /* Add to end of chain of Dst's for this GW */

    if ( gwp->first_Dst == NULL )
	gwp->first_Dst = rep;

    if ( gwp->last_Dst != NULL )
	gwp->last_Dst->next_Dst = rep;

    gwp->last_Dst = rep;

    /* Add to end of chain of GW's for this Dst */

    if ( dstp->first_Gw == NULL )
	dstp->first_Gw = rep;

    if ( dstp->last_Gw != NULL )
	{
	rep->better_Gw = dstp->last_Gw;
	dstp->last_Gw->next_Gw = rep;
	}

    dstp->last_Gw = rep;
    recnt++;

/*
    order_re( dstp, rep );
 */

    return( rep );
}


/*
 *  add_new_gw  -  Add new gw
 */

struct route_Gw *
add_new_gw ( src )
u_long	src;
{
struct route_Gw * gwp;

    gwp = (struct route_Gw *)malloc( sizeof(struct route_Gw) );
    gwp->Gw_addr.s_addr = src;
    gwp->next_Gw = NULL;
    gwp->first_Dst = NULL;
    gwp->last_Dst = NULL;
    gwcnt++;
    printf( "\n em_ripwatch:  %-16s", inet_ntoa( gwp->Gw_addr ) );
    return( gwp );
}


/*
 *  add_new_dst  - Add new route_Dst entry
 */

struct route_Dst *
add_new_dst ( gwp, dstp, rep, netinfo )
struct route_Gw * gwp;
struct route_Dst * dstp;
struct route_ent * rep;
struct netinfo * netinfo;
{
    dstp = (struct route_Dst *)malloc( sizeof(struct route_Dst) );
    bcopy( &netinfo->rip_dst.sa_data[2], &dstp->ipRouteDest, 4 );
    dstp->NetMask.s_addr = 0;
    dstp->next_Dst = NULL;
    dstp->first_Gw = NULL;
    dstp->last_Gw = NULL;
    dstcnt++;
    rep = add_new_re( gwp, dstp, netinfo );
    return( dstp );
}


/*
 *  hashit - hash the IP address
 */

hashit ( bp )
u_char * bp;
{
    return( ( bp[0] + bp[1] + bp[2] + bp[3] ) % HASHSIZE );
}


/*
 *  locate_Gw  - Find GW entry
 */

struct route_Gw *
locate_Gw ( src )
u_long src;
{
struct route_Gw * gwp;
u_int	hash;

    hash = hashit( (u_char *)&src );

    /* Check for existing entry in hash table */

    if ( gwtab[hash] == 0 )
	{
	gwtab[hash] = add_new_gw( src );
	return( gwtab[hash] );
	}

    gwp = gwtab[hash];
    while ( gwp->Gw_addr.s_addr != src && gwp->next_Gw != NULL )
	gwp = gwp->next_Gw;

    /* Found match or at end of chain */

    if ( gwp->Gw_addr.s_addr != src )
	{
	gwp->next_Gw = add_new_gw( src );
	gwp = gwp->next_Gw;
	}

    /*  At correct GW entry */
    return( gwp );
}


/*
 *  locate_Dst  - Find DST entry
 */

struct route_Dst *
locate_Dst ( gwp, rep, netinfo )
struct route_Gw * gwp;
struct route_ent * rep;
struct netinfo * netinfo;
{
u_long	dstn;
struct route_Dst * dstp;
u_int	hash;

    bcopy( &netinfo->rip_dst.sa_data[2], &dstn, 4 );

    hash = hashit( &netinfo->rip_dst.sa_data[2] );

    /* Check for existing entry in hash table */

    if ( dsttab[hash] == 0 )
	{
	dsttab[hash] = add_new_dst( gwp, dstp, rep, netinfo );
	return( dsttab[hash] );
	}

    dstp = dsttab[hash];
    while ( dstp->ipRouteDest.s_addr != dstn && dstp->next_Dst != NULL )
	dstp = dstp->next_Dst;

    /* Found match or at end of chain */

    if ( (dstp->ipRouteDest.s_addr) != dstn )
	{
	dstp->next_Dst = add_new_dst( gwp, dstp, rep, netinfo );
	dstp = dstp->next_Dst;
	}

    /*  At correct DST entry */

    return( dstp );
}


/*
 *  locate_route_ent - Locate route entry
 */

struct route_ent *
locate_route_ent ( gwp, dstp, rep, netinfo )
struct route_Gw * gwp;
struct route_Dst * dstp;
struct route_ent * rep;
struct netinfo * netinfo;
{
struct route_ent * advp;
u_long	dstn;

    bcopy( &netinfo->rip_dst.sa_data[2], &dstn, 4 );

    if ( rep != NULL )
	if ( rep->ipRouteDest.s_addr == dstn &&
		 rep->Gw_addr.s_addr == gwp->Gw_addr.s_addr )
	    return( rep );

    for ( advp = dstp->first_Gw; advp->next_Gw != NULL; advp = advp->next_Gw )
	if ( advp->ipRouteDest.s_addr == dstn &&
		 advp->Gw_addr.s_addr == gwp->Gw_addr.s_addr )
	    return ( advp );

    if ( advp->ipRouteDest.s_addr == dstn &&
	     advp->Gw_addr.s_addr == gwp->Gw_addr.s_addr )
	return ( advp );
    
    advp->next_Gw = add_new_re( gwp, dstp, netinfo );

    return( advp->next_Gw );
}


/*
 *  Update_route_ent - Update route entry
 */

struct route_ent *
update_route_ent ( gwp, dstp, rep, netinfo )
struct route_Gw * gwp;
struct route_Dst * dstp;
struct route_ent * rep;
struct netinfo * netinfo;
{
    /*  At correct DST entry  */

    rep->tv_seconds = now;
    if ( rep->ipRouteMetric1 != netinfo->rip_metric )
	{
	rep->ipRouteMetric1 = netinfo->rip_metric;
	/*
	order_re( dstp, rep );
	 */
	}

    return;
}
