/*
 * util.c - miscellaneous utility routines.
 *
 * David A. Curry				Jeffrey C. Mogul
 * SRI International				Digital Equipment Corporation
 * 333 Ravenswood Avenue			Western Research Laboratory
 * Menlo Park, CA 94025				100 Hamilton Avenue
 * davy@erg.sri.com				Palo Alto, CA 94301
 *						mogul@decwrl.dec.com
 *
 * $Log:	util.c,v $
 * Revision 3.0  91/01/23  08:23:29  davy
 * NFSWATCH Version 3.0.
 * 
 * Revision 1.4  91/01/17  10:18:08  davy
 * New features from Jeff Mogul.
 * 
 * Revision 1.6  91/01/07  15:35:20  mogul
 * Must rebuild client hash table after sorting
 * 
 * Revision 1.5  91/01/07  14:09:10  mogul
 * Improved sortbyusage stuff
 * 
 * Revision 1.4  91/01/04  14:14:27  mogul
 * Support for client counters
 * 
 * Revision 1.3  91/01/03  17:42:20  mogul
 * Support for per-procedure counters
 * 
 * Revision 1.2  90/08/17  15:47:50  davy
 * NFSWATCH Version 2.0.
 * 
 * Revision 1.1  88/11/29  11:21:05  davy
 * NFSWATCH Release 1.0
 * 
 */
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <curses.h>
#include <stdio.h>
#ifdef sun
#include <mntent.h>
#include <exportent.h>
#endif /* sun */
#ifdef ultrix
#include <sys/mount.h>
#endif /* ultrix */

#include "ripwatch.h"
#include "externs.h"

/*
 * clear_vars - set interval counters to zero.
 */
void
clear_vars()
{
	register int i;

	int_pkt_total = 0;
	int_pkt_drops = 0;
	int_dst_pkt_total = 0;

	for (i = 0; i < PKT_NCOUNTERS; i++)
		pkt_counters[i].pc_interval = 0;
}

/*
 * savestr - save a string in dynamically allocated memory.
 */
char *
savestr(s)
register char *s;
{
	char *malloc();
	register char *str;

	str = malloc(strlen(s) + 1);

	if (str == NULL) {
		(void) fprintf(stderr, "%s: out of memory.\n", pname);
		finish(-1);
	}

	(void) strcpy(str, s);
	return(str);
}

/*
 * prtime - convert a time to hh:mm:ss.
 */
char *
prtime(sec)
time_t sec;
{
	int hh, mm, ss;
	static char tbuf[16];

	hh = sec / 3600;
	sec %= 3600;

	mm = sec / 60;
	sec %= 60;

	ss = sec;

	(void) sprintf(tbuf, "%02d:%02d:%02d", hh, mm, ss);
	return(tbuf);
}

/*
 * error - print an error message preceded by the program name.
 */
void
error(str)
register char *str;
{
	char buf[BUFSIZ];

	(void) sprintf(buf, "%s: %s", pname, str);
	(void) perror(buf);
}

/*
 * finish - clean up and exit.
 */
void
finish(code)
int code;
{
	int i;
	struct timeval tv;

	/*
	 * Close the nit devices.
	 */

	for ( i=0; i<ninterfaces; i++ )
	    if (if_fd[i] >= 0)
		(void) close(if_fd[i]);

	if (logging) {
		(void) fprintf(logfp, "#\n# endlog\n#\n");
		(void) fclose(logfp);
	}

	printf( "\n" );
	dumptable();

	if (code < 0)
		exit(-code);

	gettimeofday( &tv, (struct timezone *) 0 );
	printf( " em_ripwatch took %d seconds\n\n",
						tv.tv_sec - starttime.tv_sec );
	exit(0);
}

/*
 * setup_pkt_counters - set up packet counter screen coordinates.
 */
void
setup_pkt_counters()
{
	register int i, j;

	(void) bzero((char *) pkt_counters,
		PKT_NCOUNTERS * sizeof(PacketCounter));

	/*
	 * Set up the strings.
	 */
	pkt_counters[PKT_UDP].pc_name = "UDP Packets";
	pkt_counters[PKT_ROUTING].pc_name = "Routing Control";
	pkt_counters[PKT_BROADCAST].pc_name = "Broadcast";
	pkt_counters[PKT_OTHER].pc_name = "Other Packets";

}


/*
 * usage - print a usage message and exit.
 */
void
usage()
{
	fprintf(stderr, "Usage: %s ", pname);
	fprintf(stderr, "\t[-dev device]\n\t[-allif]\n");
        fprintf(stderr, "\t[-lf logfile]\n");
	fprintf(stderr, "\t[-t timeout]\n");
        fprintf(stderr, "\t[-usage]\n\t[-fs]\n\t[-if]\n\t");
	fprintf(stderr, "[-l]\n");
	finish(-1);
}

