/*
 *  em_seqp.c  DCM Wood	93/01/17
 */

/*
 *	Sequential Ping Explorer Module
 *
 *  Get addresses of all interfaces on current system
 *  Determine connected network, subnet masks, and subnets
 *
 *  Set parameters from environment
 *  Set parameters from command line
 *  If Journal server location not set from env or cl,
 *    Get address from getpeername
 *  If input file specified,
 *    Open input file with  fd=0
 *  Read input file
 *    Set parameters from input file
 *    If EOF or blank line,
 *      Stop parameter processing
 *  If output file specified,
 *    Open output file with  fd=1
 *
 *  Read list of IP addresses and build list of destinations
 *  If no input or args, examine directly connected subnets (if small enough)
 *  For each destination, send a ping and wait one second for all replies
 *  Repeat for each destination that has received no reply.
 *
 *  Send IP addresses to the Journal Server
 */

#include <stdio.h>
#include <strings.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/sockio.h>
#include <sys/errno.h>

#include <net/if.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

#include <netdb.h>

#include "send_data.h"
#define SOCKADDR_IN
#include "data_server.h"
 
#define EM_MAIN
#include "em.h"

/*
#ifndef PIN
#include "pin.h"
#endif

char	pin[20]= PIN;
*/
extern	char	pin[];

struct in_addr	jsa;		/* Journal Server Address  */
int     jsp;			/* Journal Server Pointer  */

struct in_addr thishost;	/* Address of this host  */
struct timeval	tv;		/* Time for various uses  */
struct timezone	tz;		/* Time zone  */

#define DELAY	1000

extern char ** environ;

u_long	em_StartTime;
u_long	em_alb=0;			/* addr lower bound */
u_long	em_aub=0;			/* addr upper bound */
char	em_inputf[64]="\n";
int	em_ipi=DELAY;			/* interpacket interval in ms */
int	em_max_retries=2;		/* retry count */
int	em_max_ttl=30;
char	em_outputf[64]="";		/* output file */
char	em_prog[64];			/* program name */
int	em_ps=0;			/* packet size */
int	em_runtime;
int	em_snm=0;			/* mask request */
u_long	em_subnet=0;			/* subnet to examine */

u_long	js_addr=0;			/* journal server address */
int	js_port=DATA_PORT;

FILE *	em_ifp;
FILE *	jsfp;

struct dest {
	struct in_addr inaddr;
	struct in_addr mask;
	struct dest * previous;
	struct dest * next;
	int	nresponses;
	};

struct dest * destp;
struct dest * destFirst=NULL;
int	num_dest=0;

extern	char * strtok();

struct	ip_icmp {
	struct ip ip;
	struct icmp icmp; };

#define RBUFSIZE	1600
#define MAXPACKET	1500
#define MAX_INTF	20

struct net_intf {
	u_long	addr;
	u_long	mask;
	} net_intf[MAX_INTF];

int	datalen=64;
struct sockaddr_in from;
int	ident;
int	inflight;
int	num_net_intf=0;
int	retry;
u_char	rpacket[RBUFSIZE];
struct timeval td;
struct timeval ts;
int     len;

u_char	packet[MAXPACKET];
extern	int errno;

int	psfd;			/* packet socket file descriptor */

struct sockaddr_in whereto;	/* Who to ping */


/*
 *  AddDest
 *	Add sub-network and mask to list of destinations
 */

AddDest( addr, mask )
u_long	addr;
u_long	mask;
{
    struct dest * destp;
    struct dest * lastp;
    u_long  net;
    u_long  netmask;


    if ( destFirst == NULL )
	{
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	if ( destp == NULL )
	    {
	    fprintf( stderr, "em_seqp: could not malloc dest(1)\n" );
	    return;
	    }

	destFirst = destp;
	lastp = NULL;
	}
    else
	{
        for ( destp=destFirst; destp!=NULL; destp=destp->next )
	    {
	    if ( destp->inaddr.s_addr == addr )
		return;

	    lastp = destp;
	    }
	destp = (struct dest *)malloc( sizeof( struct dest ) );
	if ( destp == NULL )
	    {
	    fprintf( stderr, "em_seqp: could not malloc dest(2)\n" );
	    return;
	    }

	lastp->next = destp;
	}

    destp->inaddr.s_addr = addr;
    destp->mask.s_addr = mask;
    destp->previous = lastp;
    destp->next = NULL;
    destp->nresponses = 0;
    num_dest++;
    return;
}


/*
 *  GetIntf
 *	Get info for all interfaces on this system
 */

GetIntf()
{
    int     fd;
    int     host;
    int     i;
    struct ifconf  ifc;
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    u_long  max_host;
    u_long  ogw;

    /* Get some memory in which to store the ifreq entries      */

    ifc.ifc_len = MAX_INTF * sizeof( struct ifreq );

    ifc.ifc_req = (struct ifreq *)calloc( MAX_INTF, sizeof( struct ifreq ) );

    /* Open a socket to do ioctls on.                           */

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
        {        
        perror( "socket" );
        return (-1);
        }               
                        
    /* Do the ioctl, check the result.                          */
                 
    if ( ioctl( fd, SIOCGIFCONF, &ifc) == -1 )
        {
        perror( "ioctl SIOCGIFCONF");
        close( fd );
        return (-1);
        } 

    if ( ifc.ifc_len/sizeof(struct ifreq) == 1 )
        if ( debug )
	    printf( " There is only one interface on this system\n" );
    else
        if ( debug )
	    printf( " There are %d interfaces on this system\n",
                                        ifc.ifc_len/sizeof(struct ifreq) );
    ifreqp = ifc.ifc_req;
    ogw = 0;

    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
	{
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
        if ( debug )
	    printf( " %s   %-16s", ifreqp->ifr_name, inet_ntoa( inaddr ) );
	net_intf[num_net_intf].addr = inaddr.s_addr;
        if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
            {
            perror( "ioctl SIOCGIFNETMASK" );
            continue;
            }
        bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
        if ( debug )
	    printf( "   %-16s\n", inet_ntoa( inmask ) );
	net_intf[num_net_intf].mask = inmask.s_addr;
        if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
            continue;

	num_net_intf++;
        }

    if ( destFirst == NULL )
	for ( i=0; i<num_net_intf; i++ )
	    {
	    max_host = 0xffffffff & ~net_intf[i].mask;
	    if ( max_host * em_ipi / 1000 > 3600 )
		continue;

	    for ( host=1; host<max_host; host++ )
		AddDest( net_intf[i].addr & net_intf[i].mask | host,
							    net_intf[i].mask );
	    }

    close( fd );
    return( 0 );
}


/*
 *  ArgBit
 *	Process bit flag argument
 */

ArgBit( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep |= *altdflt;
    return( 0 );
}


/*
 *  ArgCS
 *	Process Character String argument
 */

ArgCS( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    strcpy( (char *)valuep, altdflt );
    return( 0 );
}


/*
 *  ArgInt
 *	Process Integer argument
 */

ArgInt( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = atoi( altdflt );
    return( 0 );
}


/*
 *  ArgIP
 *	Process Internet Protocol Address argument
 */

ArgIP( valuep, altdflt )
int *	valuep;
char *	altdflt;
{
    *valuep = inet_addr( altdflt );
    if ( *valuep < 1 )
	return ( -1 );

    return( 0 );
}


/*
 *  ProcessCL
 *	Process Command Line arguments
 */

ProcessCL( ac, av )
int	ac;
char ** av;
{
    struct arg
	{
	int	(*func)();
	char	flag;
	char *	text;
	int *	value;
	char *	altdflt;
	};

    static struct arg arg[]={
    { ArgCS,	'd',	"default directory",	(int*)defdir,	 NULL	     },
    { ArgCS,	'f',	"input file",		(int*)em_inputf, "\0"  },
    { ArgCS,	'o',	"output file",		(int*)em_outputf,"em_output" },

    { ArgIP,	'j',	"journal server addr",	(int*)&js_addr,		NULL },
    { ArgIP,	's',	"subnet to examine",	(int*)&em_subnet,	NULL },
    { ArgIP,	'l',	"addr lower bound",	(int*)&em_alb,		NULL },
    { ArgIP,	'u',	"addr upper bound",	(int*)&em_aub,		NULL },

    { ArgInt,	'p',	"port",			&js_port,	NULL	},
    { ArgInt,	'r',	"retry count",		&em_max_retries,"5"	},
    { ArgInt,	'v',	"interpacket interval",	&em_ipi,	"2500"	},
    { ArgInt,	'z',	"packet size",		&em_ps,		NULL	},
    { ArgInt,	't',	"run time",		&em_runtime,	"3"	},

    { ArgInt,	'm',	"mask request",		&em_snm,	"1"	},

    { ArgBit,	'x',	"debug flag",		&debug,		"1"	},

    { NULL,	'\0',	NULL,			NULL,		NULL	}
			};

    char   fc;
    int    i;
    char * progp;
    struct arg * ap;

    progp = av[0];
    ac--, av++;				/* Skip over program name */
    while (ac && *av[0] == '-')
	{
	fc = *++av[0];
	for ( ap=&arg[0]; ap->func!=NULL; ap++ )
	    {
	    if ( fc == ap->flag )
		{
		if ( ( ac == 1 ) || ( ( ac > 1 ) && ( *av[1] == '-' ) ) )
		    {
		    if ( ap->altdflt == NULL )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    else
			{
		        i = (*ap->func)( ap->value, ap->altdflt );
			if ( i < 0 )
			    {
			    fprintf( stderr, "%s error\n", ap->text );
			    goto arg_error;
			    }
			break;
			}
		    }
		else if ( ( ac > 1 ) && ( *av[1] != '-' ) )
		    {
		    ac--;
		    i = (*ap->func)( ap->value, av[1] );
		    if ( i < 0 )
			{
			fprintf( stderr, "%s error\n", ap->text );
			goto arg_error;
			}
		    av++;
		    break;
		    }
		else
		    {
		    fprintf( stderr, "%s error\n", ap->text );
		    goto arg_error;
		    }
		}  /* endif ( fc == ap->flag ) */
	    }  /* end for */
	ac--, av++;
	}  /* end while */

    arg_error:
	if (ac != 0)
	    {
	    fprintf( stderr, "Usage: %s\n", progp );
	    for ( i=0; arg[i].func!=NULL; i++ )
		fprintf( stderr, "	[-%c %s]\n", arg[i].flag, arg[i].text );
	    exit( -1 );
	    }

    return;
}


/*
 *  ProcessInputFile
 *	Open and read input file
 */

ProcessInputFile()
{
    int    ac;
    u_long addr;
    char * av[MAXARGS];
    int    buf[MAXLINE];
    int    cc;
    char * cp;
    int    ic;
    int    ifd;
    int    ll;
    struct sockaddr_in name;
    int    namelen;
    char * sp;

    if ( em_inputf[0] == '\n' )
	return;

    if ( em_inputf[0] == '\0' )
	{
	em_ifp = fdopen( 0, "r" );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, " Could not fdopen input.\n" );
	    exit( -1 );
	    }

	if ( jsa.s_addr == 0 )
	    {
	    namelen = sizeof( name );
	    cc = getpeername( 0, &name, namelen );
	    if ( cc == 0 && namelen != 0 )
		if ( name.sin_addr.s_addr != 0 )
		    jsa.s_addr = name.sin_addr.s_addr;
	    }
	}
    else
	{
	em_ifp = freopen( em_inputf, "r", 0 );
	if ( em_ifp == NULL )
	    {
	    fprintf( stderr, " Could not freopen input file %s.\n", em_inputf );
	    exit( -1 );
	    }
	}

    ac = 0;
    av[0] = "em_prog";
/*
    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	ll = strlen( buf );
	if ( ll <= 1 )
	    break;

	if ( strspn( buf, " \t\n" ) == ll )
	    break;

	cp = strtok( buf, " \t\n" );

	do {
	    if ( cp == NULL )
		break;

	    cc = strlen( cp );
	    sp = (char *)malloc( cc+1 );
	    strncpy( sp, cp, cc );
	    ac++;
	    av[ac] = sp;
	    cp = strtok( NULL, " \t\n" );
	    } while ( ac < MAXARGS );

	}
    
    if ( ac > 0 )
	ProcessCL( ac, av );
 */

    for (;;)
	{
	cp = fgets( buf, MAXLINE, em_ifp );
	if ( cp == NULL )
	    break;

	for(;;)
	    {
	    if ( *cp != ' ' )
		break;

	    cp++;
	    }

	addr = inet_addr( cp );
	if ( addr != -1 )
	    AddDest( addr, 0 );
	}

    fclose( em_ifp );
    return;
}


/*
 *  UpdateJS
 *	Update the Journal Server
 */

UpdateJS()
{
    struct dest * dp;
    int     i;
    int     jsfd;		/* Journal Server socket file descriptor */
    struct dest * sdp;

    jsfd = getnetconnect( js_addr, js_port );
    jsfp = fdopen( jsfd, "w" );

    i = 0;
    for ( dp=destFirst; dp!=NULL; dp=dp->next )
	{
	if ( dp->nresponses == 0 )
	    continue;

	fprintf( jsfp, "INTF  %s\n", pin );
	fprintf( jsfp, "PROT-ADDR:\tAF_INET %s\n", inet_ntoa( dp->inaddr ) );
	fprintf( jsfp, "\n\n" );
	i++;
	}

    fprintf( stderr, " %s: %d INTF entries sent to Journal Server\n",
								em_prog, i );
    fclose( jsfp );
    return;
}


main( argc, argv )
int	argc;
char **	argv;
{
    int     i;

    setlinebuf( stdout );
    strcpy( em_prog, argv[0] );
    printf( " %s: started\n", em_prog );

    ProcessCL( argc, argv );

    ProcessInputFile();

    GetIntf();

    /*  Start of code to do Exploring	*	*	*	/* */
    {								/* */
    struct dest * dp;
    int  i;
    struct dest * ldp;

    gettimeofday( &tv, &tz);
    em_StartTime = tv.tv_sec;
    fprintf( stderr, " em_seqp: %d destinations\n", num_dest );
    InitSend();

    for ( retry=0; retry<2; retry++ )
	{
	inflight = 0;
	for ( dp=destFirst; dp!=NULL; dp=dp->next )
	    {
	    if ( dp->nresponses != 0 )
		continue;

	    SendPing( &(dp->inaddr) );
	    WaitForResponses( em_ipi, dp );
	    ldp = dp;
	    }
	WaitForResponses( em_ipi, ldp );
	for ( i=0; i<10; i++ )
	    {
	    if ( inflight <= 0 )
		break;

	    WaitForResponses( em_ipi, ldp );
	    }
	}

    }								/* */
    /*  End of code to do Exploring	*	*	*	/* */

    /*  Print statistics  */

    gettimeofday( &tv, &tz);
    printf( " %s: took %d seconds\n\n", argv[0], tv.tv_sec - em_StartTime );

    UpdateJS();

    close( 0 );
    return( 0 );
}

/*
 *  WaitForResopnses
 */

WaitForResponses( delay, dp )
struct dest * dp;
int	delay;
{
    int     cc;
    fd_set  fds;
    int     fromlen;
    struct dest * sdp;

    gettimeofday( &ts, &tz);

    td.tv_sec = delay / 1000;
    td.tv_usec = delay % 1000;

    do {
	FD_ZERO( &fds );
	FD_SET( psfd, &fds );
	cc = select( psfd+1, &fds, (fd_set *)0, (fd_set *)0, &td );

	if ( cc == 0 )
	    return;

	if ( cc < 0 )
	    {
	    perror( "em_seqp: select" );
	    exit( -1 );
	    }

	len = RBUFSIZE;
	fromlen = sizeof( from );
	cc = recvfrom( psfd, rpacket, len, 0, &from, &fromlen );
	if ( cc < 0)
	    {
	    if( errno == EINTR )
				continue;
	    perror("ping: recvfrom");
	    continue;
	    }

	if ( ((struct ip_icmp *)rpacket)->icmp.icmp_type != ICMP_ECHOREPLY )
	    continue;			/* not an ICMP Echo_Reply */

	if ( ((struct ip_icmp *)rpacket)->icmp.icmp_id != ident )
	    continue;			/* not our ECHO */

	from.sin_addr.s_addr = ntohl( from.sin_addr.s_addr );
	for ( sdp=dp; sdp!=NULL; sdp=sdp->previous )
	    {
	    if ( from.sin_addr.s_addr != sdp->inaddr.s_addr )
		continue;

	    sdp->nresponses++;
	    inflight--;
	    break;
	    }
	if ( debug != 0 )
	    fprintf( stderr, "\n em_seqp: received %s ",
						inet_ntoa( from.sin_addr ) );
	} while ( SetNewDelay( delay ) > 0 );

    return;
}


SetNewDelay( delay )
int	delay;
{
    gettimeofday( &tv, &tz);
    td.tv_usec = delay % 1000 - tv.tv_usec + ts.tv_usec;
    if ( td.tv_usec < 0 )
	{
	td.tv_usec += 1000000;
	td.tv_sec = 1 + delay / 1000 - tv.tv_sec + ts.tv_sec;
	}
    else
	td.tv_sec = delay / 1000 - tv.tv_sec + ts.tv_sec;

    return( td.tv_sec * 1000000 + td.tv_usec);
}

/*
 *  SendPing
 *	Send one ICMP Echo_Request to specified address
 *  Compose and transmit an ICMP ECHO REQUEST packet.  The IP packet
 *  will be added on by the kernel.  The ID field is our UNIX process ID,
 *  and the sequence number is an ascending integer.  The first 8 bytes
 *  of the data portion are used to hold a UNIX "timeval" struct in VAX
 *  byte-order, to compute the round-trip time.
 */

SendPing( inaddrp )
struct in_addr * inaddrp;
{
    static u_char outpack[MAXPACKET];
    register struct icmp *icp = (struct icmp *) outpack;
    int i, cc;
    register struct timeval *tp = (struct timeval *) &outpack[8];
    register u_char *datap = &outpack[8+sizeof(struct timeval)];

    icp->icmp_type = ICMP_ECHO;
    icp->icmp_code = 0;
    icp->icmp_cksum = 0;
    icp->icmp_seq = retry;
    icp->icmp_id = ident;		/* ID */

    cc = datalen+8;			/* skips ICMP portion */

    gettimeofday( tp, &tz );

    for( i=8; i<datalen; i++)	/* skip 8 for time */
	*datap++ = i;

    /* Compute ICMP checksum here */

    icp->icmp_cksum = in_cksum( icp, cc );

    /* cc = sendto(s, msg, len, flags, to, tolen) */

    whereto.sin_addr.s_addr = inaddrp->s_addr;
    i = sendto( psfd, outpack, cc, 0, &whereto, sizeof(struct sockaddr) );

    if( i < 0 || i != cc )
	{
	if( i < 0 )
	    perror("em_seqp: sendto");
	else
	    fprintf( stderr, "em_seqp: wrote %d of %d chars to host %s\n",
						i, cc, inet_ntoa( *inaddrp ) );
	}
    inflight++;
    if ( debug != 0 )
        fprintf( stderr, "\n em_seqp: sent %s ",
						inet_ntoa( whereto.sin_addr ) );
}


/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
in_cksum(addr, len)
u_short *addr;
int len;
{
    register int nleft = len;
    register u_short *w = addr;
    register u_short answer;
    register int sum = 0;

    /*
     *  Our algorithm is simple, using a 32 bit accumulator (sum),
     *  we add sequential 16 bit words to it, and at the end, fold
     *  back all the carry bits from the top 16 bits into the lower
     *  16 bits.
     */

    while( nleft > 1 )
	{
	sum += *w++;
	nleft -= 2;
	}

    /* mop up an odd byte, if necessary */

    if( nleft == 1 )
	sum += *(u_char *)w;

    /* add back carry outs from top 16 bits to low 16 bits */

    sum += (sum >> 16);		/* add hi 16 to low 16 */
    answer = ~sum;		/* truncate to 16 bits */
    return (answer);
}


/*
 *  InitSend
 *	Initialize the socket and packet for sending
 */

InitSend()
{
    struct protoent *proto;

    bzero( (char *)&whereto, sizeof(struct sockaddr) );
    whereto.sin_family = AF_INET;

    datalen = 64-8;

    ident = getpid() & 0xFFFF;

    proto = getprotobyname("icmp");
    if ( proto == NULL )
	{
	fprintf(stderr, "em_seqp: icmp: unknown protocol\n");
	exit( -1 );
	}

    psfd = socket( AF_INET, SOCK_RAW, proto->p_proto );
    if ( psfd < 0 )
	{
	perror("em_seqp: socket");
	exit( -1 );
	}
}

