
#ifndef lint
static char sccsid[] = "@(#)net.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  net.c --
 *
 *	Miscellaneous subroutines for the trp program.
 */

#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

#include "trp.h"


/*
 * IsAddr --
 *
 *	Returns TRUE if the string looks like an Internet address.
 *	A string with a trailing dot is not an address, even if it looks
 *	like one.
 *
 *	XXX doesn't treat 255.255.255.255 as an address.
 */

Boolean
IsAddr(host, addrPtr)
char *host;
unsigned long *addrPtr;		/* If return TRUE, contains IP address */
{
    register char *cp;
    unsigned long addr;

    if (isdigit(host[0])) {
	    /* Make sure it has only digits and dots. */
	    for (cp = host; *cp; ++cp) {
		if (!isdigit(*cp) && *cp != '.') 
		    return FALSE;
	    }
	    /* If it has a trailing dot, don't treat it as an address. */
	    if (*--cp != '.') { 
		if ((addr = inet_addr(host)) != (unsigned long) -1) {
		    *addrPtr = addr;
		    return TRUE;
#if 0
		} else {
		    /* XXX Check for 255.255.255.255 case */
#endif
		}
	    }
    }
    return FALSE;
}


/*
 * IsNet --
 *
 *	Returns TRUE if the string looks like an Internet network number.
 *	A string with a trailing dot is not a network number,
 *	even if it looks like one.
 *
 *	XXX doesn't treat 255.255.255.255 as a network number.
 */

Boolean
IsNet( host, addrPtr )
char *host;
unsigned long *addrPtr;		/* If return TRUE, contains IP address */
{
    register char *cp;
    unsigned long addr;

    if (isdigit(host[0])) {
	    /* Make sure it has only digits and dots. */
	    for (cp = host; *cp; ++cp) {
		if (!isdigit(*cp) && *cp != '.') 
		    return FALSE;
	    }
	    /* If it has a trailing dot, don't treat it as an address. */
	    if (*--cp != '.') { 
		if ((addr = inet_network(host)) != (unsigned long) -1) {
		    while ( ( addr & 0xff000000 ) == 0  && addr != 0 )
			addr = addr << 8;
		    *addrPtr = addr;
		    return TRUE;
		}
	    }
    }
    return FALSE;
}


/*
 *  MakeMask --
 *
 *	Return a mask derived from the given address
 */

u_long
MakeMask( addr )
u_long	addr;
{
    if ( IN_CLASSA( addr ) )
	return( IN_CLASSA_NET );

    if ( IN_CLASSB( addr ) )
	return( IN_CLASSB_NET );

    if ( IN_CLASSC( addr ) )
	return( IN_CLASSC_NET );

    return( 0xfffffff0 );
}


