
#ifndef lint
static char sccsid[] = "@(#)queue.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  queue.c --
 *
 *	Miscellaneous subroutines for the trp program.
 */

#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

#include "trp.h"

#define  PQ	1
#define  WQ	2

int	pqsize = 0;
int	wqsize = 0;

/*
 *  AddToWQ
 *	Add Dest entry to Wait Queue
 */

AddToWQ( dstp )
struct Dest * dstp;
{
    if ( dstp->qNext || dstp->qPrev || dstp->qCode )
	{
	Fprintf( stderr, " Attempt to add queued entry to Wait queue\n" );
	return;
	}

    if ( ( wqIN == NULL ) && ( wqOUT == NULL ) )	/* If Q empty */
	{
	dstp->qPrev = NULL;
	wqOUT = dstp;
	}
    else
    if ( ( wqIN != NULL ) && ( wqOUT != NULL ) )	/* If Q not empty */
	{
        wqIN->qNext = dstp;
        dstp->qPrev = wqIN;
	}
    else
	{
	Fprintf( stderr, " Wait queue pointers fouled\n" );
	return;
	}
    dstp->qNext = NULL;
    dstp->qCode = WQ;
    wqIN = dstp;
    dstp->TimeToAct = now + waittime;
    wqsize++;
    return;
}


/*
 *  RemFromWQ
 *	Remove entry from Wait Queue
 */

RemFromWQ( dstp )
struct Dest * dstp;
{
    if ( dstp->qCode != WQ )
	{
	Fprintf( stderr, " Attempt to remove from, but not on, Wait queue\n" );
	return( 0 );
	}

    if ( ( wqOUT == NULL ) && ( wqIN == NULL ) )
	{
	Fprintf( stderr, " Attempt to remove from empty Wait queue\n" );
	return( 0 );
	}

    if ( wqOUT == wqIN )		/* If only entry in Q */
	wqIN = wqOUT = NULL;
    else
	{
	if ( ( dstp->qPrev == NULL ) && ( dstp->qNext == NULL ) )
	    {
	    Fprintf( stderr,
		    " Tried to remove non-queued entry from Wait queue - %s\n",
		    inet_ntoa( dstp->addr ) );
	    return( 0 );
	    }

	if ( wqOUT == dstp )		/* If first in Q */
	    wqOUT = dstp->qNext;
	else
	    dstp->qPrev->qNext = dstp->qNext;

	if ( wqIN == dstp )		/* If last in Q */
	    wqIN = dstp->qPrev;
	else
	    dstp->qNext->qPrev = dstp->qPrev;
	}
    
    dstp->qPrev = NULL;
    dstp->qNext = NULL;
    dstp->qCode = 0;
    wqsize--;
    return( 1 );
}


/*
 *  AddToPQ
 *	Add Dest entry to Process Queue
 */

AddToPQ( dstp )
struct Dest * dstp;
{
    if ( dstp->qNext || dstp->qPrev || dstp->qCode )
	{
	Fprintf( stderr, " Attempt to add queued entry to Process queue\n" );
	return;
	}

    if ( pqIN == NULL )
	{
	dstp->qPrev = NULL;
	pqOUT = dstp;
	}
    else
	{
        pqIN->qNext = dstp;
        dstp->qPrev = pqIN;
	}
    dstp->qNext = NULL;
    dstp->qCode = PQ;
    pqIN = dstp;
    pqsize++;
    return;
}


/*
 *  RemFromPQ
 *	Remove entry from Process Queue
 */

RemFromPQ( dstp )
struct Dest * dstp;
{
    if ( dstp->qCode != PQ )
	{
	Fprintf( stderr,
		    " Attempt to remove from, but not on, Process queue\n" );
	return( 0 );
	}

    if ( ( pqOUT == NULL ) && ( pqIN == NULL ) )
	{
	Fprintf( stderr, " Attempt to remove from empty Process queue\n" );
	return;
	}

    if ( pqOUT == pqIN )		/* If only entry in Q */
	pqIN = pqOUT = NULL;
    else
	{
	if ( pqOUT == dstp )		/* If first in Q */
	    pqOUT = dstp->qNext;
	else
	    dstp->qPrev->qNext = dstp->qNext;

	if ( pqIN == dstp )		/* If last in Q */
	    pqIN = dstp->qPrev;
	else
	    dstp->qNext->qPrev = dstp->qPrev;
	}
    dstp->qPrev = NULL;
    dstp->qNext = NULL;
    dstp->qCode = 0;
    pqsize--;
    return;
}


/*
 *  hash_addr
 *	Return the hash code for an address
 */

hash_addr( addr )
u_long	addr;
{
    int  hash;

    hash = ( addr & ~IN_CLASSB_NET ) >> 8;
    hash %= HASH_TABLE_SIZE;
    return( hash );
}

