#ifndef lint
static char *rcsid =
	"@(#)$Header: trace.c,v 1.17 89/02/28 21:01:13 van Exp $ (LBL)";
#endif

/*
 *  trace.c    DCM Wood   92/09/13
 *
 *  Modified traceroute to discover routes to subnets of current network.
 */

#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <sys/time.h>

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/ioctl.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/ip_icmp.h>
#include <netinet/udp.h>
#include <netdb.h>
#include <ctype.h>

#include "trp.h"

#define	NSFNET3		0x8cde0000	/*  140.222  */
#define	NSI		0x80a10000	/*  128.161  */

#ifndef FD_SET
#define NFDBITS         (8*sizeof(fd_set))
#define FD_SETSIZE      NFDBITS
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)      bzero((char *)(p), sizeof(*(p)))
#endif

u_long	HA[HAMAX]= {0, 1, 2};

struct opacket {		/* udp probe packet */
	struct ip ip;
	struct udphdr udp;
	u_char seq;		/* sequence number of this packet */
	u_char ttl;		/* ttl packet left with */
	struct timeval tv;	/* time packet left */
	};

u_char	packet[512];		/* last inbound (icmp) packet */
struct opacket * outpacket;	/* last output (udp) packet */

int on = 1;
struct hostent *	hp;
struct protoent *	pe;
int s;				/* receive (icmp) socket file descriptor */
int sndsock;			/* send (udp) socket file descriptor */

struct Dest * destp;
struct in_addr	odaddr;
struct in_addr	gw;

int	lsrr = 0;
u_char	optlist[MAX_IPOPTLEN], *oix;

struct sockaddr whereto;	/* Who to try to reach */

struct sockaddr_in	from;
struct sockaddr_in *	to = (struct sockaddr_in *) &whereto;

char *hostname;
char hnamebuf[MAXHOSTNAMELEN];


/*
 *  traceinit--
 *	Initialize the trace routine
 */

traceinit()
{
    if (datalen < 0 || datalen >= MAXPACKET - sizeof(struct opacket)) {
	    Fprintf(stderr, "traceroute: packet size must be 0 <= s < %ld\n",
		    MAXPACKET - sizeof(struct opacket));
	    exit(1);
    }
    datalen += sizeof(struct opacket);

    if ((pe = getprotobyname("icmp")) == NULL)
	{
	Fprintf(stderr, "icmp: unknown protocol\n");
	exit(10);
	}
    if ((s = socket(AF_INET, SOCK_RAW, pe->p_proto)) < 0)
	{
	perror("traceroute: icmp socket");
	exit(5);
	}
    if (options & SO_DEBUG)
	(void) setsockopt(s, SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof(on));

    if (options & SO_DONTROUTE)
	(void) setsockopt(s, SOL_SOCKET, SO_DONTROUTE, (char *)&on, sizeof(on));

    if ((sndsock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0)
	{
	perror("traceroute: raw socket");
	exit(5);
	}

#ifdef SO_SNDBUF
    if ( setsockopt(sndsock, SOL_SOCKET, SO_SNDBUF, (char *)&datalen,
							sizeof(datalen)) < 0)
	{
	perror("traceroute: SO_SNDBUF");
	exit(6);
	}
#endif SO_SNDBUF
#ifdef IP_HDRINCL
    if ( setsockopt(sndsock, IPPROTO_IP, IP_HDRINCL, (char *)&on,
							sizeof(on)) < 0)
	{
	perror("traceroute: IP_HDRINCL");
	exit(6);
	}
#endif IP_HDRINCL
    if (options & SO_DEBUG)
	(void) setsockopt(sndsock, SOL_SOCKET, SO_DEBUG,
						(char *)&on, sizeof(on));
    if (options & SO_DONTROUTE)
	(void) setsockopt(sndsock, SOL_SOCKET, SO_DONTROUTE,
						(char *)&on, sizeof(on));
    outpacket = (struct opacket *)malloc((unsigned)datalen);
    if (! outpacket)
	{
	perror("traceroute: malloc");
	exit(1);
	}
    if (source)
	{
	(void) bzero((char *)&from, sizeof(struct sockaddr));
	from.sin_family = AF_INET;
	from.sin_addr.s_addr = inet_addr(source);
	if (from.sin_addr.s_addr == -1)
	    {
	    Printf("traceroute: unknown host %s\n", source);
	    exit(1);
	    }
	outpacket->ip.ip_src = from.sin_addr;
#ifndef IP_HDRINCL
	if (bind(sndsock, (struct sockaddr *)&from, sizeof(from)) < 0)
	    {
	    perror ("traceroute: bind:");
	    exit (1);
	    }
#endif IP_HDRINCL
	}
    inflight = 0;
}


/*
 *  LaunchPacket--
 *	Build and send trace packet
 */

LaunchPacket()
{
    char  ipstr[20];
    oix = optlist;
    bzero(optlist, sizeof(optlist));


    destp = pqOUT;

/*
 *  if ( destp->addr.s_addr & netmask.s_addr != net.s_addr )
 *	{
 *	Fprintf( stderr, " Bad address - %s\n", inet_ntoa( destp->addr ) );
 *	for (;;);
 *	}
 */
    (void) bzero((char *)&whereto, sizeof(struct sockaddr));
    to->sin_family = AF_INET;
    to->sin_addr.s_addr = destp->addr.s_addr;
    to->sin_addr.s_addr |= ( HA[destp->HAseq] & ~destp->sm.s_addr );

    (void) bzero((char *)outpacket, datalen);
    outpacket->ip.ip_dst = to->sin_addr;
    outpacket->ip.ip_tos = tos;

    ident = (getpid() & 0xffff) | 0x8000;

    if (lsrr > 0)
	{
	lsrr++;
	optlist[IPOPT_OLEN]=IPOPT_MINOFF-1+(lsrr*sizeof(u_long));
	bcopy((caddr_t)&to->sin_addr, oix, sizeof(u_long));
	oix += sizeof(u_long);
	while ((oix - optlist)&3) oix++;	/* Pad to an even boundry */

	if ((pe = getprotobyname("ip")) == NULL)
	    {
	    perror("traceroute: unknown protocol ip\n");
	    exit(10);
	    }
	if ((setsockopt(sndsock, pe->p_proto, IP_OPTIONS, optlist,
							oix-optlist)) < 0)
	    {
	    perror("traceroute: lsrr options");
	    exit(5);
	    }
	}

    RemFromPQ( destp );
    destp->Seq++;
    send_probe( destp->Seq, destp->TTL );
    (void)gettimeofday( &ts, &tz );	/* Set time of last packet sent */
    AddToWQ( destp );
    PSCnt++;
    inflight++;
    strcpy( ipstr, inet_ntoa( to->sin_addr ) );
    Printf( "%-16s %3d %3d %3d Sent to %s\n", inet_ntoa( destp->addr ),
					destp->TTL,
					destp->Retries,
					inflight,
					ipstr );
}


/*
 *  TimeOut
 *	Handle the Timeout of an entry in the Wait Queue
 */

TimeOut()
{
    destp = wqOUT;
    if ( RemFromWQ( destp ) )
        inflight--;
    Printf( "%-16s %3d %3d %3d Timeout\n", inet_ntoa( destp->addr ),
					   destp->TTL,
					   destp->Retries,
					   inflight );
    destp->HAseq++;
    if ( destp->HAseq >= HAMAX )
	{
	destp->Retries++;
	if ( destp->Retries >= max_retries )
	    {
            if ( ( destp->Flags & 1 ) == 1 )
	        return;

	    if ( destp->ReplyTTL[0] > 0 )
		if ( destp->ReplyTTL[0] <= destp->TTL )
		    return;

	    destp->TTL++;
	    if ( destp->TTL >= max_ttl )
	        return;

	    if ( destp->TTL > 2 * ( destp->LastRespTTL ) + 1 )
	        return;

	    destp->Retries = 0;
	    }
	destp->HAseq = 0;
	}
    AddToPQ( destp );
    return;
}


/*
 *  WaitForReply
 *	Wait for response to any packet sent
 *	Return
 *	  0 - if no related packet received in timeout period
 *	  1 - if related packet recedived before timeout
 */

WaitForReply()
{
    struct in_addr	addr;
    int		cc;
    char *	dsnr;
    int		dt;
    char *	got_there = NULL;
    int		i;
    struct ip *	ip;
    char	ipstr[20];
    char *	stop;

    cc = wait_for_reply( s, &from );
    if ( cc == 0 )
	return( 0 );

    if ( ! ( i = packet_ok( packet, cc, &from, seq ) ) )
	return( 0 );

    dt = deltaT( &tv );
    destp->LastRespTTL = destp->TTL;

/*  Printf("  %d ms", dt);			*/
    switch(i - 1)
	{
	case ICMP_UNREACH_PORT:
#ifndef ARCHAIC
	    ip = (struct ip *)packet;
/*		if (ip->ip_ttl <= 1)
*	    Printf(" !");		*/
#endif ARCHAIC
	    got_there = Port_Unreach;
	    break;

	case ICMP_UNREACH_NET:
/*	    Printf(" !N");		*/
	    destp->Unreachable = Net_Unreach;
	    break;

	case ICMP_UNREACH_HOST:
/*	    Printf(" !H");		*/
	    destp->Unreachable = Host_Unreach;
	    break;

	case ICMP_UNREACH_PROTOCOL:
/*	    Printf(" !P");		*/
	    got_there = Proto_Unreach;
	    break;

	case ICMP_UNREACH_NEEDFRAG:
/*	    Printf(" !F");		*/
	    destp->Unreachable = Need_Frag;
	    break;

	case ICMP_UNREACH_SRCFAIL:
/*	    Printf(" !S");		*/
	    destp->Unreachable = SRC_Fail;
	    break;
	}  /* end of switch */

    if ( ( from.sin_addr.s_addr & snmask.s_addr ) ==
			( destp->addr.s_addr & snmask.s_addr ) )
	{
	dsnr = " Dest Subnet Reached";;
	stop = NULL;
	}
    else
	{
	dsnr = NULL;
        stop = AddToTrace( destp, from.sin_addr.s_addr );
	}

    strcpy( ipstr, inet_ntoa( destp->addr ) );

    if ( got_there != NULL && dsnr == NULL )
	{
	Printf( "%-16s %3d %3d %3d %-16s GotThere (%s)\n",
				ipstr,
				destp->TTL,
				destp->Retries,
				inflight,
				inet_ntoa( from.sin_addr ),
				got_there );
	AddToSNGW( from.sin_addr.s_addr, destp->addr.s_addr );
	destp->Flags |= 1;
	destp->HAseq++;
	/* destp->Retries = 0; */
	if ( destp->HAseq <= HAMAX )
	    AddToPQ( destp );
	return( 1 );
	}

    if ( dsnr != NULL )
	{
	Printf( "%-16s %3d %3d %3d %-16s DestReached (%s)\n",
				ipstr,
				destp->TTL,
				destp->Retries,
				inflight,
				inet_ntoa( from.sin_addr ),
				got_there );

	for ( i=0; i<HAMAX; i++ )
	    {
	    if ( destp->Reply[i].s_addr == from.sin_addr.s_addr )
		break;

	    if ( destp->Reply[i].s_addr == 0 )
		{
		destp->Reply[i].s_addr = from.sin_addr.s_addr;
		destp->ReplyTTL[i] = destp->TTL;
		break;
		}
	    }
	if ( i == HAMAX )
	    {
	    Printf( "%-16s %3d %3d %3d %-16s Too many replies!\n",
					ipstr,
					destp->TTL,
					destp->Retries,
					inflight,
					inet_ntoa( from.sin_addr ) );
	    return( 1 );
	    }

	destp->Flags |= 1;
	destp->HAseq++;
	/* destp->Retries = 0; */
	if ( destp->HAseq <= HAMAX )
	    AddToPQ( destp );
	return( 1 );
	}

    if ( destp->Unreachable != NULL )
	{
        Printf( "%-16s %3d %3d %3d %-16s UR %s\n",
				    ipstr,
				    destp->TTL,
				    destp->Retries,
				    inflight,
				    inet_ntoa( from.sin_addr ),
				    destp->Unreachable );
	return( 1 );
	}
    
    if ( destp->TTL > max_ttl )
        {
        Printf( "%-16s %3d %3d %3d %-16s TTL > %d\n",
				    ipstr,
				    destp->TTL,
				    destp->Retries,
				    inflight,
				    inet_ntoa( from.sin_addr ),
				    max_ttl );
	return( 1 );
	}
    if ( ( from.sin_addr.s_addr & 0xffff0000 == NSFNET3 ) ||
	 ( from.sin_addr.s_addr & 0xffff0000 == NSI     ) )
	{
        Printf( "%-16s %3d %3d %3d %-16s BackboneNet\n",
				    ipstr,
				    destp->TTL,
				    destp->Retries,
				    inflight,
				    inet_ntoa( from.sin_addr ) );
	return( 1 );
	}

    if ( stop != NULL )
	{
        Printf( "%-16s %3d %3d %3d %-16s STOP %s\n",
				    ipstr,
				    destp->TTL,
				    destp->Retries,
				    inflight,
				    inet_ntoa( from.sin_addr ),
				    stop );
	destp->Flags |= 1;
	destp->HAseq++;
	/* destp->Retries = 0; */
	if ( destp->HAseq <= HAMAX )
	    AddToPQ( destp );
	return( 1 );
	}

    Printf( "%-16s %3d %3d %3d %-16s TTL exceeded\n",
				    ipstr,
				    destp->TTL,
				    destp->Retries,
				    inflight,
				    inet_ntoa( from.sin_addr ) );
    if ( ( destp->Flags & 1 ) == 1 )
	return( 1 );

    if ( destp->ReplyTTL[0] > 0 )
	if ( destp->ReplyTTL[0] <= destp->TTL )
	    return( 1 );

    destp->TTL++;
    destp->HAseq = 0;
    destp->Retries = 0;
    destp->Code = i-1;
    AddToPQ( destp );
    return( 1 );
}

wait_for_reply( sock, from )
int sock;
struct sockaddr_in *from;
{
    fd_set fds;
    struct timeval wait;
    int cc = 0;
    int fromlen = sizeof (*from);

    FD_ZERO(&fds);
    FD_SET(sock, &fds);
    wait.tv_sec  = 0;
    wait.tv_usec = PACKET_INTERVAL * 1000;

    if ( select( sock+1, &fds, (fd_set *)0, (fd_set *)0, &wait ) > 0 )
	cc = recvfrom( s, (char *)packet, sizeof(packet), 0,
					(struct sockaddr *)from, &fromlen );

    return( cc );
}


send_probe(seq, ttl)
{
    struct opacket *op = outpacket;
    struct ip *ip = &op->ip;
    struct udphdr *up = &op->udp;
    int i;

    ip->ip_off = 0;
    ip->ip_p = IPPROTO_UDP;
    ip->ip_len = datalen;
    ip->ip_ttl = ttl;

    up->uh_sport = htons(ident);
    up->uh_dport = htons(port+seq);
    up->uh_ulen = htons((u_short)(datalen - sizeof(struct ip)));
    up->uh_sum = 0;

    op->seq = seq;
    op->ttl = ttl;
    (void) gettimeofday(&op->tv, &tz);

    i = sendto(sndsock, (char *)outpacket, datalen, 0, &whereto,
	       sizeof(struct sockaddr));
    if (i < 0 || i != datalen)  {
	    if (i<0)
		    perror("sendto");
	    Printf("traceroute: wrote %s %d chars, ret=%d\n", hostname,
		    datalen, i);
	    (void) fflush(stdout);
    }
}


deltaT( tp )
struct timeval * tp;
{
    struct timeval tv;

    (void) gettimeofday(&tv, &tz);
    tvsub( &tv, tp );
    return( tv.tv_sec*1000 + (tv.tv_usec + 500)/1000 );
}


/*
 * Convert an ICMP "type" field to a printable string.
 */

char *
pr_type( t )
u_char t;
{
    static char *ttab[] = {
    "Echo Reply",	"ICMP 1",	"ICMP 2",	"Dest Unreachable",
    "Source Quench", "Redirect",	"ICMP 6",	"ICMP 7",
    "Echo",		"ICMP 9",	"ICMP 10",	"Time Exceeded",
    "Param Problem", "Timestamp",	"Timestamp Reply", "Info Request",
    "Info Reply"
    };

    if(t > 16)
	    return("OUT-OF-RANGE");

    return(ttab[t]);
}


/*
 * Verify that packet is OK.  Return 0 if not.
 */

packet_ok( buf, cc, from, seq )
u_char * buf;
int cc;
struct sockaddr_in * from;
int seq;
{
    register struct icmp * icp;
    u_char type, code;
    int hlen;

#ifndef ARCHAIC
    struct ip * ip;

    ip = (struct ip *) buf;
    hlen = ip->ip_hl << 2;
    if (cc < hlen + ICMP_MINLEN) {
	    if (verbose)
		    Printf("packet too short (%d bytes) from %s\n", cc,
			    inet_ntoa( from->sin_addr));
	    return( 0 );
    }
    cc -= hlen;
    icp = (struct icmp *)(buf + hlen);
#else
    icp = (struct icmp *)buf;
#endif ARCHAIC

    type = icp->icmp_type;
    code = icp->icmp_code;
    if ((type == ICMP_TIMXCEED && code == ICMP_TIMXCEED_INTRANS) ||
	 type == ICMP_UNREACH) {
	    struct ip * hip;
	    struct udphdr * up;

	    hip = &icp->icmp_ip;
	    hlen = hip->ip_hl << 2;
	    up = (struct udphdr *)((u_char *)hip + hlen);
	    if ( hlen + 12 > cc ||
		 hip->ip_p != IPPROTO_UDP ||
		 up->uh_sport != htons(ident) )
		return( 0 );

	    bcopy( &hip->ip_dst, &odaddr, 4 );
	    destp = FindDest( odaddr.s_addr );
	    if ( destp == NULL )
		{
		Printf( " Could not find hash entry for addr = %s\n",
						    inet_ntoa( odaddr ) );
		return( 0 );
		}

	    if ( RemFromWQ( destp ) )
	        inflight--;
	    if ( up->uh_dport == htons(port+ destp->Seq) )
		    return( type == ICMP_TIMXCEED? -1 : code+1 );
    }

#ifndef ARCHAIC
    if (verbose) {
	    int i;
	    u_long *lp = (u_long *)&icp->icmp_ip;

	    Printf("\n%d bytes from %s to %s", cc,
		    inet_ntoa( from->sin_addr), inet_ntoa( ip->ip_dst));
	    Printf(": icmp type %d (%s) code %d\n", type, pr_type(type),
		   icp->icmp_code);
	    for (i = 4; i < cc ; i += sizeof(long))
		    Printf("%2d: x%8.8lx\n", i, *lp++);
    }
#endif ARCHAIC
    return( 0 );
}


/*
 * Subtract 2 timeval structs:  out = out - in.
 * Out is assumed to be >= in.
 */

tvsub(out, in)
register struct timeval *out, *in;
{
    if ((out->tv_usec -= in->tv_usec) < 0)   {
	    out->tv_sec--;
	    out->tv_usec += 1000000;
    }
    out->tv_sec -= in->tv_sec;
    return;
}

