#ifndef lint
static char *rcsid =
"@(#)$Header: trp.c 1.17 89/02/28 21:01:13 van Exp $ (LBL)";
#endif

/*
 *  trp.c    DCM Wood   92/09/13
 *
 *  Modified traceroute to discover routes to subnets of current network.
 *
 *	Traceroute to all the destinations specified, building a list
 *  of subnets and gateways as we go.
 */

/*
 *  Read in a list of destinations.
 *	If no list available,
 *	  Get addresses of all interfaces
 *	  Use those network numbers
 *	  Find all entries in routing table with those network numbers
 *
 *  For each destination:
 *    Build a Dest entry
 *    For TTL from 1 to 30 (or less)
 *      Send trace packets toward the destination
 *      Record the addresses for the responses
 *      If redirect, try different address with same TTL
 *      If TTL exceeded, increase TTL
 *      If port unreachable, stop this trace
 *  Dump resulting subnet and gateway data structures
 */

#include <stdio.h>
#include <errno.h>
#include <strings.h>
#include <sys/time.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_var.h>
#include <sys/file.h>
#include <netdb.h>
#include <ctype.h>

#define SOCKADDR_IN

#include "data_server.h"
#include "send_data.h"

#define MAIN

#include "trp.h"

struct data_record * make_data();
struct intf_list * make_intf();

char usage[] =
"Usage: trp [-drv] [-f [input_file]] [-j journal_server_addr] [-w wait] [-m max_ttl] [-q max_retries] [-p port] [-t tos] [-s src_addr] [data size]\n";


main( argc, argv )
int	argc;
char **	argv;
{
    struct in_addr	addr;
    char **av = argv;
    int	cc;
    int dt;
    int ic;
    int  i;
    int  fd;
    int hash;
    int n;
    int	StartTime;
    u_long	ogw;
    FILE *	dfp;
    FILE *	ifp;
    char	inputf[64];
    char	ipstr[20];
    struct ifconf  ifc;
    struct ifreq * ifreqp;
    struct in_addr inaddr;
    struct in_addr inmask;
    struct in_addr jsa;
    struct Intf * intfp;
    struct Hop * hopp;
    struct Hop * lhp;
    struct Net * netp;
    struct Dest *  dp;
    struct GWent * gwp;
    struct SNent * snp;

    struct gwPointer * gwpp;
    struct snPointer * snpp;

    setlinebuf (stdout);
    gettimeofday( &tv, &tz);
    now = tv.tv_sec;
    StartTime = tv.tv_sec;
    printf( " em_trp: started\n" );
    inputf[0] = '\0';

    argc--, av++;
    while (argc && *av[0] == '-')
	{
	while (*++av[0])
	    switch (*av[0])
		{
	    case 'd':
		options |= SO_DEBUG;
		debug++;
		break;

	    case 'f':
		if ( ( argc > 1 ) && ( *av[1] != '-' ) ) {
		    argc--, av++;
		    strcpy( inputf, av[0] );
		    }
		else
		    strcpy( inputf, "trp.input" );
		goto nextarg;

	    case 'j':
		argc--, av++;
		jsa.s_addr = inet_addr(av[0]);
		if (jsa.s_addr < 1) {
		    fprintf(stderr, "journal server address error\n");
		    exit(1);
		    }
		goto nextarg;

	    case 'm':
		argc--, av++;
		max_ttl = atoi(av[0]);
		if (max_ttl <= 1) {
		    fprintf(stderr, "max ttl must be >1\n");
		    exit(1);
		    }
		goto nextarg;

	    case 'p':
		argc--, av++;
		port = atoi(av[0]);
		if (port < 1) {
		    fprintf(stderr, "port must be >0\n");
		    exit(1);
		    }
		goto nextarg;

	    case 'q':
		argc--, av++;
		max_retries = atoi(av[0]);
		if (max_retries < 1) {
		    fprintf(stderr, "max_retries must be >0\n");
		    exit(1);
		    }
		goto nextarg;

	    case 'r':
		options |= SO_DONTROUTE;
		break;

	    case 's':
		/*
		 * set the ip source address of the outbound
		 * probe (e.g., on a multi-homed host).
		 */
		argc--, av++;
		source = av[0];
		goto nextarg;

	    case 't':
		argc--, av++;
		tos = atoi(av[0]);
		if (tos < 0 || tos > 255) {
		    fprintf(stderr, "tos must be 0 to 255\n");
		    exit(1);
		    }
		goto nextarg;

	    case 'v':
		verbose++;
		break;

	    case 'w':
		argc--, av++;
		waittime = atoi(av[0]);
		if (waittime <= 1) {
		    fprintf(stderr, "wait must be >1 sec\n");
		    exit(1);
		    }
		goto nextarg;
		}
nextarg:
	argc--, av++;
	}
    if (argc != 0)  {
	fprintf( stderr, usage );
	exit(1);
	}

    if (argc >= 1)
	datalen = atoi(av[0]);

    /* Get some memory in which to store the ifreq entries	*/

    ifc.ifc_len = 20 * sizeof( struct ifreq );

    ifc.ifc_req = (struct ifreq *)calloc( 20, sizeof( struct ifreq ) );

    /* Open a socket to do ioctls on.				*/

    if ( ( fd = socket( AF_INET, SOCK_DGRAM, 0 ) ) == -1 )
	{
	perror( "socket" );
	return (-1);
	}

    /* Do the ioctl, check the result.				*/

    if ( ioctl( fd, SIOCGIFCONF, &ifc) == -1 )
	{
	perror( "ioctl SIOCGIFCONF");
	close( fd );
	return (-1);
	}

    if ( ifc.ifc_len/sizeof(struct ifreq) == 1 )
	Printf( " There is only one interface on this system\n" );
    else
	Printf( " There are %d interfaces on this system\n",
	    ifc.ifc_len/sizeof(struct ifreq) );
    ifreqp = ifc.ifc_req;
    ogw = 0;
    for ( i=0; i<ifc.ifc_len; i+=sizeof( struct ifreq ), ifreqp++ )
	{
	bcopy( &ifreqp->ifr_addr.sa_data[2], &inaddr.s_addr, 4 );
	Printf( " %s   %-16s", ifreqp->ifr_name, inet_ntoa( inaddr ) );
	if ( ioctl( fd, SIOCGIFNETMASK, ifreqp ) == -1 )
	    {
	    perror( "ioctl SIOCGIFNETMASK" );
	    continue;
	    }
	bcopy( &ifreqp->ifr_addr.sa_data[2], &inmask.s_addr, 4 );
	Printf( "   %-16s\n", inet_ntoa( inmask ) );
	if ( ( inaddr.s_addr & 0xff000000 ) == 0x7f000000 )
	    continue;

	AddNet( inaddr.s_addr & MakeMask( inaddr.s_addr ), inmask.s_addr );
	if ( ogw != 0 )
	    AddIntfToGW( inaddr.s_addr, ogw );
	AddToSNGW( inaddr.s_addr, inaddr.s_addr & inmask.s_addr );
	ogw = inaddr.s_addr;
	}
    close( fd );

    net.s_addr = NetList->addr.s_addr;
    netmask.s_addr = MakeMask( NetList->addr.s_addr );
    snmask.s_addr  = NetList->sm.s_addr;


    if ( inputf[0] != '\0' )
	{
	ifp = fopen( inputf, "r" );
	if ( ifp == NULL )
	    {
	    fprintf( stderr, " Could not open input file %s.\n", inputf );
	    exit( 1 );
	    }

	for ( ic=0; ; ic++ )
	    {
	    cc = fscanf( ifp, "%s", ipstr );
	    if ( cc != 1 )
		break;

	    if ( IsNet( ipstr, &addr.s_addr ) || IsAddr( ipstr, &addr.s_addr ) )
		AddDestEnt( &addr );
	    }

	if ( ic == 0 )
	    {
	    fprintf( stderr, " Empty input file.\n" );
	    exit( 1 );
	    }

	fclose( ifp );
	}
    else
	{
	for ( netp=NetList; netp != NULL; netp = netp->Next )
	    {
	    int sc;
	    int snd;
	    u_long  addx;

	    snd = MakeMask( netp->addr.s_addr ) ^ netp->sm.s_addr;
	    for ( sc=0; sc<24; snd >>= 1, sc++ )
		if ( ( snd & 1 ) == 1 )
		    break;

	    for ( ic = 0; ic < snd; ic++ )
		{
		addx = netp->addr.s_addr | ( ic << sc );
		AddDestEnt( &addx );
		}
	    }
	}

    fprintf( stdout, " em_trp: searching for %d destinations\n", DestCnt );
    gettimeofday( &ts, &tz);
    traceinit();


    /*  THE MAIN LOOP  */

    while ( ( pqOUT != NULL ) || ( wqOUT != 0 ) )
	{
	gettimeofday( &tv, &tz);
	now = tv.tv_sec;
	cc = WaitForReply();			/* At least delay .125 sec */
	gettimeofday( &tv, &tz);
	now = tv.tv_sec;
	if ( wqOUT != NULL )
	    if ( wqOUT->TimeToAct <= now )
		TimeOut();
	if ( pqOUT != NULL )
	    {
	    dt = deltaT( &ts );
	    if ( dt > PACKET_INTERVAL )
		if ( ( cc == 0 ) || ( dt > 2 * PACKET_INTERVAL ) )
		    LaunchPacket();
	    }
	}

    /*  Add Replies from Destination Subnets  */

    for ( hash=0; hash<HASH_TABLE_SIZE; hash++ )
	for ( dp=DHT[hash]; dp!=NULL; dp=dp->SameHash )
	    {
	    if ( dp->Reply[0].s_addr == 0 )
		break;

	    AddDestReplies( dp );
	    }


    /*  Print statistics  */

    gettimeofday( &tv, &tz);
    fprintf( stdout, " em_trp: took %d seconds\n",
						tv.tv_sec - StartTime );
    fprintf( stdout, " %6d packets sent\n", PSCnt );
    fprintf( stdout, " %6d subnets found\n", SNCnt );
    fprintf( stdout, " %6d gateways found\n", GWCnt );
    fprintf( stdout, " %6d interfaces found\n", IntfCnt );

    dfp = fopen( "trp.dump", "w" );
    if ( dfp == NULL )
	fprintf( stderr, " Could not open dump file.\n" );

    /*  Dump Gateways  */

    Fprintf( stdout,  "\nSummary of gateways collected.\n" );
    if ( dfp != NULL )
	fprintf( dfp, "\nSummary of gateways collected.\n" );

    n = 0;
    for ( gwp=GWFirst; gwp!=NULL; gwp=gwp->Next )
	{
	struct data_record * data;
	struct intf_list ** intfs_ptr;
	struct intf_list * nintf;
	struct subnet ** con_sub_ptr;
	struct sockaddr_in * gw_addr;

	data = make_data(GATEWAY);
	intfs_ptr = &(data->dr_type.gateway->intfs);
	Fprintf( stdout,  " %s  %d\n", inet_ntoa( gwp->addr ), gwp->snCount );
	if ( dfp != NULL )
	    fprintf( dfp, " %s  %d\n", inet_ntoa( gwp->addr ), gwp->snCount );

	gw_addr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	if ( gw_addr == NULL )
	    {
	    printf("unable to allocate space for sockaddr \n");
	    exit(1);
	    }

	bcopy(&(gwp->addr),(&gw_addr->sin_addr),sizeof(struct in_addr));
	nintf = make_intf();
	nintf->subnet_intf->ip_addr->pa = gw_addr;
	nintf->subnet_intf->ip_addr->addr_family = AF_INET;
	*intfs_ptr = nintf;
	intfs_ptr = &((*intfs_ptr)->next_intf);

	for ( intfp=gwp->FirstIntf; intfp!=NULL; intfp=intfp->Next )
	    {
	    if ( intfp->addr.s_addr == gwp->addr.s_addr )
		continue;

	    Fprintf( stdout,  "    %s\n", inet_ntoa( intfp->addr ) );
	    if ( dfp != NULL )
		fprintf( dfp, "    %s\n", inet_ntoa( intfp->addr ) );

	    gw_addr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	    if ( gw_addr == NULL )
		{
		printf("unable to allocate space for sockaddr_in \n");
		exit(1);
		}

	    bcopy(&(intfp->addr),(&gw_addr->sin_addr),sizeof(struct in_addr));
	    nintf = make_intf();
	    nintf->subnet_intf->ip_addr->pa = gw_addr;
	    nintf->subnet_intf->ip_addr->addr_family = AF_INET;
	    *intfs_ptr = nintf;
	    intfs_ptr = &((*intfs_ptr)->next_intf);

	    }

	con_sub_ptr = &(data->dr_type.gateway->connected_subnets);
	for ( snpp=gwp->snppStart; snpp!=NULL; snpp=snpp->Next )
	    {
	    struct subnet * con_sub;
	    struct subnet * make_subnet();
	    struct sockaddr_in * sn_addr;

	    Fprintf( stdout,  "\t     %s\n", inet_ntoa( snpp->sn->addr ) );
	    if ( dfp != NULL )
		fprintf( dfp, "\t     %s\n", inet_ntoa( snpp->sn->addr ) );

	    sn_addr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	    if ( sn_addr == NULL )
		{
		printf("Unable to allocate memory for sn_addr\n");
		exit(1);
		}

	    bcopy(&(snpp->sn->addr),(&sn_addr->sin_addr),sizeof(struct in_addr));
	    con_sub = make_subnet();
	    con_sub->subnet_info->net_numb->pa  = sn_addr;
	    con_sub->subnet_info->net_numb->addr_family = AF_INET;
	    *con_sub_ptr = con_sub;
	    con_sub_ptr =  &((*con_sub_ptr)->next_subnet);
	    }
	pack_data(data);
	n++;
	}
    fprintf( stderr, " em_trp:%4d GATEWAY entries sent to Journal Server\n", n);

    /*  Dump Subnets  */

    Fprintf( stdout,  "\nSummary of subnets collected.\n" );
    if ( dfp != NULL )
	fprintf( dfp, "\nSummary of subnets collected.\n" );

    n = 0;
    for ( snp=SNFirst; snp!=NULL; snp=snp->Next )
	{
	struct data_record * data;
	struct sockaddr_in * sn_num;
	struct gw_pointer ** net_gw;

	data = make_data( NETWORK );
	Fprintf( stdout,  " %s  %d\n", inet_ntoa( snp->addr ), snp->gwCount );
	if ( dfp != NULL )
	    fprintf( dfp, " %s  %d\n", inet_ntoa( snp->addr ), snp->gwCount );

	data->dr_type.subnet->net_numb = (struct intf_prot_addr *)
	    malloc( sizeof(struct intf_prot_addr) );
	if ( data->dr_type.subnet->net_numb == NULL )
	    {
	    printf("Unable to allocate memory for net_numb\n");
	    exit(1);
	    }

	sn_num = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	if ( sn_num == NULL )
	    {
	    printf("Unable to allocate memory for sn_num\n");
	    exit(1);
	    }

	bcopy(&(snp->addr), (&sn_num->sin_addr),sizeof(struct in_addr));
	data->dr_type.subnet->net_numb->pa = sn_num;
	data->dr_type.subnet->net_numb->addr_family = AF_INET;
	net_gw = &(data->dr_type.subnet->gwp_start);

	for ( gwpp=snp->gwppStart; gwpp!=NULL; gwpp=gwpp->Next )
	    {
	    struct gw_pointer * gw;
	    struct sockaddr_in * gw_addr;
	    struct gw_pointer * make_gateway();

	    Fprintf( stdout,  "          %s\n", inet_ntoa( gwpp->gw->addr ) );
	    if ( dfp != NULL )
		fprintf( dfp, "          %s\n", inet_ntoa( gwpp->gw->addr ) );

	    gw_addr = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
	    if ( gw_addr == NULL )
		{
		printf("Unable to allocate memory for gw_addr\n");
		exit(1);
		}

	    bcopy(&(gwpp->gw->addr),(&gw_addr->sin_addr),sizeof(struct in_addr));
	    gw = make_gateway();
	    gw->gw_host->intfs->subnet_intf->ip_addr->pa = gw_addr;
	    gw->gw_host->intfs->subnet_intf->ip_addr->addr_family = AF_INET;
	    *net_gw = gw;
	    net_gw = &((*net_gw)->gw_next);
	    }
	pack_data(data);
	n++;
	}
    send_data( NULL );
    fprintf( stderr,
		" em_trp:%4d NETWORK entries sent to Journal Server\n", n);

    fclose( dfp );
    exit;

    /*  Dump Traces  */

    dfp = fopen( "trp.traces", "w" );
    if ( dfp == NULL )
	fprintf( stderr, " Could not open traces file.\n" );

    Fprintf( stdout,  "\nSummary of traces collected.\n" );
    if ( dfp != NULL )
	fprintf( dfp, "\nSummary of traces collected.\n" );

    for ( hash=0; hash<HASH_TABLE_SIZE; hash++ )
	for ( dp=DHT[hash]; dp!=NULL; dp=dp->SameHash )
	    {
	    if ( dp->NextHop == NULL )
		break;

	    Fprintf( stdout,
			"\n %s  %d", inet_ntoa( dp->addr ), dp->LastRespTTL );
	    if ( dfp != NULL )
		fprintf( dfp,
			"\n %s  %d", inet_ntoa( dp->addr ), dp->LastRespTTL );

	    for ( hopp=dp->NextHop; hopp!=NULL; hopp=hopp->NextHop )
		{
		Fprintf( stdout,  "\n\t   %16s  %d",
					inet_ntoa( hopp->addr ), hopp->TTL );
		if ( dfp != NULL )
		    fprintf( dfp, "\n\t   %16s  %d",
					inet_ntoa( hopp->addr ), hopp->TTL );
		lhp = hopp;
		}
	    for ( i=0; i<HAMAX; i++ )
		{
		if ( dp->Reply[i].s_addr == 0 )
		    break;

		Fprintf( stdout,  "\n\t       %16s  %d",
				inet_ntoa( dp->Reply[i] ), dp->ReplyTTL[i] );
		if ( dfp != NULL )
		    fprintf( dfp, "\n\t       %16s  %d",
				inet_ntoa( dp->Reply[i] ), dp->ReplyTTL[i] );
		}
	    if ( dp->Unreachable != NULL )
		{
		Fprintf( stdout,  "  %s", dp->Unreachable );
		if ( dfp != NULL )
		    fprintf( dfp, "  %s", dp->Unreachable );
		}
	    }
    fprintf( stdout,  "\n" );
    if ( dfp != NULL )
	fprintf( dfp, "\n" );

    fclose( dfp );
    exit;
}

