/*
 *  trp.h  DCM Wood  92/09/13
 *	Define data structures
 */

#define  HAMAX  3
#define  MAXPACKET       65535   /* max ip packet size */

#define  HASH_TABLE_SIZE  512
#define  TRUE  1
#define  FALSE 0
#define  TIMEOUT  10
#define  MAX_TTL  30
#define  MAX_RETRIES  2
#define  PACKET_INTERVAL  250

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN  64
#endif

#define Fprintf if ( debug ) (void)fprintf
#define Sprintf if ( debug ) (void)sprintf
#define Printf if ( debug ) (void)printf

struct Dest {
	struct in_addr	addr;
	struct in_addr	scr;
	struct in_addr	netmask;
	struct in_addr	sm;
	struct in_addr	Reply[HAMAX];
	short   ReplyTTL[HAMAX];
	char *	name;
	char *	domain;
	char *	Unreachable;
	struct Hop *	NextHop;
	struct Dest *	SameHash;
	struct Dest *	qNext;
	struct Dest *	qPrev;
	time_t	CreationTime;
	time_t	LastCheckedTime;
	time_t	TimeToAct;
	short	Retries;
	short	TTL;
	short	Seq;
	short	LastRespTTL;
	short	Flags;
	short	Code;
	short	HAseq;
	short   qCode;
	};

struct Hop {
	struct in_addr  addr;
	int	TTL;
	struct Hop *   NextHop;
	struct Hop *   SameTTL;
	};

struct snPointer {
	struct SNent * sn;
	struct snPointer * Next;
	};

struct gwPointer {
	struct GWent * gw;
	struct gwPointer * Next;
	};

struct TreeNode {
	struct in_addr	addr;
	struct in_addr	sm;
	struct TreeNode *	Next;
	struct TreeNode *	SameLevel;
	struct TreeNode *	ToLeaves;
	struct TreeNode *	ToRoot;
	struct Intf *	FirstIntf;
	struct sngwPointer *  sngwStart;
	struct sngwPointer *  sngwEnd;
	short	flags;
	short	snCount;
	short	ProgenyCnt;
	short	Type				/*  GS or SN  */;
	};

struct GWent {
	struct in_addr	addr;
	struct in_addr	sm;
	struct GWent *	Next;
	struct Intf *	FirstIntf;
	struct snPointer *  snppStart;
	struct snPointer *  snppEnd;
	short	flags;
	short	IntfCnt;
	short	snCount;
	short	ProgenyCnt;
	short	Type				/*  GS or SN  */;
	};

struct SNent {
	struct in_addr	addr;
	struct in_addr	sm;
	struct SNent *	Next;
	struct gwPointer * gwppStart;
	struct gwPointer * gwppEnd;
	short	flags;
	short	gwCount;
	short	ProgenyCnt;
	short	Type				/*  GS or SN  */;
	};

struct Intf {
	struct in_addr	addr;
	struct in_addr	sm;
	int	CreationTime;
	struct GWent *	GW;
	struct Intf *	Next;		/* next intf on this gw */
	struct Intf *	SameHash;
	};

struct Net {
	struct in_addr	addr;
	struct in_addr	sm;
	struct Net *	Next;
	};

typedef int Boolean;
typedef struct Dest * DestP;
typedef struct Intf * IntfP;
typedef struct GWent * GWP;
typedef struct SNent * SNP;

#ifdef MAIN

struct timeval  ts;		/* time last packet sent  */
struct timeval  tv;
struct timezone tz;
struct itimerval  value;
struct itimerval ovalue;
time_t	now;			/* Current date/time in seconds */
int	inflight;

struct in_addr	net;			/* Network being traced */
struct in_addr	netmask;		/* Network mask - w/o subnet */
struct in_addr	snmask;			/* SubNet Mask  */

struct Dest *	wqIN;		/* Wait Q  in pointer */
struct Dest *	wqOUT;		/* Wait Q out pointer */

struct Dest *	pqIN;		/* Process Q  in pointer */
struct Dest *	pqOUT;		/* Process Q out pointer */

struct Dest *	DHT[HASH_TABLE_SIZE];
struct Intf *	IHT[HASH_TABLE_SIZE];

int	DestCnt;			/* Count of Dest Entries created */
int	IntfCnt;			/* Count of Intf Entries created */
int	GWCnt;				/* Count of Gateway Entries created */
int	SNCnt;				/* Count of Subnet Entries created */
int	PSCnt;				/* Count of packets sent */
int	NetCnt;				/* Count of Networks encountered */

struct GWent * GWFirst=NULL;		/* Linked list of gateways  */
struct GWent * GWLast=NULL;

struct SNent * SNFirst=NULL;		/* Linked list of subnets  */
struct SNent * SNLast=NULL;

struct Net * NetList=NULL;		/* Linked list of networks  */
struct Net * NetLast=NULL;		/* Linked list of networks  */

int	max_ttl=MAX_TTL;
int	max_retries=MAX_RETRIES;
u_short	ident;
u_short	port=32768+666;		/* start udp dest port # for probe packets */

int	datalen;		/* How much data */
int	debug = 0;		/* debug flag */
int	options;		/* socket options */
int	verbose;
int	waittime=TIMEOUT;	/* time to wait for response (in seconds) */
int	seq=0;
int	tos=0;
char *	source=0;

char	Port_Unreach[] = "Port Unreachable";
char	Net_Unreach[] = "Net unreachable";
char	Host_Unreach[] = "Host unreachable";
char	Proto_Unreach[] = "Protocol Unreachable";
char	Need_Frag[] = "Need fragmentation";
char	SRC_Fail[] = "SRC fail";

char	Duplicate[] = "Duplicate on Path";
char	Max_ttl[] = "More than max_ttl Hops";
char	Previous[] = "Found previous address";
char	Loopback[] = "Reply from Loopback";
char	Intf_Added[] = "Interface added";

#else  /* MAIN */

extern	struct timeval  ts;		/* time last packet sent  */
extern	struct timeval  tv;
extern	struct timezone tz;
extern	struct itimerval  value;
extern	struct itimerval ovalue;
extern	time_t	now;			/* Current date/time in seconds */
extern	int	inflight;

extern	struct in_addr	net;		/* Network being traced */
extern	struct in_addr	netmask;	/* Network mask - w/o subnet */
extern	struct in_addr	snmask;		/* SubNet Mask  */

extern	struct Dest *	wqIN;		/* Wait Q  in pointer */
extern	struct Dest *	wqOUT;		/* Wait Q out pointer */

extern	struct Dest *	pqIN;		/* Process Q  in pointer */
extern	struct Dest *	pqOUT;		/* Process Q out pointer */

extern	struct Dest *	DHT[];
extern	struct Intf *	IHT[];

extern	int	DestCnt;	/* Count of Dest Entries created */
extern	int	IntfCnt;	/* Count of Intf Entries created */
extern	int	GWCnt;		/* Count of Gateway Entries created */
extern	int	SNCnt;		/* Count of Subnet Entries created */
extern	int	PSCnt;		/* Count of packets sent */
extern	int	NetCnt;		/* Count of Networks encountered */

extern	struct GWent * GWFirst;		/* Linked list of gateways  */
extern	struct GWent * GWLast;

extern	struct SNent * SNFirst;		/* Linked list of subnets  */
extern	struct SNent * SNLast;

extern	struct Net * NetList;		/* Linked list of networks  */
extern	struct Net * NetLast;		/* Linked list of networks  */

extern	int	max_ttl;
extern	int	max_retries;
extern	u_short	ident;
extern	u_short	port;		/* start udp dest port # for probe packets */

extern	int	datalen;	/* How much data */
extern	int	debug;		/* debug flag */
extern	int	options;	/* socket options */
extern	int	verbose;
extern	int	waittime;	/* time to wait for response (in seconds) */
extern	int	seq;
extern	int	tos;
extern	char*	source;

extern	char	Port_Unreach[];
extern	char	Net_Unreach[];
extern	char	Host_Unreach[];
extern	char	Proto_Unreach[];
extern	char	Need_Frag[];
extern	char	SRC_Fail[];

extern	char	Duplicate[];
extern	char	Max_ttl[];
extern	char	Previous[];
extern	char	Loopback[];
extern	char	Intf_Added[];

#endif /* MAIN */

extern  int errno;
extern  char *malloc();
extern  char *inet_ntoa();
extern  u_long inet_addr();
extern	DestP	FindDest();
extern	IntfP	FindIntf();
extern	Boolean	IsAddr();
extern	Boolean IsNet();
extern	char * AddToTrace();

