
#ifndef lint
static char sccsid[] = "@(#)util.c	5.22 (Berkeley) 8/3/90";
#endif /* not lint */

/*
 *  util.c --
 *
 *	Miscellaneous subroutines for the trp program.
 */

#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <string.h>

#include "trp.h"

struct GWent * gwp;
struct SNent * snp;

struct gwPointer * gwpp;
struct snPointer * snpp;


/*
 *  AddIntfToGW
 *	Add an interface to an existing gateway
 */

AddIntfToGW( intf, gw )
u_long	intf;
u_long	gw;
{
    char  intfipstr[64];
    char  gwipstr[64];
    struct GWent * gwp;
    struct Intf  * intfp;
    struct Intf  * intfp2;

    intfp = FindIntf( gw );
    gwp = intfp->GW;
    intfp = FindIntf( intf );
    if ( intfp->GW != NULL )
	{
	strcpy( intfipstr, inet_ntoa( intfp->addr ) );
	strcpy(   gwipstr, inet_ntoa(   gwp->addr ) );
	Printf( " Failed attempt to add interface %s to GW %s\n",
				intfipstr, gwipstr );
	return;
	}

    if ( gwp->FirstIntf == NULL )
	{
	strcpy( intfipstr, inet_ntoa( intfp->addr ) );
	strcpy(   gwipstr, inet_ntoa(   gwp->addr ) );
	Printf( " Adding interface %s to GW %s\n",
				intfipstr, gwipstr );
	gwp->FirstIntf == intfp;
	intfp->Next = NULL;
	return;
	}

    intfp->GW = gwp;
    for ( intfp2=gwp->FirstIntf; intfp2->Next != NULL; intfp2=intfp2->Next )
	;
    intfp2->Next = intfp;
    intfp->Next = NULL;
    return;
}


/*
 *  AddNet
 *	Add network and subnet mask to list of networks
 */

AddNet( addr, mask )
u_long	addr;
u_long	mask;
{
    struct Net * netp;

    if ( NetList == NULL )
	{
	netp = (struct Net *)malloc( sizeof( struct Net ) );
	netp->addr.s_addr = addr;
	netp->sm.s_addr = mask;
	netp->Next = NULL;
	NetList = netp;
	NetLast = netp;
	NetCnt++;
	return;
	}

    for ( netp = NetList; netp != NULL; netp = netp->Next )
	if ( netp->addr.s_addr == addr )
	    return;

    netp = (struct Net *)malloc( sizeof( struct Net ) );
    netp->addr.s_addr = addr;
    netp->sm.s_addr = mask;
    netp->Next = NULL;
    NetLast->Next = netp;
    NetLast = netp;
    NetCnt++;
    return;
}


/*
 *  FindIntf
 *	Find the pointer to an Interface entry.
 *	If not found, create one.
 */

IntfP
FindIntf( addr )
u_long	addr;
{
    int  hash;
    struct Intf ** pip;
    struct Intf *  intfp;

    hash = hash_addr( addr );

    pip = (struct Intf **)&IHT[hash];
    intfp = *pip;

    while ( intfp != NULL )
	{
	if ( intfp->addr.s_addr == addr )
	    return( intfp );

	pip = &(intfp->SameHash);
	intfp = intfp->SameHash;
	}

    IntfCnt++;
    intfp = (struct Intf *)calloc( 1, sizeof( struct Intf) );
    intfp->CreationTime = now;
    intfp->addr.s_addr = addr;
    *pip = intfp;
    return( intfp );
}


/*
 *  CreateGW
 *	Create a GWent
 */

GWP
CreateGW( intfp )
struct Intf * intfp;
{
    struct GWent * gwp;

    gwp = (struct GWent *)calloc( 1, sizeof( struct GWent ) );
    intfp->GW = gwp;
    gwp->addr.s_addr = intfp->addr.s_addr;
    fprintf( stdout, " %-16s gateway found\n", inet_ntoa( gwp->addr ) );
    gwp->FirstIntf = intfp;
    if ( GWFirst == NULL )
	GWFirst = GWLast = gwp;
    else
	{
	GWLast->Next = gwp;
	GWLast = gwp;
	}
    gwp->IntfCnt++;
    GWCnt++;
    return( gwp );
}


/*
 *  FindGW
 *	Find or create gateway entry
 */

GWP
FindGW( addr )
u_long	addr;
{
    struct GWent * gwp;
    struct Intf * intfp;

    if ( (gwp = GWFirst ) == NULL )
	{
	gwp = (struct GWent *)calloc( 1, sizeof( struct GWent ) );
	gwp->addr.s_addr = addr;
	GWFirst = gwp;
	GWLast  = gwp;
	GWCnt++;
        fprintf( stdout, " %-16s gateway found\n", inet_ntoa( gwp->addr ) );
	return( gwp );
	}

    for ( ; gwp != NULL; gwp = gwp->Next )
	{
	if ( gwp->addr.s_addr == addr )
	    return( gwp );

	for ( intfp = gwp->FirstIntf; intfp != NULL; intfp = intfp->Next )
	    if ( intfp->addr.s_addr == addr )
		return( gwp );
	}

    gwp = (struct GWent *)calloc( 1, sizeof( struct GWent ) );
    gwp->addr.s_addr = addr;
    GWLast->Next  = gwp;
    GWLast  = gwp;
    GWCnt++;
    fprintf( stdout, " %-16s gateway found\n", inet_ntoa( gwp->addr ) );
    return( gwp );
}


/*
 *  FindSN
 *	Find subnet entry
 */

SNP
FindSN( subnet )
u_long	subnet;
{
    struct SNent * snp;

    if ( (snp = SNFirst ) == NULL )
	{
	snp = (struct SNent *)calloc( 1, sizeof( struct SNent ) );
	snp->addr.s_addr = subnet;
	SNFirst = snp;
	SNLast  = snp;
	SNCnt++;
	fprintf( stdout, " %-16s subnet found\n", inet_ntoa( snp->addr ) );
	return( snp );
	}

    for ( ; snp != NULL; snp = snp->Next )
	if ( snp->addr.s_addr == subnet )
	    return( snp );

    snp = (struct SNent *)calloc( 1, sizeof( struct SNent ) );
    snp->addr.s_addr = subnet;
    SNLast->Next  = snp;
    SNLast  = snp;
    SNCnt++;
    fprintf( stdout, " %-16s subnet found\n", inet_ntoa( snp->addr ) );
    return( snp );
}


/*
 *  AddSNp
 *	Add subnet pointer to gateway
 */

AddSNp( gwp, subnet )
struct GWent * gwp;
u_long subnet;
{
    struct SNent * snp;
    struct snPointer * snpp;

    if ( ( snpp = gwp->snppStart ) == NULL )
	{
	snpp = (struct snPointer *)malloc( sizeof( struct snPointer ) );
	gwp->snppStart = snpp;
	gwp->snppEnd = snpp;
	snp = FindSN( subnet );
	snpp->sn = snp;
	snpp->Next = NULL;
	gwp->snCount++;
	return;
	}
    for ( ; snpp != NULL; snpp = snpp->Next )
	if ( snpp->sn->addr.s_addr == subnet )
	    return;

    snpp = (struct snPointer *)malloc( sizeof( struct snPointer ) );
    gwp->snppEnd->Next = snpp;
    gwp->snppEnd = snpp;
    snp = FindSN( subnet );
    snpp->sn = snp;
    snpp->Next = NULL;
    gwp->snCount++;
    return;
}


/*
 *  AddGWp
 *	Add gateway pointer to subnet
 */

AddGWp( snp, gwp )
struct SNent * snp;
struct GWent * gwp;
{
    struct gwPointer * gwpp;
    struct Intf * intfp;

    if ( ( gwpp = snp->gwppStart ) == NULL )
	{
	gwpp = (struct gwPointer *)malloc( sizeof( struct gwPointer ) );
	snp->gwppStart = gwpp;
	snp->gwppEnd = gwpp;
	gwpp->gw = gwp;
	gwpp->Next = NULL;
	snp->gwCount++;
	return;
	}
    for ( ; gwpp != NULL; gwpp = gwpp->Next )
	if ( gwpp->gw == gwp )
	    return;

    gwpp = (struct gwPointer *)malloc( sizeof( struct gwPointer ) );
    snp->gwppEnd->Next = gwpp;
    snp->gwppEnd = gwpp;
    gwpp->gw = gwp;
    gwpp->Next = NULL;
    snp->gwCount++;
    return;
}


/*
 *  AddToSNGW
 *	Add to the gateway/subnet/interface structures
 */

AddToSNGW( addr, subnet )
u_long	addr;
u_long	subnet;
{
    struct Intf * intfp;
    struct SNent * snp;

    /*  Find the interface  */

    intfp = FindIntf( addr );
    if ( intfp->GW == NULL )
	{
	gwp = CreateGW( intfp );
	intfp->GW = gwp;
	}
    else
        gwp = intfp->GW;

    AddSNp( gwp, subnet );

    snp = FindSN( subnet );
    AddGWp( snp, gwp );
    return;
}


/*
 *  AddToTrace( destp, addr )
 *	Add current packet to end of trace for this destination
 */

char *
AddToTrace( destp, addr )
struct Dest * destp;
u_long	addr;
{
    int i;
    struct Hop * nhp;
    struct Hop * php = NULL;
    struct Hop * xhp;
    struct Hop * yhp = NULL;
    u_long	subnet;
    u_long	ogw;

    subnet = addr & snmask.s_addr;

    if ( destp->NextHop == NULL )
	{ /*  First hop  */
	nhp = (struct Hop *)malloc( sizeof( struct Hop ) );
	nhp->addr.s_addr = addr;
	nhp->NextHop = NULL;
	nhp->TTL = destp->TTL;
	nhp->SameTTL = NULL;
	destp->NextHop = nhp;
	}
    else
	{ /*  Not first hop  */
	nhp = destp->NextHop;
	for ( i=0; i<=max_ttl; i++ )
	    {
	    if (nhp == NULL )
		break;

	    if ( ( addr == nhp->addr.s_addr ) && ( nhp->TTL != destp->TTL ) )
		return( Duplicate );

	    php = nhp;
	    nhp = nhp->NextHop;
	    }

	if ( i > max_ttl )
	    return( Max_ttl );  /* Should never happen */

	if ( php->TTL == destp->TTL )
	    for ( xhp = php; xhp->SameTTL != NULL; xhp = xhp->SameTTL )
		{
		if ( xhp->addr.s_addr == addr )
		    return( Previous );

		if ( ( addr & 0xff000000 ) == 0x7f000000 )
		    continue;
		yhp = xhp;
		}

	nhp = (struct Hop *)malloc( sizeof( struct Hop ) );
	nhp->addr.s_addr = addr;
	nhp->NextHop = NULL;
	nhp->TTL = destp->TTL;
	nhp->SameTTL = NULL;
	if ( php->TTL == destp->TTL )
	    {
	    if ( yhp == NULL )
	        php->SameTTL = nhp;
	    else
	        yhp->SameTTL = nhp;
	    }
	else
	    php->NextHop = nhp;

	if ( ( addr & 0xff000000 ) == 0x7f000000 )
	    return( Loopback );

	if ( ( php->TTL == destp->TTL ) && ( yhp != NULL ) )
	    {
	    AddIntfToGW( addr, yhp->addr.s_addr );
	    return( Intf_Added );
	    }
	ogw = php->addr.s_addr;
	AddToSNGW( ogw, subnet );
	}

/*  if ( ( addr & IN_CLASSA_NET ) == INADDR_LOOPBACK )  */
    if ( ( addr & 0xff000000 ) == 0x7f000000 )
	return( Loopback );

    AddToSNGW( addr, subnet );

    return( NULL );
}


/*
 *  AddDestReplies( destp )
 *	Add replies from destination subnets to SN/GW data structures
 */

char *
AddDestReplies( destp )
struct Dest * destp;
{
    u_long  addr;
    int     i;
    struct Hop * nhp;
    struct Hop * php = NULL;
    struct Hop * xhp;
    struct Hop * yhp = NULL;
    u_long  ogw;
    u_long  subnet;
    int     ttli = 0;

    addr = destp->Reply[ttli].s_addr;
    subnet = addr & snmask.s_addr;
    if ( subnet == addr )
	{
	ttli++;
	addr = destp->Reply[ttli].s_addr;
	if ( addr == 0 )
	    return;

	subnet = addr & snmask.s_addr;
	if ( subnet == addr )
	    return;
	}

    if ( destp->NextHop != NULL )
	{ /*  Not first hop  */
	nhp = destp->NextHop;

	for ( i=1; i<=max_ttl; i++ )
	    {
	    if (nhp->NextHop == NULL )
		break;

	    nhp = nhp->NextHop;
	    }

	if ( nhp->TTL == destp->ReplyTTL[ttli] )
	    {
	    AddIntfToGW( addr, nhp->addr.s_addr );
	    /* return( Intf_Added );  */
	    }

	ogw = nhp->addr.s_addr;
	AddToSNGW( ogw, subnet );
	}

    /* AddToSNGW( addr, subnet ); */

    return;
}

