/*
 *  data_server.h
 */

#include<sys/types.h>
#include<sys/socket.h>
#include<net/if.h>
#include<netinet/in.h>
#include<netinet/if_ether.h>

/* %W% %G% */

#define MAX_FAMILY 23
#define MAX_BUF 160 

#ifndef MAXLINE
#define MAXLINE  160
#endif

#define BUFSIZE 80
#define DNPTRSIZE 50

#define DATA_FILE "./data"
#define TESTDATA "./testdata"

#define DATA_PORT 1813
#define QUERY_PORT 1814

#define INTF		1
#define NETWORK		2
#define GATEWAY		3
#define EADDR_STORE	4
#define IN_ADDR_STORE 	5
#define IN_NAME_STORE	6
#define SUBNET		7

#define DATA_SOCK	0
#define QUERY_SOCK	1

#define READ		1
#define WRITE		2

#define GET_INTF	5
#define GET_NETWORK	6
#define GET_GW		7
#define GET_ALL		8
#define GET_EADDR	10
#define GET_IPADDR	15
#define CONFLICTS_EA	50
#define CONFLICTS_IP	51

#define NOT_FOUND	5
#define FOUND		10

#define CONFLICT	2
#define NO_CONFLICT	4


struct addr_table {
	int octet;
	union {
		struct avl_node *root_next_octet;
		struct intf *intf_data;
		struct network *subnet;
		} dir;
	struct addr_table *conflicts;
	};

struct name_table {
	char *name;
	char *root_of_subdomain;	
	char *parent;
	struct intf *intf_data;
	};

struct intf {
	struct intf_mac_addr *in_mac_addr;
	struct intf_prot_addr *ip_addr;
	struct intf_prot_addr *submask;
	struct intf_name *hostname;
	struct host *host;
	};

struct intf_mac_addr {
	struct ether_addr *mac_addr;
	time_t disc_time;
	char *disc_m;
	};

struct intf_prot_addr {
#ifdef SOCKADDR_IN
	struct sockaddr_in *pa;
#else
	struct sockaddr *pa;
#endif
	int addr_family;
	time_t disc_time;
	char *disc_m;
	};

struct intf_name {
	char *local_name;
	char *domain;
	time_t disc_time;
	char *disc_m;
	};

struct data_record {
	int data_type;
	union {
		struct intf *intf;
		struct network *subnet;
		struct host *gateway;
		} dr_type;
	};

#define  Intfp  dr_type.intf
#define  SNp    dr_type.subnet
#define  GWp    dr_type.gateway

struct network {
	struct intf_prot_addr  *net_numb;
	struct intf_prot_addr *subnet;
	int gw_count;
	struct gw_pointer *gwp_start ;
	};

struct gw_pointer {
	struct host *gw_host;
	struct gw_pointer *gw_next;
	};

struct subnet {
	struct network *subnet_info;
	struct intf *net_intf;
	struct subnet *next_subnet;
	};

struct intf_list {
	struct intf *subnet_intf;
	struct intf_list *next_intf;
	};

struct host {
	int no_intf;
	char *gw_name;
	struct subnet *connected_subnets;
	struct intf_list *intfs;
	char *device_type;
	int gateway;
	char *os;
	int nfs_fileserver;
	char *mail_host;
	};

struct netintf {
	u_long	addr;
	u_long	smask;
	u_long	nmask;
	};

