/*
 *  data_server.c
 */

/*
 *	finish		Upon receipt of signal, quit
 *	main		Journal Server main program
 *	Authenticate	Check for valid write to Journal Server
 */

#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "data_server.h"

#ifndef PIN
#include "pin.h"
#endif

extern	int	errno;
extern	int	intfcnt;
extern	struct netintf netintf[];

char *	jsfn = {TESTDATA};
FILE *	datafile;

int	debug=0;
FILE *	sfp;
char	pin[20] = PIN;
char	req[20];

int	data_port=  DATA_PORT;
int	query_port= QUERY_PORT;


/*
 *    finish
 *  On signal, quit.
 */

finish()
{
    fprintf( stderr, "\n js: signal received.  Quitting.\n" );
    if ( sfp == NULL )
        fclose( sfp );
    exit( 0 );
}


main( argc, argv )
int     argc;
char ** argv;
{
    int     cc;
    char *  cp;
    int     client_len;
    struct sockaddr_in client_addr;
    int     i;
    char    line[MAXLINE];
    int     n;
    int     nfds;		/* number of file descriptors available */
    int     real_sockfd;
    int     socket_store[3];
    int     sockfd;
    struct timeval td;
    fd_set  mask;		/* mask to select what file descriptor to
				   listen on */
    fd_set  tmpmask;		/* temporary mask to store select fields */



    /* get the command line arguments if they exist */

    while ( *(++argv) )
	{
	if ( **argv != '-' )
	    {
	    printf( "Use a - before arguments\n" );
	    exit( 1 );
	    }

	switch ( *(*argv + 1) )
	    {
	    case 'p':
		if ( *(++argv) )
		    {
		    data_port = atoi( *argv );
		    query_port = data_port + 1;
		    }

	    case 'f':
		if ( *(++argv) )
		    jsfn = argv[0];
	    }
	}
    
    GetIntf();				/* Get interfaces on this machine  */

    signal( SIGINT,  finish );		/* Quit on TERM  */
    signal( SIGTERM, finish );		/* Quit on TERM  */

    init_sockets( socket_store );	/* Initialize sockets */

    nfds = getdtablesize();
    if ( nfds > FD_SETSIZE )
	nfds = FD_SETSIZE;

    /* set the mask for the select call to pick the correct socket to connect */

    FD_ZERO( &mask );
    FD_SET( socket_store[DATA_SOCK], &mask );
    FD_SET( socket_store[QUERY_SOCK], &mask );

    /* loop forever listening for connections and processing data */

    for (;;)
	{

#ifdef DEBUG
	fprintf( stderr, "js: About to select\n" );
#endif

	/*  Wait for connect requests  */

	tmpmask = mask;
	td.tv_sec = 1;
	td.tv_usec = 0;
	n = select( nfds, &tmpmask, (fd_set *) NULL, (fd_set *) NULL, &td );
	if ( n < 0 )
	    {
	    if (errno == EINTR)
		continue;

	    perror( "select" );
	    exit(1);
	    }

	if ( n == 0 )
	    continue;

	client_len = sizeof(client_addr);
	bzero( &client_addr, sizeof(client_addr) );

	/* see if the data sock has been selected  */

	if ( FD_ISSET( socket_store[DATA_SOCK], &tmpmask ) )
	    {
	    void    store_data();

	    /* create a new socket completing the data connection  */

	    real_sockfd = accept( socket_store[DATA_SOCK],
				(struct sockaddr *) & client_addr, &client_len);
	    if ( real_sockfd < 0 )
		{
		if ( errno == EINTR )
		    continue;
		perror( "accept data" );
		exit(0);
		}

	    /* load the data from the network connection */

	    sfp = fdopen( real_sockfd, "r" );
	    cc = Authenticate( &client_addr, sfp, WRITE );
	    if ( ! cc )
		fprintf( stderr, "js: authentication failed for %s\n",
					inet_ntoa( client_addr.sin_addr ) );
	    else
		{
		for ( i=0; ; i++ )
		    {
		    cp = fgets( line, MAXLINE, sfp );
		    if ( cp == NULL )
			break;

		    fputs( line, stdout );
		    }
		printf( " js.fake: %d lines transfered from %s\n", i,
					inet_ntoa( client_addr.sin_addr ) );
		fflush( stdout );
		}
	    fclose( sfp );
	    }

	/* see if the query sock has been selected  */

	if ( FD_ISSET( socket_store[QUERY_SOCK], &tmpmask ) )
	    {
	    char   *buf;
	    u_short opcode;

	    buf = (char *) malloc(BUFSIZE);
	    if ( buf == NULL )
		{
		fprintf(  stderr, "js: Error in allocating memory for buf\n");
		exit(1);
		}

	    /* create a new socket completing the query connection  */

	    real_sockfd = accept( socket_store[QUERY_SOCK],
			(struct sockaddr *) & client_addr, &client_len );
	    if ( real_sockfd < 0 )
		{
		if (errno == EINTR)
		    continue;

		perror( "accept query" );
		exit(0);
		}

	    sfp = fdopen( real_sockfd, "w" );

    read_label:
	    n = read( real_sockfd, buf, BUFSIZE );
	    if ( n < 0 )
		{
		if ( errno = EINTR )
		    goto read_label;

		perror("read:");
		exit(1);
		}

	    opcode = ntohs(*((u_short *) buf));

	    switch (opcode)
		{
		char  * root_node;
		int     nargs;
		struct in_addr ipaddr;

		case GET_INTF:
		    printf( " js.fake: GET_INTF request\n" );
		    break;

		case GET_EADDR:
		    printf( " js.fake: GET_EADDR request\n" );
		    break;

		case GET_IPADDR:
		    printf( " js.fake: GET_IPADDR request\n" );
		    break;

		case GET_NETWORK:
		    printf( " js.fake: GET_NETWORK request\n" );
		    break;

		case GET_GW:
		    printf( " js.fake: GET_GW request\n" );
		    break;

		case GET_ALL:
		    printf( " js.fake: GET_ALL request\n" );
		    break;

		case CONFLICTS_EA:
		    printf( " js.fake: CONFLICTS_EA request\n" );
		    break;

		case CONFLICTS_IP:
		    printf( " js.fake: CONFLICTS_IP request\n" );
		    break;
		}

	    fclose( sfp );
	    }  /*  end of QUERY if  */
	}  /*  end of infinite for loop  */
}  /*  end of main  */


/*
 *    Authenticate
 *  Authenticate the data connection
 */

Authenticate( addr, sfp, mode )
struct sockaddr_in * addr;
FILE *	sfp;
int	mode;
{
    int    cc;
    char * cp;
    int    i;
    char   line[MAXLINE];
    u_long mask;
    u_long spin;

    if       ( IN_CLASSA( addr->sin_addr.s_addr ) )
	mask = IN_CLASSA_NET;
    else  if ( IN_CLASSB( addr->sin_addr.s_addr ) )
	mask = IN_CLASSB_NET;
    else  if ( IN_CLASSC( addr->sin_addr.s_addr ) )
	mask = IN_CLASSC_NET;
    else
	return( 0 );

    if ( addr->sin_addr.s_addr & mask == IN_LOOPBACKNET )
	return( 1 );

    for ( i=0; i<intfcnt; i++ )
	if ( ( ( addr->sin_addr.s_addr ^ netintf[i].addr ) & mask ) == 0 )
	    break;

    if ( i < intfcnt )
	return( 1 );

    if ( i == intfcnt )
	return( 0 );

/*
 *  WaitForInput( sfp );
 *
 *  cp = fgets( line, MAXLINE, sfp );
 *  if ( cp == NULL )
 *	return( 0 );
 *
 *  cc = sscanf( line, "%s %d", req, &spin );
 *  if ( cc == 1 )
 *	spin = 0;
 *
 *  if ( cc > 0 )
 *      if ( mode == READ )
 *	    return( req );
 *	else
 *	    if ( spin == pin )
 *	        return( req );
 *
 *  fclose( sfp );
 */
    return( 0 );
}

