/*
 *  data.c
 */

/*
 *	init_database	Initialize the data base in memory
 *	store_data
 *	store_ipaddr
 *	store_eaddr
 *	addr_compare
 *	store_name
 *	name_compare
 *	store_octet
 *	write_file	Set file to cause data file to be written later
 *	print_intf
 *	print_intf_addr
 *	print_ether_addr
 *	print_ether_conflicts
 *	print_ip_conflicts
 *	check_conflicts
 *	print_data
 */

#include <stdio.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#include "data_server.h"

/* %W% %G% */

extern	char *	jsfn;
extern	int	init_data;

int	ea_counter;
int	ea_con_counter;
int	ip_counter;
int	ip_con_counter;

int	now;			/* results of most recent time call */


/*
 * function to initialize the data server's database
 *
 * No arguments passed in and none returned
 *
 */

void
init_database()
{
    int     jsfd;		/* file descriptor for data file */
    extern char *data_store[];	/* array of headers for various data trees */
    void    load_data();	/* function to load data */
    void    print_data();	/* function to write data */
    char   *list_init();	/* initialize a linked list */
    void    store_data();

    /* open the data file where the ascii version of database lies */

    jsfd = open( jsfn, O_RDONLY );

    /* initialize the list where all data lies */

    data_store[INTF] = list_init( NULL, NULL, print_data );

    /* initialize list for gateways */

    data_store[GATEWAY] = list_init( NULL, NULL, print_data );
    data_store[NETWORK] = list_init( NULL, NULL, print_data );

    /* load the data from the data file */

    if ( jsfd >= 0 )
	{
	init_data = 1;
	get_data( jsfd, store_data, 0);
	}
    return;
}


/*
 *  store_data used to put all data into the appropriate data structures
 */

void
store_data( data )
struct data_record * data;		/* data record to store */
{
    static char *list_last_node = NULL;	/* last node in linked list for all
					   data */
    static char *gateway_last_node = NULL;	/* last node in list of
						   gateways */
    static char *subnet_last_node = NULL;
    void    store_addr();		/* routine to store addresses */
    struct name_table *store_name();	/* routine to store name */
    int     store_ipaddr();
    int     store_eaddr();
    struct addr_table *store_octet();
    char   *list_insert();		/* routine to add node to linked list */
    struct data_record *tmp_data;	/* return value from storing data */
    int     x;				/* counter */
    int     update_record = 0;

    now = time( NULL );
    if ( data->data_type == INTF )
	{
	int	data_found = 0;
	int	eaddr_status = NOT_FOUND;
	int	ip_addr_status = NOT_FOUND;
	struct addr_table *eaddr_exist = NULL;
	struct addr_table *ipaddr_exist = NULL;
	struct name_table *name_addr = NULL;
	void add_entry();

	/* store the ethernet address first.
	 * This is the first data piece stored since
	 *   this doesn't change often */

	if  ( data->dr_type.intf->in_mac_addr ) 
	    {
	    eaddr_status = store_eaddr( data->dr_type.intf, &eaddr_exist );
	    data_found = 1;
	    }

	/*  store the IP address  */
	/*  This is second item that changes infrequently. */
	/*  It is possible that this can change even though  */
	/*    the ethernet address does not change  */

	if (data->dr_type.intf->ip_addr) 
	    {
	    ip_addr_status = store_ipaddr( data->dr_type.intf, &ipaddr_exist );
	    data_found = 1;
	    }

	/* finally store the name */

	if ( data->dr_type.intf->hostname )
	    {
	    name_addr = store_name( data->dr_type.intf);
	    data_found = 1;
	    }

	/* store the aliases for a host */

	/* update the time for the data */

	/* if ( ! init_data )
	    update_data_time( data ); */

	/* put all data into a linked list */

	if ( (eaddr_status == NOT_FOUND && ip_addr_status == NOT_FOUND) &&
								data_found )
	    list_last_node =
			list_insert( data_store[INTF], list_last_node, data);
	else
	    add_entry( eaddr_exist, ipaddr_exist, name_addr, data  );
	}

    if ( data->data_type == NETWORK )
	{
	int     subnet_not_stored = 0;
	struct network * current_network;

	if ( data->dr_type.subnet->net_numb )
	    {
	    u_char * tmp_addr;
	    char  ** root;
	    struct addr_table * tmp_node;

	    root = &data_store[SUBNET];
	    tmp_addr = (u_char *) malloc(4 * sizeof(u_char));
	    if ( tmp_addr == NULL )
		{
		fprintf(stderr, "js: Error allocating space for tmp_addr\n");
		exit(1);
		}

	    bcopy(((struct sockaddr_in *)
			data->dr_type.subnet->net_numb->pa)->sin_addr,
			(u_char *) tmp_addr, 4 * sizeof(u_char));
	    tmp_node = store_octet(root, tmp_addr, NULL);
	    if ( !tmp_node->dir.subnet )
		{
		tmp_node->dir.subnet = data->dr_type.subnet;
		subnet_not_stored = 1;
		}
	    else
		current_network = tmp_node->dir.subnet;

	    }

	if ( data->dr_type.subnet->gwp_start )
	    {
	    struct gw_pointer * gw;

	    gw = data->dr_type.subnet->gwp_start;
	    while ( gw )
		{
		char  ** root;
		struct addr_table * gw_address;
		struct subnet ** subnets;
		struct intf_list ** intfs;
		u_char * tmp_addr;
		int     not_found = 1;

		root = &data_store[IN_ADDR_STORE];
		tmp_addr = (u_char *) malloc(4 * sizeof(u_char));
		if ( tmp_addr == NULL )
		    {
		    fprintf(stderr,
				"js: error allocating space for tmp_addr\n");
		    exit(1);
		    }

		bcopy( ((struct sockaddr_in *)
			gw->gw_host->intfs->subnet_intf->ip_addr->pa)->sin_addr,
			(u_char *) tmp_addr, 4 * sizeof(u_char) );
		gw_address = store_octet( root, tmp_addr, NULL );
		if ( ! gw_address->dir.intf_data )
		    {
		    struct data_record * intf_data;

		    gw_address->dir.intf_data = gw->gw_host->intfs->subnet_intf;
		    gw_address->dir.intf_data->ip_addr->disc_time = now;
		    intf_data = (struct data_record *)
					malloc(sizeof(struct data_record));
		    if ( intf_data == NULL )
			{
			fprintf( stderr,
				"js: unable to allocate space for intf_data\n");
			exit( 1 );
			}

		    bzero( intf_data, sizeof(struct data_record) );
		    intf_data->data_type = INTF;
		    intf_data->dr_type.intf = gw_address->dir.intf_data;
		    list_last_node = list_insert( data_store[INTF],
						  list_last_node, intf_data );
		    }
		if ( gw_address->dir.intf_data->host )
		    {
		    /* free( gw->gw_host ); */
		    gw->gw_host = gw_address->dir.intf_data->host;
		    }
		else
		    {
		    struct data_record * gateway;

		    gw_address->dir.intf_data->host = gw->gw_host;
		    gateway = (struct data_record *)
					malloc(sizeof(struct data_record));
		    if ( gateway == NULL )
			{
			fprintf( stderr,
				    "js: error allocating space for gateway\n");
			exit(1);
			}

		    bzero( gateway, sizeof(struct data_record) );
		    gateway->data_type = GATEWAY;
		    gateway->dr_type.gateway = gw->gw_host;
		    gateway_last_node = list_insert(data_store[GATEWAY],
						gateway_last_node, gateway);
		    }
		intfs = &gw_address->dir.intf_data->host->intfs;
		while ( *intfs )
		    {
		    if ( !bcmp( tmp_addr,
				((struct sockaddr_in *)
				  (*intfs)->subnet_intf->ip_addr->pa)->sin_addr,
				4 * sizeof(u_char)))
			{
			not_found = 0;
			break;
			}
		    intfs = &(*intfs)->next_intf;
		    }
		if ( not_found )
		    {
		    struct intf_list * new_intf;

		    new_intf = (struct intf_list *)
					malloc(sizeof(struct intf_list));
		    if ( new_intf == NULL )
			{
			fprintf( stderr,
				"js: Unable to allocate memory for new_intf\n");
			exit(1);
			}

		    bzero( new_intf, sizeof(struct intf_list) );
		    new_intf->subnet_intf = gw_address->dir.intf_data;
		    *intfs = new_intf;
		    }
		not_found = 1;
		subnets = &gw_address->dir.intf_data->host->connected_subnets;
		while ( *subnets )
		    {
		    int	cc;

		    cc = bcmp( ((struct sockaddr_in *)
			    data->dr_type.subnet->net_numb->pa)->sin_addr,
			((struct sockaddr_in *)
			    (*subnets)->subnet_info->net_numb->pa)->sin_addr,
			4 * sizeof(u_char));
		    if ( ! cc )
			{
			not_found = 0;
			break;
			}
		    subnets = &(*subnets)->next_subnet;
		    }
		if ( not_found )
		    {
		    struct subnet * new_subnet;

		    new_subnet = (struct subnet *)
						malloc(sizeof(struct subnet));
		    if ( new_subnet == NULL )
			{
			fprintf( stderr,
				"js: Unable to allocate mem for new_subnet\n");
			exit(1);
			}

		    bzero( new_subnet, sizeof(struct subnet) );
		    new_subnet->subnet_info = data->dr_type.subnet;
		    (*subnets) = new_subnet;
		    }
/*
 *  This section of code needs improving so the correct intf goes
 *  with the correct subnet
 */

/*
 *		if ( (*subnets)->net_intf )
 *		    {
 *		    free( (*subnets)->net_intf );
 */
		    /* put an error message here or ask dave about this */
/*
 *		    } 
 *		    (*subnets)->net_intf = gw_address->dir.intf_data;
 */
		gw = gw->gw_next;
		}
	    }
	if ( subnet_not_stored )
	    subnet_last_node = list_insert(data_store[NETWORK],
							subnet_last_node, data);
	else
	    {
	    struct gw_pointer * gwp = data->dr_type.subnet->gwp_start;

	    while ( gwp )
		{
		struct gw_pointer ** network_gwp;

		network_gwp = &current_network->gwp_start;
		while ( *network_gwp )
		    {
		    if ( (*network_gwp)->gw_host == gwp->gw_host )
			break;

		    network_gwp = &(*network_gwp)->gw_next;
		    }

		gwp = gwp->gw_next;
		}
	    }
	}

    if (data->data_type == GATEWAY)
	{
	struct intf_list * interfaces;
	struct subnet * subnets;
	int     gw_not_found = 0;

	if ( data->dr_type.gateway->intfs )
	    {
	    interfaces = data->dr_type.gateway->intfs;
	    while ( interfaces )
		{
		char  ** root;
		u_char * tmp_addr;
		struct addr_table * tmp_node;

		root = &data_store[IN_ADDR_STORE];
		tmp_addr = (u_char *) malloc(4 * sizeof(u_char));
		if ( tmp_addr == NULL )
		    {
		    fprintf( stderr,
				"js: Unable to allocate memory for tmp_addr\n");
		    exit(1);
		    }

		bcopy( ((struct sockaddr_in *)
				interfaces->subnet_intf->ip_addr->pa)->sin_addr,
				tmp_addr, 4 * sizeof(u_char));
		tmp_node = store_octet( root, tmp_addr, NULL );
		if ( !tmp_node->dir.intf_data )
		    {
		    struct data_record * intf_data;

		    tmp_node->dir.intf_data = interfaces->subnet_intf;
		    intf_data = (struct data_record *)
					malloc(sizeof(struct data_record));
		    if ( intf_data == NULL )
			{
			fprintf( stderr,
				"js: unable to allocate space for intf_data\n");
			exit(1);
			}

		    bzero( intf_data, sizeof(struct data_record) );
		    intf_data->data_type = INTF;
		    intf_data->dr_type.intf = tmp_node->dir.intf_data;
		    intf_data->dr_type.intf->ip_addr->disc_time = now;
		    list_last_node = list_insert(data_store[INTF],
						list_last_node, intf_data);
		    }
		else
		    {
		    /* free( interfaces->subnet_intf ); */
		    interfaces->subnet_intf = tmp_node->dir.intf_data;
		    }
		if ( !tmp_node->dir.intf_data->host )
		    {
		    tmp_node->dir.intf_data->host = data->dr_type.gateway;
		    gw_not_found = 1;
		    }
		interfaces = interfaces->next_intf;
		}
	    }

	if ( data->dr_type.gateway->connected_subnets )
	    {
	    subnets = data->dr_type.gateway->connected_subnets;
	    while ( subnets )
		{
		int     gw_pointer_not_found = 1;
		int     intf_not_found = 1;
		u_char * tmp_addr;
		char  ** root;
		struct addr_table * tmp_node;
		struct gw_pointer ** gw_node;
		struct gw_pointer * gateway;
		struct intf_list ** interfaces;

		tmp_addr = (u_char *) malloc(4 * sizeof(u_char));
		if ( tmp_addr == NULL )
		    {
		    fprintf( stderr,
				"js: Unable to allocate memory for tmp_addr\n");
		    exit(1);
		    }

		gateway = (struct gw_pointer *)
					malloc(sizeof(struct gw_pointer));
		if ( gateway == NULL )
		    {
		    fprintf( stderr,
				"js: Can't allocate space for gateway\n");
		    exit(1);
		    }

		bzero( gateway, sizeof(struct gw_pointer) );
		if ( subnets->net_intf )
		    {
		    struct addr_table * tmp_node;

		    root = &data_store[IN_ADDR_STORE];
		    bcopy( ((struct sockaddr_in *)
				subnets->net_intf->ip_addr->pa)->sin_addr,
				tmp_addr, 4 * sizeof(u_char));
		    tmp_node = store_octet( root, tmp_addr, NULL );
		    if ( !tmp_node->dir.intf_data )
			{
			struct data_record * intf_data;

			tmp_node->dir.intf_data = subnets->net_intf;
			intf_data = (struct data_record *)
				    	malloc(sizeof(struct data_record));
			if ( intf_data == NULL )
			    {
			    fprintf( stderr,
				"js: Unable to allocate mem for intf_data\n");
			    exit(1);
			    }

			bzero( intf_data, sizeof(struct data_record) );
			intf_data->data_type = INTF;
			intf_data->dr_type.intf = subnets->net_intf;
			intf_data->dr_type.intf->ip_addr->disc_time = now;
			list_last_node = list_insert(data_store[INTF],
						list_last_node, intf_data);
			}
		    else
			{
			/* free( subnets->net_intf ); */
			subnets->net_intf = tmp_node->dir.intf_data;
			}
		    if ( !tmp_node->dir.intf_data->host )
			{
			tmp_node->dir.intf_data->host = data->dr_type.gateway;
			gw_not_found = 1;
			}
		    interfaces = &data->dr_type.gateway->intfs;
		    while ( *interfaces )
			{
			int	cc;

			cc = bcmp( ((struct sockaddr_in *)
			     (*interfaces)->subnet_intf->ip_addr->pa)->sin_addr,
			    ((struct sockaddr_in *)
			     subnets->net_intf->ip_addr->pa)->sin_addr,
			    4 * sizeof(u_char) );
			if ( cc == 0 )
			    {
			    intf_not_found = 0;
			    break;
			    }
			interfaces = &(*interfaces)->next_intf;
			}
		    if ( intf_not_found )
			{
			struct intf_list * new_intf;

			new_intf = (struct intf_list *)
					malloc(sizeof(struct intf_list));
			if ( new_intf == NULL )
			    {
			    fprintf( stderr,
				"js: unable to allocate memory for new_intf\n");
			    exit(1);
			    }

			bzero( new_intf, sizeof(struct intf_list) );
			new_intf->subnet_intf = subnets->net_intf;
			*interfaces = new_intf;
			}
		    }

		root = &data_store[SUBNET];
		bcopy( ((struct sockaddr_in *)
			  subnets->subnet_info->net_numb->pa)->sin_addr,
			tmp_addr,
			4 * sizeof(u_char));
		tmp_node = store_octet( root, tmp_addr, NULL );
		if ( !tmp_node->dir.subnet )
		    {
		    struct data_record * network_data;

		    tmp_node->dir.subnet = subnets->subnet_info;
		    network_data = (struct data_record *)
				malloc(sizeof(struct data_record));
		    if (network_data == NULL)
			{
			fprintf( stderr,
			    "js: unable to allocate mem for network_data\n");
			exit(1);
			}

		    bzero( network_data, sizeof(struct data_record) );
		    network_data->data_type = NETWORK;
		    network_data->dr_type.subnet = tmp_node->dir.subnet;
		    subnet_last_node = list_insert( data_store[NETWORK],
				subnet_last_node, network_data );
		    }
		else
		    {
		    /* free( subnets->subnet_info ); */
		    subnets->subnet_info = tmp_node->dir.subnet;
		    }
		gw_node = &tmp_node->dir.subnet->gwp_start;
		while ( *gw_node )
		    {
		    struct intf_list * intfs;

		    intfs = (*gw_node)->gw_host->intfs;
		    while ( intfs )
			{
			int cc;

			cc = bcmp( ((struct sockaddr_in *)
				  data->dr_type.gateway->intfs->subnet_intf->ip_addr->pa)->sin_addr,
				((struct sockaddr_in *)
				  intfs->subnet_intf->ip_addr->pa)->sin_addr,
				4 * sizeof(u_char) );
			if ( cc == 0 )
			    {
			    gw_pointer_not_found = 0;
			    break;
			    }
			intfs = intfs->next_intf;

			}
		    gw_node = &(*gw_node)->gw_next;
		    }
		if ( gw_pointer_not_found )
		    {
		    gateway->gw_host = data->dr_type.gateway;
		    *gw_node = gateway;
		    }
		subnets = subnets->next_subnet;
		}
	    if ( gw_not_found )
		gateway_last_node = list_insert(data_store[GATEWAY],
						gateway_last_node, data);
	    }
	}
}


int
store_ipaddr( intf_rec, exist_intf )
struct intf * intf_rec;
struct addr_table **exist_intf;
{
    u_char * tmp_addr;

#ifdef DEBUG
    struct avl_node ** root;

#else
    char  ** root;

#endif

    struct addr_table * tmp_node;
    struct addr_table * store_octet();
    struct addr_table * orig_data;


    tmp_addr = (u_char *) malloc(4 * sizeof(u_char));
    if ( tmp_addr == NULL )
	{
	fprintf( stderr, "js: error allocating space for tmp_addr\n");
	exit(1);
	}

    root = &data_store[IN_ADDR_STORE];
    bcopy( ((struct sockaddr_in *) intf_rec->ip_addr->pa)->sin_addr,
		(u_char *) tmp_addr, 4 * sizeof(u_char));
    tmp_node = (struct addr_table *) store_octet( root, tmp_addr, &orig_data );
    orig_data->dir.intf_data = intf_rec;
    if ( tmp_node != orig_data )
	{
	*exist_intf = tmp_node;

	intf_rec->ip_addr->disc_time = now;
	tmp_node->dir.intf_data->ip_addr->disc_time =  now;
	if ( check_conflicts( tmp_node, orig_data ) == NO_CONFLICT )
	    return( FOUND );
	else
	    return( NOT_FOUND );
	}
    else
	{
	if ( ! intf_rec->ip_addr->disc_time )
		intf_rec->ip_addr->disc_time = now;
	*exist_intf = NULL;
	return( NOT_FOUND );
	}
}


int
store_eaddr( intf_rec, exist_intf )
struct intf * intf_rec;
struct addr_table **exist_intf;
{

    struct addr_table * tmp_node;
    struct addr_table * orig_data;
    struct addr_table *store_ea_octet();
    int     addr_compare();
    struct avl_node ** root;
    
    root = (struct avl_node **) &(data_store[EADDR_STORE]);
    tmp_node = store_ea_octet(root,intf_rec,&orig_data);
    orig_data->dir.intf_data = intf_rec;
    if ( tmp_node != orig_data )
	{
	*exist_intf = tmp_node;
	intf_rec->in_mac_addr->disc_time = now;
	tmp_node->dir.intf_data->in_mac_addr->disc_time = now;
	if ( check_conflicts( tmp_node, orig_data ) == NO_CONFLICT )
	    return( FOUND );
	else
	    return( NOT_FOUND );
	}
    else
	{
	if ( ! intf_rec->in_mac_addr->disc_time )
		intf_rec->in_mac_addr->disc_time = now;
	*exist_intf = NULL;
	return( NOT_FOUND );
	}
}


int
addr_compare( data1, data2 )
struct addr_table * data1;
struct addr_table * data2;
{
    if ( data2->octet < data1->octet )
	return( -1 );
    else if ( data1->octet < data2->octet )
	return( 1 );
    else
	return( 0 );
}


/*
 * store_name stores a full domain name.
 * It uses an avl tree to store the same level of name.
 * As a part of the data at a node the root of the next level name is stored
 */

struct name_table *
store_name( intf_rec )
struct intf *intf_rec;
{
    char  * sp;			/* pointer to working part of name */
    char ** root;		/* root node name tree */
    int     n;			/* length of name */
    int     hostname_part = 0;	/* tells us if name is part of domain or not */
    char  * work_name;		/* start of the entire working name */
    int     name_compare();	/* routine to compare names */

    struct name_table *tmp_node;/* leaf for name node */

    if ( intf_rec->hostname->domain )
	{
	work_name = (char *) malloc(
		    strlen(intf_rec->hostname->local_name) +
		    strlen(intf_rec->hostname->domain) + 2);
	if ( work_name == NULL )
	    {
	    fprintf( stderr, "js: error allocating space for work_name\n" );
	    exit(1);
	    }

	sprintf( work_name, "%s.%s", intf_rec->hostname->local_name,
		    intf_rec->hostname->domain );
	}
    else
	{
	/* work_name is used to put the current working part of the name */

	work_name = malloc(strlen(intf_rec->hostname->local_name) + 1);
	if ( work_name == NULL )
	    {
	    fprintf( stderr, "js: Can't allocate memory for work_name\n" );
	    exit(1);
	    }

	sprintf(work_name, "%s",intf_rec->hostname->local_name);

	/* name is the full domain name at this point */
	}

    /* get the address of the root node */

    root = &data_store[IN_NAME_STORE];

    /* get the length of the full name, replace the last . with null */

    n = strlen( work_name );
    if ( work_name[n] == '.' )
	work_name[n] = '\0';

    /* get the last part of the name and work from top down to host */

    sp = rindex( work_name, '.' );
    do {
	/* if sp is null, we must either be at the host part of the name or */
	/*  the full domain name was not given */

	if ( sp == NULL )
	    {
	    hostname_part = 1;		/* this will get us out of the loop */

	    /* set start pointer to workname - 1 */

	    sp = work_name;
	    sp--;
	    }

	/* now null out the . for the next word */

	else
	    *sp = '\0';

	tmp_node = (struct name_table *) malloc(sizeof(struct name_table));
	if ( tmp_node == NULL )
	    {
	    fprintf(stderr, "js: error creating space for tmp_node\n");
	    exit(1);
	    }

	bzero( tmp_node, sizeof(struct name_table) );
	tmp_node->name = (char *) malloc(strlen((sp + 1)) + 1);
	if ( tmp_node->name == NULL )
	    {
	    fprintf( stderr, "js: error allocating space for tmp_node->name\n");
	    exit(1);
	    }

	/* name table is a node in the avl tree with roots for an avl tree */
	/*  for subdomains */

	strcpy( tmp_node->name, (sp + 1) );

	/* set the data and the root of the next domain to null */

	tmp_node->intf_data = NULL;
	tmp_node->root_of_subdomain = NULL;

	/* search for this part of name and insert in current tree if not */
	/*  found */

	tmp_node = (struct name_table *) avlsearch( tmp_node, name_compare,
									root );

	/* put the address of the subdomain root to process the next part of */
	/*  the name */

	root = &(tmp_node->root_of_subdomain);
	} while ( ((sp = rindex(work_name, '.')) != NULL) ||
							hostname_part == 0);

    /* data only goes with a host */
    tmp_node->intf_data = intf_rec;
    return(tmp_node);
}


/*
 *  Compares names and returns the approriate response from strcmp
 */

int
name_compare( data1, data2 )
struct name_table * data1;	/* name table entry 1 */
struct name_table * data2;	/* name table entry 2 */
{
    /* compare the two names and return the appropriate response */

    return( strcmp( data2->name, data1->name ) );
}


struct addr_table *
store_octet( root, tmp_addr, return_addr )
struct avl_node ** root;
u_char * tmp_addr;
struct addr_table ** return_addr;
{
    struct addr_table * tmp_node;
    struct addr_table * ans_data;
    int     addr_compare();
    int     counter;

    for ( counter = 0; counter < 4; counter++ )
	{
	tmp_node = (struct addr_table *) malloc(sizeof(struct addr_table));

	if ( tmp_node == NULL )
	    {
	    fprintf( stderr, "js: error allocating space for tmp_node\n");
	    exit(1);
	    }
	bzero( tmp_node, sizeof(struct addr_table) );
	tmp_node->octet = *tmp_addr;
	ans_data = (struct addr_table *) avlsearch( tmp_node, addr_compare,
									root );
	if ( !ans_data )
	    return( ans_data );

	root = (struct avl_node **) &(ans_data->dir.root_next_octet);
	tmp_addr++;
	}

    if ( return_addr )
	*return_addr = tmp_node;

    return( ans_data );
}


struct addr_table *
store_ea_octet(root,intf_rec,return_addr)
struct avl_node **root;
struct intf *intf_rec;
struct addr_table **return_addr;
{
    struct addr_table *tmp_node;
    int     counter;
    int     manufacture_eaddr = 0;
    struct addr_table *orig_data;

    tmp_node = (struct addr_table *) malloc(sizeof(struct addr_table));
    if ( tmp_node == NULL )
	{
	fprintf( stderr, "js: Error allocating space for tmp_node\n");
	exit(1);
	}

    bzero( tmp_node, sizeof(struct addr_table) );

    for ( counter = 0; counter <= 2; counter++ )
	{
	manufacture_eaddr |= (int)
	    intf_rec->in_mac_addr->mac_addr->ether_addr_octet[counter];
	manufacture_eaddr <<= 8;
	}

    manufacture_eaddr >>= 8;
    tmp_node->octet = manufacture_eaddr;

    tmp_node = (struct addr_table *)
			avlsearch( (char *)tmp_node, addr_compare, root );

    root = &(tmp_node->dir.root_next_octet);

    orig_data = (struct addr_table *) malloc(sizeof(struct addr_table));
    if ( orig_data == NULL )
	{
	fprintf( stderr, "js: Error allocating space for tmp_node\n" );
	exit(1);
	}
    bzero(orig_data,sizeof(struct addr_table));
    manufacture_eaddr = 0;

    for ( counter = 3; counter <= 5; counter++ )
	{
	manufacture_eaddr |= (int)
	    intf_rec->in_mac_addr->mac_addr->ether_addr_octet[counter];
	manufacture_eaddr <<= 8;
	}

    orig_data->octet = manufacture_eaddr >> 8;

    tmp_node = (struct addr_table *) avlsearch( orig_data, addr_compare, root );

    if(return_addr)
	*return_addr = orig_data;

    return(tmp_node);
}


/*
 *  Set "writefile" flag
 */

void
write_file()
{
    extern int writefile;

    writefile = 1;
    return;
}

 
void
print_intf_addr( data )
struct addr_table * data;
{
    if ( ip_counter == 4 )
	{
	struct data_record *new_data;

	new_data = (struct data_record *) malloc(sizeof(struct data_record));
	if ( new_data == NULL )
	    {
	    fprintf( stderr, "js: Unable to allocate space for new_data\n");
	    exit(1);
	    }

	new_data->data_type = INTF;
	new_data->dr_type.intf = data->dir.intf_data;
	write_data( new_data, 1,1 );
	}
    else
	{
	ip_counter++;
	avl_walk( data->dir.root_next_octet, print_intf_addr );
	ip_counter--;
	}
    return;
}


void
print_ether_addr( data )
struct addr_table * data;
{
    if ( ea_counter == 2 )
	{
	struct data_record * new_data;

	new_data = (struct data_record *) malloc(sizeof(struct data_record));
	if ( new_data == NULL )
	    {
	    fprintf( stderr, "js: Unable to allocate space for new_data\n");
	    exit(1);
	    }

	bzero( new_data, sizeof(struct data_record) );
	new_data->data_type = INTF;
	new_data->dr_type.intf = data->dir.intf_data;
	write_data( new_data, 1,1 );
	}
    else
	{
	ea_counter++;
	avl_walk( data->dir.root_next_octet, print_ether_addr );
	ea_counter--;
	}
    return;
}


void
print_ether_conflicts( data )
struct addr_table * data;
{
    if ( ea_con_counter == 2 )
	{
	if ( data->conflicts )
	    {
	    struct addr_table * next_conflict;

	    next_conflict = data;
	    while ( next_conflict )
		{
		struct data_record * new_data;

		new_data = (struct data_record *)
			    		malloc(sizeof(struct data_record));
		if ( new_data == NULL )
		    {
		    fprintf( stderr,
				"js: Unable to allocate space for new_data\n");
		    exit(1);
		    }

		bzero( new_data, sizeof(struct data_record) );
		new_data->data_type = INTF;
		new_data->dr_type.intf = next_conflict->dir.intf_data;
		write_data( new_data, 1,1 );
		next_conflict = next_conflict->conflicts;
		}
	    }
	}
    else
	{
	ea_con_counter++;
	avl_walk( data->dir.root_next_octet, print_ether_conflicts );
	ea_con_counter--;
	}
    return;
}


void
print_ipaddr_conflicts( data )
struct addr_table * data;
{
    if ( ip_con_counter == 4 )
	{
	if ( data->conflicts )
	    {
	    struct addr_table * next_conflict;

	    next_conflict = data;
	    while ( next_conflict )
		{
		struct data_record * new_data;

		new_data = (struct data_record *)
			    		malloc(sizeof(struct data_record));
		if ( new_data == NULL )
		    {
		    fprintf( stderr,
				"js: Unable to allocate space for new_data\n");
		    exit(1);
		    }

		bzero( new_data, sizeof(struct data_record) );
		new_data->data_type = INTF;
		new_data->dr_type.intf = next_conflict->dir.intf_data;
		write_data( new_data, 1,1 );
		next_conflict = next_conflict->conflicts;
		}
	    }
	}
    else
	{
	ip_con_counter++;
	avl_walk( data->dir.root_next_octet, print_ipaddr_conflicts );
	ip_con_counter--;
	}
    return;
}


int
check_conflicts( tmp_data, orig_data )
struct addr_table * tmp_data;
struct addr_table * orig_data;
{
    struct addr_table * conflict_addr;
    int     founddata = 0;
    struct addr_table * next_conflict;
    struct addr_table * prv_conflict = NULL;

    /* we either have exact duplicates or we have a conflict */

    do {
	int  cc;

	cc = intf_data_cmp( tmp_data->dir.intf_data, orig_data->dir.intf_data );
	if ( cc )
	    {
	    founddata++;
	    break;
	    }

	prv_conflict = tmp_data;
	tmp_data = tmp_data->conflicts;
	} while ( !founddata && tmp_data );

    if ( ! founddata )
	{
	conflict_addr = (struct addr_table *) malloc(sizeof(struct addr_table));
	if ( conflict_addr == NULL )
	    {
	    perror( "js: can't allocate space for conflict_addr\n");
	    exit(1);
	    }

	bzero( conflict_addr, sizeof(struct addr_table) );
	tmp_data = prv_conflict;
	tmp_data->conflicts = conflict_addr;
	conflict_addr->dir.intf_data = orig_data->dir.intf_data;
	return( CONFLICT );
	}
    else
	return( NO_CONFLICT );
}


void
print_data( data )
struct data_record * data;
{
    write_data( data, 1,1 );
}


void
add_entry( eaddr, ipaddr, name_addr, data )
struct addr_table * eaddr;
struct addr_table * ipaddr;
struct name_table * name_addr;
struct data_record * data;
{
    struct addr_table * old_data;
    struct avl_node ** root;

    if ( ipaddr && ! eaddr && data->dr_type.intf->in_mac_addr )
	{
	ipaddr->dir.intf_data->in_mac_addr = data->dr_type.intf->in_mac_addr;
	root = ((struct avl_node **)&data_store[EADDR_STORE]);
	old_data = store_ea_octet( root, data->dr_type.intf, NULL );

	/* free( old_data->dir.intf_data ); */

	old_data->dir.intf_data = ipaddr->dir.intf_data;
	}
    else if (eaddr && !ipaddr && data->dr_type.intf->ip_addr)
	{
	u_char *tmp_addr;

	eaddr->dir.intf_data->ip_addr = data->dr_type.intf->ip_addr;
	root = ((struct avl_node **)&data_store[IN_ADDR_STORE]);
	bcopy( ((struct sockaddr_in *) data->dr_type.intf->ip_addr->pa)->sin_addr,
	    (u_char *) tmp_addr, 4 * sizeof(u_char));
	old_data = store_octet( root, tmp_addr, NULL );

	/* free( old_data->dir.intf_data ); */

	old_data->dir.intf_data = eaddr->dir.intf_data;
	}
    else
    if ( ipaddr )
    {
        old_data = ipaddr;
    }
    else
    {
      old_data = eaddr;
    }
    if ( data->dr_type.intf->submask )
	old_data->dir.intf_data->submask = data->dr_type.intf->submask;

    if ( name_addr )
	{
	old_data->dir.intf_data->hostname = data->dr_type.intf->hostname;
	name_addr->intf_data->hostname = data->dr_type.intf->hostname;
	}
    else
	return;
}

