/*
 * data_server.c
 */

/*
 * finish	Upon receipt of SIGTERM, write data file and quit
 * main		Journal Server main program
 * WriteJF	Dump database to disk file
 */

#include <stdio.h>
#include <signal.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <netdb.h>

#include "data_server.h"

extern	int	ea_counter;
extern	int	ea_con_counter;
extern	int	errno;
extern	int	ip_counter;
extern	int	ip_con_counter;
extern	int	intfcnt;
extern	struct netintf netintf[];
extern	char	pin[];


/* array of char * leading to tops of data lists or trees */

#ifdef AVL
struct list_head *data_store[] = {
#else
char   *data_store[] = {
#endif
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL,
    NULL
    };

int	compare_in_addr();	/* comparison of internet address */
int	name_compare();		/* comparison of internet name */
int	eaddr_compare();	/* comparison of ethernet address */

FILE *	datafile = NULL;
int	debug = 0;		/* Print debug messages if != 0 */
int	init_data;
static	jmp_buf	jmpenv;
char *	jsfn = {TESTDATA};
int	na = 0;			/* No authentication required if != 0 */
u_long  pw;
FILE *	sfp;
char	req[20];
int	writefile;
int     real_sockfd = 0;
int	data_port = DATA_PORT;
int	query_port = QUERY_PORT;
FILE *  ServerLogFile = NULL;

/*
 * finish
 *	On termination signal clean up and quit.
 */

finish()
{
    fprintf( stderr, " js: SIGTERM received.  Quitting.\n" );
    WriteJF();
    if (ServerLogFile)
	fclose(ServerLogFile);
    exit( 0 );
}


main( argc, argv )
int	argc;
char **	argv;
{
    void    init_database();	/* routine to load data base */
    void    list_print();	/* routine to print list -- used for writing
				   data file */
    void    write_file();	/* routine to set writefile flag */
    void    conflict_print();
    void    pipe_probs();

    struct sockaddr_in client_addr;
    int     client_len;
    int     n;
    int     nfds;		/* number of file descriptors available */
    int     socket_store[3];
    struct timeval td;
    fd_set  mask;		/* mask to select what file descriptor to
				   listen on */
    fd_set  tmpmask;		/* temporary mask to store select fields */
    char    *QueryString;	/* for logging purposes */


    /* get the command line arguments if they exist */

    while ( *(++argv) )
	{
	if ( **argv != '-' )
	    {
	    printf( "Use a - before arguments\n" );
	    exit( 1 );
	    }
	switch ( *(*argv + 1) )
	    {
	    case 'd':
		debug = 1;	/* set the debug flag */
		break;

	    case 'f':
		if ( *(++argv) )
		    jsfn = argv[0];	/* set journal file */
		break;

	    case 'n':
		na = 1;		/* no authorization required */
		break;

	    case 'p':
		if ( *(++argv) )
		    {
		    data_port = atoi( *argv );	/* set data and query ports */
		    query_port = data_port + 1;
		    }
		break;
	    }
	}

    if ( ! debug )
	if ( fork() != 0 )
	    exit( 0 );

    GetIntf();				/* Get interfaces on this machine  */

    init_database();			/* Load database from ascii file  */

    signal( SIGINT, write_file );	/* Set "write_file" flag on INT  */
    signal( SIGTERM, finish );		/* Write file on TERM  */
    signal( SIGPIPE, pipe_probs );  

    init_sockets( socket_store );	/* Initialize sockets */

    nfds = getdtablesize();
    if ( nfds > FD_SETSIZE )
	nfds = FD_SETSIZE;

    /* set the mask for the select call to pick the correct socket to connect */

    FD_ZERO( &mask );
    FD_SET( socket_store[DATA_SOCK], &mask );
    FD_SET( socket_store[QUERY_SOCK], &mask );

    if ( debug )
	fprintf( stderr, " js: starting main loop\n" );

   setjmp(jmpenv);
   ea_counter = 1;
   ip_counter = 1;
   ea_con_counter = 1;
   ip_con_counter = 1;

    /* loop forever listening for connections and processing data */
    
    for (;;)
	{
	/* if interrupt came, write data to file */
	
	if ( writefile )
	    WriteJF();

#ifdef DEBUG
	fprintf( stderr, " js: About to select\n" );
#endif

	/* Wait for connect requests  */

	tmpmask = mask;
	td.tv_sec = 1;
	td.tv_usec = 0;

	/* n = select( nfds, &tmpmask, (fd_set *)NULL, (fd_set *)NULL, &td ); */

	n = select( nfds, &tmpmask, (fd_set *)NULL, (fd_set *)NULL, NULL );
	if ( n < 0 )
	    {
	    if ( errno == EINTR )
		continue;

	    perror( "select" );
	    exit( 1 );
	    }
	if ( n == 0 )
	    continue;

	client_len = sizeof( client_addr );
	bzero( &client_addr, sizeof(client_addr) );

	/* see if the data sock has been selected  */

	if ( FD_ISSET( socket_store[DATA_SOCK], &tmpmask ) )
	    {
	    void    store_data();

	    /* create a new socket completing the data connection  */

	    real_sockfd = accept( socket_store[DATA_SOCK],
			(struct sockaddr *) &client_addr, &client_len );
	    if ( real_sockfd < 0 )
		{
		if ( errno == EINTR )
		    continue;
		perror( "accept data" );
		exit( 0 );
		}
	    /* load the data from the network connection */

	    init_data = 0;

	    get_data( real_sockfd, store_data, 1 );

	    close( real_sockfd );
	    real_sockfd = 0;
	    }

	/* see if the query sock has been selected  */

	if ( FD_ISSET( socket_store[QUERY_SOCK], &tmpmask ) )
	    {
	    char   *buf;
	    u_short opcode;
	    extern FILE *datafile;

	    buf = (char *) malloc( BUFSIZE );
	    if ( buf == NULL )
		{
		fprintf( stderr, "js: Error in allocating memory for buf\n" );
		exit( 1 );
		}
	    /* create a new socket completing the query connection */

	    real_sockfd = accept( socket_store[QUERY_SOCK],
			(struct sockaddr *) & client_addr, &client_len );
	    if ( real_sockfd < 0 )
		{
		if ( errno == EINTR )
		    continue;

		perror( "accept query" );
		exit( 0 );
		}

    read_label:
	    n = read( real_sockfd, buf, BUFSIZE );
	    if ( n < 0 )
		{
		if ( errno == EINTR )
		    goto read_label;

		perror( "read:" );
		exit( 1 );
		}
	    opcode = ntohs( *((u_short *) buf) );
	    datafile = fdopen( real_sockfd, "w" );

	    switch ( opcode )
		{
		    void    print_intf_addr();
		    void    print_ether_addr();
		    void    print_ether_conflicts();
		    void    print_ipaddr_conflicts();

		case GET_INTF:
		    list_print( data_store[INTF] );
		    QueryString = "GET_INTF";
		    break;

		case GET_EADDR:
		    avl_walk( data_store[EADDR_STORE], print_ether_addr );
		    QueryString = "GET_EADDR";
		    break;

		case GET_IPADDR:
		    avl_walk( data_store[IN_ADDR_STORE], print_intf_addr );
		    QueryString = "GET_IPADDR";
		    break;

		case GET_NETWORK:
		    list_print( data_store[NETWORK] );
		    QueryString = "GET_NETWORK";
		    break;

		case GET_GW:
		    list_print( data_store[GATEWAY] );
		    QueryString = "GET_GW";
		    break;

		case GET_ALL:
		    list_print( data_store[INTF] );
		    list_print( data_store[NETWORK] );
		    list_print( data_store[GATEWAY] );
		    QueryString = "GET_ALL";
		    break;

		case CONFLICTS_EA:
		    avl_walk( data_store[EADDR_STORE], print_ether_conflicts );
		    QueryString = "CONFLICTS_EA";
		    break;

		case CONFLICTS_IP:
		    avl_walk( data_store[IN_ADDR_STORE],
				print_ipaddr_conflicts );
		    QueryString = "CONFLICTS_IP";
		    break;
		}

	    /* Log the query */
            LogQuery((struct sockaddr *) & client_addr, &client_len,
		     AF_INET, QueryString);

	    fclose( datafile );
	    datafile = NULL;
	    real_sockfd = 0;
	    }			/* end of QUERY if  */
	}				/* end of infinite for loop  */
}				/* end of main  */

#define TBSIZE 30

LogQuery(client_addr, client_len, net_type, QueryString)
struct sockaddr_in *client_addr;
int client_len;
int net_type;
char *QueryString;
{
	struct hostent *gethostbyaddr(), *peer;
        struct timeval tv;
        struct timezone tz;
	struct tm *tm, *localtime();
	FILE *fopen();
	char TimeBuf[TBSIZE];
	char *HName, *inet_ntoa();

	peer = gethostbyaddr(&(client_addr->sin_addr), client_len, net_type);
	if (peer == NULL)  {
		perror("LogQuery: gethostbyaddr");
		return;
	}

	/*
	 * If we didn't get a fully-qualified host name, use IP address instead
	 */
	HName = peer->h_name;
	if (strchr(HName, '.') == NULL)
		HName =
		   inet_ntoa(*((struct in_addr *) &(client_addr->sin_addr)));

	gettimeofday(&tv, &tz);
	tm = localtime(&tv.tv_sec);
	strftime(TimeBuf, TBSIZE, "%a %D, %R", tm);

	/* Leave log file open but line buffered so see updates immediately */
	if (!ServerLogFile) {
		ServerLogFile = fopen(JS_LOGFILE, "a");
		if (ServerLogFile == NULL)  {
			perror("LogQuery: fopen");
			return;
		}
		setlinebuf(ServerLogFile);
	}

	fprintf(ServerLogFile, "%s %s %s\n", HName, QueryString, TimeBuf);
}


/*
 * WriteJF Write data base to journal file
 */

WriteJF()
{
    /* open the data file where data is being written */

    datafile = fopen( jsfn, "w" );
    if ( datafile == NULL )
	{
	fprintf( stderr, "js: Can't open %s\n", jsfn );
	exit( 1 );
	}

    /* use list print to write data to a file  */

    list_print( data_store[INTF] );
    list_print( data_store[NETWORK] );
    list_print( data_store[GATEWAY] );
    fclose( datafile );
    datafile = NULL;
    writefile = 0;

    if ( debug )
	fprintf( stderr, " js: end printing data\n" );

    return;
}
void
pipe_probs()
{
    if( datafile )
	fclose( datafile );
    if ( real_sockfd > 0 )
	close( real_sockfd );
    longjmp( jmpenv, 1 );
}
