/*
** list_insert():
**	inserts a new item 'what' into a list after the element pointed
**	to by 'where'.  if 'where' isn't specified (i.e. ==NULL) then
** 	insert places 'what' at the beginning of the list.
**
** pre: begin is a valid list, where is a pointer to a list element of
**	the list 'begin'  or NULL. 'what' is a valid DATUM.
** post: 'what' has been inserted into the list after the element 'where'.
**	  returns NULL if insert fails, otherwise returns a pointer to 
**	  new node.
**  @(#)list_insert.c	1.3  10/5/91
*/

#
#include <stdio.h>
#include "defs.h"
#include "data.h"
#include "list.h"
#
extern char *malloc();


char * list_insert(begin, where, what)
struct list_head	 *begin;
struct list_node *where;
char	 *what;
{
	struct list_node *temp; /* new node for the list */

	/* create a new node for the list */
	temp = (struct list_node *)malloc(sizeof(struct list_node));

	/* check for error from malloc */

	if(temp == NULL)
	{
		return(NULL);
	}



	temp->datum = what;
	/* increment the length of the list */
	begin->length++;

	/* Condition 1: Place node at the beginning of List */

	if(where == NULL)
	{
		/* setup the pointers to add the first item in the list */
		temp->succ = begin->list;
		temp->pred = NULL;
		begin->list = temp;

		/* consider if the list exist already, then pointers
		to temp from the old first item needs to point back to the
		new first item */

		if(temp->succ != NULL)
		{
			temp->succ->pred = temp;
			return((char *)temp);
		}

		/* list item must be the first on the list so change the
		last option on the header */

		begin->last = temp;


	}

	/* insert a node in the midst of the list */
	else
	{
		/* setup the pointers to add the next node to either the
		end of or in the middle of the list */

		temp->succ = where->succ;
		temp->pred = where;
		where->succ = temp;

		/* if the node is placed in the middle of the list,
		get the following node to point to it */
		if (temp->succ != NULL)
		{
			temp->succ->pred = temp;
			return((char *)temp);
		}


		/* list node is at the end of the list so set last pointer */
		begin->last = temp;

	}

	return((char *)temp);
} 
