/*
 *  serv_cli.c
 */

/*
 *	fill_data( databuf ,check_pin)
 *	write_data( data, follow_links, skip_pin )
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <time.h>

#include "data_server.h"

#ifndef PIN
#include "pin.h"
#endif
char	pin[20] = PIN;

extern struct intf_prot_addr * get_ipaddr();
extern struct ether_addr * ether_aton();
extern char * getkeyword();
extern char * getword();
extern char * index();

/* %W% %G% */

/*
 * fill_data
 *	Convert ASCII text into a data_record structure
 *	    Returns a pointer to the data_record
 *	    Returns a NULL pointer on error
 */

struct data_record *
fill_data( databuf ,check_pin)
char  * databuf;		/* data buffer with ascii text */
int check_pin;
{
    struct data_record * data;	/* struct to place data */
    char  * getline();		/* get a new line of data */
    char  * data_line;		/* space for line of data */
    char  * data_type;		/* type of data recieved */
    char *type;
    char ** value;

    check_pin = 0;

    /* create a data record for the new data */

    data = (struct data_record *)malloc(sizeof(struct data_record));
    if ( data == NULL )
	{
	printf("Memory for data not available\n");
	exit(1);
	}

    /* clear out the data structure */

    bzero( data, sizeof(struct data_record) );

    value = (char **)malloc(sizeof(char *));
    if ( value == NULL )
	{
	printf("Unable to allocate memory for value\n");
	exit(1);
	}

    /* read in a line of data and put the data into the proper element in the
       data structure */

    type = getline( databuf );
    data_type = getword( &type );
    databuf += strlen( data_type );
    if ( check_pin )
	if ( strncmp( getword( &type ), pin, strlen( pin ) ) )
	    return( NULL );

    if ( ! strncmp( data_type, "INTF", strlen("INTF") ) )
	{
	data->data_type = INTF;
	data->dr_type.intf = (struct intf *)malloc( sizeof(struct intf) );
	if ( data->dr_type.intf == NULL )
	    {
	    printf( "Memory for intf not available\n" );
	    exit( 1 );
	    }

	bzero( data->dr_type.intf, sizeof(struct intf) );

	while ( ( data_line = getline(databuf) ) != NULL )
	    {
	    char  * keyword;

	    databuf += strlen( data_line );
	    keyword = getkeyword( data_line, value );

	    /* Using keyword, find the value */

	    if ( !strcasecmp( keyword, "ETHER-ADDR" ) )
		{
		struct intf_mac_addr * ether_address;
		struct ether_addr * tmp_addr;	/* ethernet address */

		ether_address = (struct intf_mac_addr *)
					malloc(sizeof(struct intf_mac_addr));
		if ( ether_address == NULL )
		    {
		    printf("Can't allocate space for intf_mac_addr\n");
		    exit(1);
		    }
		bzero( ether_address, sizeof(struct intf_mac_addr) );
		ether_address->mac_addr = (struct ether_addr *)
					malloc(sizeof(struct ether_addr));
		if ( ether_address->mac_addr == NULL )
		    {
		    printf( "Can't allocate space for mac_addr\n" );
		    exit(0);
		    }

		/* convert ascii address to machine for ethernet address.
		   copy into the appropriate data structure */

		tmp_addr = ether_aton( getword( value ) );
		bcopy( tmp_addr, ether_address->mac_addr, 6 * sizeof(u_char) );
		data->dr_type.intf->in_mac_addr = ether_address;
		data->dr_type.intf->in_mac_addr->disc_time = 0;
		}

	    /* get the protocol layer address. Currently, we only support IP */
		
	    else
	    if ( ! strcasecmp( keyword, "MAC_TIME_UPDATE") &&
					(data->dr_type.intf->in_mac_addr))
		{
		/* struct tm time_of_disc;	       /* discovery date */

		/* bzero( &time_of_disc, sizeof(struct tm) ); */
      
		/* copy time from the string to the time_of_disc data struct */

		/* strptime( *value, " %A %h %e %T %Y\n", &time_of_disc ); */

		/* store time as a long seconds from sometime way back */

		/* data->dr_type.intf->in_mac_addr->disc_time =
						timegm( &time_of_disc ); */

		sscanf( *value, "%d",
			&data->dr_type.intf->in_mac_addr->disc_time );
		}
	    else
	    if ( !strcasecmp( keyword, "PROT-ADDR" ) )
		{
		struct intf_prot_addr * interface_addr;

		interface_addr = get_ipaddr( value );
		data->dr_type.intf->ip_addr = interface_addr;
		}

	    else
	    if ( ! strcasecmp( keyword, "IP_TIME_UPDATE" ) &&
					( data->dr_type.intf->ip_addr ) )
		{
		/* struct tm time_of_disc;	       /* discovery date */
      
		/* copy time from the string to the time_of_disc data struct */

		/* strptime(*value," %A %h %e %T %Y\n",&time_of_disc); */

		/* store time as a long seconds from sometime way back */

		/* data->dr_type.intf->ip_addr->disc_time
						= timegm(&time_of_disc); */

		sscanf( *value, "%d",
			&data->dr_type.intf->ip_addr->disc_time );
		}

	    /* put host name in the data record */

	    else
	    if ( ! strcasecmp( keyword, "IN-NAME" ) )
		{
		char  * str_name;	/* name of host */
		char  * local_name;
		char  * np;
		char  * domain_name = NULL;
		struct intf_name * hostname;

		/* get the name of the host */

		str_name = getword( value );

		hostname = (struct intf_name *)malloc(sizeof(struct intf_name));
		if ( hostname == NULL )
		    {
		    printf("hostname not allocated \n");
		    exit(0);
		    }

		bzero( hostname, sizeof(struct intf_name) );
		np =  index( str_name, '.' );
		if ( np != NULL )
		    {
		    *np = '\0';
		    local_name = (char *)malloc( np - str_name + 1 );
		    if ( local_name == NULL )
			{
			printf( "Can't allocate space for local_name\n" );
			exit(1);
			}

		    strcpy( local_name, str_name );
		    np++;
		    domain_name = (char *)malloc( strlen( np ) + 1 );
		    if ( domain_name == NULL )
			{
			printf( "domain_name not allocated\n" );
			exit(1);
			}

		    strcpy( domain_name, np );
/*
 *		    strncpy( domain_name, np, str_name + strlen(str_name) - np);
 *		*(domain_name + (str_name + strlen(str_name) - np + 1)) = '\0'; 
 */
		    }
		else
		    {
		    local_name = (char *)malloc( strlen(str_name) + 1 );
		    if ( local_name == NULL )
			{
			printf("Can't allocate space for local_name\n");
			exit(0);
			}

		    strcpy(local_name, str_name);
		    }

		hostname->local_name = local_name;
		if ( domain_name )
		    hostname->domain = domain_name;

		data->dr_type.intf->hostname = hostname;
		}

	    /* subnet mask  */

	    else
	    if ( ! strcasecmp( keyword, "SUBNET-MASK" ) )
		{
		struct intf_prot_addr * interface_addr;

		interface_addr = (struct intf_prot_addr *)
					malloc(sizeof(struct intf_prot_addr));
		if ( interface_addr == NULL)
		    {
		    printf( "can't allocate space for intf_addr\n" );
		    exit(1);
		    }
		bzero( interface_addr,sizeof(struct intf_prot_addr) );

		/* Determine address family for subnet mask */
		/*  note IP currently supported */

		switch ( (interface_addr->addr_family = getfamily( value ) ) )
		    {
		    struct sockaddr_in * inetaddr;	/* inet addr scheme for
						   IP address */

		case AF_INET:

		    inetaddr = (struct sockaddr_in *)
					malloc(sizeof(struct sockaddr_in));
		    if ( inetaddr == NULL )
			{
			printf("allocation for inetaddr failed\n");
			exit(1);
			}

		    /* for inet type addresses,  */
		    /*  check to see that address is a IP address  */
		    /*  and copy into appropriate data struct */

		    bzero( inetaddr, sizeof(struct sockaddr_in) );
		    inetaddr->sin_family = AF_INET;
		    inetaddr->sin_addr.s_addr = inet_addr(getword(value));
		    if ( inetaddr->sin_addr.s_addr == -1 )
			break;

		    interface_addr->pa = (struct sockaddr *) inetaddr;
		    data->dr_type.intf->submask = interface_addr;
		    }
		}
	    }
	}
    else
    if ( ! strncmp( data_type, "NETWORK", strlen("NETWORK") ) )
	{
	data->data_type = NETWORK;
	data->dr_type.subnet = (struct network *)
					malloc(sizeof(struct network));
	if ( data->dr_type.subnet == NULL )
	    {
	    printf("Memory for.subnet not available\n");
	    exit(1);
	    }

	bzero( data->dr_type.subnet, sizeof(struct network) );

	while ((data_line = getline(databuf)) != NULL)
	    {
	    char   *keyword;

	    databuf += strlen(data_line);
	    keyword = (char *)getkeyword(data_line, value);
	    if ( !strcasecmp( keyword, "net_numb" ) )
		{
		struct intf_prot_addr *net_number;

		net_number = get_ipaddr(value);
		data->dr_type.subnet->net_numb = net_number;
		}
	    else if ( !strcasecmp( keyword, "subnet_num" ) )
		{
		struct intf_prot_addr *subnet_numb;

		subnet_numb = get_ipaddr( value );
		data->dr_type.subnet->subnet = subnet_numb;
		}
	    else if ( !strcasecmp( keyword, "gateway" ) )
		{
		struct gw_pointer **next_gw;
		struct gw_pointer *gateway;
		struct intf_prot_addr *gateway_addr;
		struct intf_list *gw_intf;

		gateway_addr = get_ipaddr( value );
		gateway = (struct gw_pointer *)
					malloc(sizeof(struct gw_pointer));
		if ( gateway == NULL )
		    {
		    printf("unable to allocate memory for gateway\n");
		    exit(1);
		    }

		bzero( gateway,sizeof(struct gw_pointer) );

		gateway->gw_host = (struct host *)malloc(sizeof(struct host));
		if ( gateway->gw_host == NULL )
		    {
		    printf("Unable to allocate memory for gateway\n");
		    exit(1);
		    }
		bzero(gateway->gw_host,sizeof(struct host));
		next_gw = &data->dr_type.subnet->gwp_start;

		while ( *next_gw )
		    next_gw = &(*next_gw)->gw_next;

		*next_gw = gateway;
		gw_intf = (struct intf_list *)malloc(sizeof(struct intf_list));
		if ( gw_intf == NULL )
		    {
		    printf(" Unable to allocate memory for gw_intf\n");
		    exit(1);
		    }

		bzero( gw_intf, sizeof(struct intf_list) );

		gw_intf->subnet_intf = (struct intf *)
					malloc(sizeof(struct intf));
		if ( gw_intf->subnet_intf == NULL )
		    {
		    printf(" Unable to allocate memory for subnet_intf\n");
		    exit(1);
		    }
		bzero(gw_intf->subnet_intf,sizeof(struct intf));
		gw_intf->subnet_intf->ip_addr = gateway_addr;
		gateway->gw_host->intfs = gw_intf;
		}
	    }
	}

    else if ( !strncmp( data_type, "GATEWAY", strlen("GATEWAY") ) )
	{
	data->data_type = GATEWAY;
	data->dr_type.gateway = (struct host *)malloc(sizeof(struct host));
	if ( data->dr_type.gateway == NULL )
	    {
	    printf("Error in allocating memory for gateway\n");
	    exit(1);
	    }

	bzero( data->dr_type.gateway, sizeof(struct host) );
	while ( ( data_line = getline( databuf ) ) != NULL )
	    {
	    char   *keyword;

	    databuf += strlen(data_line);
	    keyword = (char *)getkeyword( data_line, value );
	    if ( !strcasecmp( keyword, "intf" ) )
		{
		struct intf_prot_addr *ip_prot_addr;
		struct intf_list **current_addr;
		struct intf_list *new_list;

		ip_prot_addr = get_ipaddr(value);
		current_addr = &data->dr_type.gateway->intfs;
		while ( *current_addr )
		    {
		    current_addr = &(*current_addr)->next_intf;
		    }
		new_list = (struct intf_list *)
					malloc(sizeof(struct intf_list));
		if ( new_list == NULL )
		    {
		    printf("error in creating memory for new_list\n");
		    exit(1);
		    }

		bzero(new_list,sizeof(struct intf_list));

		new_list->subnet_intf = (struct intf *)
					malloc(sizeof(struct intf));
		if ( new_list->subnet_intf == NULL )
		    {
		    printf("Unable to allocate memory for subnet_intf\n");
		    exit(1);
		    }
		bzero(new_list->subnet_intf,sizeof(struct intf));
		new_list->subnet_intf->ip_addr = ip_prot_addr;
		*current_addr = new_list;
		}
	    else if ( !strcasecmp( keyword, "subnets" ) )
		{
		struct subnet *new_subnet;
		struct intf_prot_addr *next_numb;
		struct intf_prot_addr *new_ip_addr;
		struct subnet **next_subnet;

		next_numb = get_ipaddr(value);

		next_subnet = &data->dr_type.gateway->connected_subnets;
		while ( *next_subnet )
		    next_subnet = &(*next_subnet)->next_subnet;

		new_subnet = (struct subnet *)malloc(sizeof(struct subnet));
		if ( new_subnet == NULL )
		    {
		    printf("new_subnet not  created \n");
		    exit(1);
		    }

		bzero( new_subnet, sizeof(struct subnet) );

		new_subnet->subnet_info = (struct network *)
					malloc(sizeof(struct network));
		if ( new_subnet->subnet_info == NULL )
		    {
		    printf("unable to allocate space for new_subnet\n");
		    exit(1);
		    }

		bzero( new_subnet->subnet_info, sizeof(struct network) );

		new_subnet->subnet_info->net_numb = next_numb;
		new_ip_addr = get_ipaddr( value );
		if ( new_ip_addr != NULL )
		    {
		    new_subnet->net_intf = (struct intf *)
						malloc(sizeof(struct intf));
		    if ( new_subnet->net_intf == NULL )
			{
			printf("Error in allocating space for new_subnet\n");
			exit(1);
			}
		    bzero(new_subnet->net_intf,sizeof(struct intf));
		    new_subnet->net_intf->ip_addr = new_ip_addr;
		    }

		*next_subnet = new_subnet;
		}
	    }
	/* return newly created set of data */
	}
    else 
	return(NULL);

    return( data );
}


/*
 *  write_data
 *	Convert a data_record into ASCII text
 */

void
write_data( data, follow_links, skip_pin )
struct data_record *data;
int     follow_links;
int     skip_pin;
{
    extern FILE * datafile;
    extern char * addr_family[];

    switch ( data->data_type )
	{
	struct intf *working_intf;
	struct network *working_subnet;
	struct gw_pointer *gateways;
	struct gw_pointer *gwp;
	struct intf_list *gw_intfs;
	struct subnet *gw_subnets;
	struct host *working_gw;

    case INTF:
	working_intf = data->dr_type.intf;
	if( skip_pin )
	    fprintf( datafile, "INTF \n");
	else
	    fprintf( datafile, "INTF  %s\n", pin );
	if ( working_intf->in_mac_addr )
	    {
	    fprintf( datafile,
		"ETHER-ADDR:\t%02x:%02x:%02x:%02x:%02x:%02x\n",
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[0],
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[1],
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[2],
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[3],
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[4],
		working_intf->in_mac_addr->mac_addr->ether_addr_octet[5] );
	if ( working_intf->in_mac_addr->disc_time )
	    fprintf( datafile, "MAC_TIME_UPDATE:  %d  %s",
		working_intf->in_mac_addr->disc_time,
		asctime( localtime( &working_intf->in_mac_addr->disc_time ) ) );
	    }	
/*
 *		fprintf( datafile, "ETHER-ADDR:   %s\n",
 *			    ether_ntoa(working_intf->in_mac_addr->mac_addr));
 */
	if ( working_intf->ip_addr )
	    {
	    struct sockaddr_in inet_print;

	    switch ( working_intf->ip_addr->addr_family )
		{
	    case AF_INET:
		bcopy( working_intf->ip_addr->pa, &inet_print,
						sizeof(struct sockaddr_in));
		fprintf( datafile, "PROT-ADDR:  %s %s\n",
				addr_family[working_intf->ip_addr->addr_family],
				inet_ntoa( inet_print.sin_addr ) );
		}
	    if ( working_intf->ip_addr->disc_time )
		fprintf( datafile, "IP_TIME_UPDATE:  %d  %s",
			working_intf->ip_addr->disc_time,
			asctime( localtime( &working_intf->ip_addr->disc_time )));
	    }
	if ( working_intf->hostname )
	    {
	    char * hostname;

	    if ( working_intf->hostname->local_name &&
					working_intf->hostname->domain )
		{
		hostname = (char *)malloc(
				strlen(working_intf->hostname->local_name) +
				strlen(working_intf->hostname->domain)     + 2);
		if ( hostname == NULL )
		    {
		    printf("Unable to allocate memory for hostname\n");
		    exit(1);
		    }

		sprintf( hostname, "%s.%s",
					working_intf->hostname->local_name,
					working_intf->hostname->domain );
		}
	    else
		if ( working_intf->hostname->local_name )
		    {
		    hostname = (char *)malloc(
			    strlen( working_intf->hostname->local_name ) + 1 );
		    if ( hostname == NULL )
			{
			printf( "Unable to allocate memory for hostname\n" );
			exit( 1 );
			}

		    sprintf(hostname, "%s", working_intf->hostname->local_name);
		    }
		fprintf( datafile, "IN-NAME:\t%s\n", hostname );
	    }

	if ( working_intf->submask )
	    {
	    struct sockaddr_in sub_print;

	    switch (working_intf->submask->addr_family)
		{
	    case AF_INET:
		bcopy( working_intf->submask->pa, &sub_print,
						sizeof(struct sockaddr_in));
		fprintf( datafile, "SUBNET-MASK:  %s %s\n",
				addr_family[working_intf->submask->addr_family],
				inet_ntoa( sub_print.sin_addr ) );
		}
	    }
	break;

    case NETWORK:
	working_subnet = data->dr_type.subnet;
	if(skip_pin)
	    fprintf( datafile, "NETWORK \n");
	else
	    fprintf( datafile, "NETWORK  %s\n", pin );
	if ( working_subnet->net_numb )
	    {
	    struct sockaddr_in net_print;

	    switch ( working_subnet->net_numb->addr_family )
		{
	    case AF_INET:
		bcopy( working_subnet->net_numb->pa, &net_print,
						sizeof(struct sockaddr_in));
		fprintf( datafile, "NET_NUMB:  %s %s\n",
			    addr_family[working_subnet->net_numb->addr_family],
			    inet_ntoa( net_print.sin_addr ) );
		}
	    }
	if ( working_subnet->subnet )
	    {
	    struct sockaddr_in sub_print;

	    switch ( working_subnet->subnet->addr_family )
		{
	    case AF_INET:
		bcopy( working_subnet->subnet->pa, &sub_print,
						sizeof(struct sockaddr_in));
		fprintf( datafile, "SUBNET_NUM:  %s %s\n",
			    addr_family[working_subnet->subnet->addr_family],
			    inet_ntoa( sub_print.sin_addr ) );
		}
	    }
	gwp = working_subnet->gwp_start;
	while (gwp)
	    {
	    struct subnet *next_subnet;

	    if ( follow_links )
		{
		next_subnet = gwp->gw_host->connected_subnets;
		while ( next_subnet )
		    {
		    if ( bcmp( ((struct sockaddr_in *)
			    next_subnet->subnet_info->net_numb->pa)->sin_addr,
			    ((struct sockaddr_in *)
			    working_subnet->net_numb->pa)->sin_addr,
			    4 * sizeof(u_char)) == 0)
			{
			struct sockaddr_in sub_intf;

			if (!next_subnet->net_intf)
			    {
			    struct sockaddr_in sub_intf;

			    switch ( gwp->gw_host->intfs->subnet_intf->ip_addr->
								    addr_family)
				{
			    case AF_INET:
				bcopy( (struct sockaddr_in *)
					gwp->gw_host->intfs->
					subnet_intf->ip_addr->pa,
					&sub_intf, sizeof(struct sockaddr_in));
				fprintf( datafile, "GATEWAY: %s %s\n",
					addr_family[gwp->gw_host->intfs->
					subnet_intf->ip_addr->addr_family],
					inet_ntoa( sub_intf.sin_addr ) );
				}
			    }
			else
			    switch ( next_subnet->net_intf->ip_addr->
					addr_family)
				{
			    case AF_INET:
				bcopy( (struct sockaddr_in *)
					next_subnet->net_intf->ip_addr->pa,
					&sub_intf,
					sizeof(struct sockaddr_in));
				fprintf( datafile, "GATEWAY:  %s %s\n",
					addr_family[next_subnet->
					net_intf->
					ip_addr->
					addr_family],
					inet_ntoa( sub_intf.sin_addr )
					);
				}
			}
		    next_subnet = next_subnet->next_subnet;
		    }
		}
	    else
		{
		struct sockaddr_in sub_intf;

		switch ( gwp->gw_host->intfs->subnet_intf->ip_addr->
		    addr_family)
		    {
		case AF_INET:
		    bcopy( (struct sockaddr_in *) gwp->gw_host->intfs->
			subnet_intf->ip_addr->pa,
			&sub_intf, sizeof(struct sockaddr_in));
		    fprintf( datafile, "GATEWAY: %s %s\n",
			addr_family[gwp->gw_host->intfs->
			subnet_intf->ip_addr->addr_family],
			inet_ntoa( sub_intf.sin_addr ) );
		    }
		}
	    gwp = gwp->gw_next;
	    }
	break;

    case GATEWAY:
	working_gw = data->dr_type.gateway;
	if(skip_pin)
	    fprintf( datafile, "GATEWAY\n");
	else
	    fprintf( datafile, "GATEWAY  %s\n", pin );

	gw_intfs = working_gw->intfs;
	while ( gw_intfs )
	    {
	    if ( gw_intfs->subnet_intf )
		{
		struct sockaddr_in gw_intf;

		switch ( gw_intfs->subnet_intf->ip_addr->addr_family )
		    {
		case AF_INET:
		    bcopy( gw_intfs->subnet_intf->ip_addr->pa,
			&gw_intf, sizeof(struct sockaddr_in));
		    fprintf( datafile, "INTF:  %s %s\n",
			addr_family[gw_intfs->subnet_intf->
			    ip_addr->addr_family],
			inet_ntoa( gw_intf.sin_addr ) );
		    }
		}
	    gw_intfs = gw_intfs->next_intf;
	    }
	gw_subnets = working_gw->connected_subnets;

	while ( gw_subnets )
	    {
	    if ( gw_subnets->subnet_info )
		{
		struct sockaddr_in gw_subnet_no;

		switch ( gw_subnets->subnet_info->net_numb->addr_family)
		    {
		case AF_INET:
		    bcopy( gw_subnets->subnet_info->net_numb->pa,
			&gw_subnet_no, sizeof(struct sockaddr_in));
		    fprintf( datafile, "SUBNETS:  %s %s ",
			addr_family[gw_subnets->subnet_info->
			    net_numb->addr_family],
			inet_ntoa( gw_subnet_no.sin_addr ) );
		    }
		}
	    if ( gw_subnets->net_intf )
		{
		struct sockaddr_in gw_sub_intf;

		switch (gw_subnets->net_intf->ip_addr->addr_family)
		    {
		case AF_INET:
		    bcopy( gw_subnets->net_intf->ip_addr->pa,
			&gw_sub_intf, sizeof(struct sockaddr_in));
		    fprintf( datafile, " %s %s ",
			addr_family[gw_subnets->net_intf->
			    ip_addr->addr_family],
			inet_ntoa( gw_sub_intf.sin_addr ) );
		    }
		}
	    fprintf( datafile, "\n" );
	    gw_subnets = gw_subnets->next_subnet;
	    }
	}
    fprintf( datafile, "\n\n" );
    return;
}

