/*
 *  util.c
 */

/*
 *	getfamily( string )
 *	getword( string )
 *	get_record( buf, data, leftover, bytes_read, buf_size )
 *	getline( buf )
 *	intf_data_cmp( data1, data2 )
 *	getnetconnect( data_host, port )
 *	getaddr( host, port )
 *	get_data( file_des, data_func ,authenticate)
 *	getkeyword( databuf, value )
 *	get_ipaddr( value )
 *	make_data( type )
 *	make_intf()
 *	make_subnet()
 *	make_gateway()
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <strings.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netdb.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>

#include "data_server.h"
#include "send_data.h"

/* %W% %G% */

#define BUFSIZE 80

char   *addr_family[] = {
    "AF_UNSPEC",
    "AF_UNIX",
    "AF_INET",
    "AF_IMPLINK",
    "AF_PUP",
    "AF_CHAOS",
    "AF_NS",
    "AF_NBS",
    "AF_ECMA",
    "AF_DATAKIT",
    "AF_CCITT",
    "AF_SNA",
    "AF_DECnet",
    "AF_DLI",
    "AF_LAT",
    "AF_HYLINK",
    "AF_APPLETALK",
    "AF_NIT",
    "AF_802",
    "AF_OSI",
    "AF_X25",
    "AF_OSINET",
    "AF_GOSIP",
    };


/*
 * using an array with all possible address families, determine
 * address family value. Passed the address family and returns the number
 * cooresponding to the address family.
 */

int 
getfamily( string )
char  **string;
{
    char   *family;		/* address family */
    int     x;			/* counter into address family array */
    char   *getword();		/* function to get next word  */

    /* get the name of family desired */

    family = getword( string );

    /* find the cooresponding integer value for the family */

    for ( x = 0; x < MAX_FAMILY; x++ )

	/* address family found if strings compare */

	if ( strcmp( family, addr_family[x] ) == 0 )
	    return( x );

    /* address family not found, return bad result */

    return( -1 );
}


/*
 * getword parses a line and returns individual lines. The rest of the
 * line is returned in string. The word is returned.
 */

char *
getword( string )
char  **string;		/* orig word passed in and new word passed out */
{
    char * end_of_word;	/* pointer to end of words -- moves through word */
    char * start_of_word;	/* fixed at start of word  */
    char * returned_word;	/* storage for where word is passed back */

    /* set start_of_word to the start of the string */

    start_of_word = *string;

    /* skip over blanks to the start of the word */

    while ( isspace( *start_of_word ) )
	start_of_word++;

    /* set end_of_word to start_of_word so we can parse address and find the
     end of word */

    end_of_word = start_of_word;

    /* if the letter is alphanumeric or certain types of punctuation, letter
     is part of current word */

    while ( isalnum(*end_of_word) ||
		*end_of_word == '-' ||
		*end_of_word == '.' ||
		*end_of_word == '_' ||
		*end_of_word == ':' )
	end_of_word++;

    returned_word = (char *)malloc( end_of_word - start_of_word + 2 );
    if ( returned_word == NULL )
	{
	printf( "Memory unavailable for returned_word\n" );
	exit( 1 );
	}

    /* copy word into space to return the word */

    strncpy( returned_word, start_of_word, end_of_word - start_of_word );
    *( returned_word + (end_of_word - start_of_word) ) = '\0';

    /* set the word_p pointer to space after next word */

    *string = end_of_word;

    return( returned_word );
}


/*
 * get_record retrieves records of data from the ascii stored file. A
 * record is noted to be a set of lines seperated by two newlines. If
 * the buffer is read and the entire record not read, the incomplete
 * record is returned and more data retrieved to fill out the record.
 */

int 
get_record( buf, data, leftover, bytes_read, buf_size )
char   *buf;			/* buffer where data was read */
char  **data;			/* storage spot to return record */
char  **leftover;		/* storage for leftover -- this could be
		   static */
int     bytes_read;		/* total number of bytes read into the
		   current buffer */
int     buf_size;		/* total size of buffer */
{
    char   *buf_ptr = buf;	/* pointer into the buf */
    int     size_of_record = 0;	/* total size of the individual record */
    static int     end_on_newline = 0;
    static int     short_on_newline = 0;
    for ( ;; )
	{
	int     total_read;	/* total number of bytes processed so far */

	/* find the end of the line in one element of the buffer.
	   If we reach the end of the buffer, stop moving buffer pointer */

	while ( (bytes_read + (buf_ptr - buf) + 1 < (buf_size)) &&
							(*buf_ptr != '\n') )
	    buf_ptr++;

	/* We have either reached the end of a record or we have reached the
           end of the buffer. If the later, store it for a future read. */

	total_read = bytes_read + ( buf_ptr - buf );
	if ( ( total_read == buf_size - 1 ) ||
	    ( ( total_read == buf_size -2 ) && ( *(buf_ptr + 1) == '\n' ) ) ||
	    ( end_on_newline && (*(buf_ptr + 1) == '\n')) ||
	    ( short_on_newline && *buf_ptr == '\n')||
	    (*(buf_ptr + 1) == '\n') &&
	    (*(buf_ptr + 2) == '\n'))
	    {
	    /* get the size of the data record */

	    size_of_record = buf_ptr - buf + 1;
	    short_on_newline = 0;
	    end_on_newline = 0;

	    /* if the following is true, we are at the end of the current
		buffer and and we need to store the data as leftover. This is
		guranteed because if we were at the end of a record, we would
		actually be two memory locations before the end of the buffer */

	    if ( total_read >= buf_size - 2 )
		{
		char   *leftover_tmp;	/* storage to build leftover */
		int     prev_left_length = 0;	/* amount of previous left
									data */

		/* In this situation, we have gone through the entire buffer
		   twice and still haven't found the end of the record */

		if(*buf_ptr == '\n')
		    if(total_read == buf_size -2)
			short_on_newline = 1;
		    else
			end_on_newline = 1;

		if ( *leftover )

		    /* get the length of the previous leftover */

		    prev_left_length = strlen( *leftover );

		/* create space for leftover data including the old and the
		    new */

		leftover_tmp = malloc( size_of_record + prev_left_length + 1 );
		if ( leftover_tmp == NULL )
		    {
		    printf( "Can't allocate memory for leftover_tmp\n" );
		    exit( 1 );
		    }

		/* copy in the leftover data into the tmp, copy in the new
		    data and then free the old leftover */

		if ( *leftover )
		    {
		    strcpy( leftover_tmp, *leftover );
		    strncpy( leftover_tmp + prev_left_length, buf,
			size_of_record );
		    /* free( *leftover ); */
		    }

		/* we have no previous leftovers so just copy in the existing
		    leftover */

		else
		    strncpy( leftover_tmp, buf, size_of_record );

		/* null the end of the buffer */

		*( leftover_tmp + prev_left_length + size_of_record ) = '\0';

		/* put the tmp into the leftover to be returned to the
		    calling function */

		*leftover = leftover_tmp;

		/* We have no data so set it to NULL */

		*data = NULL;
		return( buf_size );
		}

	    /* if we didn't finish filling the record when we read the last
		time as indicated by the fact that we have leftovers, copy
		the leftover first into the data and then the new data */

	    if ( *leftover )
		{
		/* allocate space for data, making sure to incorporate both
		   the leftover and the new data */

		*data = malloc( size_of_record + strlen( *leftover ) );
		if ( *data == NULL )
		    {
		    printf( "Can't allocate memory for data\n" );
		    exit( 1 );
		    }

		/* copy in the leftover */

		strcpy( *data, *leftover );

		/* copy in the remaining data  */

		strncpy( *data + strlen( *data ), buf, size_of_record );

		/* no leftover left, set to null */

		*leftover = NULL;
		return( size_of_record );
		}
	    else
		{
		/* Create space where we put the data record to return to the
		   calling function */

		*data = malloc( size_of_record );
		if ( *data == NULL )
		    {
		    printf( "Can't allocate memory for *data\n" );
		    exit( 1 );
		    }

		/* copy the data record into memory to be returned to the
		   calling function */

		strncpy( *data, buf, size_of_record );

		/* the size of the data record is returned */

		return( size_of_record );
		}
	    }
	buf_ptr++;
	}
}


/* get_line returns a line from the buf */

char   *
getline( buf )
char   *buf;			/* buffer with data in it */
{
    char   *line;		/* line contains one line of data */
    char   *end;		/* The point right before a newline */
    int     strlength;		/* length of line */

    /* new line may be absent because buffer is not full */

    end = index( buf, '\n' );
    if ( end == NULL )
	return( NULL );

    /* allocate space for a line of data */

    strlength = end - buf + 1;
    line = malloc( strlength + 1 );
    if ( line == NULL )
	{
	printf( "can't allocate memory for line\n" );
	exit( 1 );
	}

    /* put the line into line  and then put a Null at the end */

    strncpy( line, buf, strlength );
    *( line + ( strlength ) ) = '\0';
    return( line );
}


intf_data_cmp( data1, data2 )
struct intf *data1;
struct intf *data2;
{
    if ( ( data1->ip_addr ) &&
	( data2->ip_addr ) &&
	( data1->ip_addr->pa && data2->ip_addr->pa ) &&
	bcmp( data1->ip_addr->pa,
	data2->ip_addr->pa,
	sizeof(struct sockaddr) ) )
	return( 0 );

    if ( ( data1->in_mac_addr ) &&
	( data2->in_mac_addr ) &&
	( data1->in_mac_addr->mac_addr && data2->in_mac_addr->mac_addr ) &&
	bcmp( data1->in_mac_addr->mac_addr,
	data2->in_mac_addr->mac_addr,
	sizeof(struct ether_addr) ) )
	return( 0 );

    return( 1 );
}



getnetconnect( data_host, port )
char   *data_host;
int     port;
{
    int     cc;
    char  * jshenv;
    char    hostname[MAXHOSTNAMELEN];
    struct sockaddr_in * server_addr = NULL;
    struct sockaddr_in * getaddr();
    FILE  * dataserver_file;
    int     datafd = -1;

    if ( port == 0 )
	port = DATA_PORT;

    if ( data_host )
	server_addr = getaddr( data_host, port );

	/* get the hostname from the environment variable DATASERVER_HOST */

    else if ( ( jshenv = getenv( "JS_HOST" ) ) != NULL )
	{
	if ( ( server_addr = getaddr( jshenv, port ) ) == NULL )
	    {
	    printf( "Unable to get correct server\n" );
	    exit( 1 );
	    }
	}

    /* get the address from a data server file */

    else
	{
	char    buf[BUFSIZE];
	char   *buf_ptr;
	char   *buf_ptr_end;

	buf_ptr = buf;

	/* Open the data server file where the data server is listed */

	dataserver_file = fopen( DATASERVER_FILE, "r" );
	if ( dataserver_file != NULL )
	    {
	    /* Read the contents of data server file and put in the buffer */

	    fgets( buf, BUFSIZE, dataserver_file );

	    /* key word dataserver precedes the data server
     hostname or IP address */

	    if ( strncasecmp( buf, "dataserver", strlen("dataserver") ) != 0 )
		{
		printf( "dataserver file format wrong\n" );
		exit( 1 );
		}

	    /* Get the hostname for the data server */

	    sscanf( buf, "dataserver %s", hostname );
	    /*
       *	    buf_ptr += strlen( "dataserver" );
       *	    while ( isspace(*buf_ptr) )
       *		buf_ptr++;
       *	    buf_ptr_end = buf_ptr;
       *	    while ( !isspace(*buf_ptr_end) )
       *		buf_ptr_end++;
       */
	    }
	else
	    {
	    int     cc;

	    cc = gethostname( hostname, MAXHOSTNAMELEN );
	    if ( cc != 0 )
		{
		printf( "Could not get local hostname\n" );
		exit( 1 );
		}
	    }

	/* found the Host name, so stick it away and
       find the IP address for the host */

	/*
     *	    hostname = malloc( buf_ptr_end - buf_ptr + 1 );
     *	    if ( hostname == NULL )
     *		{
     *		printf( "Unable to allocate memory for hostname\n" );
     *		exit( 1 );
     *		}
     *
     *	    strncpy( hostname, buf_ptr, buf_ptr_end - buf_ptr );
     *	    *( hostname + ( buf_ptr_end - buf_ptr ) ) = '\0';
     */
	server_addr = getaddr( hostname, port );
	if ( server_addr == NULL )
	    {
	    printf( "Unable to get journal server address\n" );
	    exit( 1 );
	    }
	}

    datafd = socket( AF_INET, SOCK_STREAM, 0 );
    if ( datafd < 0 )
	{
	perror( "socket" );
	exit( 1 );
	}

    cc = connect( datafd, (struct sockaddr *)server_addr,
	sizeof(struct sockaddr) );
    if ( cc < 0 )
	{
	perror( "connect" );
	exit( 1 );
	}

    return( datafd );
}


/*
 * getaddr retrieves the ip address of the host where the data server
 * exists. The function is given the name host to look up the address.
 * It returns the sockaddr structure for the host.
 */

struct sockaddr_in *
getaddr( host, port )
char   *host;			/* name of host to get address for */
int     port;
{
    struct sockaddr_in *serv_addr;	/* inet address field */
    struct hostent *hp;		/* name entry for host */
    unsigned long inaddr;	/* holds the ip address of host */

    /* create space for serv_addr, struct in_addr for internet address  */

    serv_addr = (struct sockaddr_in *) malloc( sizeof(struct sockaddr_in) );
    if ( serv_addr == NULL )
	{
	printf( "Unable to allocate memory for serv_addr\n" );
	exit( 1 );
	}

    /* check to see if the host name is an IP address  */

    inaddr = inet_addr( host );
    if ( inaddr != -1 )
	{
	bcopy( (char *) &inaddr, serv_addr->sin_addr, sizeof(inaddr) );
	}

    /* The address must be a name, so get IP address for the name */

    else
	{
	hp = gethostbyname( host );
	if ( hp == NULL )
	    {
	    printf( "Unable to get host information\n" );
	    exit( 1 );
	    }

	/* put address into the sockaddr_in structure */

	bcopy( (char *)hp->h_addr_list[0],
	    (char *)&(serv_addr->sin_addr), hp->h_length);
	}

    /* Fill in the port and address family field */

    serv_addr->sin_port = htons( port );
    serv_addr->sin_family = AF_INET;

    return( serv_addr );
}


/*
 * This routine will take a buffer of data and convert it to a data
 * record. This data record is then stored in various formats
 */

void 
get_data( file_des, data_func ,authenticate)
int     file_des;		/* file descriptor for data file */
void     (*data_func) ();
int     authenticate;
{
    char  **datarecord;		/* pointer for getting pure data records */
    int     nread;		/* number of bytes read */
    char  **leftover;		/* data record not found */
    struct data_record *data;	/* data record with data */
    char    buf[MAX_BUF];	/* buffer to put in data */

    bzero( buf, MAX_BUF );
    datarecord = (char **) malloc( sizeof(char *) );
    if ( datarecord == NULL )
	{
	printf( "Memory not available for datarecord\n" );
	exit( 1 );
	}

    leftover = (char **) malloc( sizeof(char *) );
    if ( leftover == NULL )
	{
	printf( "Memory not available for leftover\n" );
	exit( 1 );
	}

    *leftover = NULL;
    while( ( nread = read( file_des,buf,MAX_BUF ) ) > 0 )
	{
	char   *rec_ptr = buf;	/* pointer to a data record */
	int     bytes_read = 0;	/* number of bytes read by get_record */
	struct data_record *fill_data();	/* routine to fill data
			       record */


	/* read in a buffer worth of data.
       Don't worry about records at this point */


	/* go through the data and interpret the buffer full */

	while ( bytes_read < nread )
	    {
	    /* read in a record worth of data. It is possible that the full
     record is not available. If so, the partial data is in
     leftover */

	    bytes_read += get_record( rec_ptr, datarecord, leftover,
		bytes_read, nread);

	    /* if get_record doesn't return a record, we need to move the ptr
     and continue reading in new data */

	    if ( !*datarecord )
		{
		rec_ptr = buf;
		continue;
		}

	    /* skip over the two newlines */
	    rec_ptr = buf + bytes_read;
	    while ((rec_ptr < buf + nread ) &&(*rec_ptr == '\n'))
		{
		rec_ptr++;
		bytes_read++;
		}


	    /* get a data record filled with data */

	    data = fill_data( *datarecord,authenticate );

	    /* store the data into internal memory */

	    if( data )
		( *data_func ) ( data );
	    }

	/* do this while we fill our buffer with data each time */

	}

    /* No more data in pipe and we have data to put away so do it */

    if ( *leftover )
	{
	/* fill a data record */

	data = fill_data( *leftover,authenticate );

	/* store the data */

	if( data )
	    ( *data_func ) ( data );
	}
}


char   *
getkeyword( databuf, value )
char   *databuf;
char  **value;
{
    char   *keyword;		/* magic token */
    char   *getword();		/* function to grab word */


    /* allocate space for a double pointer to get a word and set to null */

    *value = NULL;

    /* get the keyword. If the : is not present, the line is not good */

    keyword = strtok( databuf, ":" );
    if ( keyword == NULL )
	return( NULL );

    /* set value to the value of keyword */

    *value = keyword + strlen( keyword ) + 1;

    return( keyword );
}


struct intf_prot_addr *
get_ipaddr( value )
char  **value;
{
    struct intf_prot_addr *interface_addr;
    struct sockaddr_in *inetaddr;

    interface_addr = (struct intf_prot_addr *)
	malloc( sizeof(struct intf_prot_addr) );
    if ( interface_addr == NULL )
	{
	printf( "unable to allocate memory for interface_addr\n" );
	exit( 1 );
	}

    bzero( interface_addr, sizeof(struct intf_prot_addr) );

    /* determine the type of address sought from the addr family field */

    switch ( interface_addr->addr_family = getfamily( value ) )
	{

    case AF_INET:		/* Handle the case for Internet addresses */

	inetaddr = (struct sockaddr_in *)
	    malloc( sizeof(struct sockaddr_in) );
	if ( inetaddr == NULL )
	    {
	    printf( "error in getting space for protocol address\n" );
	    exit( 1 );
	    }

	/* put the IP address into a IP specific data structure. put
       address into data structure */

	bzero( inetaddr, sizeof(struct sockaddr_in) );
	inetaddr->sin_family = AF_INET;

	/* possible that IP address not present, so break */

	inetaddr->sin_addr.s_addr = inet_addr( getword( value ) );
	if ( inetaddr->sin_addr.s_addr == -1 )
	    break;

	interface_addr->pa = (struct sockaddr *) inetaddr;
	break;

    default:
	interface_addr = NULL;

	}
    return( interface_addr );
}



struct data_record *
make_data( type )
int     type;
{
    struct data_record *data;

    data = (struct data_record *) malloc( sizeof(struct data_record) );
    if ( data == NULL )
	{
	printf( "Error in allocating space for data\n" );
	exit( 1 );
	}

    bzero( data, sizeof(struct data_record) );

    switch ( type )
	{

    case INTF:
	data->data_type = INTF;
	data->dr_type.intf = (struct intf *) malloc( sizeof(struct intf) );
	if ( data->dr_type.intf == NULL )
	    {
	    printf( "Error in allocating intf\n" );
	    exit( 1 );
	    }

	bzero( data->dr_type.intf, sizeof(struct intf) );
	break;

    case GATEWAY:
	data->data_type = GATEWAY;
	data->dr_type.gateway = (struct host *) malloc(sizeof(struct host));
	if ( data->dr_type.gateway == NULL )
	    {
	    printf( "Unable to allocate space for gateway\n" );
	    exit( 1 );
	    }

	bzero( data->dr_type.gateway, sizeof(struct host) );
	break;

    case NETWORK:
	data->data_type = NETWORK;
	data->dr_type.subnet = (struct network *)
	    malloc( sizeof(struct network) );
	if ( data->dr_type.subnet == NULL )
	    {
	    printf( "Unable to allocate memory for subnet\n" );
	    exit( 1 );
	    }

	bzero( data->dr_type.subnet, sizeof(struct network) );
	}

    return( data );
}



struct intf_list *
make_intf()
{
    struct intf_list *new_intf;

    new_intf = (struct intf_list *) malloc( sizeof(struct intf_list) );
    if ( new_intf == NULL )
	{
	printf( "Unable to allocate memory for new_intf\n" );
	exit( 1 );
	}

    bzero( new_intf, sizeof(struct intf_list) );
    new_intf->subnet_intf = (struct intf *) malloc( sizeof(struct intf) );
    if ( new_intf->subnet_intf == NULL )
	{
	printf( "Unable to allocate memory for new_intf->subnet_intf\n" );
	exit( 1 );
	}

    bzero( new_intf->subnet_intf, sizeof(struct intf) );

    new_intf->subnet_intf->ip_addr = (struct intf_prot_addr *)
	malloc(sizeof(struct intf_prot_addr));
    if ( new_intf->subnet_intf->ip_addr == NULL )
	{
	printf( "unable to allocate memory for ip_addr\n" );
	exit( 1 );
	}

    bzero( new_intf->subnet_intf->ip_addr, sizeof(struct intf_prot_addr) );
    return( new_intf );
}



struct subnet *
make_subnet()
{
    struct subnet *new_subnet;

    new_subnet = (struct subnet *) malloc(sizeof(struct subnet));
    if ( new_subnet == NULL )
	{
	printf( "Unable to allocate memory for new_subnet\n" );
	exit( 1 );
	}

    bzero( new_subnet, sizeof(struct subnet) );
    new_subnet->subnet_info = (struct network *) malloc(sizeof(struct network));
    if ( new_subnet->subnet_info == NULL )
	{
	printf( "Unable to allocate memory for subnet_info\n" );
	exit( 1 );
	}

    bzero( new_subnet->subnet_info, sizeof(struct network) );
    new_subnet->subnet_info->net_numb = (struct intf_prot_addr *)
	malloc(sizeof(struct intf_prot_addr));
    if ( new_subnet->subnet_info->net_numb == NULL )
	{
	printf( "Unable to allocate memory for net_numb\n" );
	exit( 1 );
	}

    bzero( new_subnet->subnet_info->net_numb, sizeof(struct intf_prot_addr));
    return( new_subnet );
}



struct gw_pointer *
make_gateway()
{
    struct gw_pointer *new_gateway;

    new_gateway = (struct gw_pointer *) malloc(sizeof(struct gw_pointer));
    if ( new_gateway == NULL )
	{
	printf( "Unable to allocate memory for new_gateway\n" );
	exit( 1 );
	}

    bzero( new_gateway, sizeof(struct gw_pointer) );
    new_gateway->gw_host = (struct host *) malloc(sizeof(struct host));
    if ( new_gateway->gw_host == NULL )
	{
	printf( "Unable to allocate space for gw_host\n" );
	exit( 1 );
	}

    bzero( new_gateway->gw_host, sizeof(struct host) );
    new_gateway->gw_host->intfs = make_intf();

    /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
     
     new_gateway->gw_host->intfs = (struct intf_list *)
     malloc( sizeof(struct intf_list) );
     if ( new_gateway->gw_host->intfs == NULL )
     {
     printf( "Unable to allocate memory for intfs\n" );
     exit( 1 );
     }
     
     bzero( new_gateway->gw_host->intfs,sizeof(struct intf_list) );
     if ( ( new_gateway->gw_host->intfs->subnet_intf = (struct intf *)
     malloc( sizeof(struct intf) ) ) == NULL )
     {
     printf( "Unable to allocate memory for subnet_intfs\n" );
     exit( 1 );
     }
     
     bzero( new_gateway->gw_host->intfs->subnet_intf, sizeof(struct intf) );
     new_gateway->gw_host->intfs->subnet_intf->ip_addr =
     (struct intf_prot_addr *)
     malloc(sizeof(struct intf_prot_addr) );
     if ( new_gateway->gw_host->intfs->subnet_intf->ip_addr == NULL )
     {
     printf( "Unable to allocate memory for ip_addr\n" );
     exit( 1 );
     }
     
     bzero( new_gateway->gw_host->intfs->subnet_intf->ip_addr,
     sizeof(struct intf_prot_addr));
     * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

    return( new_gateway );
}

